/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"

/* Header removal size limited to 128B (64 words) */
#define MLX5HWS_DV_ACTION_REMOVE_HEADER_MAX_SIZE 128

/*
 * A write operation uses 2 WQEBBs (control and data). We can fit a maximum of
 * 16 consecutive 64B arguments in a single write operation.
 */
#define ACTION_ARG_MAX_WRITE_SIZE (MLX5_IFC_ARG_SZ_DATA * SEND_MAX_WQES_PER_RULE / 2)

static bool action_flags_is_root(uint64_t action_flags)
{
	return action_flags & MLX5DV_HWS_ACTION_FLAG_ROOT;
}

static bool action_flags_is_fdb_any(uint64_t action_flags)
{
	return action_flags & (MLX5DV_HWS_ACTION_FLAG_FDB |
			       MLX5DV_HWS_ACTION_FLAG_FDB_TX |
			       MLX5DV_HWS_ACTION_FLAG_FDB_RX);
}

static bool action_flags_is_fdb_rx_or_tx(uint64_t action_flags)
{
	return action_flags & (MLX5DV_HWS_ACTION_FLAG_FDB_TX |
			       MLX5DV_HWS_ACTION_FLAG_FDB_RX);
}

static int action_flags_get_tbl_type(uint64_t action_flags,
				     enum mlx5dv_hws_table_type *tbl_type)
{
	uint32_t tbl_type_value = (MLX5DV_HWS_ACTION_FLAG_NIC_RX |
				   MLX5DV_HWS_ACTION_FLAG_NIC_TX |
				   MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_RX |
				   MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_TX |
				   MLX5DV_HWS_ACTION_FLAG_FDB_RX |
				   MLX5DV_HWS_ACTION_FLAG_FDB_TX |
				   MLX5DV_HWS_ACTION_FLAG_FDB);

	/* Calculate table_type */
	switch (action_flags & tbl_type_value) {
	case MLX5DV_HWS_ACTION_FLAG_NIC_RX:
		*tbl_type = MLX5DV_HWS_TABLE_TYPE_NIC_RX;
		break;
	case MLX5DV_HWS_ACTION_FLAG_NIC_TX:
		*tbl_type = MLX5DV_HWS_TABLE_TYPE_NIC_TX;
		break;
	case MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_RX:
		*tbl_type = MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX;
		break;
	case MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_TX:
		*tbl_type = MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX;
		break;
	case MLX5DV_HWS_ACTION_FLAG_FDB_RX:
		*tbl_type = MLX5DV_HWS_TABLE_TYPE_FDB_RX;
		break;
	case MLX5DV_HWS_ACTION_FLAG_FDB_TX:
		*tbl_type = MLX5DV_HWS_TABLE_TYPE_FDB_TX;
		break;
	case MLX5DV_HWS_ACTION_FLAG_FDB:
		*tbl_type = MLX5DV_HWS_TABLE_TYPE_FDB;
		break;
	default:
		/* Allow only one tbl_type per action */
		HWS_ERR("Invalid table type, flags: 0x%lx", action_flags);
		errno = EINVAL;
		return -errno;
	}

	return 0;
}

static uint32_t action_flags_get_ib_port(uint64_t action_flags)
{
	return (action_flags >> BITS_IN_DW);
}

static uint32_t
action_tbl_type_to_action_flags(enum mlx5dv_hws_table_type tbl_type)
{
	switch (tbl_type) {
	case MLX5DV_HWS_TABLE_TYPE_NIC_RX:
		return MLX5DV_HWS_ACTION_FLAG_NIC_RX;
	case MLX5DV_HWS_TABLE_TYPE_NIC_TX:
		return MLX5DV_HWS_ACTION_FLAG_NIC_TX;
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX:
		return MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_RX;
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX:
		return MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_TX;
	case MLX5DV_HWS_TABLE_TYPE_FDB_RX:
		return MLX5DV_HWS_ACTION_FLAG_FDB_RX;
	case MLX5DV_HWS_TABLE_TYPE_FDB_TX:
		return MLX5DV_HWS_ACTION_FLAG_FDB_TX;
	case MLX5DV_HWS_TABLE_TYPE_FDB:
		return MLX5DV_HWS_ACTION_FLAG_FDB;
	default:
		/* Allow only one tbl_type per action */
		HWS_ERR("Invalid table type, tbl_type: %d", tbl_type);
		errno = EINVAL;
		return -errno;
	}
}

int action_get_default_stc(struct mlx5dv_hws_context *ctx,
			   uint8_t tbl_type,
			   uint32_t ib_port)
{
	struct cmd_stc_modify_attr stc_attr = {0};
	struct action_default_stc *default_stc;
	struct context_common_res *common_res;
	int ret;

	common_res = context_get_common_res(ctx, tbl_type, ib_port);
	if (!common_res) {
		HWS_ERR("Failed to get common resource");
		return -errno;
	}

	if (common_res->default_stc) {
		common_res->default_stc->refcount++;
		return 0;
	}

	default_stc = calloc(1, sizeof(*default_stc));
	if (!default_stc) {
		HWS_ERR("Failed to allocate memory for default STCs");
		errno = ENOMEM;
		return -errno;
	}

	stc_attr.action_type = MLX5_IFC_STC_ACTION_TYPE_NOP;
	stc_attr.action_offset = ACTION_OFFSET_DW0;
	stc_attr.reparse_mode = MLX5_IFC_STC_REPARSE_IGNORE;

	ret = action_alloc_single_stc(ctx, &stc_attr, tbl_type, ib_port,
				      &default_stc->nop_ctr);
	if (ret) {
		HWS_ERR("Failed to allocate default counter STC");
		goto free_default_stc;
	}

	stc_attr.action_offset = ACTION_OFFSET_DW5;
	ret = action_alloc_single_stc(ctx, &stc_attr, tbl_type, ib_port,
				      &default_stc->nop_dw5);
	if (ret) {
		HWS_ERR("Failed to allocate default NOP DW5 STC");
		goto free_nop_ctr;
	}

	stc_attr.action_offset = ACTION_OFFSET_DW6;
	ret = action_alloc_single_stc(ctx, &stc_attr, tbl_type, ib_port,
				      &default_stc->nop_dw6);
	if (ret) {
		HWS_ERR("Failed to allocate default NOP DW6 STC");
		goto free_nop_dw5;
	}

	stc_attr.action_offset = ACTION_OFFSET_DW7;
	ret = action_alloc_single_stc(ctx, &stc_attr, tbl_type, ib_port,
				      &default_stc->nop_dw7);
	if (ret) {
		HWS_ERR("Failed to allocate default NOP DW7 STC");
		goto free_nop_dw6;
	}

	stc_attr.action_offset = ACTION_OFFSET_HIT;
	if (!context_shared_gvmi_used(ctx)) {
		stc_attr.action_type = MLX5_IFC_STC_ACTION_TYPE_ALLOW;
	} else {
		/* On shared gvmi the default hit behavior is jump to alias end ft */
		stc_attr.action_type = MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_FT;
		stc_attr.dest_table_id = common_res->gvmi_res->aliased_end_ft->object_id;
	}

	ret = action_alloc_single_stc(ctx, &stc_attr, tbl_type, ib_port,
				      &default_stc->default_hit);
	if (ret) {
		HWS_ERR("Failed to allocate default allow STC");
		goto free_nop_dw7;
	}

	common_res->default_stc = default_stc;
	common_res->default_stc->refcount++;

	return 0;

free_nop_dw7:
	action_free_single_stc(ctx, tbl_type, ib_port, &default_stc->nop_dw7);
free_nop_dw6:
	action_free_single_stc(ctx, tbl_type, ib_port, &default_stc->nop_dw6);
free_nop_dw5:
	action_free_single_stc(ctx, tbl_type, ib_port, &default_stc->nop_dw5);
free_nop_ctr:
	action_free_single_stc(ctx, tbl_type, ib_port, &default_stc->nop_ctr);
free_default_stc:
	free(default_stc);
	return -errno;
}

void action_put_default_stc(struct mlx5dv_hws_context *ctx,
			    uint8_t tbl_type,
			    uint32_t ib_port)
{
	struct action_default_stc *default_stc;
	struct context_common_res *common_res;

	common_res = context_get_common_res(ctx, tbl_type, ib_port);
	default_stc = common_res->default_stc;
	if (--default_stc->refcount)
		return;

	action_free_single_stc(ctx, tbl_type, ib_port, &default_stc->default_hit);
	action_free_single_stc(ctx, tbl_type, ib_port, &default_stc->nop_dw7);
	action_free_single_stc(ctx, tbl_type, ib_port, &default_stc->nop_dw6);
	action_free_single_stc(ctx, tbl_type, ib_port, &default_stc->nop_dw5);
	action_free_single_stc(ctx, tbl_type, ib_port, &default_stc->nop_ctr);

	free(default_stc);

	common_res->default_stc = NULL;
}

static bool
action_validate_hws_action(struct mlx5dv_hws_context *ctx, uint64_t action_flags)
{
	if (action_flags_is_fdb_any(action_flags) && !ctx->caps->eswitch_manager) {
		HWS_ERR("Cannot create FDB action for non-eswitch-manager");
		return false;
	}

	if (action_flags_is_root(action_flags)) {
		if (action_flags_is_fdb_rx_or_tx(action_flags)) {
			HWS_ERR("Root table doesn't support FDB RX/TX");
			return false;
		}

		return true;
	}

	if (!(ctx->flags & CONTEXT_FLAG_HWS_SUPPORT)) {
		HWS_ERR("Cannot create HWS action since HWS is not supported");
		return false;
	}

	return true;
}

static struct mlx5dv_hws_action *
action_create_generic(struct mlx5dv_hws_context *ctx,
		      enum action_type action_type,
		      uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;

	if (!action_validate_hws_action(ctx, action_flags)) {
		errno = ENOTSUP;
		return NULL;
	}

	action = calloc(1, sizeof(*action));
	if (!action) {
		HWS_ERR("Failed to allocate memory for action [%d]", action_type);
		errno = ENOMEM;
		return NULL;
	}

	action->ctx = ctx;
	action->type = action_type;
	action->flags = action_flags;

	return action;
}

static void
action_dest_fill_stc_attr(struct mlx5dv_hws_action *action,
			  enum mlx5dv_hws_table_type tbl_type,
			  struct mlx5dv_devx_obj *obj,
			  struct cmd_stc_modify_attr *attr)
{
	struct cmd_query_caps *caps = action->ctx->caps;

	attr->reparse_mode = MLX5_IFC_STC_REPARSE_IGNORE;

	switch (action->type) {
	case ACTION_TYP_DEST_DROP:
		attr->action_type = MLX5_IFC_STC_ACTION_TYPE_DROP;
		attr->action_offset = ACTION_OFFSET_HIT;
		break;
	case ACTION_TYP_DEST_MISS:
		attr->action_offset = ACTION_OFFSET_HIT;

		if (table_is_fw_fdb_any(tbl_type)) {
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_VPORT;
			attr->vport.esw_owner_vhca_id = action->ctx->caps->vhca_id;
			attr->vport.vport_num =
				action->ctx->caps->eswitch_manager_vport_number;
			attr->vport.eswitch_owner_vhca_id_valid =
				action->ctx->caps->merged_eswitch;
		} else {
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_ALLOW;
		}
		break;
	case ACTION_TYP_DEST_TIR:
		attr->action_type = MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_TIR;
		attr->action_offset = ACTION_OFFSET_HIT;
		attr->dest_tir_num = obj->object_id;
		break;
	case ACTION_TYP_DEST_TBL:
		attr->action_type = MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_FT;
		attr->action_offset = ACTION_OFFSET_HIT;
		attr->dest_table_id = obj->object_id;

		/* Cross pipe to RX only for unified FDB Rx dest */
		if ((action->dest_tbl.tbl->type == MLX5DV_HWS_TABLE_TYPE_FDB_RX)
		    && caps->stc_jump_to_fdb_rx)
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_FT_FDB_RX;
		break;
	case ACTION_TYP_DEST_ARRAY:
		attr->action_type = MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_FT;
		attr->action_offset = ACTION_OFFSET_HIT;
		attr->dest_table_id = obj->object_id;
		break;
	case ACTION_TYP_DEST_ROOT:
		attr->action_type = MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_FT;
		attr->action_offset = ACTION_OFFSET_HIT;
		attr->dest_table_id = action->root_tbl.sa->id;
		break;
	case ACTION_TYP_DEST_VPORT:
		attr->action_type = MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_VPORT;
		attr->action_offset = ACTION_OFFSET_HIT;
		attr->vport.vport_num = action->vport.vport_num;
		if (attr->vport.vport_num == WIRE_PORT)
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_UPLINK;
		attr->vport.esw_owner_vhca_id =	action->vport.esw_owner_vhca_id;
		attr->vport.eswitch_owner_vhca_id_valid = caps->merged_eswitch;
		break;
	case ACTION_TYP_DEST_MATCHER:
		attr->action_type = MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_STE_TABLE;
		attr->action_offset = ACTION_OFFSET_HIT;
		attr->ste_table.ste_obj_id =
			action->dest_matcher.matcher->match_ste.ste_0->object_id;
		attr->ste_table.match_definer_id = caps->trivial_match_definer;
		break;
	default:
		HWS_ERR("Invalid action type %d", action->type);
		assert(false);
	}
}

int action_alloc_single_stc(struct mlx5dv_hws_context *ctx,
			    struct cmd_stc_modify_attr *stc_attr,
			    enum mlx5dv_hws_table_type tbl_type,
			    uint32_t ib_port,
			    struct pool_chunk *stc)
{
	struct cmd_stc_modify_attr cur_stc_attr = *stc_attr;
	struct mlx5dv_devx_obj *devx_obj;
	struct pool *stc_pool;
	int ret;

	stc_pool = context_get_stc_pool(ctx, tbl_type, ib_port);
	if (!stc_pool) {
		HWS_ERR("Failed to get STC pool");
		return -errno;
	}

	ret = pool_chunk_alloc(stc_pool, stc);
	if (ret) {
		HWS_ERR("Failed to allocate single action STC");
		return ret;
	}

	cur_stc_attr.stc_offset = stc->offset;

	/* Dynamic reparse not supported, overwrite and use default */
	if (!context_cap_dynamic_reparse(ctx))
		cur_stc_attr.reparse_mode = MLX5_IFC_STC_REPARSE_IGNORE;

	devx_obj = pool_chunk_get_base_devx_obj(stc_pool, stc);

	/* Parallel devx obj modify over the same base not supported */
	pthread_spin_lock(&ctx->stc_lock);
	ret = cmd_stc_modify(devx_obj, &cur_stc_attr);
	pthread_spin_unlock(&ctx->stc_lock);
	if (ret) {
		HWS_ERR("Failed to modify STC action_type %d tbl_type %d",
			stc_attr->action_type, tbl_type);
		goto free_chunk;
	}

	return 0;

free_chunk:
	pool_chunk_free(stc_pool, stc);
	return errno;
}

void action_free_single_stc(struct mlx5dv_hws_context *ctx,
			    enum mlx5dv_hws_table_type tbl_type,
			    uint32_t ib_port,
			    struct pool_chunk *stc)
{
	struct cmd_stc_modify_attr stc_attr = {0};
	struct mlx5dv_devx_obj *devx_obj;
	struct pool *stc_pool;

	stc_pool = context_get_stc_pool(ctx, tbl_type, ib_port);

	/* Modify the STC not to point to an object */
	stc_attr.action_type = MLX5_IFC_STC_ACTION_TYPE_DROP;
	stc_attr.action_offset = ACTION_OFFSET_HIT;
	stc_attr.stc_offset = stc->offset;

	devx_obj = pool_chunk_get_base_devx_obj(stc_pool, stc);

	/* Parallel devx obj modify over the same base not supported */
	pthread_spin_lock(&ctx->stc_lock);
	cmd_stc_modify(devx_obj, &stc_attr);
	pthread_spin_unlock(&ctx->stc_lock);

	pool_chunk_free(stc_pool, stc);
}

static int
action_dest_create_stc(struct mlx5dv_hws_action *action,
		       struct mlx5dv_devx_obj *obj)
{
	struct mlx5dv_hws_context *ctx = action->ctx;
	struct cmd_stc_modify_attr stc_attr = {0};
	enum mlx5dv_hws_table_type tbl_type;
	uint32_t ib_port;
	int ret;

	ib_port = action_flags_get_ib_port(action->flags);

	ret = action_flags_get_tbl_type(action->flags, &tbl_type);
	if (ret) {
		HWS_ERR("Failed to get a single table type from action_flags");
		return ret;
	}

	action_dest_fill_stc_attr(action, tbl_type, obj, &stc_attr);

	pthread_spin_lock(&ctx->ctrl_lock);
	ret = action_alloc_single_stc(ctx, &stc_attr, tbl_type, ib_port, action->stc);
	pthread_spin_unlock(&ctx->ctrl_lock);
	if (ret) {
		HWS_ERR("Failed to allocate STC table type [%d]", tbl_type);
		return ret;
	}

	return 0;
}

static void action_dest_destroy_stc(struct mlx5dv_hws_action *action)
{
	struct mlx5dv_hws_context *ctx = action->ctx;
	enum mlx5dv_hws_table_type tbl_type;
	uint32_t ib_port;
	int ret;

	ret = action_flags_get_tbl_type(action->flags, &tbl_type);
	if (ret) {
		assert(false);
		HWS_ERR("Failed to find action tbl_type");
		return;
	}

	ib_port = action_flags_get_ib_port(action->flags);
	action_free_single_stc(ctx, tbl_type, ib_port, action->stc);
}

int action_stc_cache_init(struct action_stc_cache **cache)
{
	struct action_stc_cache *new_cache;

	new_cache = calloc(1, sizeof(*new_cache));
	if (!new_cache) {
		errno = ENOMEM;
		return -errno;
	}

	list_head_init(&new_cache->head);

	*cache = new_cache;

	return 0;
}

void action_stc_cache_uninit(struct action_stc_cache *cache)
{
	if (!list_empty(&cache->head))
		HWS_ERR("STC cache not empty !");

	free(cache);
}

static struct action_stc_cache_entry *
action_stc_cache_search(struct action_stc_cache *cache,
			struct action_stc_cache_data *data)
{
	struct action_stc_cache_entry *cache_entry;
	struct action_stc_cache_data *cached_data;

	list_for_each(&cache->head, cache_entry, next) {
		cached_data = &cache_entry->data;
		if (cached_data->tbl_type == data->tbl_type &&
		    cached_data->ib_port == data->ib_port &&
		    cached_data->attr.action_type == data->attr.action_type &&
		    !memcmp(&cached_data->attr, &data->attr,
			    sizeof(struct cmd_stc_modify_attr)))
			return cache_entry;
	}

	return NULL;
}

static int
action_stc_cache_add(struct action_stc_cache *cache,
		     struct action_stc_cache_data *cache_data,
		     struct pool_chunk *stc)
{
	struct action_stc_cache_entry *cache_entry;

	cache_entry = calloc(1, sizeof(*cache_entry));
	if (!cache_entry) {
		HWS_ERR("Failed to allocate cached_stc");
		errno = ENOMEM;
		return -errno;
	}

	cache_entry->stc = *stc;
	cache_entry->refcount = 1;
	memcpy(&cache_entry->data, cache_data, sizeof(*cache_data));

	list_add(&cache->head, &cache_entry->next);

	return 0;
}

static int
action_stc_cache_get(struct mlx5dv_hws_context *ctx,
		     struct action_stc_cache_data *cache_data,
		     struct mlx5dv_devx_obj *obj,
		     struct pool_chunk *stc)
{
	struct action_stc_cache_entry *cached_entry;
	int ret;

	cached_entry = action_stc_cache_search(ctx->stc_cache, cache_data);
	if (cached_entry) {
		cached_entry->refcount++;
		*stc = cached_entry->stc;
		return 0;
	}

	ret = action_alloc_single_stc(ctx,
				      &cache_data->attr,
				      cache_data->tbl_type,
				      cache_data->ib_port,
				      stc);
	if (ret) {
		HWS_ERR("Failed to create new STC for action ifc type: %d",
			cache_data->attr.action_type);
		return ret;
	}

	ret = action_stc_cache_add(ctx->stc_cache, cache_data, stc);
	if (ret) {
		HWS_ERR("Failed to add stc to cache");
		goto free_stc;
	}

	return 0;

free_stc:
	action_free_single_stc(ctx, cache_data->tbl_type, cache_data->ib_port, stc);
	return ret;
}

static void action_stc_cache_put(struct mlx5dv_hws_context *ctx,
				 enum mlx5dv_hws_table_type tbl_type,
				 uint32_t ib_port,
				 struct pool_chunk *stc)
{
	struct action_stc_cache_entry *cache_entry = NULL;
	struct action_stc_cache *cache = ctx->stc_cache;
	bool found = false;

	list_for_each(&cache->head, cache_entry, next) {
		if (cache_entry->data.tbl_type == tbl_type &&
		    cache_entry->data.ib_port == ib_port &&
		    cache_entry->stc.offset == stc->offset) {
			found = true;
			break;
		}
	}

	if (!found) {
		assert(found);
		HWS_ERR("Fail to find cached STC (tbl_type : %d offset: %d)",
			tbl_type, stc->offset);
		return;
	}

	if (--cache_entry->refcount)
		return;

	action_free_single_stc(ctx, tbl_type, ib_port, stc);
	list_del(&cache_entry->next);
	free(cache_entry);
}

int mlx5dv_hws_action_flags_set_ib_port(struct mlx5dv_hws_context *ctx,
					uint32_t ib_port,
					uint64_t *action_flags)
{
	uint16_t vport_num;
	bool other_vport;
	int ret;

	if (!(*action_flags & MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_RX) &&
	    !(*action_flags & MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_TX)) {
		HWS_ERR("Only PRDMA action flags require ib-port");
		errno = EINVAL;
		return -errno;
	}

	ret = context_query_ib_port(ctx, ib_port, &vport_num, &other_vport);
	if (ret) {
		HWS_ERR("Failed querying ib-port [%d]", ib_port);
		return ret;
	}

	/* Format: [IB Port 32b][FLAGS 32b] */
	*action_flags |= (uint64_t)ib_port << BITS_IN_DW;

	return 0;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_dest_drop(struct mlx5dv_hws_context *ctx,
				   uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;
	int ret;

	action = action_create_generic(ctx, ACTION_TYP_DEST_DROP, action_flags);
	if (!action)
		return NULL;

	if (!action_flags_is_root(action_flags)) {
		ret = action_dest_create_stc(action, NULL);
		if (ret)
			goto free_action;
	}

	return action;

free_action:
	free(action);
	return NULL;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_dest_default_miss(struct mlx5dv_hws_context *ctx,
					   uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;
	int ret;

	if (context_shared_gvmi_used(ctx)) {
		HWS_ERR("Cannot use default-miss action in shared GVMI context");
		errno = ENOTSUP;
		return NULL;
	}

	action = action_create_generic(ctx, ACTION_TYP_DEST_MISS, action_flags);
	if (!action)
		return NULL;

	if (!action_flags_is_root(action_flags)) {
		ret = action_dest_create_stc(action, NULL);
		if (ret)
			goto free_action;
	}

	return action;

free_action:
	free(action);
	return NULL;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_dest_table(struct mlx5dv_hws_context *ctx,
				    struct mlx5dv_hws_table *tbl,
				    uint64_t action_flags)
{
	enum mlx5dv_hws_table_type tbl_type;
	struct mlx5dv_hws_action *action;
	int ret;

	ret = action_flags_get_tbl_type(action_flags, &tbl_type);
	if (ret)
		return NULL;

	if (table_is_root(tbl)) {
		HWS_ERR("Root table cannot be set as destination");
		errno = ENOTSUP;
		return NULL;
	}

	if (ctx != tbl->ctx) {
		HWS_ERR("Table belong to other context!");
		errno = ENOTSUP;
		return NULL;
	}

	action = action_create_generic(ctx, ACTION_TYP_DEST_TBL, action_flags);
	if (!action)
		return NULL;

	action->dest_tbl.tbl = tbl;

	if (action_flags_is_root(action_flags)) {
		if (context_shared_gvmi_used(ctx))
			action->dest_tbl.devx_obj = tbl->local_ft;
		else
			action->dest_tbl.devx_obj = tbl->ft;
	} else {
		action->dest_tbl.devx_obj = tbl->ft;
		ret = action_dest_create_stc(action, tbl->ft);
		if (ret)
			goto free_action;
	}

	return action;

free_action:
	free(action);
	return NULL;
}

static int
action_create_dest_vport_hws(struct mlx5dv_hws_context *ctx,
			     struct mlx5dv_hws_action *action,
			     uint32_t ib_port_num)
{
	uint16_t vport_num, esw_vhca_id;
	int ret;

	ret = cmd_ib_port_query(ctx->ibv_ctx, ib_port_num, &vport_num, &esw_vhca_id);
	if (ret) {
		HWS_ERR("Failed querying ib-port %d", ib_port_num);
		return ret;
	}

	if (!ctx->caps->merged_eswitch && ctx->caps->vhca_id != esw_vhca_id) {
		HWS_ERR("Not merged-eswitch (%d) forbidden send other vhca_id (%d)",
			ctx->caps->vhca_id, esw_vhca_id);
		errno = ENOTSUP;
		return -errno;
	}

	action->vport.vport_num = vport_num;
	action->vport.esw_owner_vhca_id = esw_vhca_id;

	ret = action_dest_create_stc(action, NULL);
	if (ret) {
		HWS_ERR("Failed creating stc for port %d", ib_port_num);
		return ret;
	}

	return 0;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_dest_vport(struct mlx5dv_hws_context *ctx,
				    uint32_t ib_port_num,
				    uint64_t action_flags)
{
	enum mlx5dv_hws_table_type tbl_type;
	struct mlx5dv_hws_action *action;
	int ret;

	if (context_shared_gvmi_used(ctx)) {
		HWS_ERR("Cannot use vport action in shared GVMI context");
		errno = ENOTSUP;
		return NULL;
	}

	ret = action_flags_get_tbl_type(action_flags, &tbl_type);
	if (ret)
		return NULL;

	if (!table_is_fdb_any(tbl_type)) {
		HWS_ERR("Vport action is supported for FDB only");
		errno = ENOTSUP;
		return NULL;
	}

	if (action_flags_is_root(action_flags)) {
		HWS_ERR("Vport action is supported for non-root only");
		errno = ENOTSUP;
		return NULL;
	}

	action = action_create_generic(ctx, ACTION_TYP_DEST_VPORT, action_flags);
	if (!action)
		return NULL;

	ret = action_create_dest_vport_hws(ctx, action, ib_port_num);
	if (ret) {
		HWS_ERR("Failed to create vport action HWS");
		goto free_action;
	}

	return action;

free_action:
	free(action);
	return NULL;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_dest_tir(struct mlx5dv_hws_context *ctx,
				  struct mlx5dv_devx_obj *obj,
				  bool is_local,
				  uint64_t action_flags)
{
	enum mlx5dv_hws_table_type tbl_type;
	struct mlx5dv_hws_action *action;
	int ret;

	ret = action_flags_get_tbl_type(action_flags, &tbl_type);
	if (ret)
		return NULL;

	if (tbl_type != MLX5DV_HWS_TABLE_TYPE_NIC_RX &&
	    tbl_type != MLX5DV_HWS_TABLE_TYPE_FDB_RX &&
	    tbl_type != MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX) {
		HWS_ERR("TIR action can be supported on NIC/RDMA/FDB RX");
		errno = ENOTSUP;
		return NULL;
	}

	if (!is_local) {
		HWS_ERR("TIR should be created on local ibv_device");
		errno = ENOTSUP;
		return NULL;
	}

	action = action_create_generic(ctx, ACTION_TYP_DEST_TIR, action_flags);
	if (!action)
		return NULL;

	if (!action_flags_is_root(action_flags)) {
		if (context_shared_gvmi_used(ctx)) {
			obj = cmd_allow_and_create_aliased_obj(ctx->local_ibv_ctx,
							       ctx->ibv_ctx,
							       ctx->caps->vhca_id,
							       obj->object_id,
							       MLX5_GENERAL_OBJ_TYPE_TIR_ALIAS);
			if (!obj) {
				HWS_ERR("Failed to create aliased object");
				goto free_action;
			}
		}

		ret = action_dest_create_stc(action, obj);
		if (ret)
			goto clean_alias_obj;
	}

	action->devx_dest.devx_obj = obj;

	return action;

clean_alias_obj:
	if (context_shared_gvmi_used(ctx))
		cmd_destroy_obj(obj);
free_action:
	free(action);
	return NULL;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_dest_matcher(struct mlx5dv_hws_context *ctx,
				      struct mlx5dv_hws_matcher *matcher,
				      enum mlx5dv_hws_action_dest_matcher_type type,
				      uint64_t action_flags)
{
	struct mlx5dv_hws_matcher_attr *attr = &matcher->attr;
	struct mlx5dv_hws_action *action;
	int ret;

	if (type != MLX5DV_HWS_ACTION_DEST_MATCHER_BY_INDEX) {
		HWS_ERR("Only dest matcher by index is supported");
		goto not_supported;
	}

	if (action_flags_is_root(action_flags)) {
		HWS_ERR("Dest matcher action not supported over root");
		goto not_supported;
	}

	if (table_is_root(matcher->tbl)) {
		HWS_ERR("Root matcher cannot be set as destination");
		goto not_supported;
	}

	if (attr->resizable) {
		HWS_ERR("Resizeble matcher is not supported as dest");
		goto not_supported;
	}

	if (!attr->isolated) {
		HWS_ERR("Only isolated matcher supported as destination");
		goto not_supported;
	}

	action = action_create_generic(ctx, ACTION_TYP_DEST_MATCHER, action_flags);
	if (!action)
		return NULL;

	action->dest_matcher.matcher = matcher;

	ret = action_dest_create_stc(action, NULL);
	if (ret) {
		HWS_ERR("Failed to create action jump to matcher STC");
		free(action);
		return NULL;
	}

	return action;

not_supported:
	errno = ENOTSUP;
	return NULL;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_dest_root(struct mlx5dv_hws_context *ctx,
				   uint16_t priority,
				   uint64_t action_flags)
{
	struct mlx5dv_steering_anchor_attr attr = {0};
	enum mlx5dv_hws_table_type tbl_type;
	struct mlx5dv_steering_anchor *sa;
	struct mlx5dv_hws_action *action;
	int ret;

	if (context_shared_gvmi_used(ctx)) {
		HWS_ERR("Cannot use this action in shared GVMI context");
		errno = ENOTSUP;
		return NULL;
	}

	ret = action_flags_get_tbl_type(action_flags, &tbl_type);
	if (ret)
		return NULL;

	if (action_flags_is_root(action_flags)) {
		HWS_ERR("Dest root supported only over HWS tables");
		errno = ENOTSUP;
		return NULL;
	}

	if (table_conv_hws_tbl_type_to_dv_ft(tbl_type, &attr.ft_type))
		return NULL;

	attr.priority = priority;

	sa = mlx5dv_create_steering_anchor(ctx->ibv_ctx, &attr);
	if (!sa)
		return NULL;

	action = action_create_generic(ctx, ACTION_TYP_DEST_ROOT, action_flags);
	if (!action)
		goto free_steering_anchor;

	action->root_tbl.sa = sa;

	ret = action_dest_create_stc(action, NULL);
	if (ret)
		goto free_action;

	return action;

free_action:
	free(action);
free_steering_anchor:
	mlx5dv_destroy_steering_anchor(sa);
	return NULL;

}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_dest_array(struct mlx5dv_hws_context *ctx,
				    size_t num_dest,
				    struct mlx5dv_hws_action_dest_attr dests[],
				    uint64_t action_flags)
{
	struct cmd_packet_reformat_create_attr reformat_attr = {0};
	enum mlx5_ifc_ft_type ft_type = MLX5_IFC_FT_NIC_RX; /* Avoid static check warn */
	struct cmd_ft_create_attr ft_attr = {0};
	struct cmd_set_fte_attr fte_attr = {0};
	struct mlx5dv_devx_obj *reformat_obj;
	enum mlx5dv_hws_table_type tbl_type;
	struct cmd_set_fte_dest *fte_dest;
	struct cmd_forward_tbl *fw_island;
	struct mlx5dv_hws_action *action;
	uint16_t vport_num;
	bool other_vport;
	uint32_t ib_port;
	int i, ret;

	if (num_dest <= 1) {
		errno = EINVAL;
		HWS_ERR("Action must have multiple dests");
		return NULL;
	}

	/* This is a SW limitation, some dests can be allowed in the future */
	if (context_shared_gvmi_used(ctx)) {
		HWS_ERR("Cannot use dest array in shared GVMI context");
		errno = ENOTSUP;
		return NULL;
	}

	ret = action_flags_get_tbl_type(action_flags, &tbl_type);
	if (ret) {
		HWS_ERR("Failed to get a single table type from action_flags");
		return NULL;
	}

	switch (tbl_type) {
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX:
		ib_port = action_flags_get_ib_port(action_flags);
		ret = context_query_ib_port(ctx, ib_port, &vport_num, &other_vport);
		if (ret) {
			HWS_ERR("Failed query IB port [%d]", ib_port);
			return NULL;
		}

		ft_attr.vport = vport_num;
		ft_attr.other_vport = other_vport;
		ft_attr.level = MLX5_IFC_MULTI_PATH_FT_MAX_LEVEL - 1;
		break;
	case MLX5DV_HWS_TABLE_TYPE_NIC_RX:
		ft_attr.level = MLX5_IFC_MULTI_PATH_FT_MAX_LEVEL - 1;
		break;
	case MLX5DV_HWS_TABLE_TYPE_FDB:
	case MLX5DV_HWS_TABLE_TYPE_FDB_RX:
	case MLX5DV_HWS_TABLE_TYPE_FDB_TX:
		ft_attr.level = ctx->caps->fdb_ft.max_level - 1;
		break;
	default:
		HWS_ERR("Destination action not supported over required table");
		errno = ENOTSUP;
		return NULL;
	}

	action = action_create_generic(ctx, ACTION_TYP_DEST_ARRAY, action_flags);
	if (!action)
		return NULL;

	fte_dest = calloc(num_dest, sizeof(*fte_dest));
	if (!fte_dest) {
		HWS_ERR("Failed to allocate memory for destinations");
		errno = ENOMEM;
		goto free_action;
	}

	fte_attr.dests_num = num_dest;
	fte_attr.dests = fte_dest;

	/* Get FW table type */
	table_get_fw_ft_type(tbl_type, &ft_type);
	ft_attr.type = (uint8_t) ft_type;

	/* Process destination */
	for (i = 0; i < num_dest; i++) {
		switch (dests[i].dest_action->type) {
		case ACTION_TYP_DEST_TIR:
			fte_dest[i].destination_type = MLX5_FLOW_DESTINATION_TYPE_TIR;
			fte_dest[i].destination_id = dests[i].dest_action->devx_dest.devx_obj->object_id;
			fte_attr.action_flags |= MLX5_IFC_FLOW_CONTEXT_ACTION_FWD_DEST;
			break;
		case ACTION_TYP_DEST_DROP:
			fte_dest[i].destination_type = MLX5_FLOW_DESTINATION_TYPE_NOP;
			fte_dest[i].destination_id = 0;
			fte_attr.action_flags |= MLX5_IFC_FLOW_CONTEXT_ACTION_FWD_DEST;
			break;
		case ACTION_TYP_DEST_TBL:
			fte_dest[i].destination_type = MLX5_FLOW_DESTINATION_TYPE_FT;
			fte_dest[i].destination_id = dests[i].dest_action->dest_tbl.devx_obj->object_id;
			fte_attr.action_flags |= MLX5_IFC_FLOW_CONTEXT_ACTION_FWD_DEST;
			fte_attr.ignore_flow_level = true;
			break;
		case ACTION_TYP_DEST_MISS:
			if (!table_is_fdb_any(tbl_type)) {
				HWS_ERR("Default miss action supported for FDB only");
				errno = ENOTSUP;
				goto free_dest_list;
			}
			/* Go to esw manager vport */
			fte_dest[i].destination_type = MLX5_FLOW_DESTINATION_TYPE_VPORT;
			fte_dest[i].destination_id = ctx->caps->eswitch_manager_vport_number;
			fte_attr.action_flags |= MLX5_IFC_FLOW_CONTEXT_ACTION_FWD_DEST;
			break;
		case ACTION_TYP_DEST_VPORT:
			fte_dest[i].destination_type = MLX5_FLOW_DESTINATION_TYPE_VPORT;
			fte_dest[i].destination_id = dests[i].dest_action->vport.vport_num;

			if (ctx->caps->merged_eswitch) {
				fte_dest[i].ext_flags |= CMD_EXT_DEST_ESW_OWNER_VHCA_ID;
				fte_dest[i].esw_owner_vhca_id = dests[i].dest_action->vport.esw_owner_vhca_id;
			}

			if (fte_dest[i].destination_id == WIRE_PORT && table_is_fdb_any(tbl_type)) {
				if (ctx->caps->fdb_uplink_hairpin) {
					fte_attr.uplink_hairpin_en = 1;
				} else {
					HWS_ERR("No support for wire to wire (hairpin)");
					errno = ENOTSUP;
					goto free_dest_list;
				}
			}
			fte_attr.action_flags |= MLX5_IFC_FLOW_CONTEXT_ACTION_FWD_DEST;
			break;
		default:
			HWS_ERR("Unsupported dest action type [%d] idx [%d]",
				dests[i].dest_action->type ,i);
			errno = ENOTSUP;
			goto free_dest_list;
		}

		switch (dests[i].reformat.type) {
		case MLX5DV_HWS_ACTION_DEST_REFORMAT_TYPE_NONE:
			continue;
		case MLX5DV_HWS_ACTION_DEST_REFORMAT_TYPE_L2_TO_TNL_L2:
			reformat_attr.type = MLX5_IFC_PACKET_REFORMAT_CONTEXT_REFORMAT_TYPE_L2_TO_L2_TUNNEL;
			break;
		case MLX5DV_HWS_ACTION_DEST_REFORMAT_TYPE_L2_TO_TNL_L3:
			reformat_attr.type = MLX5_IFC_PACKET_REFORMAT_CONTEXT_REFORMAT_TYPE_L2_TO_L3_TUNNEL;
			break;
		default:
			HWS_ERR("Invalid reformat action provided type [%d]", dests[i].reformat.type);
			errno = EINVAL;
			goto free_dest_list;
		}

		reformat_attr.reformat_param_0 = 0;
		reformat_attr.data_sz = dests[i].reformat.hdr_data.sz;
		reformat_attr.data = dests[i].reformat.hdr_data.hdr_data;

		reformat_obj = cmd_packet_reformat_create(ctx->ibv_ctx, &reformat_attr);
		if (!reformat_obj) {
			HWS_ERR("Failed to allocate reformat for dest idx [%d]", i);
			goto free_dest_list;
		}

		fte_dest[i].ext_flags |= CMD_EXT_DEST_REFORMAT;
		fte_dest[i].ext_reformat = reformat_obj;
		fte_dest[i].ext_reformat_hdr_sz = reformat_attr.data_sz;
		ft_attr.reformat_en = true;
		fte_attr.extended_dest = 1;
	}

	fw_island = cmd_forward_tbl_create(ctx->ibv_ctx, &ft_attr, &fte_attr);
	if (!fw_island) {
		HWS_ERR("Failed to create multi dest FW forward tbl");
		goto free_dest_list;
	}

	/* Create STC for non root tables */
	if (!(action_flags & MLX5DV_HWS_ACTION_FLAG_ROOT)) {
		ret = action_dest_create_stc(action, fw_island->ft);
		if (ret) {
			HWS_ERR("Failed creating stc for destination array");
			goto free_fw_island;
		}
	}

	action->dest_array.fw_island = fw_island;
	action->dest_array.num_dest = num_dest;
	action->dest_array.dests = fte_dest;

	return action;

free_fw_island:
	cmd_forward_tbl_destroy(fw_island);
free_dest_list:
	for (i = 0; i < num_dest; i++) {
		if (fte_dest[i].ext_reformat)
			cmd_destroy_obj(fte_dest[i].ext_reformat);
	}
	free(fte_dest);
free_action:
	free(action);
	return NULL;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_tag(struct mlx5dv_hws_context *ctx, uint64_t action_flags)
{
	if (action_flags &
	    (MLX5DV_HWS_ACTION_FLAG_FDB |
	     MLX5DV_HWS_ACTION_FLAG_FDB_TX |
	     MLX5DV_HWS_ACTION_FLAG_NIC_TX |
	     MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_RX |
	     MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_TX)) {
		HWS_ERR("TAG action supported only on NIC_RX/FDB_RX");
		errno = ENOTSUP;
		return NULL;
	}

	return action_create_generic(ctx, ACTION_TYP_TAG, action_flags);
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_pop_vlan(struct mlx5dv_hws_context *ctx, uint64_t action_flags)
{
	if (action_flags_is_root(action_flags)) {
		HWS_ERR("Pop vlan action not supported for root");
		errno = ENOTSUP;
		return NULL;
	}

	return action_create_generic(ctx, ACTION_TYP_POP_VLAN, action_flags);
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_push_vlan(struct mlx5dv_hws_context *ctx,
				   uint64_t action_flags)
{
	if (action_flags_is_root(action_flags)) {
		HWS_ERR("Push vlan action not supported for root");
		errno = ENOTSUP;
		return NULL;
	}

	return action_create_generic(ctx, ACTION_TYP_PUSH_VLAN, action_flags);
}

static int
action_conv_reformat_to_verbs(enum action_type action_type,
			      uint32_t *verb_reformat_type)
{
	switch (action_type) {
	case ACTION_TYP_REFORMAT_TNL_L2_TO_L2:
		*verb_reformat_type =
			MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TUNNEL_TO_L2;
		return 0;
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
		*verb_reformat_type =
			MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L2_TUNNEL;
		return 0;
	case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
		*verb_reformat_type =
			MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L3_TUNNEL_TO_L2;
		return 0;
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
		*verb_reformat_type =
			MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L3_TUNNEL;
		return 0;
	default:
		HWS_ERR("Invalid root reformat action type");
		errno = EINVAL;
		return -errno;
	}
}

static int
action_create_reformat_root(struct mlx5dv_hws_action *action,
			    size_t data_sz,
			    void *data)
{
	enum mlx5dv_flow_table_type dv_ft_type;
	enum mlx5dv_hws_table_type tbl_type;
	uint32_t verb_reformat_type = 0;
	struct ibv_context *ibv_ctx;
	int ret;

	ret = action_flags_get_tbl_type(action->flags, &tbl_type);
	if (ret)
		return ret;

	/* Convert action to FT type and verbs reformat type */
	ret = table_conv_hws_tbl_type_to_dv_ft(tbl_type, &dv_ft_type);
	if (ret)
		return ret;

	ret = action_conv_reformat_to_verbs(action->type, &verb_reformat_type);
	if (ret)
		return -errno;

	/* Create the reformat type for root table */
	ibv_ctx = context_get_local_ibv(action->ctx);

	action->flow_action =
		mlx5dv_create_flow_action_packet_reformat(ibv_ctx,
							  data_sz,
							  data,
							  verb_reformat_type,
							  dv_ft_type);
	if (!action->flow_action) {
		HWS_ERR("Failed to create dv_create_flow reformat");
		return -errno;
	}

	return 0;
}

static int
action_create_modify_header_root(struct mlx5dv_hws_action *action,
				 enum mlx5dv_hws_table_type tbl_type,
				 size_t data_sz,
				 void *data)
{
	enum mlx5dv_flow_table_type ft_type;
	struct ibv_context *ibv_ctx;
	int ret;

	/* Convert action to FT type and verbs reformat type */
	ret = table_conv_hws_tbl_type_to_dv_ft(tbl_type, &ft_type);
	if (ret)
		return ret;

	/* Create the reformat type for root table */
	ibv_ctx = context_get_local_ibv(action->ctx);

	action->flow_action =
		mlx5dv_create_flow_action_modify_header(ibv_ctx,
							data_sz,
							data,
							ft_type);
	if (!action->flow_action) {
		HWS_ERR("Failed to create dv_create_flow modify-header");
		return -errno;
	}

	return 0;
}

static bool action_is_resource_valid(enum mlx5dv_hws_resource_type rsc_type,
				     enum action_type action_type)
{
	switch (action_type) {
	case ACTION_TYP_ASO_METER:
		return (rsc_type == MLX5DV_HWS_RESOURCE_TYPE_ASO_METER);
	case ACTION_TYP_ASO_CT:
		return (rsc_type == MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK);
	case ACTION_TYP_ASO_FIRST_HIT:
		return (rsc_type == MLX5DV_HWS_RESOURCE_TYPE_ASO_FLOW_HIT);
	case ACTION_TYP_CTR:
		return (rsc_type == MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER);
	case ACTION_TYP_ASO_IPSEC:
		return (rsc_type == MLX5DV_HWS_RESOURCE_TYPE_IPSEC_OFFLOAD);
	case ACTION_TYP_MODIFY_HDR:
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
	case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
	case ACTION_TYP_INSERT_HEADER:
		return (resource_type_is_arg(rsc_type));
	default:
		return false;
	}
}

static bool action_check_resource(struct mlx5dv_hws_resource *resource[],
				  uint8_t num_of_resources,
				  enum action_type action_type,
				  uint64_t action_flags)
{
	/* Root resource allowed for counter only */
	bool allow_root_resource = action_type == ACTION_TYP_CTR;
	int i;

	if (action_flags_is_root(action_flags) && !allow_root_resource) {
		if (resource) {
			HWS_ERR("Root action [%d] doesn't require resource", action_type);
			errno = ENOTSUP;
			return false;
		}
		return true;
	}

	if (!num_of_resources || num_of_resources > ACTION_STC_MAX_RESOURCES_FOR_ACTION) {
		HWS_ERR("Action [%d] can't support num of resources [%d]",
			action_type, num_of_resources);
		errno = ENOTSUP;
		return false;
	}

	for (i = 0; i < num_of_resources; i++) {
		if (!action_is_resource_valid(resource[i]->type, action_type)) {
			HWS_ERR("Invalid resource type [%d] for action [%d]",
				resource[i]->type, action_type);
			errno = ENOTSUP;
			return false;
		}
	}

	return true;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_reformat_tnl_l2_to_l2(struct mlx5dv_hws_context *ctx,
					       uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;
	int ret;

	action = action_create_generic(ctx,
				       ACTION_TYP_REFORMAT_TNL_L2_TO_L2,
				       action_flags);
	if (!action)
		return NULL;

	if (action_flags_is_root(action_flags)) {
		ret = action_create_reformat_root(action, 0, NULL);
		if (ret)
			goto out_err;
	}

	return action;

out_err:
	free(action);
	return NULL;
}

static struct mlx5dv_hws_resource **
action_dup_resources(struct mlx5dv_hws_resource *resource[],
		     uint8_t num_of_resources)
{
	struct mlx5dv_hws_resource **dup;
	int i;

	dup = calloc(sizeof(struct mlx5dv_hws_resource *), num_of_resources);
	if (!dup) {
		errno = ENOMEM;
		return NULL;
	}

	for (i = 0; i < num_of_resources; i++)
		dup[i] = resource[i];

	return dup;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_reformat_l2_to_tnl_l2(struct mlx5dv_hws_context *ctx,
					       struct mlx5dv_hws_resource *resource[],
					       uint8_t num_of_resources,
					       struct mlx5dv_hws_action_hdr_data *hdr_data,
					       uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;
	int ret;

	if (!hdr_data->sz) {
		HWS_ERR("Invalid header size [%ld]", hdr_data->sz);
		errno = ENOTSUP;
		return NULL;
	}

	if (!action_check_resource(resource,
				   num_of_resources,
				   ACTION_TYP_REFORMAT_L2_TO_TNL_L2,
				   action_flags)) {
		HWS_ERR("Action resource checks failed");
		return NULL;
	}

	action = action_create_generic(ctx,
				       ACTION_TYP_REFORMAT_L2_TO_TNL_L2,
				       action_flags);
	if (!action)
		return NULL;

	action->reformat.header_size = hdr_data->sz;
	action->reformat.num_of_resources = num_of_resources;

	if (action_flags_is_root(action_flags)) {
		ret = action_create_reformat_root(action,
						  action->reformat.header_size,
						  hdr_data->hdr_data);
		if (ret)
			goto out_err;
	} else {
		action->reformat.resource = action_dup_resources(resource, num_of_resources);
		if (!action->reformat.resource)
			goto out_err;
	}

	return action;

out_err:
	free(action);
	return NULL;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_reformat_l2_to_tnl_l3(struct mlx5dv_hws_context *ctx,
					       struct mlx5dv_hws_resource *resource[],
					       uint8_t num_of_resources,
					       struct mlx5dv_hws_action_hdr_data *hdr_data,
					       uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;
	int ret;

	if (!hdr_data->sz) {
		HWS_ERR("Invalid reformat header data size");
		errno = ENOTSUP;
		return NULL;
	}

	if (!action_check_resource(resource,
				   num_of_resources,
				   ACTION_TYP_REFORMAT_L2_TO_TNL_L3,
				   action_flags)) {
		HWS_ERR("Action resource checks failed");
		return NULL;
	}

	action = action_create_generic(ctx,
				       ACTION_TYP_REFORMAT_L2_TO_TNL_L3,
				       action_flags);
	if (!action)
		return NULL;

	action->reformat.header_size = hdr_data->sz;
	action->reformat.num_of_resources = num_of_resources;

	if (action_flags_is_root(action_flags)) {
		ret = action_create_reformat_root(action,
						  action->reformat.header_size,
						  hdr_data->hdr_data);
		if (ret)
			goto out_err;
	} else {
		action->reformat.resource = action_dup_resources(resource, num_of_resources);
		if (!action->reformat.resource)
			goto out_err;
	}

	return action;

out_err:
	free(action);
	return NULL;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_reformat_trailer(struct mlx5dv_hws_context *ctx,
					  struct mlx5dv_hws_action_trailer_attr *attr,
					  uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;

	if (action_flags_is_root(action_flags)) {
		HWS_ERR("Reformat trailer action not supported for root");
		errno = ENOTSUP;
		return NULL;
	}

	if (attr->type != MLX5DV_HWS_ACTION_TRAILER_TYPE_IPSEC &&
	    attr->type != MLX5DV_HWS_ACTION_TRAILER_TYPE_PSP) {
		HWS_ERR("Only reformat trailer of type IPsec or PSP is supported");
		errno = ENOTSUP;
		return NULL;
	}

	if (attr->op == MLX5DV_HWS_ACTION_TRAILER_OP_INSERT) {
		if (action_flags & (MLX5DV_HWS_ACTION_FLAG_NIC_RX |
				    MLX5DV_HWS_ACTION_FLAG_FDB_RX |
				    MLX5DV_HWS_ACTION_FLAG_FDB |
				    MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_RX)) {
			HWS_ERR("Trailer insertion is only supported in TX");
			errno = EINVAL;
			return NULL;
		}
	} else if (attr->op == MLX5DV_HWS_ACTION_TRAILER_OP_REMOVE) {
		if (action_flags & (MLX5DV_HWS_ACTION_FLAG_NIC_TX |
				    MLX5DV_HWS_ACTION_FLAG_FDB_TX |
				    MLX5DV_HWS_ACTION_FLAG_FDB |
				    MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_TX)) {
			HWS_ERR("Trailer removal is only supported in RX");
			errno = EINVAL;
			return NULL;
		}
	} else {
		errno = ENOTSUP;
		return NULL;
	}

	if (attr->size > (ctx->caps->stc_max_trailer * DW_SIZE)) {
		HWS_ERR("Wrong trailer size, size should not exceed %u",
			ctx->caps->stc_max_trailer * DW_SIZE);
		errno = EINVAL;
		return NULL;
	}

	if (attr->size % DW_SIZE) {
		HWS_ERR("Trailer size supported in DW granularity (%u)", DW_SIZE);
		errno = ENOTSUP;
		return NULL;
	}

	action = action_create_generic(ctx, ACTION_TYP_REFORMAT_TRAILER, action_flags);
	if (!action)
		return NULL;

	action->reformat_trailer.type = attr->type;
	action->reformat_trailer.op = attr->op;
	action->reformat_trailer.size = attr->size / DW_SIZE;

	return action;
}

static void
action_prepare_decap_l3_actions(size_t data_sz, uint8_t *mh_data,
				int *num_of_actions)
{
	int actions;
	uint32_t i;

	/* Remove L2L3 outer headers */
	DEVX_SET(stc_ste_param_remove, mh_data, action_type,
		 MLX5_MODIFICATION_TYPE_REMOVE);
	DEVX_SET(stc_ste_param_remove, mh_data, decap, 0x1);
	DEVX_SET(stc_ste_param_remove, mh_data, remove_start_anchor,
		 MLX5_HEADER_ANCHOR_PACKET_START);
	DEVX_SET(stc_ste_param_remove, mh_data, remove_end_anchor,
		 MLX5_HEADER_ANCHOR_INNER_IPV6_IPV4);
	mh_data += ACTION_MH_DOUBLE_SIZE; /* Assume every action is 2 dw */
	actions = 1;

	/* Add the new header using inline action 4Byte at a time, the header
	 * is added in reversed order to the beginning of the packet to avoid
	 * incorrect parsing by the HW. Since header is 14B or 18B an extra
	 * two bytes are padded and later removed.
	 */
	for (i = 0; i < data_sz / ACTION_MH_INLINE_DATA_SIZE + 1; i++) {
		DEVX_SET(stc_ste_param_insert, mh_data, action_type,
			 MLX5_MODIFICATION_TYPE_INSERT);
		DEVX_SET(stc_ste_param_insert, mh_data, inline_data, 0x1);
		DEVX_SET(stc_ste_param_insert, mh_data, insert_anchor,
			 MLX5_HEADER_ANCHOR_PACKET_START);
		DEVX_SET(stc_ste_param_insert, mh_data, insert_size, 2);
		mh_data += ACTION_MH_DOUBLE_SIZE;
		actions++;
	}

	/* Remove first 2 extra bytes */
	DEVX_SET(stc_ste_param_remove_words, mh_data, action_type,
		 MLX5_MODIFICATION_TYPE_REMOVE_WORDS);
	DEVX_SET(stc_ste_param_remove_words, mh_data, remove_start_anchor,
		 MLX5_HEADER_ANCHOR_PACKET_START);
	/* The hardware expects here size in words (2 bytes) */
	DEVX_SET(stc_ste_param_remove_words, mh_data, remove_size, 1);
	actions++;

	*num_of_actions = actions;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_reformat_tnl_l3_to_l2(struct mlx5dv_hws_context *ctx,
					       struct mlx5dv_hws_resource *resource[],
					       uint8_t num_of_resources,
					       struct mlx5dv_hws_action_hdr_data *hdr_data,
					       uint64_t action_flags)
{
	uint8_t mh_data[ACTION_MH_REFORMAT_DATA_SIZE] = {0};
	struct mlx5dv_hws_action *action;
	struct mlx5dv_devx_obj *pat_obj;
	int num_of_actions, mh_data_sz;
	int ret;

	if (hdr_data->sz != ACTION_HDR_LEN_L2 &&
	    hdr_data->sz != ACTION_HDR_LEN_L2_W_VLAN) {
		HWS_ERR("Header size is not supported for decap-l3");
		errno = EINVAL;
		return NULL;
	}

	if (!action_check_resource(resource,
				   num_of_resources,
				   ACTION_TYP_REFORMAT_TNL_L3_TO_L2,
				   action_flags)) {
		HWS_ERR("Action resource checks failed");
		return NULL;
	}

	action = action_create_generic(ctx, ACTION_TYP_REFORMAT_TNL_L3_TO_L2, action_flags);
	if (!action)
		return NULL;

	if (action_flags_is_root(action_flags)) {
		ret = action_create_reformat_root(action,
						  hdr_data->sz,
						  hdr_data->hdr_data);
		if (ret)
			goto out_err;
	} else {
		action_prepare_decap_l3_actions(hdr_data->sz,
						mh_data,
						&num_of_actions);

		mh_data_sz = num_of_actions * ACTION_MH_DOUBLE_SIZE;

		/* Create new pattern */
		pat_obj = pat_get_pattern(ctx, (__be64 *)mh_data, mh_data_sz);
		if (!pat_obj) {
			HWS_ERR("Failed to create decapL3 pattern");
			goto out_err;
		}

		action->modify_header.resource = action_dup_resources(resource, num_of_resources);
		if (!action->modify_header.resource)
			goto out_err;

		action->modify_header.num_of_resources = num_of_resources;
		action->modify_header.num_of_actions = num_of_actions;
		action->modify_header.pattern_obj = pat_obj;
		action->modify_header.reparse =
			pat_require_reparse((__be64 *)mh_data, num_of_actions);
	}

	return action;

out_err:
	free(action);
	return NULL;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_modify_header(struct mlx5dv_hws_context *ctx,
				       struct mlx5dv_hws_resource *resource[],
				       uint8_t num_of_resources,
				       struct mlx5dv_hws_action_modify_header *pattern,
				       bool force_reparse,
				       uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;
	int ret;

	if (!pattern->sz || pattern->sz % ACTION_MH_DOUBLE_SIZE) {
		HWS_ERR("Invalid pattern size [%ld]", pattern->sz);
		errno = EINVAL;
		return NULL;
	}

	if (!action_check_resource(resource,
				   num_of_resources,
				   ACTION_TYP_MODIFY_HDR,
				   action_flags)) {
		HWS_ERR("Action resource checks failed");
		return NULL;
	}

	action = action_create_generic(ctx, ACTION_TYP_MODIFY_HDR, action_flags);
	if (!action)
		return NULL;

	action->reformat.num_of_resources = num_of_resources;
	action->modify_header.num_of_actions = pattern->sz / ACTION_MH_DOUBLE_SIZE;

	if (action_flags_is_root(action_flags)) {
		enum mlx5dv_hws_table_type tbl_type;

		if (force_reparse) {
			HWS_ERR("Force reparse not support over root");
			errno = ENOTSUP;
			goto free_action;
		}

		ret = action_flags_get_tbl_type(action->flags, &tbl_type);
		if (ret)
			goto free_action;

		ret = action_create_modify_header_root(action,
						       tbl_type,
						       pattern->sz,
						       pattern->modify_actions);
		if (ret)
			goto free_action;
	} else {
		/* Create new pattern */
		action->modify_header.pattern_obj = pat_get_pattern(ctx,
								    pattern->modify_actions,
								    pattern->sz);
		if (!action->modify_header.pattern_obj) {
			HWS_ERR("Failed to create pattern, size: [%lu]", pattern->sz);
			goto free_action;
		}

		action->modify_header.resource = action_dup_resources(resource, num_of_resources);
		if (!action->modify_header.resource) {
			HWS_ERR("Failed to duplicate resources");
			goto pattern_put;
		}

		action->modify_header.reparse = force_reparse ||
			pat_require_reparse(pattern->modify_actions,
					    pattern->sz / ACTION_MH_DOUBLE_SIZE);
	}

	return action;

pattern_put:
	pat_put_pattern(ctx, action->modify_header.pattern_obj);
free_action:
	free(action);
	return NULL;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_counter(struct mlx5dv_hws_context *ctx,
				 struct mlx5dv_hws_resource *resource[],
				 uint8_t num_of_resources,
				 uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;

	if (!action_check_resource(resource,
				   num_of_resources,
				   ACTION_TYP_CTR,
				   action_flags)) {
		HWS_ERR("Action resource checks failed");
		return NULL;
	}

	action = action_create_generic(ctx, ACTION_TYP_CTR, action_flags);
	if (!action)
		return NULL;

	action->counter.resource = action_dup_resources(resource, num_of_resources);
	if (!action->counter.resource)
		goto out_err;

	action->counter.num_of_resources = num_of_resources;

	return action;
out_err:
	free(action);
	return NULL;
}

static struct mlx5dv_hws_action *
action_create_aso_action(struct mlx5dv_hws_context *ctx,
			 struct mlx5dv_hws_resource *resource[],
			 uint8_t num_of_resources,
			 enum action_type action_type,
			 uint8_t return_reg_id,
			 uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;

	if (action_flags_is_root(action_flags)) {
		HWS_ERR("ASO action not supported for root");
		errno = ENOTSUP;
		return NULL;
	}

	if (!action_check_resource(resource,
				   num_of_resources,
				   action_type,
				   action_flags)) {
		HWS_ERR("Action resource checks failed");
		return NULL;
	}

	action = action_create_generic(ctx, action_type, action_flags);
	if (!action)
		return NULL;

	action->aso.resource = action_dup_resources(resource, num_of_resources);
	if (!action->aso.resource)
		goto out_err;

	action->aso.num_of_resources = num_of_resources;
	action->aso.return_reg_id = return_reg_id;
	return action;

out_err:
	free(action);
	return NULL;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_aso_meter(struct mlx5dv_hws_context *ctx,
				   struct mlx5dv_hws_resource *resource[],
				   uint8_t num_of_resources,
				   uint8_t return_reg_id,
				   uint64_t action_flags)
{
	return action_create_aso_action(ctx, resource,
					num_of_resources,
					ACTION_TYP_ASO_METER,
					return_reg_id,
					action_flags);
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_aso_ct(struct mlx5dv_hws_context *ctx,
				struct mlx5dv_hws_resource *resource[],
				uint8_t num_of_resources,
				uint8_t return_reg_id,
				uint64_t action_flags)
{
	return action_create_aso_action(ctx, resource, num_of_resources,
					ACTION_TYP_ASO_CT,
					return_reg_id, action_flags);
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_aso_first_hit(struct mlx5dv_hws_context *ctx,
				       struct mlx5dv_hws_resource *resource[],
				       uint8_t num_of_resources,
				       uint8_t return_reg_id,
				       uint64_t action_flags)
{
	return action_create_aso_action(ctx, resource, num_of_resources,
					ACTION_TYP_ASO_FIRST_HIT,
					return_reg_id, action_flags);
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_aso_ipsec(struct mlx5dv_hws_context *ctx,
				   struct mlx5dv_hws_resource *resource[],
				   uint8_t num_of_resources,
				   uint8_t return_reg_id,
				   uint64_t action_flags)
{
	/* Parameter return_reg_id must be zero for ipsec aso (taken from ipsec offload object)
	 * ASO operation is supported only for ipsec with full offload enabled
	 * TODO: Query and check that full offload enabled
	 */
	if (return_reg_id) {
		HWS_ERR("Unsupported non-zero return_reg_id value");
		errno = EINVAL;
		return NULL;
	}

	return action_create_aso_action(ctx, resource, num_of_resources,
					ACTION_TYP_ASO_IPSEC,
					return_reg_id, action_flags);
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_barrier(struct mlx5dv_hws_context *ctx,
				 uint64_t action_flags)
{
	if (action_flags_is_root(action_flags)) {
		HWS_ERR("Barrier action is not supported for root");
		errno = ENOTSUP;
		return NULL;
	}

	return action_create_generic(ctx, ACTION_TYP_BARRIER, action_flags);
}

static void
action_nat64_write_arg_data(struct mlx5dv_hws_context *ctx,
			    struct mlx5dv_hws_resource *resource,
			    enum action_nat64_stages nat_action_stage,
			    uint8_t *arg_data,
			    size_t data_size)
{
	uint32_t arg_gran, arg_idx;

	arg_gran = resource_arg_gran_get(resource->type);
	arg_idx = nat_action_stage * arg_gran;

	/* Use control queue for writing */
	arg_write(&ctx->send_queue[ctx->queues - 1], NULL,
		  resource->devx_obj->object_id + arg_idx,
		  arg_data,
		  data_size);

	/* Flush the queue */
	mlx5dv_hws_queue_execute_op(ctx,
				    ctx->queues - 1,
				    MLX5DV_HWS_QUEUE_OP_DRAIN_SYNC);
}

static void
action_nat64_zero_ip_addr(uint8_t **action_ptr, bool is_v4_to_v6)
{
	if (is_v4_to_v6) {
		DEVX_SET(set_action_in, *action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
		DEVX_SET(set_action_in, *action_ptr, field, MLX5_MODI_OUT_SIPV4);
		DEVX_SET(set_action_in, *action_ptr, data, 0);
		*action_ptr += ACTION_MH_DOUBLE_SIZE;

		DEVX_SET(set_action_in, *action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
		DEVX_SET(set_action_in, *action_ptr, field, MLX5_MODI_OUT_DIPV4);
		DEVX_SET(set_action_in, *action_ptr, data, 0);
		*action_ptr += ACTION_MH_DOUBLE_SIZE;
	} else {
		DEVX_SET(set_action_in, *action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
		DEVX_SET(set_action_in, *action_ptr, field, MLX5_MODI_OUT_SIPV6_127_96);
		DEVX_SET(set_action_in, *action_ptr, data, 0);
		*action_ptr += ACTION_MH_DOUBLE_SIZE;

		DEVX_SET(set_action_in, *action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
		DEVX_SET(set_action_in, *action_ptr, field, MLX5_MODI_OUT_SIPV6_95_64);
		DEVX_SET(set_action_in, *action_ptr, data, 0);
		*action_ptr += ACTION_MH_DOUBLE_SIZE;

		DEVX_SET(set_action_in, *action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
		DEVX_SET(set_action_in, *action_ptr, field, MLX5_MODI_OUT_SIPV6_63_32);
		DEVX_SET(set_action_in, *action_ptr, data, 0);
		*action_ptr += ACTION_MH_DOUBLE_SIZE;

		DEVX_SET(set_action_in, *action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
		DEVX_SET(set_action_in, *action_ptr, field, MLX5_MODI_OUT_SIPV6_31_0);
		DEVX_SET(set_action_in, *action_ptr, data, 0);
		*action_ptr += ACTION_MH_DOUBLE_SIZE;

		DEVX_SET(set_action_in, *action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
		DEVX_SET(set_action_in, *action_ptr, field, MLX5_MODI_OUT_DIPV6_127_96);
		DEVX_SET(set_action_in, *action_ptr, data, 0);
		*action_ptr += ACTION_MH_DOUBLE_SIZE;

		DEVX_SET(set_action_in, *action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
		DEVX_SET(set_action_in, *action_ptr, field, MLX5_MODI_OUT_DIPV6_95_64);
		DEVX_SET(set_action_in, *action_ptr, data, 0);
		*action_ptr += ACTION_MH_DOUBLE_SIZE;

		DEVX_SET(set_action_in, *action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
		DEVX_SET(set_action_in, *action_ptr, field, MLX5_MODI_OUT_DIPV6_63_32);
		DEVX_SET(set_action_in, *action_ptr, data, 0);
		*action_ptr += ACTION_MH_DOUBLE_SIZE;

		DEVX_SET(set_action_in, *action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
		DEVX_SET(set_action_in, *action_ptr, field, MLX5_MODI_OUT_DIPV6_31_0);
		DEVX_SET(set_action_in, *action_ptr, data, 0);
		*action_ptr += ACTION_MH_DOUBLE_SIZE;
	}
}

/* The function takes all the fields that need to be converted and keep them
 * in dedicated register.
 * If the user want the original source addresses to be backup 2 more registers
 * are need to be supplied.
 * As well, since HW checksum is incremental, the addresses need to be set to zero.
 */
static struct mlx5dv_devx_obj *
action_nat64_create_copy_state_ptrn(struct mlx5dv_hws_context *ctx,
				    struct mlx5dv_hws_action_nat64_attr *attr,
				    struct mlx5dv_hws_resource *resource,
				    bool *need_reparse)
{
	__be64 modify_action_data[ACTION_NAT64_MAX_MODIFY_ACTIONS] = {0};
	uint32_t tos_field, src_addr_field, dst_addr_field, ecn_field;
	struct mlx5dv_devx_obj *ptrn;
	uint32_t packet_len_field;
	uint8_t *action_ptr;
	uint32_t tos_size;
	bool is_v4_to_v6;
	size_t size;

	is_v4_to_v6 = attr->flags & MLX5DV_HWS_ACTION_NAT64_V4_TO_V6;

	if (is_v4_to_v6) {
		packet_len_field = MLX5_MODI_OUT_IPV4_TOTAL_LEN;
		tos_field = MLX5_MODI_OUT_IP_DSCP;
		tos_size = 6;
		ecn_field = MLX5_MODI_OUT_IP_ECN;
		src_addr_field = MLX5_MODI_OUT_SIPV4;
		dst_addr_field = MLX5_MODI_OUT_DIPV4;
	} else {
		packet_len_field = MLX5_MODI_OUT_IPV6_PAYLOAD_LEN;
		tos_field = MLX5_MODI_OUT_IPV6_TRAFFIC_CLASS;
		tos_size = 8;
		ecn_field = 0;
		src_addr_field = MLX5_MODI_OUT_SIPV6_31_0;
		dst_addr_field = MLX5_MODI_OUT_DIPV6_31_0;
	}

	action_ptr = (uint8_t *) modify_action_data;

	if (attr->flags & MLX5DV_HWS_ACTION_NAT64_BACKUP_ADDR) {
		DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_COPY);
		DEVX_SET(copy_action_in, action_ptr, src_field, src_addr_field);
		DEVX_SET(copy_action_in, action_ptr, dst_field,
			 attr->registers[ACTION_NAT64_REG_SRC_IP]);
		action_ptr += ACTION_MH_DOUBLE_SIZE;

		DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_COPY);
		DEVX_SET(copy_action_in, action_ptr, src_field, dst_addr_field);
		DEVX_SET(copy_action_in, action_ptr, dst_field,
			 attr->registers[ACTION_NAT64_REG_DST_IP]);
		action_ptr += ACTION_MH_DOUBLE_SIZE;
	}

	/* | 8 bit - 8 bit     - 16 bit     |
	 * | TOS   - protocol  - packet-len |
	 */
	DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_COPY);
	DEVX_SET(copy_action_in, action_ptr, src_field, packet_len_field);
	DEVX_SET(copy_action_in, action_ptr, dst_field,
		 attr->registers[ACTION_NAT64_REG_CONTROL]);
	DEVX_SET(copy_action_in, action_ptr, dst_offset, 0);/* 16 bits in the lsb */
	DEVX_SET(copy_action_in, action_ptr, length, 16);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_NOP);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_COPY);
	DEVX_SET(copy_action_in, action_ptr, src_field, MLX5_MODI_OUT_IP_PROTOCOL);
	DEVX_SET(copy_action_in, action_ptr, dst_field,
		 attr->registers[ACTION_NAT64_REG_CONTROL]);
	DEVX_SET(copy_action_in, action_ptr, dst_offset, 16);
	DEVX_SET(copy_action_in, action_ptr, length, 8);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_NOP);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_COPY);
	DEVX_SET(copy_action_in, action_ptr, src_field, tos_field);
	DEVX_SET(copy_action_in, action_ptr, dst_field,
		 attr->registers[ACTION_NAT64_REG_CONTROL]);
	DEVX_SET(copy_action_in, action_ptr, dst_offset, 24);
	DEVX_SET(copy_action_in, action_ptr, length, tos_size);
	action_ptr += ACTION_MH_DOUBLE_SIZE;
	/* In ipv4 TOS = {dscp (6bits) - ecn_field (2bits) }*/
	if (ecn_field) {
		DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_NOP);
		action_ptr += ACTION_MH_DOUBLE_SIZE;

		DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_COPY);
		DEVX_SET(copy_action_in, action_ptr, src_field, ecn_field);
		DEVX_SET(copy_action_in, action_ptr, dst_field,
			attr->registers[ACTION_NAT64_REG_CONTROL]);
		DEVX_SET(copy_action_in, action_ptr, dst_offset, 24 + tos_size);
		DEVX_SET(copy_action_in, action_ptr, length, ACTION_NAT64_ECN_SIZE);
		action_ptr += ACTION_MH_DOUBLE_SIZE;
	}

	/* Set sip and dip to 0, in order to have new csum */
	action_nat64_zero_ip_addr(&action_ptr, is_v4_to_v6);

	size = (action_ptr - (uint8_t *) modify_action_data);

	*need_reparse = pat_require_reparse(modify_action_data,
					    size / ACTION_MH_DOUBLE_SIZE);

	action_nat64_write_arg_data(ctx, resource, ACTION_NAT64_STAGE_COPY,
				    (uint8_t *) modify_action_data, size);

	ptrn = pat_get_pattern(ctx, modify_action_data, size);
	if (!ptrn)
		HWS_ERR("Failed to create copy for NAT64: action_sz: %zu, flags: 0x%x\n",
			size, attr->flags);

	return ptrn;
}

/* The function prepares the packet to its new type, by replacing to the new header.
 * The address was already set to zero, now need to push address of the new type,
 * with the relevant fields.
 */
static struct mlx5dv_devx_obj *
action_nat64_replace_state_ptrn(struct mlx5dv_hws_context *ctx,
				struct mlx5dv_hws_action_nat64_attr *attr,
				struct mlx5dv_hws_resource *resource,
				bool *need_reparse)
{
	__be64 modify_action_data[ACTION_NAT64_MAX_MODIFY_ACTIONS] = {0};
	uint32_t address_prefix[ACTION_NAT64_HEADER_MINUS_ONE] = {0};
	struct mlx5dv_devx_obj *ptrn;
	uint8_t header_size_in_dw;
	uint8_t *action_ptr;
	uint32_t eth_type;
	bool is_v4_to_v6;
	uint32_t ip_ver;
	size_t size;
	int i;

	is_v4_to_v6 = attr->flags & MLX5DV_HWS_ACTION_NAT64_V4_TO_V6;

	if (is_v4_to_v6) {
		uint32_t nat64_well_known_pref[] = {0x00010000,
						    0x9bff6400, 0x0, 0x0, 0x0,
						    0x9bff6400, 0x0, 0x0, 0x0};

		header_size_in_dw = ACTION_NAT64_IPV6_HEADER;
		ip_ver = IPV6_VERSION_BE;
		eth_type = IPV6_ETHERTYPE;
		memcpy(address_prefix, nat64_well_known_pref,
		       ACTION_NAT64_HEADER_MINUS_ONE * sizeof(uint32_t));
	} else {
		/* We make a buffer that contain the right HW csum and the DF bit set.
		 * 0x00400000 - contain the DF bit in the right place.
		 * 0xff7a0000 - is the csum for that buffer, calculated via external tool.
		 */
		uint32_t ipv4_pref[] = {0x00400000, 0xff7a0000, 0x0, 0x0, 0x0};

		header_size_in_dw = ACTION_NAT64_IPV4_HEADER_SIZE;
		ip_ver = IPV4_VER_IHL_BE;
		eth_type = IPV4_ETHERTYPE;
		memcpy(address_prefix, ipv4_pref,
		       ACTION_NAT64_IPV4_HEADER_SIZE * sizeof(uint32_t));
	}

	action_ptr = (uint8_t *) modify_action_data;

	DEVX_SET(set_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
	DEVX_SET(set_action_in, action_ptr, field, MLX5_MODI_OUT_ETHERTYPE);
	DEVX_SET(set_action_in, action_ptr, length, 16);
	DEVX_SET(set_action_in, action_ptr, data, eth_type);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	/* Push empty header with ipv6 as version */
	DEVX_SET(stc_ste_param_insert, action_ptr, action_type,
		 MLX5_MODIFICATION_TYPE_INSERT);
	DEVX_SET(stc_ste_param_insert, action_ptr, inline_data, 0x1);
	DEVX_SET(stc_ste_param_insert, action_ptr, insert_anchor,
		 MLX5_HEADER_ANCHOR_IPV6_IPV4);
	DEVX_SET(stc_ste_param_insert, action_ptr, insert_size, 2);
	DEVX_SET(stc_ste_param_insert, action_ptr, insert_argument, ip_ver);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	for (i = 0; i < header_size_in_dw - 1; i++) {
		DEVX_SET(stc_ste_param_insert, action_ptr, action_type,
			 MLX5_MODIFICATION_TYPE_INSERT);
		DEVX_SET(stc_ste_param_insert, action_ptr, inline_data, 0x1);
		DEVX_SET(stc_ste_param_insert, action_ptr, insert_anchor,
			 MLX5_HEADER_ANCHOR_IPV6_IPV4);
		DEVX_SET(stc_ste_param_insert, action_ptr, insert_size, 2);
		DEVX_SET(stc_ste_param_insert, action_ptr, insert_argument,
			(__force uint32_t) htobe32(address_prefix[i]));
		action_ptr += ACTION_MH_DOUBLE_SIZE;
	}

	/* Remove orig src/dst addr (8 bytes, 4 words) */
	DEVX_SET(stc_ste_param_remove, action_ptr, action_type,
		 MLX5_MODIFICATION_TYPE_REMOVE);
	DEVX_SET(stc_ste_param_remove, action_ptr, remove_start_anchor,
		 MLX5_HEADER_ANCHOR_IPV6_IPV4);
	DEVX_SET(stc_ste_param_remove, action_ptr, remove_end_anchor,
		 MLX5_HEADER_ANCHOR_TCP_UDP);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	size = (action_ptr - (uint8_t *) modify_action_data);

	*need_reparse = pat_require_reparse(modify_action_data,
					    size / ACTION_MH_DOUBLE_SIZE);

	action_nat64_write_arg_data(ctx, resource, ACTION_NAT64_STAGE_REPLACE,
				    (uint8_t *) modify_action_data, size);

	ptrn = pat_get_pattern(ctx, modify_action_data, size);
	if (!ptrn)
		HWS_ERR("Failed to create ptrn for replace/NAT64: sz: %zu", size);

	return ptrn;
}

/* The protocol of the packet should be copied in a different stage, due to HW
 * limitation, that wants to know the protocol before the rest of the fields.
 */
static struct mlx5dv_devx_obj *
action_nat64_copy_proto_state_ptrn(struct mlx5dv_hws_context *ctx,
				   struct mlx5dv_hws_action_nat64_attr *attr,
				   struct mlx5dv_hws_resource *resource,
				   bool *need_reparse)
{
	__be64 modify_action_data[ACTION_NAT64_MAX_MODIFY_ACTIONS] = {0};
	struct mlx5dv_devx_obj *ptrn;
	uint8_t *action_ptr;
	size_t size;

	action_ptr = (uint8_t *)modify_action_data;

	DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_COPY);
	DEVX_SET(copy_action_in, action_ptr, src_field,
		 attr->registers[ACTION_NAT64_REG_CONTROL]);
	DEVX_SET(copy_action_in, action_ptr, dst_field,
		 MLX5_MODI_OUT_IP_PROTOCOL);
	DEVX_SET(copy_action_in, action_ptr, src_offset, 16);
	DEVX_SET(copy_action_in, action_ptr, dst_offset, 0);
	DEVX_SET(copy_action_in, action_ptr, length, 8);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_NOP);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	size = (action_ptr - (uint8_t *) modify_action_data);

	/* We must reparse here, the HW should know the protocol */
	*need_reparse = true;

	action_nat64_write_arg_data(ctx, resource, ACTION_NAT64_STAGE_COPY_PROTOCOL,
				    (uint8_t *) modify_action_data, size);

	ptrn = pat_get_pattern(ctx, modify_action_data, size);
	if (!ptrn)
		HWS_ERR("Failed to create ptrn for proto/NAT64: sz: %zu, flags: 0x%x",
			size, attr->flags);
	return ptrn;
}

/* The function returns back all the fields that were kept in the registers, in
 * order to have them fit the new protocol they are having some processing here.
 */
static struct mlx5dv_devx_obj *
action_nat64_copy_back_ptrn(struct mlx5dv_hws_context *ctx,
			    struct mlx5dv_hws_action_nat64_attr *attr,
			    struct mlx5dv_hws_resource *resource,
			    bool *need_reparse)
{
	__be64 modify_action_data[ACTION_NAT64_MAX_MODIFY_ACTIONS] = {0};
	struct mlx5dv_devx_obj *ptrn;
	uint32_t packet_len_field;
	uint32_t packet_len_add;
	uint8_t *action_ptr;
	uint32_t tos_field;
	uint32_t ttl_field;
	uint32_t tos_size;
	uint32_t src_addr;
	uint32_t dst_addr;
	bool is_v4_to_v6;
	uint32_t ecn;
	size_t size;

	is_v4_to_v6 = attr->flags & MLX5DV_HWS_ACTION_NAT64_V4_TO_V6;

	if (is_v4_to_v6) {
		packet_len_field = MLX5_MODI_OUT_IPV6_PAYLOAD_LEN;
		 /* 2' comp to 20, to get -20 in add operation */
		packet_len_add = ACTION_NAT64_DEC_20;
		ttl_field = MLX5_MODI_OUT_IPV6_HOPLIMIT;
		src_addr = MLX5_MODI_OUT_SIPV6_31_0;
		dst_addr = MLX5_MODI_OUT_DIPV6_31_0;
		tos_field = MLX5_MODI_OUT_IPV6_TRAFFIC_CLASS;
		tos_size = 8;
		ecn = 0;
	} else {
		packet_len_field = MLX5_MODI_OUT_IPV4_TOTAL_LEN;
		/* IPv4 packet len is including 20 bytes of the header + data,
		 * IPv6 len is only data, so the diff is 20 bytes.
		 */
		packet_len_add = ACTION_NAT64_ADD_20;
		ttl_field = MLX5_MODI_OUT_IPV4_TTL;
		src_addr = MLX5_MODI_OUT_SIPV4;
		dst_addr = MLX5_MODI_OUT_DIPV4;
		tos_field = MLX5_MODI_OUT_IP_DSCP;
		tos_size = 6;
		ecn = MLX5_MODI_OUT_IP_ECN;
	}

	action_ptr = (uint8_t *) modify_action_data;

	DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_COPY);
	DEVX_SET(copy_action_in, action_ptr, src_field,
		 attr->registers[ACTION_NAT64_REG_CONTROL]);
	DEVX_SET(copy_action_in, action_ptr, dst_field,
		 packet_len_field);
	DEVX_SET(copy_action_in, action_ptr, src_offset, 32);
	DEVX_SET(copy_action_in, action_ptr, length, 16);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	DEVX_SET(set_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_SET);
	DEVX_SET(set_action_in, action_ptr, field, ttl_field);
	DEVX_SET(set_action_in, action_ptr, length, 8);
	DEVX_SET(set_action_in, action_ptr, data, ACTION_NAT64_TTL_DEFAULT_VAL);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	/* Copy TOS */
	DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_COPY);
	DEVX_SET(copy_action_in, action_ptr, src_field,
		 attr->registers[ACTION_NAT64_REG_CONTROL]);
	DEVX_SET(copy_action_in, action_ptr, dst_field, tos_field);
	DEVX_SET(copy_action_in, action_ptr, src_offset,
		 24 + (ecn ? ACTION_NAT64_ECN_SIZE : 0));
	DEVX_SET(copy_action_in, action_ptr, length, tos_size);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	if (ecn) {
		DEVX_SET(copy_action_in, action_ptr, action_type,
			 MLX5_MODIFICATION_TYPE_NOP);
		action_ptr += ACTION_MH_DOUBLE_SIZE;

		DEVX_SET(copy_action_in, action_ptr, action_type,
			 MLX5_MODIFICATION_TYPE_COPY);
		DEVX_SET(copy_action_in, action_ptr, src_field,
			 attr->registers[ACTION_NAT64_REG_CONTROL]);
		DEVX_SET(copy_action_in, action_ptr, dst_field, ecn);
		DEVX_SET(copy_action_in, action_ptr, src_offset, 24);
		DEVX_SET(copy_action_in, action_ptr, length, ACTION_NAT64_ECN_SIZE);
		action_ptr += ACTION_MH_DOUBLE_SIZE;
	}

	DEVX_SET(copy_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_NOP);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	/* If required Copy original addresses */
	if (attr->flags & MLX5DV_HWS_ACTION_NAT64_BACKUP_ADDR) {
		DEVX_SET(copy_action_in, action_ptr, action_type,
			 MLX5_MODIFICATION_TYPE_COPY);
		DEVX_SET(copy_action_in, action_ptr, src_field,
			 attr->registers[ACTION_NAT64_REG_SRC_IP]);
		DEVX_SET(copy_action_in, action_ptr, dst_field, src_addr);
		DEVX_SET(copy_action_in, action_ptr, src_offset, 0);
		DEVX_SET(copy_action_in, action_ptr, length, 32);
		action_ptr += ACTION_MH_DOUBLE_SIZE;

		DEVX_SET(copy_action_in, action_ptr, action_type,
			 MLX5_MODIFICATION_TYPE_COPY);
		DEVX_SET(copy_action_in, action_ptr, src_field,
			 attr->registers[ACTION_NAT64_REG_DST_IP]);
		DEVX_SET(copy_action_in, action_ptr, dst_field, dst_addr);
		DEVX_SET(copy_action_in, action_ptr, src_offset, 0);
		DEVX_SET(copy_action_in, action_ptr, length, 32);
		action_ptr += ACTION_MH_DOUBLE_SIZE;
	}

	/* Take/add off 20 bytes ipv4/6 from/to the total size */
	DEVX_SET(set_action_in, action_ptr, action_type, MLX5_MODIFICATION_TYPE_ADD);
	DEVX_SET(set_action_in, action_ptr, field, packet_len_field);
	DEVX_SET(set_action_in, action_ptr, data, packet_len_add);
	DEVX_SET(set_action_in, action_ptr, length, 16);
	action_ptr += ACTION_MH_DOUBLE_SIZE;

	size = (action_ptr - (uint8_t *) modify_action_data);

	*need_reparse = pat_require_reparse(modify_action_data,
					    size / ACTION_MH_DOUBLE_SIZE);

	action_nat64_write_arg_data(ctx, resource, ACTION_NAT64_STAGE_COPYBACK,
				    (uint8_t *) modify_action_data, size);

	ptrn = pat_get_pattern(ctx, modify_action_data, size);
	if (!ptrn)
		HWS_ERR("Failed to create ptrn for copy-back/NAT64: sz: %zu, flags: 0x%x",
			size, attr->flags);
	return ptrn;
}

static bool
action_nat64_validate_param(struct mlx5dv_hws_action_nat64_attr *attr,
			    uint64_t action_flags)
{
	if (action_flags_is_root(action_flags)) {
		HWS_ERR("Nat64 action not supported for root");
		errno = ENOTSUP;
		return false;
	}

	if (attr->num_of_registers > ACTION_NAT64_REG_MAX) {
		HWS_ERR("Nat64 action doesn't support more than %d registers",
		       ACTION_NAT64_REG_MAX);
		errno = EINVAL;
		return false;
	}

	if (attr->flags & MLX5DV_HWS_ACTION_NAT64_BACKUP_ADDR &&
	    attr->num_of_registers != ACTION_NAT64_REG_MAX) {
		HWS_ERR("Nat64 backup addr requires %d registers",
		       ACTION_NAT64_REG_MAX);
		errno = EINVAL;
		return false;
	}

	if (!(attr->flags & MLX5DV_HWS_ACTION_NAT64_V4_TO_V6 ||
	      attr->flags & MLX5DV_HWS_ACTION_NAT64_V6_TO_V4)) {
		HWS_ERR("Nat64 backup addr requires one mode at least");
		errno = EINVAL;
		return false;
	}

	return true;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_nat64(struct mlx5dv_hws_context *ctx,
			       struct mlx5dv_hws_action_nat64_attr *attr,
			       uint64_t action_flags)
{
	struct mlx5dv_hws_resource_attr res_attr = {0};
	struct mlx5dv_hws_resource *resource;
	struct mlx5dv_hws_action *action;
	bool need_reparse;

	if (!action_nat64_validate_param(attr, action_flags))
		return NULL;

	res_attr.bulk_log_size = 2; /* For 4 nat64 stages */
	res_attr.type = MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_128B;

	resource = mlx5dv_hws_resource_alloc(ctx, &res_attr);
	if (!resource) {
		HWS_ERR("Nat64 Failed to create resource");
		return NULL;
	}

	pthread_spin_lock(&ctx->ctrl_lock);

	action = action_create_generic(ctx, ACTION_TYP_NAT64, action_flags);
	if (!action)
		goto free_resource;

	action->nat64.resource = resource;

	action->nat64.pattern_obj = calloc(ACTION_NAT64_STAGES,
					   sizeof(struct mlx5dv_devx_obj *));
	if (!action->nat64.pattern_obj) {
		errno = ENOMEM;
		goto free_action;
	}

	action->nat64.pattern_obj[ACTION_NAT64_STAGE_COPY] =
		action_nat64_create_copy_state_ptrn(ctx, attr, resource, &need_reparse);
	if (!action->nat64.pattern_obj[ACTION_NAT64_STAGE_COPY]) {
		HWS_ERR("Nat64 failed creating copy state");
		goto free_ptrn_arr;
	}

	if (need_reparse)
		action->nat64.need_reparse |= BIT(ACTION_NAT64_STAGE_COPY);

	action->nat64.pattern_obj[ACTION_NAT64_STAGE_REPLACE] =
		action_nat64_replace_state_ptrn(ctx, attr, resource, &need_reparse);
	if (!action->nat64.pattern_obj[ACTION_NAT64_STAGE_REPLACE]) {
		HWS_ERR("Nat64 failed creating replace state");
		goto free_copy;
	}

	if (need_reparse)
		action->nat64.need_reparse |= BIT(ACTION_NAT64_STAGE_REPLACE);

	action->nat64.pattern_obj[ACTION_NAT64_STAGE_COPY_PROTOCOL] =
		action_nat64_copy_proto_state_ptrn(ctx, attr, resource, &need_reparse);
	if (!action->nat64.pattern_obj[ACTION_NAT64_STAGE_COPY_PROTOCOL]) {
		HWS_ERR("Nat64 failed creating copy protocol state");
		goto free_replace;
	}

	if (need_reparse)
		action->nat64.need_reparse |= BIT(ACTION_NAT64_STAGE_COPY_PROTOCOL);

	action->nat64.pattern_obj[ACTION_NAT64_STAGE_COPYBACK] =
		action_nat64_copy_back_ptrn(ctx, attr, resource, &need_reparse);
	if (!action->nat64.pattern_obj[ACTION_NAT64_STAGE_COPYBACK]) {
		HWS_ERR("Nat64 failed creating copyback state");
		goto free_copy_proto;
	}

	if (need_reparse)
		action->nat64.need_reparse |= BIT(ACTION_NAT64_STAGE_COPYBACK);

	pthread_spin_unlock(&ctx->ctrl_lock);

	return action;

free_copy_proto:
	pat_put_pattern(ctx, action->nat64.pattern_obj[ACTION_NAT64_STAGE_COPY_PROTOCOL]);
free_replace:
	pat_put_pattern(ctx, action->nat64.pattern_obj[ACTION_NAT64_STAGE_REPLACE]);
free_copy:
	pat_put_pattern(ctx, action->nat64.pattern_obj[ACTION_NAT64_STAGE_COPY]);
free_ptrn_arr:
	free(action->nat64.pattern_obj);
free_action:
	free(action);
free_resource:
	mlx5dv_hws_resource_free(resource);
	pthread_spin_unlock(&ctx->ctrl_lock);
	return NULL;
}

int action_root_build_attr(struct mlx5dv_hws_action_data actions_data[],
			   struct mlx5dv_hws_at *at,
			   struct mlx5dv_hws_action *dest_action,
			   struct mlx5dv_flow_action_attr *attr)
{
	uint32_t num_rule_actions = at->num_actions;
	struct mlx5dv_hws_action *action = NULL;
	uint8_t resource_idx;
	uint32_t i;

	for (i = 0; i < num_rule_actions + 1; i++) {
		if (i == num_rule_actions)
			action = dest_action;
		else
			action = &at->actions[i];

		switch (action->type) {
		case ACTION_TYP_DEST_TBL:
			attr[i].type = MLX5DV_FLOW_ACTION_DEST_DEVX;
			attr[i].obj = action->dest_tbl.devx_obj;
			break;
		case ACTION_TYP_DEST_TIR:
			attr[i].type = MLX5DV_FLOW_ACTION_DEST_DEVX;
			attr[i].obj = action->devx_dest.devx_obj;
			break;
		case ACTION_TYP_DEST_MISS:
			attr[i].type = MLX5DV_FLOW_ACTION_DEFAULT_MISS;
			break;
		case ACTION_TYP_DEST_DROP:
			attr[i].type = MLX5DV_FLOW_ACTION_DROP;
			break;
		case ACTION_TYP_DEST_ARRAY:
			attr[i].type = MLX5DV_FLOW_ACTION_DEST_DEVX;
			attr[i].obj = action->dest_array.fw_island->ft;
			break;
		case ACTION_TYP_TAG:
			attr[i].type = MLX5DV_FLOW_ACTION_TAG;
			attr[i].tag_value = (__force uint32_t)actions_data[i].tag.value;
			break;
		case ACTION_TYP_REFORMAT_TNL_L2_TO_L2:
		case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
		case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
		case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
		case ACTION_TYP_MODIFY_HDR:
			attr[i].type = MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION;
			attr[i].action = action->flow_action;
			break;
		case ACTION_TYP_CTR:
			attr[i].type = MLX5DV_FLOW_ACTION_COUNTERS_DEVX;
			resource_idx = actions_data[i].counter.resource_idx;
			attr[i].obj = action->counter.resource[resource_idx]->devx_obj;
			if (actions_data[i].counter.resource_offset) {
				HWS_ERR("Counter offset not supported over root");
				errno = ENOTSUP;
				return -errno;
			}
			break;
		default:
			HWS_ERR("Found unsupported action type: %d", action->type);
			errno = ENOTSUP;
			return -errno;
		}
	}

	return 0;
}

static struct actions_wqe_setter *
action_setter_find_first(struct actions_wqe_setter *setter,
			 uint8_t req_flags)
{
	/* Use a new setter if requested flags are taken */
	while (setter->flags & req_flags)
		setter++;

	/* Use current setter in required flags are not used */
	return setter;
}

static void
action_apply_stc(struct actions_apply_data *apply,
		 enum action_stc_idx stc_idx,
		 uint8_t action_idx,
		 uint8_t action_stc_idx)
{
	struct mlx5dv_hws_action *action = &apply->at->actions[action_idx];

	apply->wqe_ctrl->stc_ix[stc_idx] =
		htobe32(action->stc[action_stc_idx].offset);
}

static void
action_setter_push_vlan(struct actions_apply_data *apply,
			struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action_data *action_data;

	action_data = &apply->action_data[setter->idx_double];
	apply->wqe_data[ACTION_OFFSET_DW6] = 0;
	apply->wqe_data[ACTION_OFFSET_DW7] = action_data->push_vlan.vlan_hdr;

	action_apply_stc(apply, ACTION_STC_IDX_DW6, setter->idx_double, 0);
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW7] = 0;
}

static void
action_setter_set_inline(struct actions_apply_data *apply,
			 struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action_data *action_data;
	__be32 *data;

	action_data = &apply->action_data[setter->idx_double];
	data = (__force __be32 *)&action_data->inline_action.data;
	apply->wqe_data[ACTION_OFFSET_DW6] = data[0];
	apply->wqe_data[ACTION_OFFSET_DW7] = data[1];

	action_apply_stc(apply, ACTION_STC_IDX_DW6, setter->idx_double, 0);
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW7] = 0;
}

static void
action_setter_modify_header(struct actions_apply_data *apply,
			    struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action_data *action_data;
	struct mlx5dv_hws_action *action;
	uint32_t arg_gran, arg_idx;
	size_t resource_idx;
	size_t data_sz;
	__be32 stc_idx;

	action_data = &apply->action_data[setter->idx_double];
	action = &apply->at->actions[setter->idx_double];

	resource_idx = action_data->modify_header.resource_idx;
	stc_idx = htobe32(action->stc->offset);
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW6] = stc_idx;
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW7] = 0;

	apply->wqe_data[ACTION_OFFSET_DW6] = 0;

	/* Argument offset per the specific resource size */
	arg_gran = resource_arg_gran_get(action->modify_header.resource[resource_idx]->type);
	arg_idx = action_data->modify_header.resource_offset * arg_gran;

	data_sz = action->modify_header.num_of_actions * MLX5_IFC_ARG_SZ_ACTION;
	if (!arg_check_idx_aligned(arg_idx, data_sz))
		HWS_ERR("Un-alinged arg index [%d] provided for [%zu] actions",
			arg_idx, action->modify_header.num_of_actions);

	apply->wqe_data[ACTION_OFFSET_DW7] = htobe32(arg_idx);

	if (action_data->modify_header.actions_data) {
		apply->require_dep = 1;
		arg_write(apply->queue, NULL,
			  action->modify_header.resource[resource_idx]->devx_obj->object_id +
				arg_idx,
			  (uint8_t *)action_data->modify_header.actions_data,
			  action->modify_header.num_of_actions * ACTION_MH_DOUBLE_SIZE);
	}
}

static void
action_setter_insert_ptr(struct actions_apply_data *apply,
			 struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action_data *action_data;
	struct mlx5dv_hws_action *action;
	uint32_t arg_gran, arg_idx;
	size_t resource_idx;
	__be32 stc_idx;

	action_data = &apply->action_data[setter->idx_double];
	action = &apply->at->actions[setter->idx_double];
	resource_idx = action_data->modify_header.resource_idx;

	/* Argument offset per the specific resource size */
	arg_gran = resource_arg_gran_get(action->reformat.resource[resource_idx]->type);
	arg_idx = action_data->reformat_l2_to_tnl_l2.resource_offset * arg_gran;

	if (unlikely(!arg_check_idx_aligned(arg_idx, action->reformat.header_size))) {
		HWS_ERR("Un-alinged arg index [%d] provided for [%zu] header sz",
			arg_idx, action->reformat.header_size);
		/* Fail the rule */
		apply->rule->status = RULE_STATUS_FAILING;
	}

	apply->wqe_data[ACTION_OFFSET_DW6] = 0;
	apply->wqe_data[ACTION_OFFSET_DW7] = htobe32(arg_idx);

	stc_idx = htobe32(action->stc->offset);
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW6] = stc_idx;
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW7] = 0;

	/* Assume all arg write structs share the same field layout.
	 * reformat_l2_to_tnl_l2, reformat_l2_to_tnl_l3, insert_header.
	 */
	if (action_data->reformat_l2_to_tnl_l2.hdr_data) {
		apply->require_dep = 1;
		arg_write(apply->queue, NULL,
			  action->reformat.resource[resource_idx]->devx_obj->object_id +
			  arg_idx,
			  action_data->reformat_l2_to_tnl_l2.hdr_data,
			  action->reformat.header_size);
	}
}

static void
action_setter_tnl_l3_to_l2(struct actions_apply_data *apply,
			   struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action_data *action_data;
	struct mlx5dv_hws_resource *resource;
	struct mlx5dv_hws_action *action;
	uint32_t arg_gran, arg_idx;
	size_t resource_idx;
	__be32 stc_idx;

	action_data = &apply->action_data[setter->idx_double];
	action = &apply->at->actions[setter->idx_double];
	resource_idx = action_data->modify_header.resource_idx;
	resource = action->modify_header.resource[resource_idx];

	/* Argument offset per the specific resource size */
	arg_gran = resource_arg_gran_get(resource->type);
	arg_idx = action_data->reformat_tnl_l3_to_l2.resource_offset * arg_gran;

	apply->wqe_data[ACTION_OFFSET_DW6] = 0;
	apply->wqe_data[ACTION_OFFSET_DW7] = htobe32(arg_idx);

	stc_idx = htobe32(action->stc->offset);
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW6] = stc_idx;
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW7] = 0;

	if (action_data->reformat_tnl_l3_to_l2.hdr_data) {
		apply->require_dep = 1;
		arg_decapl3_write(apply->queue,
				  resource->devx_obj->object_id + arg_idx,
				  action_data->reformat_tnl_l3_to_l2.hdr_data,
				  action->modify_header.num_of_actions);
	}
}

static void action_send_aso(struct actions_apply_data *apply,
			    struct mlx5dv_hws_resource *resource,
			    uint32_t offset,
			    __be32 *aso_wqe,
			    size_t wqe_len)
{
	struct mlx5dv_hws_resource_enqueue_aso_attr attr = {0};
	int ret;

	apply->require_dep = 1;
	attr.offset = offset;
	attr.burst = 1; /* No doorbell, let the rule to define when db */

	ret = resource_enqueue_aso_wqe(resource,
				       apply->queue,
				       aso_wqe,
				       wqe_len,
				       &attr);
	if (unlikely(ret))
		/* No write, so make sure the user will get completion with error here */
		apply->rule->status = RULE_STATUS_FAILING;
}

static void
action_setter_aso(struct actions_apply_data *apply,
		  struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action_data *action_data;
	struct mlx5dv_hws_action *action;
	uint32_t aso_offset, aso_ctrl;
	uint32_t resource_offset;
	uint8_t resource_idx;
	__be32 *aso_wqe;
	size_t wqe_len;

	action_data = &apply->action_data[setter->idx_double];
	action = &apply->at->actions[setter->idx_double];

	switch (action->type) {
	case ACTION_TYP_ASO_METER:
		/* ASO control format:
		 * [STC only and reserved bits 29b][init_color 2b][meter_id 1b]
		 */
		resource_idx = action_data->aso_meter.resource_idx;
		resource_offset = action_data->aso_meter.resource_offset;
		aso_offset = resource_offset / MLX5_ASO_METER_NUM_PER_OBJ;
		aso_ctrl = action_data->aso_meter.resource_offset % MLX5_ASO_METER_NUM_PER_OBJ;
		aso_ctrl |= action_data->aso_meter.init_color <<
				ACTION_ASO_OFFSET_METER_INIT_COLOR;
		aso_wqe = action_data->aso_meter.aso_wqe;
		wqe_len = action_data->aso_meter.wqe_size;
		break;
	case ACTION_TYP_ASO_CT:
		/* ASO control CT format:
		 * [STC only and reserved bits 31b][direction 1b]
		 */
		resource_idx = action_data->aso_ct.resource_idx;
		resource_offset = action_data->aso_ct.resource_offset;
		aso_offset = resource_offset / MLX5_ASO_CT_NUM_PER_OBJ;
		aso_ctrl = action_data->aso_ct.direction;
		aso_wqe = action_data->aso_ct.aso_wqe;
		wqe_len = action_data->aso_ct.wqe_size;
		break;
	case ACTION_TYP_ASO_FIRST_HIT:
		/* ASO control FIRST HIT format:
		 * [STC only and reserved bits 22b][set 1b][offset 9b]
		 */
		resource_idx = action_data->aso_first_hit.resource_idx;
		resource_offset = action_data->aso_first_hit.resource_offset;
		aso_offset = resource_offset / MLX5_ASO_FIRST_HIT_NUM_PER_OBJ;
		aso_ctrl = action_data->aso_first_hit.resource_offset %
			       MLX5_ASO_FIRST_HIT_NUM_PER_OBJ;
		aso_ctrl |= action_data->aso_first_hit.read <<
				ACTION_ASO_OFFSET_FIRST_HIT_SET;
		aso_wqe = action_data->aso_first_hit.aso_wqe;
		wqe_len = action_data->aso_first_hit.wqe_size;
		break;
	case ACTION_TYP_ASO_IPSEC:
		resource_idx = action_data->aso_ipsec.resource_idx;
		resource_offset = action_data->aso_ipsec.resource_offset;
		aso_offset = resource_offset / MLX5_ASO_IPSEC_NUM_PER_OBJ;
		aso_ctrl = 0;
		aso_wqe = action_data->aso_ipsec.aso_wqe;
		wqe_len = action_data->aso_ipsec.wqe_size;
		break;
	default:
		HWS_ERR("Unsupported ASO action type: %d", action->type);
		errno = ENOTSUP;
		return;
	}

	/* aso_object_offset format: [24B] */
	apply->wqe_data[ACTION_OFFSET_DW6] = htobe32(aso_offset);
	apply->wqe_data[ACTION_OFFSET_DW7] = htobe32(aso_ctrl);

	action_apply_stc(apply, ACTION_STC_IDX_DW6, setter->idx_double, resource_idx);
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW7] = 0;

	if (aso_wqe)
		action_send_aso(apply,
				action->aso.resource[resource_idx],
				resource_offset,
				aso_wqe,
				wqe_len);
}

static void
action_setter_tag(struct actions_apply_data *apply,
		  struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action_data *action_data;

	action_data = &apply->action_data[setter->idx_single];
	apply->wqe_data[ACTION_OFFSET_DW5] = action_data->tag.value;
	action_apply_stc(apply, ACTION_STC_IDX_DW5, setter->idx_single, 0);
}

static void
action_setter_ctrl_ctr(struct actions_apply_data *apply,
		       struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action_data *action_data;
	struct mlx5dv_hws_action *action;

	action = &apply->at->actions[setter->idx_ctr];
	action_data = &apply->action_data[setter->idx_ctr];
	apply->wqe_data[ACTION_OFFSET_DW0] =
		htobe32(action_data->counter.resource_offset);
	action_apply_stc(apply, ACTION_STC_IDX_CTRL, setter->idx_ctr,
			 action_data->counter.resource_idx);

	if (action_data->counter.aso_wqe)
		action_send_aso(apply,
				action->counter.resource[action_data->counter.resource_idx],
				action_data->counter.resource_offset,
				action_data->counter.aso_wqe,
				action_data->counter.wqe_size);
}

static void
action_setter_single(struct actions_apply_data *apply,
		     struct actions_wqe_setter *setter)
{
	apply->wqe_data[ACTION_OFFSET_DW5] = 0;
	action_apply_stc(apply, ACTION_STC_IDX_DW5, setter->idx_single, 0);
}

static void
action_setter_single_scnd(struct actions_apply_data *apply,
			  struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action *action;

	action = &apply->at->actions[setter->idx_single];
	apply->wqe_data[ACTION_OFFSET_DW5] = 0;
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW5] =
		htobe32(action->stc[action->num_stc - 1].offset);
}

static void
action_setter_hit_dest_action(struct actions_apply_data *apply,
			      struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action *action = apply->dest_action;
	struct mlx5dv_hws_action_data *action_data;

	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_HIT] = htobe32(action->stc->offset);
	apply->wqe_data[ACTION_OFFSET_HIT_LSB] = 0;

	/* Set additonal destination data */
	if (unlikely(action->type == ACTION_TYP_DEST_MATCHER)) {
		action_data = &apply->action_data[apply->at->num_actions];
		apply->wqe_data[ACTION_OFFSET_HIT_LSB] =
			htobe32(action_data->dest_matcher.index << 6);
	}
}

static void
action_setter_hit_next_action(struct actions_apply_data *apply,
			      struct actions_wqe_setter *setter)
{
	apply->wqe_data[ACTION_OFFSET_HIT_LSB] =
		htobe32(apply->next_direct_idx << 6);
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_HIT] =
		htobe32(apply->jump_to_action_stc);
}

static void
action_setter_nat64(struct actions_apply_data *apply,
		    struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action *action;
	uint32_t arg_gran, arg_idx;
	__be32 stc_idx;

	action = &apply->at->actions[setter->idx_double];

	stc_idx = htobe32(action->stc[setter->stage_idx].offset);

	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW6] = stc_idx;

	/* Argument offset per the specific resource size */
	arg_gran = resource_arg_gran_get(action->nat64.resource->type);
	arg_idx = setter->stage_idx * arg_gran;
	apply->wqe_data[ACTION_OFFSET_DW6] = 0;
	apply->wqe_data[ACTION_OFFSET_DW7] = htobe32(arg_idx);

	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW7] = 0;
}

static void
action_setter_reformat_trailer(struct actions_apply_data *apply,
			       struct actions_wqe_setter *setter)
{
	action_apply_stc(apply, ACTION_STC_IDX_DW5, setter->idx_triple, 0);
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW6] = 0;
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW7] = 0;
	apply->wqe_data[ACTION_OFFSET_DW5] = 0;
	apply->wqe_data[ACTION_OFFSET_DW6] = 0;
	apply->wqe_data[ACTION_OFFSET_DW7] = 0;
}

static void
action_setter_crypto_encryption(struct actions_apply_data *apply,
				struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action_data *action_data;

	action_data = &apply->action_data[setter->idx_single];
	apply->wqe_data[ACTION_OFFSET_DW5] = htobe32(action_data->crypto.resource_offset);
	action_apply_stc(apply, ACTION_STC_IDX_DW5,
			 setter->idx_single, action_data->crypto.resource_idx);
}

static void
action_setter_crypto_decryption(struct actions_apply_data *apply,
				struct actions_wqe_setter *setter)
{
	struct mlx5dv_hws_action_data *action_data;
	struct mlx5dv_hws_action *action;
	uint32_t offset = 0;
	uint8_t res_idx = 0;

	action = &apply->at->actions[setter->idx_triple];
	action_data = &apply->action_data[setter->idx_triple];

	/* In IPsec pass the crypto object offset in the WQE,
	 * for PSP no DEK object needed in STC or WQE, HW knows which DEK is suitable
	 * for the packet from it's PSP header.
	 */
	if (action->crypto.crypto_type == MLX5DV_HWS_ACTION_CRYPTO_TYPE_IPSEC) {
		offset = action_data->crypto.resource_offset;
		res_idx = action_data->crypto.resource_idx;
	}

	action_apply_stc(apply, ACTION_STC_IDX_DW5, setter->idx_triple, res_idx);
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW6] = 0;
	apply->wqe_ctrl->stc_ix[ACTION_STC_IDX_DW7] = 0;
	apply->wqe_data[ACTION_OFFSET_DW5] = htobe32(offset);
	apply->wqe_data[ACTION_OFFSET_DW6] = 0;
	apply->wqe_data[ACTION_OFFSET_DW7] = 0;
}

static int action_template_process_setters(struct mlx5dv_hws_at *at)
{
	struct actions_wqe_setter *start_setter = at->setters + 1;
	struct actions_wqe_setter *barrier_setter = start_setter;
	struct actions_wqe_setter *setter = at->setters;
	struct actions_wqe_setter *last_setter;
	struct mlx5dv_hws_action *action;
	int i, j;

	for (i = 0; i < ACTION_MAX_STE; i++)
		setter[i].set_hit = &action_setter_hit_next_action;

	/* The same action template setters can be used with jumbo or match
	 * STE, to support both cases we reserve the first setter for cases
	 * with jumbo STE to allow jump to the first action STE.
	 * This extra setter can be reduced in some cases on rule creation.
	 */
	setter = start_setter;
	last_setter = start_setter;

	for (i = 0; i < at->num_actions; i++) {
		action = &at->actions[i];

		/* Prevent overflow, check assumes each action consume a single STE */
		if (last_setter - start_setter >= ACTION_MAX_STE - 1) {
			HWS_ERR("Requested action template exceeds max amount of resources");
			errno = ENOTSUP;
			return errno;
		}

		switch (action->type) {
		case ACTION_TYP_POP_VLAN:
			/* Single remove header to header */
			setter = action_setter_find_first(last_setter,
							  ASF_SINGLE1 |
							  ASF_MODIFY |
							  ASF_INSERT);
			setter->flags |= ASF_SINGLE1 | ASF_REMOVE;
			setter->set_single = &action_setter_single;
			setter->idx_single = i;
			break;
		case ACTION_TYP_PUSH_VLAN:
			/* Double insert inline */
			setter = action_setter_find_first(last_setter, ASF_DOUBLE | ASF_REMOVE);
			setter->flags |= ASF_DOUBLE | ASF_INSERT;
			setter->set_double = &action_setter_push_vlan;
			setter->idx_double = i;
			break;
		case ACTION_TYP_MODIFY_HDR:
			/* Double modify header list */
			setter = action_setter_find_first(last_setter, ASF_DOUBLE | ASF_REMOVE);
			setter->flags |= ASF_DOUBLE | ASF_MODIFY;
			setter->set_double = &action_setter_modify_header;
			setter->idx_double = i;
			break;
		case ACTION_TYP_NAT64:
			/* NAT64 requires 3 setters, each of them does specific modify header */
			for (j = 0; j < ACTION_NAT64_STAGES; j++) {
				setter = action_setter_find_first(last_setter,
								  ASF_DOUBLE | ASF_REMOVE);
				setter->flags |= ASF_DOUBLE | ASF_MODIFY;
				setter->set_double = &action_setter_nat64;
				setter->idx_double = i;
				/* The stage indicates which stc modify-header to push */
				setter->stage_idx = j;
			}
			break;
		case ACTION_TYP_ASO_METER:
		case ACTION_TYP_ASO_CT:
			/* Double ASO action */
			setter = action_setter_find_first(last_setter, ASF_DOUBLE | ASF_REMOVE | ASF_INSERT);
			/* Prevent parallel insert or remove */
			setter->flags |= ASF_DOUBLE | ASF_REMOVE | ASF_INSERT;
			setter->set_double = &action_setter_aso;
			setter->idx_double = i;
			break;
		case ACTION_TYP_ASO_FIRST_HIT:
		case ACTION_TYP_ASO_IPSEC:
			/* Double ASO action */
			setter = action_setter_find_first(last_setter, ASF_DOUBLE);
			setter->flags |= ASF_DOUBLE;
			setter->set_double = &action_setter_aso;
			setter->idx_double = i;
			break;
		case ACTION_TYP_REMOVE_HEADER:
		case ACTION_TYP_REFORMAT_TNL_L2_TO_L2:
			/* Single remove header to header */
			setter = action_setter_find_first(last_setter,
							  ASF_SINGLE1 |
							  ASF_MODIFY |
							  ASF_INSERT);
			setter->flags |= ASF_SINGLE1 | ASF_REMOVE;
			setter->set_single = &action_setter_single;
			setter->idx_single = i;
			break;
		case ACTION_TYP_INSERT_HEADER:
		case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
			/* Double insert header with pointer */
			setter = action_setter_find_first(last_setter, ASF_DOUBLE | ASF_REMOVE);
			setter->flags |= ASF_DOUBLE | ASF_INSERT;
			setter->set_double = &action_setter_insert_ptr;
			setter->idx_double = i;
			break;
		case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
			/* Single remove + Double insert header with pointer */
			setter = action_setter_find_first(last_setter, ASF_SINGLE1 | ASF_DOUBLE);
			setter->flags |= ASF_SINGLE1 | ASF_DOUBLE;
			setter->set_double = &action_setter_insert_ptr;
			setter->idx_double = i;
			setter->set_single = &action_setter_single_scnd;
			setter->idx_single = i;
			break;
		case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
			setter = action_setter_find_first(last_setter, ASF_DOUBLE | ASF_REMOVE);
			setter->flags |= ASF_DOUBLE | ASF_MODIFY | ASF_INSERT;
			setter->set_double = &action_setter_tnl_l3_to_l2;
			setter->idx_double = i;
			break;
		case ACTION_TYP_TAG:
			/* Single TAG action, search for any room from the start */
			setter = action_setter_find_first(start_setter, ASF_SINGLE1);
			setter->flags |= ASF_SINGLE1;
			setter->set_single = &action_setter_tag;
			setter->idx_single = i;
			break;
		case ACTION_TYP_CTR:
			/* Control counter action
			 * TODO: Current counter executed first. Support is needed
			 *	 for single ation counter action which is done last.
			 *	 Example: Decap + CTR
			 */
			setter = action_setter_find_first(barrier_setter, ASF_CTR);
			setter->flags |= ASF_CTR;
			setter->set_ctr = &action_setter_ctrl_ctr;
			setter->idx_ctr = i;
			break;
		case ACTION_TYP_BARRIER:
			last_setter++;
			barrier_setter = last_setter;
			break;
		case ACTION_TYP_INLINE_SET:
		case ACTION_TYP_INLINE_ADD:
		case ACTION_TYP_INLINE_ADD_FIELD:
		case ACTION_TYP_INLINE_COPY:
		case ACTION_TYP_INLINE_INSERT:
			/* Double insert inline */
			setter = action_setter_find_first(last_setter, ASF_DOUBLE | ASF_REMOVE);
			setter->flags |= ASF_DOUBLE | ASF_MODIFY;
			setter->set_double = &action_setter_set_inline;
			setter->idx_double = i;
			break;
		case ACTION_TYP_CRYPTO_ENCRYPT:
			/* Single encryption action, consume triple due to HW limitations */
			setter = action_setter_find_first(last_setter, ASF_TRIPLE);
			setter->flags |= ASF_TRIPLE;
			setter->set_single = &action_setter_crypto_encryption;
			setter->idx_single = i;
			break;
		case ACTION_TYP_CRYPTO_DECRYPT:
			/* Triple decryption action */
			setter = action_setter_find_first(last_setter, ASF_TRIPLE);
			setter->flags |= ASF_TRIPLE;
			setter->set_triple = &action_setter_crypto_decryption;
			setter->idx_triple = i;
			break;
		case ACTION_TYP_REFORMAT_TRAILER:
			/* Trailer action consume triple due to HW limitations */
			setter = action_setter_find_first(last_setter, ASF_TRIPLE);
			setter->flags |= ASF_TRIPLE;
			setter->set_triple = &action_setter_reformat_trailer;
			setter->idx_triple = i;
			break;
		default:
			HWS_ERR("Unsupported action type: %d", action->type);
			errno = ENOTSUP;
			assert(false);
			return -errno;
		}

		last_setter = max(setter, last_setter);
	}

	at->num_of_action_stes = last_setter - start_setter + 1;

	/* Check if action template doesn't require any action DWs */
	at->only_term = (at->num_of_action_stes == 1) &&
		!(last_setter->flags & ~(ASF_CTR));

	last_setter->set_hit = &action_setter_hit_dest_action;

	return 0;
}

static bool action_verify_restriction(struct mlx5dv_hws_matcher *matcher,
				      struct mlx5dv_hws_action *action)
{
	/* Check declared flags in actions fit matcher table type */
	if (!(action->flags & action_tbl_type_to_action_flags(matcher->tbl->type))) {
		HWS_ERR("Declared action-flags (0x%lx) doesn't fit matcher type %d",
			action->flags,
			matcher->tbl->type);
		errno = ENOTSUP;
		return false;
	}

	if (!(table_is_root(matcher->tbl) == action_flags_is_root(action->flags))) {
		HWS_ERR("%s action cannot be used over %s matcher",
			action_flags_is_root(action->flags) ?  "Root" : "HWS",
			table_is_root(matcher->tbl) ? "root" : "HWS");
		errno = ENOTSUP;
		return false;
	}

	switch (action->type) {
	case ACTION_TYP_DEST_ROOT:
	case ACTION_TYP_DEST_ARRAY:
	case ACTION_TYP_DEST_DROP:
	case ACTION_TYP_DEST_TIR:
	case ACTION_TYP_DEST_TBL:
	case ACTION_TYP_DEST_VPORT:
	case ACTION_TYP_DEST_MISS:
	case ACTION_TYP_DEST_MATCHER:
		HWS_ERR("Destination action %d not supported for action-template",
			action->type);
		errno = ENOTSUP;
		return false;
	case ACTION_TYP_TAG:
	case ACTION_TYP_POP_VLAN:
	case ACTION_TYP_PUSH_VLAN:
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
	case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
	case ACTION_TYP_MODIFY_HDR:
	case ACTION_TYP_INSERT_HEADER:
	case ACTION_TYP_CRYPTO_ENCRYPT:
	case ACTION_TYP_CRYPTO_DECRYPT:
	case ACTION_TYP_REFORMAT_TRAILER:
		/* Checks were done in action creation */
		break;
	case ACTION_TYP_ASO_METER:
	case ACTION_TYP_ASO_CT:
	case ACTION_TYP_ASO_FIRST_HIT:
	case ACTION_TYP_ASO_IPSEC:
		if (table_is_root(matcher->tbl)) {
			HWS_ERR("ASO action not supported for root");
			errno = ENOTSUP;
			return false;
		}
		break;
	default:
		break;
	}

	return true;
}

static enum mlx5_ifc_stc_action_type
action_inline_type_to_stc_type(enum action_type action_type)
{
	switch (action_type) {
	case ACTION_TYP_INLINE_SET:
		return MLX5_IFC_STC_ACTION_TYPE_SET;
	case ACTION_TYP_INLINE_ADD:
		return MLX5_IFC_STC_ACTION_TYPE_ADD;
	case ACTION_TYP_INLINE_ADD_FIELD:
		return MLX5_IFC_STC_ACTION_TYPE_ADD_FIELD;
	case ACTION_TYP_INLINE_COPY:
		return MLX5_IFC_STC_ACTION_TYPE_COPY;
	default:
		break;
	}

	return MLX5_IFC_STC_ACTION_TYPE_NOP;
}

static enum mlx5_ifc_stc_action_type
action_crypto_get_stc_type(enum mlx5dv_hws_action_crypto_type crypto_type,
			   enum action_type action_type)
{
	if (action_type == ACTION_TYP_CRYPTO_ENCRYPT) {
		switch (crypto_type) {
		case MLX5DV_HWS_ACTION_CRYPTO_TYPE_IPSEC:
			return MLX5_IFC_STC_ACTION_TYPE_CRYPTO_IPSEC_ENCRYPTION;
		case MLX5DV_HWS_ACTION_CRYPTO_TYPE_PSP:
			return MLX5_IFC_STC_ACTION_TYPE_CRYPTO_PSP_ENCRYPTION;
		default:
			break;
		}
	}

	if (action_type == ACTION_TYP_CRYPTO_DECRYPT) {
		switch (crypto_type) {
		case MLX5DV_HWS_ACTION_CRYPTO_TYPE_IPSEC:
			return MLX5_IFC_STC_ACTION_TYPE_CRYPTO_IPSEC_DECRYPTION;
		case MLX5DV_HWS_ACTION_CRYPTO_TYPE_PSP:
			return MLX5_IFC_STC_ACTION_TYPE_CRYPTO_PSP_DECRYPTION;
		default:
			break;
		}
	}

	return MLX5_IFC_STC_ACTION_TYPE_NOP;
}

static int
action_template_bind_hws(struct mlx5dv_hws_matcher *matcher,
			 struct mlx5dv_hws_at *at)
{
	struct action_stc_cache_data cache_data = {0};
	struct mlx5dv_hws_table *tbl = matcher->tbl;
	struct mlx5dv_hws_context *ctx = tbl->ctx;
	struct cmd_stc_modify_attr *attr;
	struct mlx5dv_hws_action *action;
	struct pool_chunk *stc;
	int i = 0, j;
	int ret = 0;

	cache_data.ib_port = matcher->tbl->ib_port;
	cache_data.tbl_type = matcher->tbl->type;
	attr = &cache_data.attr;

	for (i = 0; i < at->num_actions; i++) {
		/* Make sure attr is zeroed (not {0}) since we use memcmp */
		memset(attr, 0, sizeof(struct cmd_stc_modify_attr));
		action = &at->actions[i];
		stc = action->stc;

		switch (action->type) {
		case ACTION_TYP_TAG:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_TAG;
			attr->action_offset = ACTION_OFFSET_DW5;
			ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
			if (ret)
				goto roll_back;

			action->num_stc++;
			break;
		case ACTION_TYP_POP_VLAN:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_REMOVE_WORDS;
			attr->action_offset = ACTION_OFFSET_DW5;
			attr->reparse_mode = MLX5_IFC_STC_REPARSE_ALWAYS;
			attr->remove_words.start_anchor = MLX5_HEADER_ANCHOR_FIRST_VLAN_START;
			attr->remove_words.num_of_words = ACTION_HDR_LEN_L2_VLAN / 2;
			ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
			if (ret)
				goto roll_back;

			action->num_stc++;
			break;
		case ACTION_TYP_PUSH_VLAN:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_HEADER_INSERT;
			attr->action_offset = ACTION_OFFSET_DW6;
			attr->reparse_mode = MLX5_IFC_STC_REPARSE_ALWAYS;
			attr->insert_header.encap = 0;
			attr->insert_header.push_esp = 0;
			attr->insert_header.is_inline = 1;
			attr->insert_header.insert_anchor = MLX5_HEADER_ANCHOR_PACKET_START;
			attr->insert_header.insert_offset = ACTION_HDR_LEN_L2_MACS;
			attr->insert_header.header_size = ACTION_HDR_LEN_L2_VLAN;
			ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
			if (ret)
				goto roll_back;

			action->num_stc++;
			break;
		case ACTION_TYP_INSERT_HEADER:
			if (action->insert_header.reparse)
				attr->reparse_mode = MLX5_IFC_STC_REPARSE_ALWAYS;
			else
				attr->reparse_mode = MLX5_IFC_STC_REPARSE_IGNORE;

			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_HEADER_INSERT;
			attr->action_offset = ACTION_OFFSET_DW6;
			attr->insert_header.encap = action->insert_header.encap;
			attr->insert_header.push_esp = action->insert_header.push_esp;
			attr->insert_header.insert_anchor = action->insert_header.anchor;
			attr->insert_header.insert_offset = action->insert_header.offset;
			attr->insert_header.header_size = action->insert_header.header_size;

			for (j = 0; j < action->reformat.num_of_resources; j++) {
				attr->insert_header.arg_id =
					action->reformat.resource[j]->devx_obj->object_id;
				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				stc++;
				action->num_stc++;
			}
			break;
		case ACTION_TYP_REMOVE_HEADER:
			if (action->remove_header.reparse)
				attr->reparse_mode = MLX5_IFC_STC_REPARSE_ALWAYS;
			else
				attr->reparse_mode = MLX5_IFC_STC_REPARSE_IGNORE;

			if (action->remove_header.type ==
			    MLX5DV_HWS_ACTION_REMOVE_HEADER_TYPE_BY_HEADER) {
				attr->action_type = MLX5_IFC_STC_ACTION_TYPE_HEADER_REMOVE;
				attr->remove_header.decap =
					action->remove_header.decap;
				attr->remove_header.start_anchor =
					action->remove_header.start_anchor;
				attr->remove_header.end_anchor =
					action->remove_header.end_anchor;
			} else {
				attr->action_type = MLX5_IFC_STC_ACTION_TYPE_REMOVE_WORDS;
				attr->remove_words.start_anchor =
					action->remove_header.start_anchor;
				attr->remove_words.num_of_words =
					action->remove_header.num_of_words;
			}
			attr->action_offset = ACTION_OFFSET_DW5;

			ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
			if (ret)
				goto roll_back;

			action->num_stc++;
			break;
		case ACTION_TYP_REFORMAT_TNL_L2_TO_L2:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_HEADER_REMOVE;
			attr->action_offset = ACTION_OFFSET_DW5;
			attr->reparse_mode = MLX5_IFC_STC_REPARSE_ALWAYS;
			attr->remove_header.decap = 1;
			attr->remove_header.start_anchor = MLX5_HEADER_ANCHOR_PACKET_START;
			attr->remove_header.end_anchor = MLX5_HEADER_ANCHOR_INNER_MAC;
			ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
			if (ret)
				goto roll_back;

			action->num_stc++;
			break;
		case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
			/* Create the l3 new header STC like done in L2_TO_TNL_L2 */
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_HEADER_INSERT;
			attr->action_offset = ACTION_OFFSET_DW6;
			attr->reparse_mode = MLX5_IFC_STC_REPARSE_ALWAYS;
			attr->insert_header.encap = 1;
			attr->insert_header.push_esp = 0;
			attr->insert_header.insert_anchor = MLX5_HEADER_ANCHOR_PACKET_START;
			attr->insert_header.header_size = action->reformat.header_size;
			attr->insert_header.insert_offset = 0;

			for (j = 0; j < action->reformat.num_of_resources; j++) {
				attr->insert_header.arg_id =
					action->reformat.resource[j]->devx_obj->object_id;
				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				stc++;
				action->num_stc++;
			}

			/* We have 2 STC's for that action starting with the remove l2 header */
			memset(attr, 0, sizeof(struct cmd_stc_modify_attr));

			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_HEADER_REMOVE;
			attr->action_offset = ACTION_OFFSET_DW5;
			attr->reparse_mode = MLX5_IFC_STC_REPARSE_IGNORE;
			attr->remove_header.decap = 0;
			attr->remove_header.start_anchor = MLX5_HEADER_ANCHOR_PACKET_START;
			attr->remove_header.end_anchor = MLX5_HEADER_ANCHOR_IPV6_IPV4;
			ret = action_stc_cache_get(ctx,
						   &cache_data,
						   NULL,
						   &action->stc[action->num_stc]);
			if (ret)
				goto roll_back;

			action->num_stc++;
			break;
		case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_HEADER_INSERT;
			attr->action_offset = ACTION_OFFSET_DW6;
			attr->reparse_mode = MLX5_IFC_STC_REPARSE_ALWAYS;
			attr->insert_header.encap = 1;
			attr->insert_header.push_esp = 0;
			attr->insert_header.insert_anchor = MLX5_HEADER_ANCHOR_PACKET_START;
			attr->insert_header.header_size = action->reformat.header_size;
			attr->insert_header.insert_offset = 0;

			for (j = 0; j < action->reformat.num_of_resources; j++) {
				attr->insert_header.arg_id =
					action->reformat.resource[j]->devx_obj->object_id;
				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				stc++;
				action->num_stc++;
			}

			break;
		case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
		case ACTION_TYP_MODIFY_HDR:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_ACC_MODIFY_LIST;
			attr->action_offset = ACTION_OFFSET_DW6;
			if (action->modify_header.reparse)
				attr->reparse_mode = MLX5_IFC_STC_REPARSE_ALWAYS;
			attr->modify_header.pattern_id =
				action->modify_header.pattern_obj->object_id;

			for (j = 0; j < action->reformat.num_of_resources; j++) {
				attr->modify_header.arg_id =
					action->modify_header.resource[j]->devx_obj->object_id;
				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				stc++;
				action->num_stc++;
			}
			break;
		case ACTION_TYP_NAT64:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_ACC_MODIFY_LIST;
			attr->action_offset = ACTION_OFFSET_DW6;
			attr->modify_header.arg_id =
				action->nat64.resource->devx_obj->object_id;

			for (j = 0; j < ACTION_NAT64_STAGES; j++) {
				if (action->nat64.need_reparse & BIT(j))
					attr->reparse_mode = MLX5_IFC_STC_REPARSE_ALWAYS;
				else
					attr->reparse_mode = MLX5_IFC_STC_REPARSE_IGNORE;

				attr->modify_header.pattern_id =
					action->nat64.pattern_obj[j]->object_id;

				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				stc++;
				action->num_stc++;
			}

			break;
		case ACTION_TYP_ASO_METER:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_ASO;
			attr->action_offset = ACTION_OFFSET_DW6;
			attr->aso.aso_type = MLX5_IFC_ASO_OPC_MOD_POLICER;
			attr->aso.return_reg_id = action->aso.return_reg_id;

			for (j = 0; j < action->counter.num_of_resources; j++) {
				attr->aso.devx_obj_id =
					action->aso.resource[j]->devx_obj->object_id;

				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				stc++;
				action->num_stc++;
			}

			break;
		case ACTION_TYP_ASO_CT:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_ASO;
			attr->action_offset = ACTION_OFFSET_DW6;
			attr->aso.aso_type = MLX5_IFC_ASO_OPC_MOD_CT;
			attr->aso.return_reg_id = action->aso.return_reg_id;

			for (j = 0; j < action->counter.num_of_resources; j++) {
				attr->aso.devx_obj_id =
					action->aso.resource[j]->devx_obj->object_id;

				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				stc++;
				action->num_stc++;
			}

			break;
		case ACTION_TYP_ASO_FIRST_HIT:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_ASO;
			attr->action_offset = ACTION_OFFSET_DW6;
			attr->aso.aso_type = MLX5_IFC_ASO_OPC_MOD_FLOW_HIT;
			attr->aso.return_reg_id = action->aso.return_reg_id;

			for (j = 0; j < action->counter.num_of_resources; j++) {
				attr->aso.devx_obj_id =
					action->aso.resource[j]->devx_obj->object_id;

				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				stc++;
				action->num_stc++;
			}

			break;
		case ACTION_TYP_ASO_IPSEC:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_ASO;
			attr->action_offset = ACTION_OFFSET_DW6;
			attr->aso.aso_type = MLX5_IFC_ASO_OPC_MOD_IPSEC;
			attr->aso.return_reg_id = action->aso.return_reg_id;

			for (j = 0; j < action->aso.num_of_resources; j++) {
				attr->aso.devx_obj_id =
					action->aso.resource[j]->devx_obj->object_id;

				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				stc++;
				action->num_stc++;
			}
			break;
		case ACTION_TYP_CTR:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_COUNTER;
			attr->action_offset = ACTION_OFFSET_DW0;

			for (j = 0; j < action->counter.num_of_resources; j++) {
				attr->id = action->counter.resource[j]->devx_obj->object_id;
				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				stc++;
				action->num_stc++;
			}
			break;
		case ACTION_TYP_BARRIER:
			break;
		case ACTION_TYP_INLINE_SET:
		case ACTION_TYP_INLINE_ADD:
		case ACTION_TYP_INLINE_ADD_FIELD:
		case ACTION_TYP_INLINE_COPY:
			if (action->inline_modify_header.reparse)
				attr->reparse_mode = MLX5_IFC_STC_REPARSE_ALWAYS;
			else
				attr->reparse_mode = MLX5_IFC_STC_REPARSE_IGNORE;
			attr->action_type = action_inline_type_to_stc_type(action->type);
			attr->action_offset = ACTION_OFFSET_DW6;
			attr->modify_action.data = action->inline_modify_header.stc_param;

			ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
			if (ret)
				goto roll_back;

			action->num_stc++;
			break;
		case ACTION_TYP_INLINE_INSERT:
			attr->reparse_mode = MLX5_IFC_STC_REPARSE_ALWAYS;
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_HEADER_INSERT;
			attr->action_offset = ACTION_OFFSET_DW6;
			attr->insert_header.encap = 0;
			attr->insert_header.push_esp = action->insert_header.push_esp;
			attr->insert_header.is_inline = 1;
			attr->insert_header.insert_anchor = action->insert_header.anchor;
			attr->insert_header.insert_offset = action->insert_header.offset;

			attr->insert_header.header_size = action->insert_header.header_size;
			ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
			if (ret)
				goto roll_back;

			action->num_stc++;
			break;
		case ACTION_TYP_CRYPTO_ENCRYPT:
			attr->action_offset = ACTION_OFFSET_DW5;
			attr->action_type = action_crypto_get_stc_type(action->crypto.crypto_type,
								       ACTION_TYP_CRYPTO_ENCRYPT);
			for (j = 0; j < action->crypto.num_of_resources; j++) {
				attr->id = action->crypto.resource[j]->devx_obj->object_id;
				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				stc++;
				action->num_stc++;
			}
			break;
		case ACTION_TYP_CRYPTO_DECRYPT:
			attr->action_offset = ACTION_OFFSET_DW5;
			attr->action_type = action_crypto_get_stc_type(action->crypto.crypto_type,
								       ACTION_TYP_CRYPTO_DECRYPT);
			for (j = 0; j < action->crypto.num_of_resources; j++) {
				attr->id = action->crypto.resource[j]->devx_obj->object_id;
				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				stc++;
				action->num_stc++;
			}

			/* No resources for PSP protocol on decryption */
			if (action->crypto.crypto_type == MLX5DV_HWS_ACTION_CRYPTO_TYPE_PSP) {
				ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
				if (ret)
					goto roll_back;
				action->num_stc++;
			}
			break;
		case ACTION_TYP_REFORMAT_TRAILER:
			attr->action_type = MLX5_IFC_STC_ACTION_TYPE_TRAILER;
			attr->action_offset = ACTION_OFFSET_DW5;
			attr->reformat_trailer.type = action->reformat_trailer.type;
			attr->reformat_trailer.op = action->reformat_trailer.op;
			attr->reformat_trailer.size = action->reformat_trailer.size;
			ret = action_stc_cache_get(ctx, &cache_data, NULL, stc);
			if (ret)
				goto roll_back;

			action->num_stc++;
			break;
		default:
			HWS_ERR("Not supported action type: %d", action->type);
			goto roll_back;
		}
	}

	ret = action_template_process_setters(at);
	if (ret)
		goto roll_back;

	return 0;

roll_back:
	HWS_ERR("Failed to get STC for action index %d", i);
	for (; i >= 0; i--) {
		action = &at->actions[i];
		for (j = 0; j < at->actions[i].num_stc; j++)
			action_stc_cache_put(ctx,
					     tbl->type,
					     tbl->ib_port,
					     &action->stc[j]);
	}
	errno = ENOENT;
	return -errno;
}

int action_template_verify_actions(struct mlx5dv_hws_matcher *matcher,
				   struct mlx5dv_hws_at *at)
{
	struct mlx5dv_hws_action *action;
	int i;

	for (i = 0; i < at->num_actions; i++) {
		action = &at->actions[i];
		if (!action_verify_restriction(matcher, action)) {
			errno = ENOTSUP;
			return -errno;
		}
	}

	return 0;
}

int action_template_bind(struct mlx5dv_hws_matcher *matcher,
			 struct mlx5dv_hws_at *at)
{
	if (!table_is_root(matcher->tbl))
		return action_template_bind_hws(matcher, at);

	return 0;
}

static void
action_template_unbind_hws(struct mlx5dv_hws_matcher *matcher,
			   struct mlx5dv_hws_at *at)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	int i, j;

	for (i = 0; i < at->num_actions; i++) {
		if (at->actions[i].type == ACTION_TYP_BARRIER)
			continue;
		for (j = 0; j < at->actions[i].num_stc; j++)
			action_stc_cache_put(ctx,
					     matcher->tbl->type,
					     matcher->tbl->ib_port,
					     &at->actions[i].stc[j]);
	}
}

void action_template_unbind(struct mlx5dv_hws_matcher *matcher,
			    struct mlx5dv_hws_at *at)
{
	if (!table_is_root(matcher->tbl))
		return action_template_unbind_hws(matcher, at);
}

static void action_destroy_hws(struct mlx5dv_hws_action *action)
{
	uint32_t i;

	switch (action->type) {
	case ACTION_TYP_DEST_TIR:
		action_dest_destroy_stc(action);
		if (context_shared_gvmi_used(action->ctx))
			cmd_destroy_obj(action->devx_dest.devx_obj);
		break;
	case ACTION_TYP_DEST_MISS:
	case ACTION_TYP_DEST_DROP:
	case ACTION_TYP_DEST_VPORT:
	case ACTION_TYP_DEST_TBL:
	case ACTION_TYP_DEST_MATCHER:
		action_dest_destroy_stc(action);
		break;
	case ACTION_TYP_DEST_ROOT:
		action_dest_destroy_stc(action);
		mlx5dv_destroy_steering_anchor(action->root_tbl.sa);
		break;
	case ACTION_TYP_DEST_ARRAY:
		if (!(action->flags & MLX5DV_HWS_ACTION_FLAG_ROOT))
			action_dest_destroy_stc(action);
		cmd_forward_tbl_destroy(action->dest_array.fw_island);
		for (i = 0; i < action->dest_array.num_dest; i++) {
			if (action->dest_array.dests[i].ext_reformat)
				cmd_destroy_obj(action->dest_array.dests[i].ext_reformat);
		}
		free(action->dest_array.dests);
		break;
	case ACTION_TYP_REMOVE_HEADER:
	case ACTION_TYP_BARRIER:
	case ACTION_TYP_INLINE_SET:
	case ACTION_TYP_INLINE_ADD:
	case ACTION_TYP_INLINE_ADD_FIELD:
	case ACTION_TYP_INLINE_COPY:
	case ACTION_TYP_INLINE_INSERT:
	case ACTION_TYP_TAG:
	case ACTION_TYP_PUSH_VLAN:
	case ACTION_TYP_POP_VLAN:
	case ACTION_TYP_REFORMAT_TNL_L2_TO_L2:
	case ACTION_TYP_REFORMAT_TRAILER:
		break;
	case ACTION_TYP_ASO_METER:
	case ACTION_TYP_ASO_CT:
	case ACTION_TYP_ASO_IPSEC:
	case ACTION_TYP_ASO_FIRST_HIT:
		free(action->aso.resource);
		break;
	case ACTION_TYP_CRYPTO_ENCRYPT:
	case ACTION_TYP_CRYPTO_DECRYPT:
		free(action->crypto.resource);
		break;
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
		free(action->reformat.resource);
		break;
	case ACTION_TYP_INSERT_HEADER:
		free(action->insert_header.resource);
		break;
	case ACTION_TYP_MODIFY_HDR:
	case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
		pat_put_pattern(action->ctx, action->modify_header.pattern_obj);
		free(action->modify_header.resource);
		break;
	case ACTION_TYP_CTR:
		free(action->counter.resource);
		break;
	case ACTION_TYP_NAT64:
		for (i = 0; i < ACTION_NAT64_STAGES; i++)
			pat_put_pattern(action->ctx, action->nat64.pattern_obj[i]);

		free(action->nat64.pattern_obj);
		mlx5dv_hws_resource_free(action->nat64.resource);
		break;
	default:
		HWS_ERR("Not supported action type: %d", action->type);
		assert(false);
	}
}

static void action_destroy_root(struct mlx5dv_hws_action *action)
{
	switch (action->type) {
	case ACTION_TYP_TAG:
	case ACTION_TYP_DEST_MISS:
	case ACTION_TYP_DEST_DROP:
	case ACTION_TYP_DEST_VPORT:
	case ACTION_TYP_DEST_TBL:
	case ACTION_TYP_DEST_TIR:
		break;
	case ACTION_TYP_REFORMAT_TNL_L2_TO_L2:
	case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
	case ACTION_TYP_MODIFY_HDR:
		ibv_destroy_flow_action(action->flow_action);
		break;
	case ACTION_TYP_CTR:
		free(action->counter.resource);
		break;
	default:
		HWS_ERR("Not supported action type: %d", action->type);
		assert(false);
	}
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_inline(struct mlx5dv_hws_context *ctx,
				__be64 inline_action,
				bool force_reparse,
				uint64_t action_flags)
{
	enum action_type action_type = ACTION_TYP_MAX;
	struct mlx5dv_hws_action *action;
	uint8_t prm_action_type;

	if (action_flags_is_root(action_flags)) {
		HWS_LOG("Inline action not supported over root");
		errno = ENOTSUP;
		return NULL;
	}

	prm_action_type = DEVX_GET(set_action_in, &inline_action, action_type);
	switch (prm_action_type) {
	case MLX5_MODIFICATION_TYPE_SET:
		action_type = ACTION_TYP_INLINE_SET;
		break;
	case MLX5_MODIFICATION_TYPE_ADD:
		action_type = ACTION_TYP_INLINE_ADD;
		break;
	case MLX5_MODIFICATION_TYPE_ADD_FIELD:
		action_type = ACTION_TYP_INLINE_ADD_FIELD;
		break;
	case MLX5_MODIFICATION_TYPE_COPY:
		action_type = ACTION_TYP_INLINE_COPY;
		break;
	case MLX5_MODIFICATION_TYPE_INSERT:
		action_type = ACTION_TYP_INLINE_INSERT;
		break;
	default:
		HWS_ERR("Not supported inline action type: %d", action_type);
		errno = ENOTSUP;
		return NULL;
	}

	action = action_create_generic(ctx, action_type, action_flags);
	if (!action)
		return NULL;

	if (action_type == ACTION_TYP_INLINE_INSERT) {
		action->insert_header.push_esp =
			DEVX_GET(stc_ste_param_insert, &inline_action, push_esp);
		action->insert_header.anchor =
			DEVX_GET(stc_ste_param_insert, &inline_action, insert_anchor);
		action->insert_header.offset =
			DEVX_GET(stc_ste_param_insert, &inline_action, insert_offset);
		/* stc_set_header_insert_action: action_type of insert must be 4 */
		action->insert_header.header_size = 4;
		action->insert_header.reparse = 1;
	} else {
		force_reparse |= pat_require_reparse(&inline_action, 1);
		action->inline_modify_header.stc_param = inline_action;
		action->inline_modify_header.reparse = force_reparse;
	}

	return action;
}

static int
hws_action_create_crypto_valid(struct mlx5dv_hws_resource *resource[],
			       uint8_t num_of_resources,
			       enum mlx5dv_hws_action_crypto_type crypto_type,
			       enum mlx5dv_hws_action_crypto_op crypto_op,
			       uint64_t action_flags)
{
	bool crypto_need_resource = true;
	int i;

	if (action_flags_is_root(action_flags)) {
		HWS_ERR("Action flags must be only non root (HWS)");
		errno = ENOTSUP;
		return -errno;
	}

	if (crypto_type != MLX5DV_HWS_ACTION_CRYPTO_TYPE_IPSEC &&
	    crypto_type != MLX5DV_HWS_ACTION_CRYPTO_TYPE_PSP) {
		HWS_ERR("Crypto type %u is not supported", crypto_type);
		errno = ENOTSUP;
		return -errno;
	}

	/* For PSP decryption, no explicit DEK object is needed as the hardware
	 * determines the appropriate key from the PSP header.
	 */
	if (crypto_op == MLX5DV_HWS_ACTION_CRYPTO_OP_DECRYPT &&
	    crypto_type == MLX5DV_HWS_ACTION_CRYPTO_TYPE_PSP)
		crypto_need_resource = false;

	if (crypto_need_resource &&
	    (!num_of_resources || num_of_resources > ACTION_STC_MAX_RESOURCES_FOR_ACTION)) {
		HWS_ERR("Crypto op %u of type %u can't support that amount of resources %d",
			crypto_op, crypto_type, num_of_resources);
		errno = ENOTSUP;
		return -errno;
	}

	for (i = 0; i < num_of_resources; i++) {
		if (crypto_type == MLX5DV_HWS_ACTION_CRYPTO_TYPE_IPSEC &&
		    resource[i]->type != MLX5DV_HWS_RESOURCE_TYPE_IPSEC_OFFLOAD) {
			HWS_ERR("Crypto type IPSEC expects resource IPSEC_OFFLOAD");
			errno = EINVAL;
			return -errno;
		    }

		if (crypto_type == MLX5DV_HWS_ACTION_CRYPTO_TYPE_PSP &&
		    resource[i]->type != MLX5DV_HWS_RESOURCE_TYPE_DEK) {
			HWS_ERR("Crypto type PSP expects resource DEK");
			errno = EINVAL;
			return -errno;
		    }
	}

	switch (crypto_op) {
	case MLX5DV_HWS_ACTION_CRYPTO_OP_ENCRYPT:
		if (action_flags & (MLX5DV_HWS_ACTION_FLAG_NIC_RX |
				    MLX5DV_HWS_ACTION_FLAG_FDB_RX |
				    MLX5DV_HWS_ACTION_FLAG_FDB |
				    MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_RX)) {
			HWS_ERR("Encrypt operation is not supported in RX");
			errno = EINVAL;
			return -errno;
		}
		break;
	case MLX5DV_HWS_ACTION_CRYPTO_OP_DECRYPT:
		if (action_flags & (MLX5DV_HWS_ACTION_FLAG_NIC_TX |
				    MLX5DV_HWS_ACTION_FLAG_FDB_TX |
				    MLX5DV_HWS_ACTION_FLAG_FDB |
				    MLX5DV_HWS_ACTION_FLAG_RDMA_TRANSPORT_TX)) {
			HWS_ERR("Decrypt operation is not supported in TX");
			errno = EINVAL;
			return -errno;
		}
		break;
	default:
		HWS_ERR("Crypto op is not supported");
		errno = ENOTSUP;
		return -errno;
	}

	return 0;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_crypto(struct mlx5dv_hws_context *ctx,
				struct mlx5dv_hws_resource *resource[],
				uint8_t num_of_resources,
				enum mlx5dv_hws_action_crypto_type crypto_type,
				enum mlx5dv_hws_action_crypto_op crypto_op,
				uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;
	enum action_type action_type;

	if (hws_action_create_crypto_valid(resource,
					   num_of_resources,
					   crypto_type,
					   crypto_op,
					   action_flags))
		return NULL;

	action_type = crypto_op == MLX5DV_HWS_ACTION_CRYPTO_OP_ENCRYPT ?
		      ACTION_TYP_CRYPTO_ENCRYPT : ACTION_TYP_CRYPTO_DECRYPT;

	action = action_create_generic(ctx, action_type, action_flags);
	if (!action)
		return NULL;

	action->crypto.crypto_type = crypto_type;
	action->crypto.num_of_resources = num_of_resources;
	action->crypto.resource = action_dup_resources(resource, num_of_resources);
	if (!action->crypto.resource)
		goto free_action;

	return action;

free_action:
	free(action);
	return NULL;
}

int mlx5dv_hws_action_destroy(struct mlx5dv_hws_action *action)
{
	if (!action_flags_is_root(action->flags))
		action_destroy_hws(action);
	else
		action_destroy_root(action);

	free(action);

	return 0;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_insert_header(struct mlx5dv_hws_context *ctx,
				       struct mlx5dv_hws_resource *resource[],
				       uint8_t num_of_resources,
				       struct mlx5dv_hws_action_insert_header *hdr,
				       uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;

	if (action_flags_is_root(action_flags)) {
		HWS_LOG("Dynamic reformat action not supported over root");
		errno = ENOTSUP;
		return NULL;
	}

	if (!hdr->hdr_data.sz) {
		HWS_ERR("Header size cannot be zero");
		errno = EINVAL;
		return NULL;
	}

	if ((hdr->hdr_data.sz & 1) || (hdr->offset & 1)) {
		HWS_ERR("Header size and offset have to be in granularity of 2 Byte");
		errno = EINVAL;
		return NULL;
	}

	if (!action_check_resource(resource,
				   num_of_resources,
				   ACTION_TYP_INSERT_HEADER,
				   action_flags)) {
		HWS_ERR("Action resource checks failed");
		return NULL;
	}

	action = action_create_generic(ctx, ACTION_TYP_INSERT_HEADER, action_flags);
	if (!action)
		return NULL;

	action->insert_header.resource = action_dup_resources(resource, num_of_resources);
	if (!action->insert_header.resource)
		goto out_err;

	action->insert_header.num_of_resources = num_of_resources;
	action->insert_header.anchor = hdr->anchor;
	action->insert_header.offset = hdr->offset;
	action->insert_header.encap = hdr->encap;
	action->insert_header.push_esp = hdr->push_esp;
	action->insert_header.reparse = !hdr->skip_reparse;
	action->insert_header.header_size = hdr->hdr_data.sz;

	return action;

out_err:
	free(action);
	return NULL;
}

struct mlx5dv_hws_action *
mlx5dv_hws_action_create_remove_header(struct mlx5dv_hws_context *ctx,
				       struct mlx5dv_hws_action_remove_header_attr *attr,
				       uint64_t action_flags)
{
	struct mlx5dv_hws_action *action;

	if (action_flags_is_root(action_flags)) {
		HWS_LOG("Remove header action not supported over root");
		errno = ENOTSUP;
		return NULL;
	}

	action = action_create_generic(ctx, ACTION_TYP_REMOVE_HEADER, action_flags);
	if (!action)
		return NULL;

	switch (attr->type) {
	case MLX5DV_HWS_ACTION_REMOVE_HEADER_TYPE_BY_HEADER:
		action->remove_header.type = MLX5DV_HWS_ACTION_REMOVE_HEADER_TYPE_BY_HEADER;
		action->remove_header.start_anchor = attr->by_header.start_anchor;
		action->remove_header.end_anchor = attr->by_header.end_anchor;
		action->remove_header.decap = attr->by_header.decap;
		action->remove_header.reparse = !attr->skip_reparse;
		break;
	case MLX5DV_HWS_ACTION_REMOVE_HEADER_TYPE_BY_OFFSET:
		if (attr->by_offset.size % W_SIZE != 0) {
			HWS_LOG("Invalid size, HW supports header remove in WORD granularity");
			errno = EINVAL;
			goto free_action;
		}

		if (attr->by_offset.size > MLX5HWS_DV_ACTION_REMOVE_HEADER_MAX_SIZE) {
			HWS_LOG("Header removal size limited to %u bytes",
				MLX5HWS_DV_ACTION_REMOVE_HEADER_MAX_SIZE);
			errno = EINVAL;
			goto free_action;
		}

		action->remove_header.type = MLX5DV_HWS_ACTION_REMOVE_HEADER_TYPE_BY_OFFSET;
		action->remove_header.start_anchor = attr->by_offset.start_anchor;
		action->remove_header.num_of_words = attr->by_offset.size / W_SIZE;
		action->remove_header.reparse = !attr->skip_reparse;
		break;
	default:
		HWS_LOG("Unsupported remove header type %u", attr->type);
		errno = ENOTSUP;
		goto free_action;
	}

	return action;

free_action:
	free(action);
	return NULL;
}

struct mlx5dv_hws_at *
mlx5dv_hws_action_template_create(struct mlx5dv_hws_action *actions[],
				  size_t num_of_actions,
				  struct mlx5dv_hws_at_attr *attr)
{
	struct mlx5dv_hws_at *at;
	int i;

	if (attr->comp_mask) {
		HWS_ERR("Invalid comp mask provided");
		errno = ENOTSUP;
		return NULL;
	}

	at = calloc(1, sizeof(*at));
	if (!at) {
		HWS_ERR("Failed to allocate action template");
		errno = ENOMEM;
		return NULL;
	}

	at->num_actions = num_of_actions;
	if (num_of_actions) {
		at->actions = calloc(num_of_actions, sizeof(struct mlx5dv_hws_action));
		if (!at->actions) {
			HWS_ERR("Failed to allocate action type array");
			errno = ENOMEM;
			goto free_at;
		}
	} else {
		at->num_of_action_stes = 1;
	}

	for (i = 0; i < num_of_actions; i++)
		at->actions[i] = *actions[i];

	return at;

free_at:
	free(at);
	return NULL;
}

int mlx5dv_hws_action_template_destroy(struct mlx5dv_hws_at *at)
{
	if (at->actions)
		free(at->actions);

	free(at);
	return 0;
}

static void action_arg_write(struct send_engine *queue,
			     struct send_engine_post_attr *send_attr,
			     uint8_t *arg_data,
			     size_t data_size)
{
	struct send_wqe_gta_data_seg_arg *wqe_arg;
	struct send_wqe_gta_ctrl_seg *wqe_ctrl;
	struct send_engine_post_ctrl ctrl;
	int i, full_iter, leftover;
	uint8_t notify_hw;
	uint32_t arg_idx;
	void *user_data;
	size_t wqe_len;
	uint32_t id;

	/* Each WQE can hold 64B of data, it might require multiple iteration */
	full_iter = data_size / MLX5_IFC_ARG_SZ_DATA;
	leftover = data_size & (MLX5_IFC_ARG_SZ_DATA - 1);
	notify_hw = send_attr->notify_hw;
	user_data = send_attr->user_data;
	id = send_attr->id;
	arg_idx = send_attr->id;

	send_attr->notify_hw = 0;
	send_attr->user_data = NULL;
	for (i = 0; i < full_iter; i++) {
		ctrl = send_engine_post_start(queue);
		send_engine_post_req_wqe(&ctrl, (void *)&wqe_ctrl, &wqe_len);
		memset(wqe_ctrl, 0, wqe_len);
		send_engine_post_req_wqe(&ctrl, (void *)&wqe_arg, &wqe_len);
		memcpy(wqe_arg, arg_data, wqe_len);
		send_attr->id = arg_idx++;

		/* Last wqe: mark notify-hw */
		if (!leftover && (i == full_iter - 1)) {
			send_attr->notify_hw = notify_hw;
			send_attr->user_data = user_data;
		}

		send_engine_post_end(&ctrl, send_attr);
		/* Move to next argument data */
		arg_data += MLX5_IFC_ARG_SZ_DATA;
	}

	if (leftover) {
		ctrl = send_engine_post_start(queue);
		send_engine_post_req_wqe(&ctrl, (void *)&wqe_ctrl, &wqe_len);
		memset(wqe_ctrl, 0, wqe_len);
		send_engine_post_req_wqe(&ctrl, (void *)&wqe_arg, &wqe_len);
		memcpy(wqe_arg, arg_data, leftover);
		memset(wqe_arg + leftover, 0, MLX5_SEND_WQE_BB - leftover);
		send_attr->id = arg_idx;
		send_attr->notify_hw = notify_hw;
		send_attr->user_data = user_data;
		send_engine_post_end(&ctrl, send_attr);
	}

	send_engine_inc_entry(queue);
	send_attr->notify_hw = notify_hw;
	send_attr->user_data = user_data;
	send_attr->id = id;
}

static void action_arg_decapl3_write(struct send_engine *queue,
				     struct send_engine_post_attr *send_attr,
				     uint8_t *arg_data,
				     uint16_t num_of_actions)
{
	struct wqe_gta_data_seg_arg *wqe_arg;
	struct send_engine_post_ctrl ctrl;
	struct wqe_gta_ctrl_seg *wqe_ctrl;
	size_t wqe_len;

	ctrl = send_engine_post_start(queue);
	send_engine_post_req_wqe(&ctrl, (void *)&wqe_ctrl, &wqe_len);
	memset(wqe_ctrl, 0, wqe_len);
	send_engine_post_req_wqe(&ctrl, (void *)&wqe_arg, &wqe_len);
	memset(wqe_arg, 0, wqe_len);
	arg_prepare_decap_l3_data(arg_data, (uint8_t *)wqe_arg, num_of_actions);
	send_engine_post_end(&ctrl, send_attr);
	send_engine_inc_entry(queue);
}

static int
action_arg_write_precheck(struct mlx5dv_hws_resource *resource,
			  struct mlx5dv_hws_action_enqueue_arg_write_attr *attr,
			  struct send_engine *queue)
{
	if (unlikely(attr->comp_mask)) {
		HWS_LOG("Invalid comp mask provided");
		errno = ENOTSUP;
		return -errno;
	}

	if (unlikely(resource->type != MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B &&
		     resource->type != MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_128B &&
		     resource->type != MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_256B)) {
		HWS_LOG("Invalid resource type");
		errno = EINVAL;
		return -errno;
	}

	if (unlikely(attr->data_size > ACTION_ARG_MAX_WRITE_SIZE)) {
		HWS_LOG("Invalid data size");
		errno = EINVAL;
		return -errno;
	}

	if (unlikely(send_engine_err(queue))) {
		HWS_LOG("Queue error detected");
		errno = EIO;
		return -errno;
	}

	if (unlikely(send_engine_full(queue))) {
		HWS_LOG("Queue is full");
		errno = EBUSY;
		return -errno;
	}

	return 0;
}

int
mlx5dv_hws_action_enqueue_arg_write(struct mlx5dv_hws_action *action,
				    struct mlx5dv_hws_resource *resource,
				    uint16_t qid,
				    struct mlx5dv_hws_action_enqueue_arg_write_attr *attr)
{
	struct send_engine_post_attr send_attr = {0};
	struct send_engine *queue;
	int granularity;

	queue = &resource->ctx->send_queue[qid];
	granularity = resource_arg_gran_get(resource->type);

	if (unlikely(action_arg_write_precheck(resource, attr, queue)))
		return -errno;

	send_attr.user_data = attr->user_data;
	send_attr.notify_hw = !attr->burst;
	send_attr.opcode = SEND_WQE_OPCODE_TBL_ACCESS;
	send_attr.opmod = SEND_WQE_GTA_OPMOD_MOD_ARG;
	send_attr.len = SEND_WQE_SZ_GTA_CTRL + SEND_WQE_SZ_GTA_DATA;
	send_attr.id =  resource->devx_obj->object_id + attr->resource_offset * granularity;
	send_attr.user_data = attr->user_data;

	switch (action->type) {
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
	case ACTION_TYP_INSERT_HEADER:
	case ACTION_TYP_MODIFY_HDR:
		action_arg_write(queue,
				 &send_attr,
				 attr->arg_data,
				 attr->data_size);
		break;
	case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
		action_arg_decapl3_write(queue,
					 &send_attr,
					 attr->arg_data,
					 action->modify_header.num_of_actions);
		break;
	default:
		HWS_LOG("Action type not supported: %d", action->type);
		errno = ENOTSUP;
		return -errno;
	}

	/* Send dependent WQEs */
	if (!attr->burst)
		send_all_dep_wqe(queue);

	return 0;
}
