/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"
#include "hws_buddy.h"

static unsigned long buddy_find_first_bit(unsigned long *addr,
					  uint32_t size)
{
	return bitmap_find_first_bit(addr, 0, size);
}

int buddy_init(struct buddy_mem *buddy, uint32_t max_order)
{
	int i, s;

	buddy->max_order = max_order;

	buddy->bits = calloc(buddy->max_order + 1, sizeof(long *));
	if (!buddy->bits) {
		errno = ENOMEM;
		return -1;
	}

	buddy->num_free = calloc(buddy->max_order + 1, sizeof(*buddy->num_free));
	if (!buddy->num_free) {
		errno = ENOMEM;
		goto err_out_free_bits;
	}

	for (i = 0; i <= (int)buddy->max_order; ++i) {
		s = 1 << (buddy->max_order - i);
		buddy->bits[i] = bitmap_alloc0(s);
		if (!buddy->bits[i])
			goto err_out_free_num_free;
	}

	bitmap_set_bit(buddy->bits[buddy->max_order], 0);

	buddy->num_free[buddy->max_order] = 1;

	return 0;

err_out_free_num_free:
	for (i = 0; i <= (int)buddy->max_order; ++i)
		free(buddy->bits[i]);

	free(buddy->num_free);

err_out_free_bits:
	free(buddy->bits);
	return -1;
}

struct buddy_mem *buddy_create(uint32_t max_order)
{
	struct buddy_mem *buddy;

	buddy = calloc(1, sizeof(*buddy));
	if (!buddy) {
		errno = ENOMEM;
		return NULL;
	}

	if (buddy_init(buddy, max_order))
		goto free_buddy;

	return buddy;

free_buddy:
	free(buddy);
	return NULL;
}

void buddy_cleanup(struct buddy_mem *buddy)
{
	int i;

	for (i = 0; i <= (int)buddy->max_order; ++i)
		free(buddy->bits[i]);

	free(buddy->num_free);
	free(buddy->bits);
}

int buddy_alloc_mem(struct buddy_mem *buddy, int order)
{
	int seg;
	int o, m;

	for (o = order; o <= (int)buddy->max_order; ++o)
		if (buddy->num_free[o]) {
			m = 1 << (buddy->max_order - o);
			seg = buddy_find_first_bit(buddy->bits[o], m);
			if (m <= seg)
				return -1;

			goto found;
		}

	return -1;

found:
	bitmap_clear_bit(buddy->bits[o], seg);
	--buddy->num_free[o];

	while (o > order) {
		--o;
		seg <<= 1;
		bitmap_set_bit(buddy->bits[o], seg ^ 1);
		++buddy->num_free[o];
	}

	seg <<= order;

	return seg;
}

void buddy_free_mem(struct buddy_mem *buddy, uint32_t seg, int order)
{
	seg >>= order;

	while (bitmap_test_bit(buddy->bits[order], seg ^ 1)) {
		bitmap_clear_bit(buddy->bits[order], seg ^ 1);
		--buddy->num_free[order];
		seg >>= 1;
		++order;
	}

	bitmap_set_bit(buddy->bits[order], seg);

	++buddy->num_free[order];
}
