/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"

static uint32_t cmd_get_syndrome(uint32_t *out)
{
	/* Assumption: syndrome is always the second u32 */
	return be32toh((__force __be32)out[1]);
}

int cmd_destroy_obj(struct mlx5dv_devx_obj *devx_obj)
{
	return mlx5dv_devx_obj_destroy(devx_obj);
}

struct mlx5dv_devx_obj *
cmd_flow_table_create(struct ibv_context *ctx,
		      struct cmd_ft_create_attr *ft_attr)
{
	uint32_t out[DEVX_ST_SZ_DW(create_flow_table_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_flow_table_in)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	void *ft_ctx;

	DEVX_SET(create_flow_table_in, in, opcode, MLX5_CMD_OP_CREATE_FLOW_TABLE);
	DEVX_SET(create_flow_table_in, in, table_type, ft_attr->type);
	DEVX_SET(create_flow_table_in, in, other_vport, ft_attr->other_vport);
	DEVX_SET(create_flow_table_in, in, vport_number, ft_attr->vport);

	ft_ctx = DEVX_ADDR_OF(create_flow_table_in, in, flow_table_context);
	DEVX_SET(flow_table_context, ft_ctx, level, ft_attr->level);
	DEVX_SET(flow_table_context, ft_ctx, rtc_valid, ft_attr->rtc_valid);
	DEVX_SET(flow_table_context, ft_ctx, reformat_en, ft_attr->reformat_en);

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create FT (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

int cmd_flow_table_modify(struct mlx5dv_devx_obj *devx_obj,
			  struct cmd_ft_modify_attr *ft_attr)
{
	uint32_t out[DEVX_ST_SZ_DW(modify_flow_table_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(modify_flow_table_in)] = {0};
	void *ft_ctx;
	int ret;

	DEVX_SET(modify_flow_table_in, in, opcode, MLX5_CMD_OP_MODIFY_FLOW_TABLE);
	DEVX_SET(modify_flow_table_in, in, table_type, ft_attr->type);
	DEVX_SET(modify_flow_table_in, in, other_vport, ft_attr->other_vport);
	DEVX_SET(modify_flow_table_in, in, vport_number, ft_attr->vport);
	DEVX_SET(modify_flow_table_in, in, modify_field_select, ft_attr->modify_fs);
	DEVX_SET(modify_flow_table_in, in, table_id, devx_obj->object_id);

	ft_ctx = DEVX_ADDR_OF(modify_flow_table_in, in, flow_table_context);

	DEVX_SET(flow_table_context, ft_ctx, table_miss_action,
		 ft_attr->table_miss_action);
	DEVX_SET(flow_table_context, ft_ctx, table_miss_id, ft_attr->table_miss_id);
	DEVX_SET(flow_table_context, ft_ctx, rtc_id_0, ft_attr->rtc_id_0);
	DEVX_SET(flow_table_context, ft_ctx, rtc_id_1, ft_attr->rtc_id_1);

	ret = mlx5dv_devx_obj_modify(devx_obj, in, sizeof(in), out, sizeof(out));
	if (ret)
		HWS_ERR("Failed to modify FT (syndrome: %#x)",
			cmd_get_syndrome(out));

	return ret;
}

int cmd_flow_table_modify_rtc(struct mlx5dv_devx_obj *ft,
			      uint32_t fw_ft_type,
			      bool other_vport,
			      uint16_t vport,
			      struct mlx5dv_devx_obj *rtc_0,
			      struct mlx5dv_devx_obj *rtc_1)
{
	struct cmd_ft_modify_attr ft_attr = {0};

	ft_attr.modify_fs = MLX5_IFC_MODIFY_FLOW_TABLE_RTC_ID;
	ft_attr.type = fw_ft_type;
	ft_attr.vport = vport;
	ft_attr.other_vport = other_vport;
	ft_attr.rtc_id_0 = rtc_0 ? rtc_0->object_id : 0;
	ft_attr.rtc_id_1 = rtc_1 ? rtc_1->object_id : 0;

	return cmd_flow_table_modify(ft, &ft_attr);
}

int cmd_flow_table_query(struct mlx5dv_devx_obj *devx_obj,
			 struct cmd_ft_query_attr *ft_attr,
			 uint64_t *icm_addr_0,
			 uint64_t *icm_addr_1)
{
	uint32_t out[DEVX_ST_SZ_DW(query_flow_table_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(query_flow_table_in)] = {0};
	void *ft_ctx;
	int ret;

	DEVX_SET(query_flow_table_in, in, opcode, MLX5_CMD_OP_QUERY_FLOW_TABLE);
	DEVX_SET(query_flow_table_in, in, table_type, ft_attr->type);
	DEVX_SET(query_flow_table_in, in, table_id, devx_obj->object_id);

	ret = mlx5dv_devx_obj_query(devx_obj, in, sizeof(in), out, sizeof(out));
	if (ret) {
		HWS_ERR("Failed to query FT (syndrome: %#x)",
			cmd_get_syndrome(out));
		return ret;
	}

	ft_ctx = DEVX_ADDR_OF(query_flow_table_out, out, flow_table_context);
	*icm_addr_0 = DEVX_GET64(flow_table_context, ft_ctx, sw_owner_icm_root_0);
	*icm_addr_1 = DEVX_GET64(flow_table_context, ft_ctx, sw_owner_icm_root_1);

	return ret;
}

static struct mlx5dv_devx_obj *
cmd_flow_group_create(struct ibv_context *ctx,
		      struct cmd_fg_attr *fg_attr)
{
	uint32_t out[DEVX_ST_SZ_DW(create_flow_group_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_flow_group_in)] = {0};
	struct mlx5dv_devx_obj *devx_obj;

	DEVX_SET(create_flow_group_in, in, opcode, MLX5_CMD_OP_CREATE_FLOW_GROUP);
	DEVX_SET(create_flow_group_in, in, table_type, fg_attr->table_type);
	DEVX_SET(create_flow_group_in, in, table_id, fg_attr->table_id);

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create Flow group(syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

struct mlx5dv_devx_obj *cmd_set_fte(struct ibv_context *ctx,
				    uint32_t table_type,
				    uint32_t table_id,
				    uint32_t group_id,
				    struct cmd_set_fte_attr *fte_attr)
{
	uint32_t out[DEVX_ST_SZ_DW(set_fte_out)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	uint32_t dest_entry_sz;
	uint32_t total_dest_sz;
	void *in_flow_context;
	uint32_t action_flags;
	uint8_t *in_dests;
	uint32_t inlen;
	uint32_t *in;
	uint32_t i;

	dest_entry_sz = fte_attr->extended_dest ?
			DEVX_ST_SZ_BYTES(extended_dest_format) :
			DEVX_ST_SZ_BYTES(dest_format);
	total_dest_sz = dest_entry_sz * fte_attr->dests_num;
	inlen = align((DEVX_ST_SZ_BYTES(set_fte_in) + total_dest_sz), DW_SIZE);

	in = calloc(1, inlen);
	if (!in) {
		HWS_ERR("Failed to allocate memory for fte object");
		errno = ENOMEM;
		return NULL;
	}

	DEVX_SET(set_fte_in, in, opcode, MLX5_CMD_OP_SET_FLOW_TABLE_ENTRY);
	DEVX_SET(set_fte_in, in, table_type, table_type);
	DEVX_SET(set_fte_in, in, table_id, table_id);

	in_flow_context = DEVX_ADDR_OF(set_fte_in, in, flow_context);
	DEVX_SET(flow_context, in_flow_context, group_id, group_id);
	DEVX_SET(flow_context, in_flow_context, flow_source,
		 fte_attr->flow_source);
	DEVX_SET(flow_context, in_flow_context, extended_destination,
		 fte_attr->extended_dest);
	DEVX_SET(flow_context, in_flow_context, uplink_hairpin_en,
		 fte_attr->uplink_hairpin_en);
	DEVX_SET(set_fte_in, in, ignore_flow_level,
		 fte_attr->ignore_flow_level);

	action_flags = fte_attr->action_flags;
	DEVX_SET(flow_context, in_flow_context, action, action_flags);

	if (action_flags & MLX5_IFC_FLOW_CONTEXT_ACTION_REFORMAT)
		DEVX_SET(flow_context, in_flow_context,
			 packet_reformat_id, fte_attr->packet_reformat_id);

	if (action_flags & MLX5_IFC_FLOW_CONTEXT_ACTION_FWD_DEST) {
		in_dests = (uint8_t *)DEVX_ADDR_OF(flow_context,
						   in_flow_context,
						   destination);

		for (i = 0; i < fte_attr->dests_num; i++) {
			struct cmd_set_fte_dest *dest = &fte_attr->dests[i];

			switch (dest->destination_type) {
			case MLX5_FLOW_DESTINATION_TYPE_VPORT:
				if (dest->ext_flags & CMD_EXT_DEST_ESW_OWNER_VHCA_ID) {
					DEVX_SET(dest_format,
						 in_dests,
						 destination_eswitch_owner_vhca_id_valid,
						 1);
					DEVX_SET(dest_format, in_dests,
						 destination_eswitch_owner_vhca_id,
						 dest->esw_owner_vhca_id);
				}
				/* Fall through */
			case MLX5_FLOW_DESTINATION_TYPE_TIR:
			case MLX5_FLOW_DESTINATION_TYPE_FT:
				DEVX_SET(dest_format, in_dests, destination_type,
					 dest->destination_type);
				DEVX_SET(dest_format, in_dests, destination_id,
					 dest->destination_id);
				if (dest->ext_flags & CMD_EXT_DEST_REFORMAT) {
					DEVX_SET(dest_format, in_dests, packet_reformat, 1);
					DEVX_SET(extended_dest_format, in_dests, packet_reformat_id,
						 dest->ext_reformat->object_id);
				}
				break;
			case MLX5_FLOW_DESTINATION_TYPE_NOP:
				DEVX_SET(dest_format, in_dests, destination_type,
					 dest->destination_type);
				break;
			default:
				errno = EOPNOTSUPP;
				goto free_in;
			}

			in_dests = in_dests + dest_entry_sz;
		}
		DEVX_SET(flow_context, in_flow_context, destination_list_size,
			 fte_attr->dests_num);
	}

	devx_obj = mlx5dv_devx_obj_create(ctx, in, inlen, out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create FTE (syndrome: %#x)",
			cmd_get_syndrome(out));
		goto free_in;
	}

	free(in);
	return devx_obj;

free_in:
	free(in);
	return NULL;
}

struct cmd_forward_tbl *
cmd_forward_tbl_create(struct ibv_context *ctx,
		       struct cmd_ft_create_attr *ft_attr,
		       struct cmd_set_fte_attr *fte_attr)
{
	struct cmd_fg_attr fg_attr = {0};
	struct cmd_forward_tbl *fwd_tbl;

	fwd_tbl = calloc(1, sizeof(*fwd_tbl));
	if (!fwd_tbl) {
		HWS_ERR("Failed to allocate memory");
		errno = ENOMEM;
		return NULL;
	}

	fwd_tbl->ft = cmd_flow_table_create(ctx, ft_attr);
	if (!fwd_tbl->ft) {
		HWS_ERR("Failed to create FT");
		goto free_tbl;
	}

	fg_attr.table_id = fwd_tbl->ft->object_id;
	fg_attr.table_type = ft_attr->type;

	fwd_tbl->fg = cmd_flow_group_create(ctx, &fg_attr);
	if (!fwd_tbl->fg) {
		HWS_ERR("Failed to create FG");
		goto free_ft;
	}

	fwd_tbl->fte = cmd_set_fte(ctx,
				   ft_attr->type,
				   fwd_tbl->ft->object_id,
				   fwd_tbl->fg->object_id,
				   fte_attr);
	if (!fwd_tbl->fte) {
		HWS_ERR("Failed to create FTE");
		goto free_fg;
	}
	return fwd_tbl;

free_fg:
	cmd_destroy_obj(fwd_tbl->fg);
free_ft:
	cmd_destroy_obj(fwd_tbl->ft);
free_tbl:
	free(fwd_tbl);
	return NULL;
}

void cmd_forward_tbl_destroy(struct cmd_forward_tbl *fwd_tbl)
{
	cmd_destroy_obj(fwd_tbl->fte);
	cmd_destroy_obj(fwd_tbl->fg);
	cmd_destroy_obj(fwd_tbl->ft);
	free(fwd_tbl);
}

void cmd_set_attr_connect_miss_tbl(struct mlx5dv_hws_context *ctx,
				   uint32_t fw_ft_type,
				   enum mlx5dv_hws_table_type type,
				   bool other_vport,
				   uint16_t vport,
				   struct cmd_ft_modify_attr *ft_attr)
{
	struct mlx5dv_devx_obj *default_miss_tbl;
	struct context_common_res *common_res;

	if (!table_is_fdb_any(type) && !context_shared_gvmi_used(ctx))
		return;

	ft_attr->type = fw_ft_type;
	ft_attr->modify_fs = MLX5_IFC_MODIFY_FLOW_TABLE_MISS_ACTION;
	ft_attr->table_miss_action = MLX5_IFC_MODIFY_FLOW_TABLE_MISS_ACTION_GOTO_TBL;
	ft_attr->other_vport = other_vport;
	ft_attr->vport = vport;

	/* IB Port is hardcoded to zero (since always FDB) */
	common_res = context_get_common_res(ctx, type, 0);

	if (table_is_fdb_any(type)) {
		default_miss_tbl = common_res->default_miss->ft;
		ft_attr->table_miss_id = default_miss_tbl->object_id;
	} else {
		ft_attr->table_miss_id =
			common_res->gvmi_res->aliased_end_ft->object_id;
	}
}

struct mlx5dv_devx_obj *cmd_rtc_create(struct ibv_context *ctx,
				       struct cmd_rtc_create_attr *rtc_attr)
{
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_rtc_in)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	void *attr;

	attr = DEVX_ADDR_OF(create_rtc_in, in, hdr);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, obj_type, MLX5_GENERAL_OBJ_TYPE_RTC);

	attr = DEVX_ADDR_OF(create_rtc_in, in, rtc);
	if (rtc_attr->is_compare) {
		DEVX_SET(rtc, attr, ste_format_0,
			 MLX5_IFC_RTC_STE_FORMAT_4DW_RANGE);
	} else {
		DEVX_SET(rtc, attr, ste_format_0, rtc_attr->is_frst_jumbo ?
			 MLX5_IFC_RTC_STE_FORMAT_11DW :
			 MLX5_IFC_RTC_STE_FORMAT_8DW);
	}

	if (rtc_attr->is_scnd_range) {
		DEVX_SET(rtc, attr, ste_format_1, MLX5_IFC_RTC_STE_FORMAT_RANGE);
		DEVX_SET(rtc, attr, num_match_ste, 2);
	} else if (rtc_attr->is_scnd_compare) {
		DEVX_SET(rtc, attr, ste_format_1, MLX5_IFC_RTC_STE_FORMAT_4DW_RANGE);
		DEVX_SET(rtc, attr, num_match_ste, 2);
	}

	DEVX_SET(rtc, attr, pd, rtc_attr->pd);
	DEVX_SET(rtc, attr, update_method, rtc_attr->fw_gen_wqe);
	DEVX_SET(rtc, attr, update_index_mode, rtc_attr->update_index_mode);
	DEVX_SET(rtc, attr, access_index_mode, rtc_attr->access_index_mode);
	DEVX_SET(rtc, attr, log_depth, rtc_attr->log_depth);
	DEVX_SET(rtc, attr, log_hash_size, rtc_attr->log_size);
	DEVX_SET(rtc, attr, table_type, rtc_attr->table_type);
	DEVX_SET(rtc, attr, num_hash_definer, rtc_attr->num_hash_definer);
	DEVX_SET(rtc, attr, match_definer_0, rtc_attr->match_definer_0);
	DEVX_SET(rtc, attr, match_definer_1, rtc_attr->match_definer_1);
	DEVX_SET(rtc, attr, stc_id, rtc_attr->stc_base);
	DEVX_SET(rtc, attr, ste_table_base_id, rtc_attr->ste_base);
	DEVX_SET(rtc, attr, ste_table_offset, rtc_attr->ste_offset);
	DEVX_SET(rtc, attr, miss_flow_table_id, rtc_attr->miss_ft_id);
	DEVX_SET(rtc, attr, reparse_mode, rtc_attr->reparse_mode);
	DEVX_SET(rtc, attr, other_vport, rtc_attr->other_vport);
	DEVX_SET(rtc, attr, vport_number, rtc_attr->vport);

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create RTC (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

struct mlx5dv_devx_obj *cmd_stc_create(struct ibv_context *ctx,
				       struct cmd_stc_create_attr *stc_attr)
{
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_stc_in)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	void *attr;

	attr = DEVX_ADDR_OF(create_stc_in, in, hdr);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, obj_type, MLX5_GENERAL_OBJ_TYPE_STC);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, log_obj_range, stc_attr->log_obj_range);

	attr = DEVX_ADDR_OF(create_stc_in, in, stc);
	DEVX_SET(stc, attr, table_type, stc_attr->table_type);
	DEVX_SET(stc, attr, other_vport, stc_attr->other_vport);
	DEVX_SET(stc, attr, vport_number, stc_attr->vport);

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create STC (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

static int
cmd_stc_modify_set_stc_param(struct cmd_stc_modify_attr *stc_attr,
			     void *stc_param)
{
	switch (stc_attr->action_type) {
	case MLX5_IFC_STC_ACTION_TYPE_COUNTER:
		DEVX_SET(stc_ste_param_flow_counter, stc_param,
			 flow_counter_id, stc_attr->id);
		break;
	case MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_TIR:
		DEVX_SET(stc_ste_param_tir, stc_param,
			 tirn, stc_attr->dest_tir_num);
		break;
	case MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_FT:
	case MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_FT_FDB_RX:
		DEVX_SET(stc_ste_param_table, stc_param,
			 table_id, stc_attr->dest_table_id);
		break;
	case MLX5_IFC_STC_ACTION_TYPE_ACC_MODIFY_LIST:
		DEVX_SET(stc_ste_param_header_modify_list, stc_param,
			 header_modify_pattern_id,
			 stc_attr->modify_header.pattern_id);
		DEVX_SET(stc_ste_param_header_modify_list, stc_param,
			 header_modify_argument_id,
			 stc_attr->modify_header.arg_id);
		break;
	case MLX5_IFC_STC_ACTION_TYPE_HEADER_REMOVE:
		DEVX_SET(stc_ste_param_remove, stc_param, action_type,
			 MLX5_MODIFICATION_TYPE_REMOVE);
		DEVX_SET(stc_ste_param_remove, stc_param, decap,
			 stc_attr->remove_header.decap);
		DEVX_SET(stc_ste_param_remove, stc_param, remove_start_anchor,
			 stc_attr->remove_header.start_anchor);
		DEVX_SET(stc_ste_param_remove, stc_param, remove_end_anchor,
			 stc_attr->remove_header.end_anchor);
		break;
	case MLX5_IFC_STC_ACTION_TYPE_HEADER_INSERT:
		DEVX_SET(stc_ste_param_insert, stc_param, action_type,
			 MLX5_MODIFICATION_TYPE_INSERT);
		DEVX_SET(stc_ste_param_insert, stc_param, encap,
			 stc_attr->insert_header.encap);
		DEVX_SET(stc_ste_param_insert, stc_param, push_esp,
			 stc_attr->insert_header.push_esp);
		DEVX_SET(stc_ste_param_insert, stc_param, inline_data,
			 stc_attr->insert_header.is_inline);
		DEVX_SET(stc_ste_param_insert, stc_param, insert_anchor,
			 stc_attr->insert_header.insert_anchor);
		/* HW gets the next 2 sizes in words */
		DEVX_SET(stc_ste_param_insert, stc_param, insert_size,
			 stc_attr->insert_header.header_size / W_SIZE);
		DEVX_SET(stc_ste_param_insert, stc_param, insert_offset,
			 stc_attr->insert_header.insert_offset / W_SIZE);
		DEVX_SET(stc_ste_param_insert, stc_param, insert_argument,
			 stc_attr->insert_header.arg_id);
		break;
	case MLX5_IFC_STC_ACTION_TYPE_COPY:
	case MLX5_IFC_STC_ACTION_TYPE_SET:
	case MLX5_IFC_STC_ACTION_TYPE_ADD:
	case MLX5_IFC_STC_ACTION_TYPE_ADD_FIELD:
		*(__be64 *)stc_param = stc_attr->modify_action.data;
		break;
	case MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_VPORT:
	case MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_UPLINK:
		DEVX_SET(stc_ste_param_vport, stc_param, vport_number,
			 stc_attr->vport.vport_num);
		DEVX_SET(stc_ste_param_vport, stc_param, eswitch_owner_vhca_id,
			 stc_attr->vport.esw_owner_vhca_id);
		DEVX_SET(stc_ste_param_vport, stc_param,
			 eswitch_owner_vhca_id_valid,
			 stc_attr->vport.eswitch_owner_vhca_id_valid);
		break;
	case MLX5_IFC_STC_ACTION_TYPE_DROP:
	case MLX5_IFC_STC_ACTION_TYPE_NOP:
	case MLX5_IFC_STC_ACTION_TYPE_TAG:
	case MLX5_IFC_STC_ACTION_TYPE_ALLOW:
		break;
	case MLX5_IFC_STC_ACTION_TYPE_ASO:
		DEVX_SET(stc_ste_param_execute_aso, stc_param, aso_object_id,
			 stc_attr->aso.devx_obj_id);
		DEVX_SET(stc_ste_param_execute_aso, stc_param, return_reg_id,
			 stc_attr->aso.return_reg_id);
		DEVX_SET(stc_ste_param_execute_aso, stc_param, aso_type,
			 stc_attr->aso.aso_type);
		break;
	case MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_STE_TABLE:
		DEVX_SET(stc_ste_param_ste_table, stc_param, ste_obj_id,
			 stc_attr->ste_table.ste_obj_id);
		DEVX_SET(stc_ste_param_ste_table, stc_param, match_definer_id,
			 stc_attr->ste_table.match_definer_id);
		DEVX_SET(stc_ste_param_ste_table, stc_param, log_hash_size,
			 stc_attr->ste_table.log_hash_size);
		break;
	case MLX5_IFC_STC_ACTION_TYPE_REMOVE_WORDS:
		DEVX_SET(stc_ste_param_remove_words, stc_param, action_type,
			 MLX5_MODIFICATION_TYPE_REMOVE_WORDS);
		DEVX_SET(stc_ste_param_remove_words,
			 stc_param, remove_start_anchor,
			 stc_attr->remove_words.start_anchor);
		DEVX_SET(stc_ste_param_remove_words, stc_param,
			 remove_size, stc_attr->remove_words.num_of_words);
		break;
	case MLX5_IFC_STC_ACTION_TYPE_CRYPTO_PSP_ENCRYPTION:
	case MLX5_IFC_STC_ACTION_TYPE_CRYPTO_IPSEC_ENCRYPTION:
		DEVX_SET(stc_ste_param_ipsec_encrypt, stc_param,
			 ipsec_object_id, stc_attr->id);
		break;
	case MLX5_IFC_STC_ACTION_TYPE_CRYPTO_IPSEC_DECRYPTION:
		DEVX_SET(stc_ste_param_ipsec_decrypt, stc_param,
			 ipsec_object_id, stc_attr->id);
		break;
	case MLX5_IFC_STC_ACTION_TYPE_CRYPTO_PSP_DECRYPTION:
		/* Empty STC param */
		break;
	case MLX5_IFC_STC_ACTION_TYPE_TRAILER:
		DEVX_SET(stc_ste_param_trailer, stc_param, command,
			 stc_attr->reformat_trailer.op);
		DEVX_SET(stc_ste_param_trailer, stc_param, type,
			 stc_attr->reformat_trailer.type);
		DEVX_SET(stc_ste_param_trailer, stc_param, length,
			 stc_attr->reformat_trailer.size);
		break;
	default:
		HWS_ERR("Not supported type %d", stc_attr->action_type);
		errno = EINVAL;
		return -errno;
	}
	return 0;
}

int cmd_stc_modify(struct mlx5dv_devx_obj *devx_obj,
		   struct cmd_stc_modify_attr *stc_attr)
{
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_stc_in)] = {0};
	void *stc_param;
	void *attr;
	int ret;

	attr = DEVX_ADDR_OF(create_stc_in, in, hdr);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, opcode, MLX5_CMD_OP_MODIFY_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, obj_type, MLX5_GENERAL_OBJ_TYPE_STC);
	DEVX_SET(general_obj_in_cmd_hdr, in, obj_id, devx_obj->object_id);
	DEVX_SET(general_obj_in_cmd_hdr, in, obj_offset, stc_attr->stc_offset);

	attr = DEVX_ADDR_OF(create_stc_in, in, stc);
	DEVX_SET(stc, attr, ste_action_offset, stc_attr->action_offset);
	DEVX_SET(stc, attr, action_type, stc_attr->action_type);
	DEVX_SET(stc, attr, reparse_mode, stc_attr->reparse_mode);
	DEVX_SET64(stc, attr, modify_field_select,
		   MLX5_IFC_MODIFY_STC_FIELD_SELECT_NEW_STC);

	/* Set destination TIRN, TAG, FT ID, STE ID */
	stc_param = DEVX_ADDR_OF(stc, attr, stc_param);
	ret = cmd_stc_modify_set_stc_param(stc_attr, stc_param);
	if (ret)
		return ret;

	ret = mlx5dv_devx_obj_modify(devx_obj, in, sizeof(in), out, sizeof(out));
	if (ret)
		HWS_ERR("Failed to modify STC FW action_type %d (syndrome: %#x)",
			stc_attr->action_type,
			cmd_get_syndrome(out));

	return ret;
}

struct mlx5dv_devx_obj *cmd_arg_create(struct ibv_context *ctx,
				       uint16_t log_obj_range,
				       uint32_t pdn)
{
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_arg_in)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	void *attr;

	attr = DEVX_ADDR_OF(create_arg_in, in, hdr);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, obj_type, MLX5_GENERAL_OBJ_TYPE_ARG);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, log_obj_range, log_obj_range);

	attr = DEVX_ADDR_OF(create_arg_in, in, arg);
	DEVX_SET(arg, attr, access_pd, pdn);

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create ARG (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

struct mlx5dv_devx_obj *cmd_flow_counter_create(struct ibv_context *ctx,
						uint16_t log_obj_size,
						uint32_t pdn)
{

	uint32_t out[DEVX_ST_SZ_DW(alloc_flow_counter_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(alloc_flow_counter_in)] = {0};
	struct mlx5dv_devx_obj *devx_obj;

	DEVX_SET(alloc_flow_counter_in, in, opcode,
		 MLX5_CMD_OP_ALLOC_FLOW_COUNTER);
	DEVX_SET(alloc_flow_counter_in, in, log_obj_size, log_obj_size);
	DEVX_SET(alloc_flow_counter_in, in, pd, pdn);

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create FLOW_COUNTER (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

struct mlx5dv_devx_obj *cmd_conn_track_create(struct ibv_context *ctx,
					      uint16_t log_obj_size,
					      uint32_t pdn)
{
	uint32_t in[DEVX_ST_SZ_DW(create_conn_track_aso_in)] = {0};
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	void *ptr;

	ptr = DEVX_ADDR_OF(create_conn_track_aso_in, in, hdr);
	DEVX_SET(general_obj_in_cmd_hdr, ptr, opcode,
		 MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr, ptr, obj_type,
		 MLX5_GENERAL_OBJ_TYPE_CONN_TRACK_OFFLOAD);
	DEVX_SET(general_obj_in_cmd_hdr, ptr, log_obj_range, log_obj_size);
	ptr = DEVX_ADDR_OF(create_conn_track_aso_in, in, conn_track_offload);
	DEVX_SET(conn_track_offload, ptr, conn_track_aso_access_pd, pdn);

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create CONN_TRACK_ASO (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

struct mlx5dv_devx_obj *cmd_flow_hit_create(struct ibv_context *ctx,
					    uint16_t log_obj_size,
					    uint32_t pdn)
{
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_flow_hit_aso_in)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	void *ptr = NULL;

	ptr = DEVX_ADDR_OF(create_flow_hit_aso_in, in, hdr);
	DEVX_SET(general_obj_in_cmd_hdr, ptr, opcode,
		 MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr, ptr, obj_type,
		 MLX5_GENERAL_OBJ_TYPE_FLOW_HIT_ASO);
	DEVX_SET(general_obj_in_cmd_hdr, ptr, log_obj_range, log_obj_size);
	ptr = DEVX_ADDR_OF(create_flow_hit_aso_in, in, flow_hit_aso);
	DEVX_SET(flow_hit_aso, ptr, access_pd, pdn);

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create FLOW_HIT_ASO (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

struct mlx5dv_devx_obj *cmd_flow_meter_create(struct ibv_context *ctx,
					      uint16_t log_obj_size,
					      uint32_t pdn)
{
	uint32_t in[DEVX_ST_SZ_DW(create_flow_meter_aso_in)] = {0};
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	void *ptr;

	ptr = DEVX_ADDR_OF(create_flow_meter_aso_in, in, hdr);
	DEVX_SET(general_obj_in_cmd_hdr, ptr, opcode,
		MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr, ptr, obj_type,
		MLX5_GENERAL_OBJ_TYPE_FLOW_METER_ASO);
	DEVX_SET(general_obj_in_cmd_hdr, ptr, log_obj_range,
		log_obj_size);
	ptr = DEVX_ADDR_OF(create_flow_meter_aso_in, in, flow_meter_aso);
	DEVX_SET(flow_meter_aso, ptr, access_pd, pdn);

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create FLOW_METER_ASO (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

struct mlx5dv_devx_obj *
cmd_header_modify_pattern_create(struct ibv_context *ctx,
				 uint32_t pattern_length,
				 uint8_t *actions)
{
	uint32_t in[DEVX_ST_SZ_DW(create_header_modify_pattern_in)] = {0};
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	uint64_t *pattern_data;
	int num_of_actions;
	void *pattern;
	void *attr;
	int i;

	if (pattern_length > MLX5_IFC_MAX_ACTIONS_DATA_IN_HEADER_MODIFY) {
		HWS_ERR("Pattern length %d exceeds limit %d",
			pattern_length,
			MLX5_IFC_MAX_ACTIONS_DATA_IN_HEADER_MODIFY);
		errno = EINVAL;
		return NULL;
	}

	attr = DEVX_ADDR_OF(create_header_modify_pattern_in, in, hdr);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, obj_type, MLX5_GENERAL_OBJ_TYPE_MODIFY_HEADER_PATTERN);

	pattern = DEVX_ADDR_OF(create_header_modify_pattern_in, in, pattern);
	DEVX_SET(header_modify_pattern_in, pattern,
		 pattern_length, pattern_length / QW_SIZE);

	pattern_data = (uint64_t *)DEVX_ADDR_OF(header_modify_pattern_in,
						pattern, pattern_data);
	memcpy(pattern_data, actions, pattern_length);

	num_of_actions = pattern_length / MLX5_IFC_ARG_SZ_ACTION;
	for (i = 0; i < num_of_actions; i++) {
		int type;

		type = DEVX_GET(set_action_in, &pattern_data[i], action_type);
		if (type != MLX5_MODIFICATION_TYPE_COPY &&
		    type != MLX5_MODIFICATION_TYPE_ADD_FIELD)
			/* Action typ-copy use all bytes for control */
			DEVX_SET(set_action_in, &pattern_data[i], data, 0);
	}

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create header_modify_pattern (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

struct mlx5dv_devx_obj *cmd_ste_create(struct ibv_context *ctx,
				       struct cmd_ste_create_attr *ste_attr)
{
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_ste_in)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	void *attr;

	attr = DEVX_ADDR_OF(create_ste_in, in, hdr);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, obj_type, MLX5_GENERAL_OBJ_TYPE_STE);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, log_obj_range, ste_attr->log_obj_range);

	attr = DEVX_ADDR_OF(create_ste_in, in, ste);
	DEVX_SET(ste, attr, table_type, ste_attr->table_type);
	DEVX_SET(ste, attr, other_vport, ste_attr->other_vport);
	DEVX_SET(ste, attr, vport_number, ste_attr->vport);

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create STE (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

struct mlx5dv_devx_obj *
cmd_definer_create(struct ibv_context *ctx,
		   struct cmd_definer_create_attr *def_attr)
{
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_definer_in)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	void *ptr;

	DEVX_SET(general_obj_in_cmd_hdr,
		 in, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr,
		 in, obj_type, MLX5_GENERAL_OBJ_TYPE_DEFINER);

	ptr = DEVX_ADDR_OF(create_definer_in, in, definer);
	DEVX_SET(definer, ptr, format_id, MLX5_IFC_DEFINER_FORMAT_ID_SELECT);

	DEVX_SET(definer, ptr, format_select_dw0, def_attr->dw_selector[0]);
	DEVX_SET(definer, ptr, format_select_dw1, def_attr->dw_selector[1]);
	DEVX_SET(definer, ptr, format_select_dw2, def_attr->dw_selector[2]);
	DEVX_SET(definer, ptr, format_select_dw3, def_attr->dw_selector[3]);
	DEVX_SET(definer, ptr, format_select_dw4, def_attr->dw_selector[4]);
	DEVX_SET(definer, ptr, format_select_dw5, def_attr->dw_selector[5]);
	DEVX_SET(definer, ptr, format_select_dw6, def_attr->dw_selector[6]);
	DEVX_SET(definer, ptr, format_select_dw7, def_attr->dw_selector[7]);
	DEVX_SET(definer, ptr, format_select_dw8, def_attr->dw_selector[8]);

	DEVX_SET(definer, ptr, format_select_byte0, def_attr->byte_selector[0]);
	DEVX_SET(definer, ptr, format_select_byte1, def_attr->byte_selector[1]);
	DEVX_SET(definer, ptr, format_select_byte2, def_attr->byte_selector[2]);
	DEVX_SET(definer, ptr, format_select_byte3, def_attr->byte_selector[3]);
	DEVX_SET(definer, ptr, format_select_byte4, def_attr->byte_selector[4]);
	DEVX_SET(definer, ptr, format_select_byte5, def_attr->byte_selector[5]);
	DEVX_SET(definer, ptr, format_select_byte6, def_attr->byte_selector[6]);
	DEVX_SET(definer, ptr, format_select_byte7, def_attr->byte_selector[7]);

	ptr = DEVX_ADDR_OF(definer, ptr, match_mask);
	memcpy(ptr, def_attr->match_mask,
	       DEVX_FLD_SZ_BYTES(definer, match_mask));

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create Definer (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

struct mlx5dv_devx_obj *cmd_sq_create(struct ibv_context *ctx,
				      struct cmd_sq_create_attr *attr)
{
	uint32_t out[DEVX_ST_SZ_DW(create_sq_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_sq_in)] = {0};
	void *sqc = DEVX_ADDR_OF(create_sq_in, in, ctx);
	void *wqc = DEVX_ADDR_OF(sqc, sqc, wq);
	struct mlx5dv_devx_obj *devx_obj;

	DEVX_SET(create_sq_in, in, opcode, MLX5_CMD_OP_CREATE_SQ);
	DEVX_SET(sqc, sqc, cqn, attr->cqn);
	DEVX_SET(sqc, sqc, flush_in_error_en, 1);
	DEVX_SET(sqc, sqc, non_wire, 1);
	DEVX_SET(sqc, sqc, ts_format, attr->ts_format);
	DEVX_SET(wq, wqc, wq_type, MLX5_WQ_TYPE_CYCLIC);
	DEVX_SET(wq, wqc, pd, attr->pdn);
	DEVX_SET(wq, wqc, uar_page, attr->page_id);
	DEVX_SET(wq, wqc, log_wq_stride, ilog32(MLX5_SEND_WQE_BB - 1));
	DEVX_SET(wq, wqc, log_wq_sz, attr->log_wq_sz);
	DEVX_SET(wq, wqc, dbr_umem_id, attr->dbr_id);
	DEVX_SET(wq, wqc, wq_umem_id, attr->wq_id);

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj)
		return NULL;

	return devx_obj;
}

struct mlx5dv_devx_obj *
cmd_packet_reformat_create(struct ibv_context *ctx,
			   struct cmd_packet_reformat_create_attr *attr)
{
	uint32_t out[DEVX_ST_SZ_DW(alloc_packet_reformat_out)] = {0};
	size_t insz, cmd_data_sz, cmd_total_sz;
	struct mlx5dv_devx_obj *devx_obj;
	void *prctx;
	void *pdata;
	void *in;

	cmd_total_sz = DEVX_ST_SZ_BYTES(alloc_packet_reformat_context_in);
	cmd_total_sz += DEVX_ST_SZ_BYTES(packet_reformat_context_in);
	cmd_data_sz = DEVX_FLD_SZ_BYTES(packet_reformat_context_in, reformat_data);
	insz = align(cmd_total_sz + attr->data_sz - cmd_data_sz, DW_SIZE);

	in = calloc(1, insz);
	if (!in) {
		errno = ENOMEM;
		return NULL;
	}

	DEVX_SET(alloc_packet_reformat_context_in, in, opcode,
		 MLX5_CMD_OP_ALLOC_PACKET_REFORMAT_CONTEXT);

	prctx = DEVX_ADDR_OF(alloc_packet_reformat_context_in, in,
			     packet_reformat_context);
	pdata = DEVX_ADDR_OF(packet_reformat_context_in, prctx, reformat_data);

	DEVX_SET(packet_reformat_context_in, prctx, reformat_type, attr->type);
	DEVX_SET(packet_reformat_context_in, prctx,
		 reformat_param_0, attr->reformat_param_0);
	DEVX_SET(packet_reformat_context_in, prctx,
		 reformat_data_size, attr->data_sz);
	memcpy(pdata, attr->data, attr->data_sz);

	devx_obj = mlx5dv_devx_obj_create(ctx, in, insz, out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create packet reformat (syndrome: %#x)",
			cmd_get_syndrome(out));
		goto out_free_in;
	}

	free(in);

	return devx_obj;

out_free_in:
	free(in);
	return NULL;
}

int cmd_sq_modify_rdy(struct mlx5dv_devx_obj *devx_obj)
{
	uint32_t out[DEVX_ST_SZ_DW(modify_sq_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(modify_sq_in)] = {0};
	void *sqc = DEVX_ADDR_OF(modify_sq_in, in, ctx);
	int ret;

	DEVX_SET(modify_sq_in, in, opcode, MLX5_CMD_OP_MODIFY_SQ);
	DEVX_SET(modify_sq_in, in, sqn, devx_obj->object_id);
	DEVX_SET(modify_sq_in, in, sq_state, MLX5_SQC_STATE_RST);
	DEVX_SET(sqc, sqc, state, MLX5_SQC_STATE_RDY);

	ret = mlx5dv_devx_obj_modify(devx_obj, in, sizeof(in), out, sizeof(out));
	if (ret)
		HWS_ERR("Failed to modify SQ (syndrome: %#x)",
			cmd_get_syndrome(out));

	return ret;
}

int cmd_allow_other_vhca_access(struct ibv_context *ctx,
				struct cmd_allow_other_vhca_access_attr *attr)
{
	uint32_t out[DEVX_ST_SZ_DW(allow_other_vhca_access_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(allow_other_vhca_access_in)] = {0};
	void *key;
	int ret;

	DEVX_SET(allow_other_vhca_access_in,
		 in, opcode, MLX5_CMD_OP_ALLOW_OTHER_VHCA_ACCESS);
	DEVX_SET(allow_other_vhca_access_in,
		 in, object_type_to_be_accessed, attr->obj_type);
	DEVX_SET(allow_other_vhca_access_in,
		 in, object_id_to_be_accessed, attr->obj_id);

	key = DEVX_ADDR_OF(allow_other_vhca_access_in, in, access_key);
	memcpy(key, attr->access_key, sizeof(attr->access_key));

	ret = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
	if (ret)
		HWS_ERR("Failed to execute ALLOW_OTHER_VHCA_ACCESS command");

	return ret;
}

struct mlx5dv_devx_obj *
cmd_alias_obj_create(struct ibv_context *ctx,
		     struct cmd_alias_obj_create_attr *alias_attr)
{
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_alias_obj_in)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	void *attr;
	void *key;

	attr = DEVX_ADDR_OF(create_alias_obj_in, in, hdr);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr,
		 attr, obj_type, alias_attr->obj_type);
	DEVX_SET(general_obj_in_cmd_hdr, attr, alias_object, 1);

	attr = DEVX_ADDR_OF(create_alias_obj_in, in, alias_ctx);
	DEVX_SET(alias_context, attr, vhca_id_to_be_accessed,
		 alias_attr->vhca_id);
	DEVX_SET(alias_context, attr, object_id_to_be_accessed,
		 alias_attr->obj_id);

	key = DEVX_ADDR_OF(alias_context, attr, access_key);
	memcpy(key, alias_attr->access_key, sizeof(alias_attr->access_key));

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create ALIAS OBJ (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

struct mlx5dv_devx_obj *
cmd_allow_and_create_aliased_obj(struct ibv_context *ibv_owner,
				 struct ibv_context *ibv_allowed,
				 uint16_t vhca_id_to_be_accessed,
				 uint32_t aliased_object_id,
				 uint16_t object_type)
{
	struct cmd_allow_other_vhca_access_attr allow_attr = {0};
	struct cmd_alias_obj_create_attr alias_attr = {0};
	struct mlx5dv_devx_obj *devx_obj;
	char key[MLX5_ACCESS_KEY_LEN];
	int i;

	for (i = 0; i < MLX5_ACCESS_KEY_LEN; i++)
		key[i] = rand() & 0xFF;

	memcpy(allow_attr.access_key, key, MLX5_ACCESS_KEY_LEN);
	allow_attr.obj_type = object_type;
	allow_attr.obj_id = aliased_object_id;

	if (cmd_allow_other_vhca_access(ibv_owner, &allow_attr)) {
		HWS_ERR("Failed to allow object to be aliased");
		return NULL;
	}

	memcpy(alias_attr.access_key, key, MLX5_ACCESS_KEY_LEN);
	alias_attr.obj_id = aliased_object_id;
	alias_attr.obj_type = object_type;
	alias_attr.vhca_id = vhca_id_to_be_accessed;

	devx_obj = cmd_alias_obj_create(ibv_allowed, &alias_attr);
	if (!devx_obj) {
		HWS_ERR("Failed to create alias object");
		return NULL;
	}

	return devx_obj;
}

int cmd_generate_wqe(struct ibv_context *ctx,
		     struct cmd_generate_wqe_attr *attr,
		     struct mlx5_cqe64 *ret_cqe)
{
	uint32_t out[DEVX_ST_SZ_DW(generate_wqe_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(generate_wqe_in)] = {0};
	uint8_t status;
	void *ptr;
	int ret;

	DEVX_SET(generate_wqe_in, in, opcode, MLX5_CMD_OP_GENERATE_WQE);
	DEVX_SET(generate_wqe_in, in, pdn, attr->pdn);

	ptr = DEVX_ADDR_OF(generate_wqe_in, in, wqe_ctrl);
	memcpy(ptr, attr->wqe_ctrl,
	       DEVX_FLD_SZ_BYTES(generate_wqe_in, wqe_ctrl));

	ptr = DEVX_ADDR_OF(generate_wqe_in, in, wqe_gta_ctrl);
	memcpy(ptr, attr->gta_ctrl,
	       DEVX_FLD_SZ_BYTES(generate_wqe_in, wqe_gta_ctrl));

	ptr = DEVX_ADDR_OF(generate_wqe_in, in, wqe_gta_data_0);
	memcpy(ptr, attr->gta_data_0,
	       DEVX_FLD_SZ_BYTES(generate_wqe_in, wqe_gta_data_0));

	if (attr->gta_data_1) {
		ptr = DEVX_ADDR_OF(generate_wqe_in, in, wqe_gta_data_1);
		memcpy(ptr, attr->gta_data_1,
		       DEVX_FLD_SZ_BYTES(generate_wqe_in, wqe_gta_data_1));
	}

	ret = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
	if (ret) {
		HWS_ERR("Failed to write GTA WQE using FW (syndrome: %#x)",
			cmd_get_syndrome(out));
		return ret;
	}

	status = DEVX_GET(generate_wqe_out, out, status);
	if (status) {
		HWS_ERR("Invalid FW CQE status %d", status);
		errno = EINVAL;
		return -errno;
	}

	ptr = DEVX_ADDR_OF(generate_wqe_out, out, cqe_data);
	memcpy(ret_cqe, ptr, sizeof(*ret_cqe));

	return 0;
}

int cmd_query_caps(struct ibv_context *ctx,
		   struct cmd_query_caps *caps)
{
	uint32_t out[DEVX_ST_SZ_DW(query_hca_cap_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(query_hca_cap_in)] = {0};
	uint64_t general_obj_types_supported;
	struct ibv_device_attr_ex attr_ex;
	u32 res;
	int ret;

	DEVX_SET(query_hca_cap_in, in, opcode, MLX5_CMD_OP_QUERY_HCA_CAP);
	DEVX_SET(query_hca_cap_in, in, op_mod,
		 MLX5_GET_HCA_CAP_OP_MOD_GENERAL_DEVICE |
		 MLX5_HCA_CAP_OPMOD_GET_CUR);

	ret = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
	if (ret) {
		HWS_ERR("Failed to query device caps");
		return ret;
	}

	caps->flow_counter_access_aso =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap.flow_counter_access_aso);

	if (caps->flow_counter_access_aso)
		caps->flow_counter_access_aso_opc_mod =
			DEVX_GET(query_hca_cap_out, out,
				capability.cmd_hca_cap.flow_counter_access_aso_opc_mod);

	general_obj_types_supported = DEVX_GET64(query_hca_cap_out, out,
						 capability.cmd_hca_cap.general_obj_types);

	caps->wqe_based_update =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap.wqe_based_flow_table_update_cap);

	caps->eswitch_manager =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap.eswitch_manager);

	caps->flex_protocols =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap.flex_parser_protocols);

	caps->flex_parser_id_geneve_opt_0 =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap.flex_parser_id_geneve_opt_0);

	caps->log_header_modify_argument_granularity =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap.log_header_modify_argument_granularity);

	caps->log_header_modify_argument_granularity -=
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap.log_header_modify_argument_granularity_offset);

	caps->log_header_modify_argument_max_alloc =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap.log_header_modify_argument_max_alloc);

	caps->definer_format_sup =
		DEVX_GET64(query_hca_cap_out, out,
			   capability.cmd_hca_cap.match_definer_format_supported);

	caps->vhca_id =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap.vhca_id);

	caps->sq_ts_format =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap.sq_ts_format);

	caps->ipsec_offload =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap.ipsec_offload);

	caps->roce =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap.roce);

	DEVX_SET(query_hca_cap_in, in, op_mod,
		 MLX5_GET_HCA_CAP_OP_MOD_GENERAL_DEVICE_2 |
		 MLX5_HCA_CAP_OPMOD_GET_CUR);

	ret = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
	if (ret) {
		HWS_ERR("Failed to query device caps");
		return ret;
	}

	caps->full_dw_jumbo_support =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap_2.format_select_dw_8_6_ext);

	caps->format_select_gtpu_dw_0 =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap_2.format_select_dw_gtpu_dw_0);

	caps->format_select_gtpu_dw_1 =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap_2.format_select_dw_gtpu_dw_1);

	caps->format_select_gtpu_dw_2 =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap_2.format_select_dw_gtpu_dw_2);

	caps->format_select_gtpu_ext_dw_0 =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap_2.format_select_dw_gtpu_first_ext_dw_0);

	caps->supp_type_gen_wqe =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap_2.generate_wqe_type);

	/* check cross-VHCA support in cap2 */
	res = DEVX_GET(query_hca_cap_out, out,
		       capability.cmd_hca_cap_2.cross_vhca_object_to_object_supported);

	caps->cross_vhca_resources =
		(res & MLX5_CROSS_VHCA_OBJ_TO_OBJ_TYPE_STC_TO_TIR) &&
		(res & MLX5_CROSS_VHCA_OBJ_TO_OBJ_TYPE_STC_TO_FT) &&
		(res & MLX5_CROSS_VHCA_OBJ_TO_OBJ_TYPE_FT_TO_RTC);

	res = DEVX_GET(query_hca_cap_out, out,
		       capability.cmd_hca_cap_2.
		       allowed_object_for_other_vhca_access);

	caps->cross_vhca_resources &=
		(res & MLX5_CROSS_VHCA_ALLOWED_OBJS_TIR) &&
		(res & MLX5_CROSS_VHCA_ALLOWED_OBJS_FT) &&
		(res & MLX5_CROSS_VHCA_ALLOWED_OBJS_RTC);

	caps->flow_table_hash_type =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap_2.flow_table_hash_type);

	caps->encap_entropy_hash_type =
		DEVX_GET(query_hca_cap_out, out,
			 capability.cmd_hca_cap_2.encap_entropy_hash_type);

	DEVX_SET(query_hca_cap_in, in, op_mod,
		 MLX5_GET_HCA_CAP_OP_MOD_NIC_FLOW_TABLE |
		 MLX5_HCA_CAP_OPMOD_GET_CUR);

	ret = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
	if (ret) {
		HWS_ERR("Failed to query flow table caps");
		return ret;
	}

	caps->nic_ft.max_level =
		DEVX_GET(query_hca_cap_out, out,
			 capability.flow_table_nic_cap.ft_properties_nic_receive.max_ft_level);

	caps->nic_ft.reparse =
		DEVX_GET(query_hca_cap_out, out,
			 capability.flow_table_nic_cap.ft_properties_nic_receive.reparse);

	caps->nic_ft.ignore_flow_level_rtc_valid =
		DEVX_GET(query_hca_cap_out, out,
			 capability.flow_table_nic_cap.ft_properties_nic_receive.ignore_flow_level_rtc_valid);

	/* Check cross-VHCA support in flow table properties */
	caps->cross_vhca_resources &=
		DEVX_GET(query_hca_cap_out, out,
			capability.flow_table_nic_cap.ft_properties_nic_receive.cross_vhca_object);

	if (caps->wqe_based_update) {
		DEVX_SET(query_hca_cap_in, in, op_mod,
			 MLX5_GET_HCA_CAP_OP_MOD_WQE_BASED_FLOW_TABLE |
			 MLX5_HCA_CAP_OPMOD_GET_CUR);

		ret = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
		if (ret) {
			HWS_ERR("Failed to query WQE based FT caps");
			return ret;
		}

		caps->rtc_reparse_mode =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.rtc_reparse_mode);

		caps->ste_format =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.ste_format);

		caps->rtc_index_mode =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.rtc_index_mode);

		caps->rtc_log_depth_max =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.rtc_log_depth_max);

		caps->ste_alloc_log_max =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.ste_alloc_log_max);

		caps->ste_alloc_log_gran =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.ste_alloc_log_granularity);

		caps->trivial_match_definer =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.trivial_match_definer);

		caps->stc_alloc_log_max =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.stc_alloc_log_max);

		caps->stc_alloc_log_gran =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.stc_alloc_log_granularity);

		caps->rtc_hash_split_table =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.rtc_hash_split_table);

		caps->rtc_linear_lookup_table =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.rtc_linear_lookup_table);

		caps->stc_max_trailer =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.stc_max_trailer);

		caps->access_index_mode =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.access_index_mode);

		caps->linear_match_definer =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.linear_match_definer_reg_c3);

		caps->rtc_max_hash_def_gen_wqe =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.rtc_max_num_hash_definer_gen_wqe);

		caps->supp_ste_format_gen_wqe =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.ste_format_gen_wqe);

		caps->fdb_tir_stc =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.fdb_jump_to_tir_stc);

		caps->stc_action_type_63_0 =
			DEVX_GET64(query_hca_cap_out, out,
				   capability.wqe_based_flow_table_cap.stc_action_type_63_0);

		caps->stc_action_type_127_64 =
			DEVX_GET64(query_hca_cap_out, out,
				   capability.wqe_based_flow_table_cap.stc_action_type_127_64);

		caps->stc_jump_to_fdb_rx =
			!!(caps->stc_action_type_127_64 &
			   BIT(MLX5_IFC_STC_ACTION_TYPE_CAP_JUMP_FT_FDB_RX % 64));

		caps->stc_push_esp =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.stc_insert_action_push_esp);

		caps->fdb_unified_en =
			DEVX_GET(query_hca_cap_out, out,
				 capability.wqe_based_flow_table_cap.fdb_unified_en);
	}

	if (caps->eswitch_manager) {
		DEVX_SET(query_hca_cap_in, in, op_mod,
			 MLX5_GET_HCA_CAP_OP_MOD_ESW_FLOW_TABLE |
			 MLX5_HCA_CAP_OPMOD_GET_CUR);

		ret = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
		if (ret) {
			HWS_ERR("Failed to query flow table esw caps");
			return ret;
		}

		caps->fdb_ft.max_level =
			DEVX_GET(query_hca_cap_out, out,
				 capability.flow_table_esw_cap.ft_properties_nic_esw_fdb.max_ft_level);

		caps->fdb_ft.reparse =
			DEVX_GET(query_hca_cap_out, out,
				 capability.flow_table_esw_cap.ft_properties_nic_esw_fdb.reparse);

		caps->fdb_uplink_hairpin =
			DEVX_GET(query_hca_cap_out, out, capability.flow_table_esw_cap.fdb_uplink_hairpin);

		DEVX_SET(query_hca_cap_in, in, op_mod,
			 MLX5_SET_HCA_CAP_OP_MOD_ESW | MLX5_HCA_CAP_OPMOD_GET_CUR);

		ret = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
		if (ret) {
			HWS_ERR("Query eswitch capabilities failed %d", ret);
			return ret;
		}

		if (DEVX_GET(query_hca_cap_out, out, capability.esw_cap.esw_manager_vport_number_valid))
			caps->eswitch_manager_vport_number =
				DEVX_GET(query_hca_cap_out, out,
					 capability.esw_cap.esw_manager_vport_number);

		caps->merged_eswitch =
			DEVX_GET(query_hca_cap_out, out, capability.esw_cap.merged_eswitch);
	}

	if (caps->roce) {
		DEVX_SET(query_hca_cap_in, in, op_mod,
			 MLX5_GET_HCA_CAP_OP_MOD_ROCE |
			 MLX5_HCA_CAP_OPMOD_GET_CUR);

		ret = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
		if (ret) {
			HWS_ERR("Failed to query roce caps");
			return ret;
		}

		caps->roce_max_src_udp_port =
			DEVX_GET(query_hca_cap_out, out,
				 capability.roce_caps.r_roce_max_src_udp_port);

		caps->roce_min_src_udp_port =
			DEVX_GET(query_hca_cap_out, out,
				 capability.roce_caps.r_roce_min_src_udp_port);
	}

	if (general_obj_types_supported & (1ULL << MLX5_GENERAL_OBJ_TYPE_FLEX_PARSE_GRAPH)) {
		DEVX_SET(query_hca_cap_in, in, op_mod,
		MLX5_GET_HCA_CAP_OP_MOD_PARSE_GRAPH_NODE_CAP | MLX5_HCA_CAP_OPMOD_GET_CUR);

		ret = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
		if (ret) {
			HWS_ERR("Failed to query parse graph caps - error (%d) %s",
			       errno, strerror(errno));
			HWS_ERR("DEVX command status: %u, syndrome 0x%x",
				DEVX_GET(query_hca_cap_out, out, status),
				DEVX_GET(query_hca_cap_out, out, syndrome));
			return ret;
		}

		caps->parse_graph_header_length_field_mask_width =
			DEVX_GET(query_hca_cap_out, out,
				 capability.parse_graph_node_cap.header_length_mask_width);

		caps->parse_graph_tlv_option_length_field_mask_width =
			DEVX_GET(query_hca_cap_out, out,
				 capability.parse_graph_node_cap.parse_graph_tlv_option_length_field_mask_width);
	}

	caps->geneve_tlv_opt = !!(general_obj_types_supported &
				  (1ULL << MLX5_GENERAL_OBJ_TYPE_GENEVE_TLV_OPT));

	ret = ibv_query_device_ex(ctx, NULL, &attr_ex);
	if (ret) {
		HWS_ERR("Failed to query device attributes");
		return ret;
	}

	caps->num_ib_ports = attr_ex.phys_port_cnt_ex;

	snprintf(caps->fw_ver, sizeof(caps->fw_ver),
		 "%s", attr_ex.orig_attr.fw_ver);

	return ret;
}

int cmd_ib_port_query(struct ibv_context *ctx,
		      uint32_t port_num,
		      uint16_t *vport_num,
		      uint16_t *esw_vhca_id)
{
	struct mlx5dv_port info = {0};
	uint64_t required_flags;
	int ret;

	ret = mlx5dv_query_port(ctx, port_num, &info);
	if (ret) {
		HWS_ERR("Failed to query required ib port data");
		return ret;
	}

	required_flags = MLX5DV_QUERY_PORT_VPORT |
			 MLX5DV_QUERY_PORT_ESW_OWNER_VHCA_ID;

	if ((info.flags & required_flags) != required_flags) {
		HWS_ERR("Missing port information from query port");
		errno = EINVAL;
		return -errno;
	}

	*vport_num = info.vport;
	*esw_vhca_id = info.esw_owner_vhca_id;

	return 0;
}

int cmd_flex_sample_info_query(struct ibv_context *ctx,
			       uint32_t sampler_id,
			       struct cmd_flex_sample_info_query *info_query)
{
	uint32_t out[DEVX_ST_SZ_DW(query_match_sample_info_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(query_match_sample_info_in)] = {0};
	int ret;

	DEVX_SET(query_match_sample_info_in, in, opcode,
		 MLX5_CMD_OP_QUERY_MATCH_SAMPLE_INFO);
	DEVX_SET(query_match_sample_info_in, in, op_mod, 0);
	DEVX_SET(query_match_sample_info_in, in, sample_field_id, sampler_id);

	ret = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
	if (ret) {
		HWS_ERR("Failed to query sample info (syndrome: %#x)",
		       cmd_get_syndrome(out));
		return -errno;
	}

	info_query->modify_field_id =
		DEVX_GET(query_match_sample_info_out, out, modify_field_id);
	info_query->field_format_select_dw =
		DEVX_GET(query_match_sample_info_out, out, field_format_select_dw);
	info_query->ok_bit_format_select_dw =
		DEVX_GET(query_match_sample_info_out, out, ok_bit_format_select_dw);
	info_query->ok_bit_offset =
		DEVX_GET(query_match_sample_info_out, out, ok_bit_offset);

	return 0;
}

static int cmd_flex_node_query(struct mlx5dv_devx_obj *obj,
			       struct mlx5dv_hws_parser_node *node)
{
	uint32_t out[DEVX_ST_SZ_DW(create_flex_parser_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(general_obj_in_cmd_hdr)] = {0};
	struct mlx5dv_hws_parser_sampler *sampler;
	void *sample;
	void *flex;
	void *hdr;
	int i = 0;
	int ret;

	hdr = DEVX_ADDR_OF(create_flex_parser_out, in, hdr);
	flex = DEVX_ADDR_OF(create_flex_parser_out, out, flex);
	sample = DEVX_ADDR_OF(parse_graph_flex, flex, sample_table);

	DEVX_SET(general_obj_in_cmd_hdr, hdr, opcode,
		 MLX5_CMD_OP_QUERY_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr, hdr, obj_type,
		 MLX5_GENERAL_OBJ_TYPE_FLEX_PARSE_GRAPH);
	DEVX_SET(general_obj_in_cmd_hdr, hdr, obj_id, obj->object_id);

	ret = mlx5dv_devx_obj_query(obj, in, sizeof(in), out, sizeof(out));
	if (ret)
		return ret;

	list_for_each(&node->sampler_head, sampler, next) {
		uint32_t en;
		uint32_t id;

		void *s_off = (void *)((char *)sample + i *
					   DEVX_ST_SZ_BYTES(parse_graph_flow_match_sample));
		i++;

		en = DEVX_GET(parse_graph_flow_match_sample, s_off,
				  flow_match_sample_en);
		if (!en)
			break;

		id = DEVX_GET(parse_graph_flow_match_sample, s_off,
				  flow_match_sample_field_id);

		sampler->id = id;
		sampler->binded++;
	}

	return 0;
}

static void cmd_flex_node_get_length_field_mask_and_shift(
	uint32_t bit_length, uint32_t field_multiplier, uint32_t field_offset,
	uint8_t cap_field_mask_width, uint32_t *size_mask, uint32_t *shift, uint32_t *offset)
{
#define DW_LOG_SZ 2
#define DW_LOG_BIT_SZ 5
	/* Offset of the field within the containing DW (in bits) */
	uint32_t length_field_offset_inside_dw;
	/* The DW number that contains the length field (basically its offset in DWs) */
	uint32_t length_field_dw;
	uint32_t length_field_shift;
	int size_granularity;
	uint32_t mask;

	/*
	 * Convert the multiplier into a shift. Since validation has already been performed
	 * and we know the multiplier is of the form 2 ^ n, the value needed is the position
	 * of the bit set in the multiplier field.
	 *
	 * FORMULA: NUMBER OF BITS IN FIELD - 1 - NUMBER OF LEADING ZEROS BEFORE SET BIT
	 *
	 * Example:
	 * For a multiplier of 0x0080:
	 * NUMBER OF BITS IN FIELD = 16
	 * NUMBER OF LEADING ZEROS BEFORE SET BIT = 8
	 *
	 * Calculation:
	 * 16 - 1 - 8 = 7
	 *
	 * Result:
	 * 1 << 7 = 0x80
	 */
	length_field_shift = sizeof(field_multiplier) *
		BITS_IN_BYTE - 1 - __builtin_clz(field_multiplier);

	/* Make sure the mask is aligned to the shift (multiplier) and that it
	 * will not exceed device's maximum value.
	 * Note that the final field length is in DW multiples.
	 */
	mask = (1 << bit_length) - 1;
	size_granularity = length_field_shift - DW_LOG_SZ;
	if (size_granularity < 0) {
		/* FW needs to shift-right the mask to simulate the needed multiplier,
		 * so make sure lower bits are zero in the mask
		 * (as they are going to be ignored anyway).
		 */
		int absolute_size_granularity = DW_LOG_SZ - length_field_shift;

		mask = mask & ~((1 << absolute_size_granularity) - 1);
	} else {
		/* FW might shift-left the mask by size_granularity, so make sure higher bits in
		 * the mask are zero to make sure the length value cannot exceed device's max value.
		 */
		mask = mask & ~(((1 << size_granularity) - 1) <<
				(cap_field_mask_width - size_granularity));
	}

	/* W/A for FW bug RM#3918910 (currently without a fix):
	 * FW does not calculate the cyclic-shift correctly when the length field
	 * is less than 6 (current cap value for max mask width).
	 * Therefore, until FW is fixed, adjust the offset with needed right-shift
	 * so that after cyclic-shift the length field will be placed correctly at
	 * the 26-31 bits where HW expects to see the length field.
	 */
	length_field_offset_inside_dw = field_offset % BITS_IN_DW;
	length_field_dw = field_offset >> DW_LOG_BIT_SZ;
	if (bit_length < cap_field_mask_width) {
		length_field_offset_inside_dw += (BITS_IN_DW -
						  (cap_field_mask_width - bit_length));
		length_field_offset_inside_dw %= BITS_IN_DW;
	}

	*shift = length_field_shift;
	*size_mask = mask;
	*offset = (length_field_dw * BITS_IN_DW) + length_field_offset_inside_dw;
}

static void cmd_flex_node_create_set_mode_length_options(struct mlx5dv_hws_context *ctx,
							 struct mlx5dv_hws_parser_node *node,
							 void *flex)
{
	uint32_t length_field_offset;
	uint32_t length_field_shift;
	uint32_t length_field_mask;

	cmd_flex_node_get_length_field_mask_and_shift(
		node->cfg.option_len.field.field.bit_length,
		node->cfg.option_len.field.multiplier,
		node->cfg.option_len.field.field.bit_offset,
		ctx->caps->parse_graph_tlv_option_length_field_mask_width,
		&length_field_mask, &length_field_shift, &length_field_offset);

	DEVX_SET(parse_graph_flex, flex, tlv_option_length_field_mask,
		 length_field_mask);
	DEVX_SET(parse_graph_flex, flex, tlv_option_length_field_offset,
		 length_field_offset);
	DEVX_SET(parse_graph_flex, flex, tlv_option_length_base_value,
		 node->cfg.option_len.field.addition);
	DEVX_SET(parse_graph_flex, flex, tlv_option_length_field_shift,
		 length_field_shift);
}

static void cmd_flex_node_create_set_mode_length_header(struct mlx5dv_hws_context *ctx,
							struct mlx5dv_hws_parser_node *node,
							void *flex)
{
	uint32_t length_field_offset;
	uint32_t length_field_shift;
	uint32_t length_field_mask;

	cmd_flex_node_get_length_field_mask_and_shift(
		node->cfg.hdr_len.field.field.bit_length,
		node->cfg.hdr_len.field.multiplier,
		node->cfg.hdr_len.field.field.bit_offset,
		ctx->caps->parse_graph_header_length_field_mask_width,
		&length_field_mask, &length_field_shift, &length_field_offset);

	DEVX_SET(parse_graph_flex, flex, header_length_field_mask,
		 length_field_mask);
	DEVX_SET(parse_graph_flex, flex, header_length_field_offset,
		 length_field_offset);
	DEVX_SET(parse_graph_flex, flex, header_length_base_value,
		 node->cfg.hdr_len.field.addition);
	DEVX_SET(parse_graph_flex, flex, header_length_field_shift,
		 length_field_shift);
}

static void cmd_flex_node_create_fill_options(struct mlx5dv_hws_context *ctx,
					      struct mlx5dv_hws_parser_node *node,
					      void *flex)
{
	DEVX_SET(parse_graph_flex, flex, tlv_option_enable, true);
	DEVX_SET(parse_graph_flex, flex, tlv_option_length_mode,
		 node->cfg.option_len.mode);
	DEVX_SET(parse_graph_flex, flex, tlv_option_offset,
		 node->cfg.option_offset);
	DEVX_SET(parse_graph_flex, flex, tlv_option_type_field_offset,
		 node->cfg.option_type_field.bit_offset);
	DEVX_SET(parse_graph_flex, flex, tlv_option_type_field_size,
		 node->cfg.option_type_field.bit_length);

	if (node->cfg.option_len.mode == MLX5DV_HWS_PARSER_HDR_LEN_MODE_FIXED) {
		DEVX_SET(parse_graph_flex, flex, tlv_option_length_base_value,
			 node->cfg.option_len.fixed.hdr_len);
	} else {
		cmd_flex_node_create_set_mode_length_options(ctx, node, flex);
	}
}

static int cmd_flex_node_create_set_arc_src(struct mlx5dv_hws_context *ctx,
					    struct mlx5dv_hws_parser_node *node,
					    void *flex,
					    struct mlx5dv_devx_obj *devx_obj)
{
	struct mlx5dv_hws_parser_arc *arc;
	void *in_arc;
	int i;

	i = 0;
	in_arc = DEVX_ADDR_OF(parse_graph_flex, flex, input_arc);
	list_for_each(&node->arc_src_head, arc, src_next) {
		void *in_off;

		if (i == CMD_GRAPH_NODE_MAX_ARC_NUM) {
			HWS_ERR("Max number of in arcs per node is:%d\n",
				   CMD_GRAPH_NODE_MAX_ARC_NUM);
			errno = EINVAL;
			return -errno;
		}

		in_off = (void *)((char *)in_arc + i++ *
				  DEVX_ST_SZ_BYTES(parse_graph_arc));
		if (arc->cfg.src_node.type == MLX5DV_HWS_PARSER_NODE_TYPE_FLEX) {
			DEVX_SET(parse_graph_arc, in_off, parse_graph_node_handle,
				 arc->cfg.src_node.flex.node->obj->object_id);
			DEVX_SET(parse_graph_arc, in_off, arc_parse_graph_node,
				 0x1F); /* node type: PROGRAMMABLE */
		} else {
			DEVX_SET(parse_graph_arc, in_off, arc_parse_graph_node,
				 arc->cfg.src_node.native.node + 2);
		}

		if (arc->cfg.src_node.type == MLX5DV_HWS_PARSER_NODE_TYPE_NATIVE ||
			arc->cfg.src_node.flex.node->cfg.has_selection_field)
			DEVX_SET(parse_graph_arc, in_off, compare_condition_value,
				 arc->cfg.selection_field_data);

		DEVX_SET(parse_graph_arc, in_off, start_inner_tunnel,
			 arc->cfg.is_inner_transition);
	}

	return 0;
}

static int cmd_flex_node_create_set_arc_dst(struct mlx5dv_hws_context *ctx,
					    struct mlx5dv_hws_parser_node *node,
					    void *flex,
					    struct mlx5dv_devx_obj *devx_obj)
{
	struct mlx5dv_hws_parser_arc *arc;
	void *out_arc;
	int i;

	i = 0;
	out_arc = DEVX_ADDR_OF(parse_graph_flex, flex, output_arc);
	list_for_each(&node->arc_dst_head, arc, dst_next) {
		void *out_off;

		if (i == CMD_GRAPH_NODE_MAX_ARC_NUM) {
			HWS_ERR("Max number of out arcs per node is:%d\n",
				CMD_GRAPH_NODE_MAX_ARC_NUM);
			errno = EINVAL;
			return -errno;
		}

		if (arc->cfg.dst_node.type == MLX5DV_HWS_PARSER_NODE_TYPE_FLEX)
			continue;

		out_off = (void *)((char *)out_arc + i++ *
				   DEVX_ST_SZ_BYTES(parse_graph_arc));

		DEVX_SET(parse_graph_arc, out_off, arc_parse_graph_node,
			 arc->cfg.dst_node.native.node + 2);

		if (node->cfg.has_selection_field)
			DEVX_SET(parse_graph_arc, out_off, compare_condition_value,
				 arc->cfg.selection_field_data);

		DEVX_SET(parse_graph_arc, out_off, start_inner_tunnel,
			 arc->cfg.is_inner_transition);
	}

	return 0;
}

static int cmd_flex_node_create_set_sampler(struct mlx5dv_hws_context *ctx,
					    struct mlx5dv_hws_parser_node *node,
					    void *flex,
					    struct mlx5dv_devx_obj *devx_obj)
{
	struct mlx5dv_hws_parser_sampler *sampler;
	void *sample_ext;
	void *sample;
	int i;

	i = 0;
	sample = DEVX_ADDR_OF(parse_graph_flex, flex, sample_table);
	sample_ext = DEVX_ADDR_OF(parse_graph_flex, flex, sample_ext_table);
	list_for_each(&node->sampler_head, sampler, next) {
		void *sample_ext_off;
		void *sample_off;

		if (i == CMD_GRAPH_NODE_MAX_ARC_NUM) {
			HWS_ERR("Max number of samplers per node is:%d\n",
				   CMD_GRAPH_NODE_MAX_SAMPLE_NUM);
			errno = EINVAL;
			return -errno;
		}

		sample_off = (void *)((char *)sample + i *
					  DEVX_ST_SZ_BYTES(parse_graph_flow_match_sample));

		DEVX_SET(parse_graph_flow_match_sample, sample_off,
			 flow_match_sample_tunnel_mode,
			 sampler->cfg.node_encap_level);

		DEVX_SET(parse_graph_flow_match_sample, sample_off,
			 flow_match_sample_offset_mode, 0x0);

		DEVX_SET(parse_graph_flow_match_sample, sample_off,
			 flow_match_sample_field_base_offset,
			 sampler->cfg.field_cfg.bit_offset / BITS_IN_BYTE);

		DEVX_SET(parse_graph_flow_match_sample, sample_off,
			 flow_match_sample_en, true);

		if (!sampler->cfg.sample_from_option)
			goto next_sampler;

		DEVX_SET(parse_graph_flow_match_sample, sample_off,
			 tlv_option_sample, true);

		sample_ext_off = (void *)((char *)sample_ext +
				 i * DEVX_ST_SZ_BYTES(parse_graph_flow_match_sample_ext));

		DEVX_SET(parse_graph_flow_match_sample_ext, sample_ext_off,
			 tlv_option_type, sampler->cfg.option_type);
next_sampler:
		i++;
	}

	return 0;
}

struct mlx5dv_devx_obj *cmd_flex_node_create(struct mlx5dv_hws_context *ctx,
					     struct mlx5dv_hws_parser_node *node)
{
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(create_flex_parser_in)] = {0};
	struct mlx5dv_devx_obj *devx_obj;
	void *flex;
	void *hdr;
	int ret;

	devx_obj = malloc(sizeof(*devx_obj));
	if (!devx_obj) {
		HWS_ERR("Failed to allocate memory for node object");
		errno = ENOMEM;
		return NULL;
	}

	hdr = DEVX_ADDR_OF(create_flex_parser_in, in, hdr);
	flex = DEVX_ADDR_OF(create_flex_parser_in, in, flex);

	DEVX_SET(general_obj_in_cmd_hdr, hdr, opcode,
		 MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr, hdr, obj_type,
		 MLX5_GENERAL_OBJ_TYPE_FLEX_PARSE_GRAPH);
	DEVX_SET(parse_graph_flex, flex, header_length_mode,
		 node->cfg.hdr_len.mode);

	if (node->cfg.hdr_len.mode == MLX5DV_HWS_PARSER_HDR_LEN_MODE_FIELD)
		cmd_flex_node_create_set_mode_length_header(ctx, node, flex);
	else if (node->cfg.hdr_len.mode == MLX5DV_HWS_PARSER_HDR_LEN_MODE_FIXED)
		DEVX_SET(parse_graph_flex, flex, header_length_base_value,
			 node->cfg.hdr_len.fixed.hdr_len);

	if (node->cfg.has_selection_field) {
		DEVX_SET(parse_graph_flex, flex, next_header_field_offset,
			 node->cfg.selection_field.bit_offset);
		DEVX_SET(parse_graph_flex, flex, next_header_field_size,
			 node->cfg.selection_field.bit_length);
	}

	if (node->cfg.has_options)
		cmd_flex_node_create_fill_options(ctx, node, flex);

	if (cmd_flex_node_create_set_arc_src(ctx, node, flex, devx_obj))
		goto out_err;

	if (cmd_flex_node_create_set_arc_dst(ctx, node, flex, devx_obj))
		goto out_err;

	if (cmd_flex_node_create_set_sampler(ctx, node, flex, devx_obj))
		goto out_err;

	devx_obj = mlx5dv_devx_obj_create(ctx->ibv_ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to allocate flex node (syndrome: %#x)",
		       cmd_get_syndrome(out));
		goto out_err;
	}

	if (!list_empty(&node->sampler_head)) {
		ret = cmd_flex_node_query(devx_obj, node);
		if (ret) {
			HWS_ERR("Failed to query node for sampler id");
			goto out_err;
			return NULL;
			cmd_destroy_obj(devx_obj);
		}
	}

	return devx_obj;

out_err:
	free(devx_obj);
	return NULL;
}

struct mlx5dv_devx_obj *
cmd_geneve_tlv_option_create(struct ibv_context *ctx,
			     uint16_t option_class,
			     uint8_t option_type,
			     uint8_t option_data_len,
			     uint8_t option_class_ignore,
			     uint8_t offset_valid,
			     uint8_t sample_offset)
{
	uint32_t in[DEVX_ST_SZ_DW(create_geneve_tlv_option_in)] = {0};
	uint32_t out[DEVX_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	struct mlx5dv_devx_obj *devx_obj;

	void *hdr = DEVX_ADDR_OF(create_geneve_tlv_option_in, in, hdr);
	void *opt = DEVX_ADDR_OF(create_geneve_tlv_option_in, in,
				 geneve_tlv_opt);
	DEVX_SET(general_obj_in_cmd_hdr, hdr, opcode,
		 MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr, hdr, obj_type,
		 MLX5_GENERAL_OBJ_TYPE_GENEVE_TLV_OPT);
	DEVX_SET(geneve_tlv_option, opt, option_type, option_type);
	DEVX_SET(geneve_tlv_option, opt, option_data_length,
		 option_data_len);

	if (option_class_ignore)
		DEVX_SET(geneve_tlv_option, opt, option_class_ignore,
			 option_class_ignore);
	else
		DEVX_SET(geneve_tlv_option, opt, option_class, option_class);

	if (offset_valid) {
		DEVX_SET(geneve_tlv_option, opt, sample_offset_valid,
			 offset_valid);
		DEVX_SET(geneve_tlv_option, opt, sample_offset,
			 sample_offset);
	}

	devx_obj = mlx5dv_devx_obj_create(ctx, in, sizeof(in), out, sizeof(out));
	if (!devx_obj) {
		HWS_ERR("Failed to create GENEVE TLV option (syndrome: %#x)",
			cmd_get_syndrome(out));
		return NULL;
	}

	return devx_obj;
}

int
cmd_geneve_tlv_option_query(struct ibv_context *ctx,
			    struct mlx5dv_devx_obj *devx_obj,
			    uint32_t *sampler_id)
{
	uint32_t out[DEVX_ST_SZ_DW(query_geneve_tlv_option_out)] = {0};
	uint32_t in[DEVX_ST_SZ_DW(general_obj_in_cmd_hdr)] = {0};
	void *hdr, *opt;
	int ret;

	hdr = DEVX_ADDR_OF(query_geneve_tlv_option_out, in, hdr);
	opt  = DEVX_ADDR_OF(query_geneve_tlv_option_out, out, geneve_tlv_opt);

	DEVX_SET(general_obj_in_cmd_hdr, hdr, opcode,
		 MLX5_CMD_OP_QUERY_GENERAL_OBJECT);
	DEVX_SET(general_obj_in_cmd_hdr, hdr, obj_type,
		 MLX5_GENERAL_OBJ_TYPE_GENEVE_TLV_OPT);
	DEVX_SET(general_obj_in_cmd_hdr, hdr, obj_id, devx_obj->object_id);
	/* Call first query to get sample handle. */
	ret = mlx5dv_devx_obj_query(devx_obj, in, sizeof(in), out,
				    sizeof(out));
	if (ret) {
		HWS_ERR("Failed to query GENEVE TLV option (syndrome: %#x)",
			cmd_get_syndrome(out));
		return -errno;
	}

	if (!DEVX_GET(geneve_tlv_option, opt, sample_id_valid)) {
		HWS_LOG("GENEVE TLV option sample isn't valid.");
		errno = EIO;
		return -errno;
	}

	*sampler_id = DEVX_GET(geneve_tlv_option, opt, geneve_sample_field_id);

	return 0;
}
