/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"

bool context_cap_dynamic_reparse(struct mlx5dv_hws_context *ctx)
{
	return IS_BIT_SET(ctx->caps->rtc_reparse_mode, MLX5_IFC_RTC_REPARSE_BY_STC);
}

uint8_t context_get_reparse_mode(struct mlx5dv_hws_context *ctx)
{
	/* Prefer to use dynamic reparse, reparse only specific actions */
	if (context_cap_dynamic_reparse(ctx))
		return MLX5_IFC_RTC_REPARSE_NEVER;

	/* Otherwise use less efficient static */
	return MLX5_IFC_RTC_REPARSE_ALWAYS;
}

int context_query_ib_port(struct mlx5dv_hws_context *ctx,
			  uint32_t ib_port,
			  uint16_t *vport_num,
			  bool *other_vport)
{
	uint16_t esw_vhca_id;
	int ret;

	if (ctx->caps->eswitch_manager && ctx->caps->num_ib_ports == 1) {
		if (ib_port != 1) {
			HWS_ERR("Invalid ib port number [%d]", ib_port);
			errno = EINVAL;
			return -errno;
		}

		/* WA for IB port query fail if not in switch-dev mode. */
		esw_vhca_id = ctx->caps->vhca_id;
		*vport_num = ctx->caps->eswitch_manager_vport_number;
	} else {
		ret = cmd_ib_port_query(ctx->ibv_ctx, ib_port, vport_num, &esw_vhca_id);
		if (ret) {
			HWS_ERR("Failed querying ib-port [%d]", ib_port);
			return ret;
		}
	}

	/* If vport_id and esw_id equal to context this is not other vport */
	*other_vport = !(ctx->caps->eswitch_manager &&
			 ctx->caps->vhca_id == esw_vhca_id &&
			 ctx->caps->eswitch_manager_vport_number == *vport_num);

	/* On multiport-ESW FW allows creating FT only on master FDB */
	if (ctx->caps->vhca_id != esw_vhca_id) {
		HWS_ERR("FW doesn't support IB ports on other ESW");
		errno = ENOTSUP;
		return -errno;
	}

	return 0;
}

struct context_common_res *
context_get_common_res(struct mlx5dv_hws_context *ctx,
		       enum mlx5dv_hws_table_type tbl_type,
		       uint32_t ib_port)
{
	/* Should be called only on slow-path with ctx ctrl lock */
	struct context_obj *ctx_common_res;

	list_for_each(&ctx->common_res, ctx_common_res, next) {
		if (ctx_common_res->tbl_type == tbl_type &&
		    ctx_common_res->ib_port == ib_port)
			return ctx_common_res->obj;
	}

	ctx_common_res = calloc(1, sizeof(*ctx_common_res));
	if (!ctx_common_res) {
		errno = ENOMEM;
		return NULL;
	}

	ctx_common_res->obj = calloc(1, sizeof(struct context_common_res));
	if (!ctx_common_res->obj) {
		errno = ENOMEM;
		goto free_ctx_common_res;
	}

	ctx_common_res->ib_port = ib_port;
	ctx_common_res->tbl_type = tbl_type;
	list_add(&ctx->common_res, &ctx_common_res->next);

	return ctx_common_res->obj;

free_ctx_common_res:
	free(ctx_common_res);
	return NULL;
}

static void
context_free_all_common_res(struct mlx5dv_hws_context *ctx)
{
	struct context_obj *ctx_common_res, *tmp;

	list_for_each_safe(&ctx->common_res, ctx_common_res, tmp, next) {
		list_del(&ctx_common_res->next);
		free(ctx_common_res->obj);
		free(ctx_common_res);
	}
}

struct pool *
context_get_stc_pool(struct mlx5dv_hws_context *ctx,
		     enum mlx5dv_hws_table_type tbl_type,
		     uint32_t ib_port)
{
	/* Should be called only on slow-path with ctx ctrl lock */
	struct pool_attr pool_attr = {0};
	struct context_obj *ctx_pool;
	uint8_t max_log_sz;
	int ret;

	list_for_each(&ctx->stc_pool, ctx_pool, next) {
		if (ctx_pool->tbl_type == tbl_type &&
		    ctx_pool->ib_port == ib_port)
			return ctx_pool->obj;
	}

	ctx_pool = calloc(1, sizeof(*ctx_pool));
	if (!ctx_pool) {
		errno = ENOMEM;
		return NULL;
	}

	if (table_is_rdma_transport(tbl_type)) {
		ret = context_query_ib_port(ctx,
					    ib_port,
					    &pool_attr.vport,
					    &pool_attr.other_vport);
		if (ret) {
			HWS_ERR("Failed to query ib port [%d]", ib_port);
			goto free_ctx_pool;
		}
	}

	pool_attr.table_type = tbl_type;
	pool_attr.pool_type = POOL_TYPE_STC;
	pool_attr.flags = POOL_FLAGS_FOR_STC_POOL;
	max_log_sz = min_t(uint8_t, POOL_STC_LOG_SZ, ctx->caps->stc_alloc_log_max);
	pool_attr.alloc_log_sz = max(max_log_sz, ctx->caps->stc_alloc_log_gran);

	ctx_pool->obj = pool_create(ctx, &pool_attr);
	if (!ctx_pool->obj) {
		HWS_ERR("Failed to allocate STC pool type [%d]", tbl_type);
		errno = ENOMEM;
		goto free_ctx_pool;
	}

	ctx_pool->ib_port = ib_port;
	ctx_pool->tbl_type = tbl_type;
	list_add(&ctx->stc_pool, &ctx_pool->next);

	return ctx_pool->obj;

free_ctx_pool:
	free(ctx_pool);
	return NULL;
}

static void
context_free_all_stc_pools(struct mlx5dv_hws_context *ctx)
{
	struct context_obj *ctx_pool, *tmp_pool;

	list_for_each_safe(&ctx->stc_pool, ctx_pool, tmp_pool, next) {
		list_del(&ctx_pool->next);
		pool_destroy(ctx_pool->obj);
		free(ctx_pool);
	}
}

static int context_pools_init(struct mlx5dv_hws_context *ctx)
{
	int ret;

	ret = pat_cache_init(&ctx->pattern_cache);
	if (ret)
		return ret;

	ret = definer_cache_init(&ctx->definer_cache);
	if (ret)
		goto uninit_pat_cache;

	ret = action_stc_cache_init(&ctx->stc_cache);
	if (ret) {
		HWS_ERR("Failed to allocate STC cache");
		goto uninit_definer_cache;
	}

	return 0;

uninit_definer_cache:
	definer_cache_uninit(ctx->definer_cache);
uninit_pat_cache:
	pat_cache_uninit(ctx->pattern_cache);
	return -errno;
}

static void context_pools_uninit(struct mlx5dv_hws_context *ctx)
{
	context_free_all_common_res(ctx);
	context_free_all_stc_pools(ctx);
	action_stc_cache_uninit(ctx->stc_cache);
	definer_cache_uninit(ctx->definer_cache);
	pat_cache_uninit(ctx->pattern_cache);
}

static void context_check_hws_supp(struct mlx5dv_hws_context *ctx)
{
	struct cmd_query_caps *caps = ctx->caps;

	/* HWS not supported on device / FW */
	if (!caps->wqe_based_update) {
		HWS_DBG("Required HWS WQE based insertion cap not supported");
		return;
	}

	/* Current solution requires all rules to set reparse bit */
	if ((!caps->nic_ft.reparse ||
	     (!caps->fdb_ft.reparse && caps->eswitch_manager)) ||
	    !IS_BIT_SET(caps->rtc_reparse_mode, MLX5_IFC_RTC_REPARSE_ALWAYS)) {
		HWS_DBG("Required HWS reparse cap not supported");
		return;
	}

	/* FW/HW must support 8DW STE */
	if (!IS_BIT_SET(caps->ste_format, MLX5_IFC_RTC_STE_FORMAT_8DW)) {
		HWS_DBG("Required HWS STE format not supported");
		return;
	}

	/* Adding rules by hash and by offset are requirements */
	if (!IS_BIT_SET(caps->rtc_index_mode,
			MLX5_IFC_RTC_STE_UPDATE_MODE_BY_HASH) ||
	    !IS_BIT_SET(caps->rtc_index_mode,
			MLX5_IFC_RTC_STE_UPDATE_MODE_BY_OFFSET)) {
		HWS_DBG("Required HWS RTC update mode not supported");
		return;
	}

	/* Support for SELECT definer ID is required */
	if (!IS_BIT_SET(caps->definer_format_sup,
			MLX5_IFC_DEFINER_FORMAT_ID_SELECT)) {
		HWS_DBG("Required HWS Dynamic definer not supported");
		return;
	}

	ctx->flags |= CONTEXT_FLAG_HWS_SUPPORT;
}

static int context_set_pdn(struct mlx5dv_hws_context *ctx,
			   struct mlx5dv_hws_context_attr *attr)
{
	struct mlx5dv_pd mlx5_pd = {0};
	struct mlx5dv_obj obj;
	int ret;

	if (!attr->pd) {
		HWS_ERR("PD object must be provided");
		errno = EINVAL;
		return -errno;
	}

	obj.pd.in = attr->pd;
	obj.pd.out = &mlx5_pd;

	ret = mlx5dv_init_obj(&obj, MLX5DV_OBJ_PD);
	if (ret) {
		HWS_ERR("Failed to get PDN from PD");
		return ret;
	}

	ctx->pd_num = mlx5_pd.pdn;

	return 0;
}

static int context_init_hws(struct mlx5dv_hws_context *ctx,
			    struct mlx5dv_hws_context_attr *attr)
{
	int ret;

	context_check_hws_supp(ctx);

	if (!(ctx->flags & CONTEXT_FLAG_HWS_SUPPORT))
		return 0;

	ret = context_set_pdn(ctx, attr);
	if (ret)
		return ret;

	ret = context_pools_init(ctx);
	if (ret)
		return ret;

	ret = send_queues_open(ctx, attr->queues, attr->queue_size);
	if (ret)
		goto pools_uninit;

	return 0;

pools_uninit:
	context_pools_uninit(ctx);
	return ret;
}

static void context_uninit_hws(struct mlx5dv_hws_context *ctx)
{
	if (!(ctx->flags & CONTEXT_FLAG_HWS_SUPPORT))
		return;

	send_queues_close(ctx);
	context_pools_uninit(ctx);
}

static int context_set_shared_ctx(struct mlx5dv_hws_context *ctx,
				  struct ibv_context *ibv_ctx,
				  struct mlx5dv_hws_context_attr *attr)
{
	struct cmd_query_caps shared_caps = {0};
	int ret;

	if (!attr->shared_ibv_ctx) {
		/* In case no resource sharing:
		 * -ibv_ctx: the port context
		 * -local_ibv_ctx: NULL
		 */
		ctx->ibv_ctx = ibv_ctx;
		ctx->local_ibv_ctx = NULL;
	} else {
		/* In case of resource sharing:
		 * -ibv_ctx: shared context
		 * -local_ibv_ctx: the port context
		 */
		ctx->ibv_ctx = attr->shared_ibv_ctx;
		ctx->local_ibv_ctx = ibv_ctx;

		/* Check shared ibv_ctx supports resource sharing */
		ret = cmd_query_caps(attr->shared_ibv_ctx, &shared_caps);
		if (ret || !shared_caps.cross_vhca_resources) {
			HWS_ERR("No cross_vhca_resources cap for shared ibv");
			errno = ENOTSUP;
			return -errno;
		}
		ctx->caps->shared_vhca_id = shared_caps.vhca_id;

		/* Check local ibv_ctx supports resource sharing */
		if (ctx->local_ibv_ctx && !ctx->caps->cross_vhca_resources) {
			HWS_ERR("No cross_vhca_resources cap for local ibv");
			errno = ENOTSUP;
			return -errno;
		}
	}

	return 0;
}

struct mlx5dv_hws_context *
mlx5dv_hws_context_open(struct ibv_context *ibv_ctx,
			struct mlx5dv_hws_context_attr *attr)
{
	struct mlx5dv_hws_context *ctx;

	if (attr->comp_mask) {
		HWS_ERR("Invalid comp mask provided");
		errno = ENOTSUP;
		return NULL;
	}

	ctx = calloc(1, sizeof(*ctx));
	if (!ctx) {
		errno = ENOMEM;
		return NULL;
	}

	list_head_init(&ctx->head);
	list_head_init(&ctx->stc_pool);
	list_head_init(&ctx->common_res);
	pthread_spin_init(&ctx->ctrl_lock, PTHREAD_PROCESS_PRIVATE);
	pthread_spin_init(&ctx->stc_lock, PTHREAD_PROCESS_PRIVATE);
	ctx->caps = calloc(1, sizeof(*ctx->caps));
	if (!ctx->caps)
		goto free_ctx;

	if (cmd_query_caps(ibv_ctx, ctx->caps))
		goto free_caps;

	if (context_set_shared_ctx(ctx, ibv_ctx, attr))
		goto free_caps;

	if (context_init_hws(ctx, attr))
		goto free_caps;

	return ctx;

free_caps:
	free(ctx->caps);
free_ctx:
	pthread_spin_destroy(&ctx->stc_lock);
	pthread_spin_destroy(&ctx->ctrl_lock);
	free(ctx);
	return NULL;
}

int mlx5dv_hws_context_close(struct mlx5dv_hws_context *ctx)
{
	context_uninit_hws(ctx);
	free(ctx->caps);
	pthread_spin_destroy(&ctx->stc_lock);
	pthread_spin_destroy(&ctx->ctrl_lock);
	free(ctx);
	return 0;
}
