/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"
#include "mlx5dv_hws_ct.h"

#define CT_MATCHER_V4_FIELD_NUM 7
#define CT_MATCHER_V6_FIELD_NUM 13
#define CT_MATCH_STE_IPV4	0x4
#define CT_MATCH_STE_IPV6	0x6

#define CT_SET(p, v, bit_off, bit_sz) \
do { \
	uint32_t _v = v; \
	uint32_t _m = __devx_mask(bit_sz); \
	uint32_t _dw_bit_off = (BITS_IN_DW - bit_sz - bit_off % BITS_IN_BYTE); \
	__be32 *_p = (__be32 *)((uint8_t *)(p) + ((bit_off) / BITS_IN_BYTE)); \
	*_p = htobe32((be32toh(*(_p)) & (~((_m) << (_dw_bit_off)))) | ((_v & _m) << _dw_bit_off)); \
} while (0)

#define CT_SET_BE32(p, v, bit_off) \
	(*((__be32 *)((uint8_t *)(p) + (bit_off / BITS_IN_BYTE))) = (v))

#define CT_SET_BE16(p, v, bit_off) \
	(*((__be16 *)((uint8_t *)(p) + (bit_off / BITS_IN_BYTE))) = (v))

#define CT_SET_BYTE(p, v, bit_off) \
	(*((uint8_t *)(p) + (bit_off / BITS_IN_BYTE)) = (v))

static void
ct_rule_apply_modify_action(struct mlx5dv_hws_ct_action_data *actions_data,
			    struct mlx5dv_hws_table *tbl,
			    struct send_engine *queue,
			    struct send_wqe_gta_ctrl_seg *wqe_ctrl,
			    struct send_wqe_gta_data_seg_ste *wqe_data,
			    struct send_engine_post_attr *send_attr,
			    uint8_t reverse,
			    uint8_t tbl_type)
{
	enum mlx5dv_hws_resource_type res_type;
	__be32 *stc_ix = wqe_ctrl->stc_ix;
	uint32_t modify_value;
	uint32_t mh_arg_id;
	uint32_t arg_gran;
	__be32 modify_stc;
	uint8_t *data;

	/* Init like if we have double action */
	stc_ix[ACTION_STC_IDX_DW7] = 0;

	switch (actions_data->mh.mode) {
	case MLX5DV_HWS_CT_MH_MODE_DATA_OPTIMIZED:
		if (reverse)
			modify_stc = actions_data->mh.action_reverse->stc->offset;
		else
			modify_stc = actions_data->mh.action->stc->offset;

		modify_value = reverse ? actions_data->mh.info.optimized.modify_reverse_value :
				actions_data->mh.info.optimized.modify_value;

		/* Modify value comes from user in BE32. */
		((uint32_t *)wqe_data)[ACTION_OFFSET_DW7] = modify_value;
		break;
	case MLX5DV_HWS_CT_MH_MODE_DATA_PRE_WRITTEN:
		if (reverse) {
			/* Assume one resource only */
			res_type = actions_data->mh.action_reverse->modify_header.resource[0]->type;
			arg_gran = resource_arg_gran_get(res_type);
			modify_stc = actions_data->mh.action_reverse->stc->offset;
			modify_value =
			actions_data->mh.info.data.modify_reverse_offset * arg_gran;
		} else {
			/* Assume one resource only */
			res_type = actions_data->mh.action->modify_header.resource[0]->type;
			arg_gran = resource_arg_gran_get(res_type);
			modify_stc = actions_data->mh.action->stc->offset;

			modify_value = actions_data->mh.info.data.modify_offset * arg_gran;
		}

		((uint32_t *)wqe_data)[ACTION_OFFSET_DW7] = htobe32(modify_value);
		break;
	case MLX5DV_HWS_CT_MH_MODE_DATA_INLINE:
		if (reverse) {
			/* Assume one resource only */
			res_type = actions_data->mh.action_reverse->modify_header.resource[0]->type;
			arg_gran = resource_arg_gran_get(res_type);
			modify_stc = actions_data->mh.action_reverse->stc->offset;
			modify_value = actions_data->mh.info.data.modify_reverse_offset * arg_gran;

			mh_arg_id =
			actions_data->mh.action_reverse->modify_header.resource[0]->devx_obj->object_id;
			data = actions_data->mh.info.data.modify_reverse_data;
		} else {
			/* Assume one resource only */
			res_type = actions_data->mh.action->modify_header.resource[0]->type;
			arg_gran = resource_arg_gran_get(res_type);
			modify_stc = actions_data->mh.action->stc->offset;
			modify_value = actions_data->mh.info.data.modify_offset * arg_gran;

			mh_arg_id =
			actions_data->mh.action->modify_header.resource[0]->devx_obj->object_id;
			data = actions_data->mh.info.data.modify_data;
		}
		/* Write the data on the same queue as the rule */
		arg_write(queue, NULL, mh_arg_id + modify_value,
			  data, actions_data->mh.info.data.data_size);

		((uint32_t *)wqe_data)[ACTION_OFFSET_DW7] = htobe32(modify_value);
		break;
	case MLX5DV_HWS_CT_MH_MODE_DATA_NONE:
	default:
		modify_stc = htobe32(tbl->common_res->default_stc->nop_dw6.offset);
		/* Instead of double, 2 singles */
		stc_ix[ACTION_STC_IDX_DW7] =
			htobe32(tbl->common_res->default_stc->nop_dw7.offset);
		((uint32_t *)wqe_data)[ACTION_OFFSET_DW7] = 0;
		break;
	}
	/* The bind function already moved it to be32 */
	stc_ix[ACTION_STC_IDX_DW6] = modify_stc;
}

static void
ct_rule_actions_apply(struct  mlx5dv_hws_matcher *matcher,
		      struct mlx5dv_hws_ct_action_data *actions_data,
		      struct send_engine *queue,
		      struct send_wqe_gta_ctrl_seg *wqe_ctrl,
		      struct send_wqe_gta_data_seg_ste *wqe_data,
		      struct send_engine_post_attr *send_attr,
		      bool reverse)
{
	struct mlx5dv_hws_table *tbl = matcher->tbl;
	uint8_t tbl_type = matcher->tbl->type;
	__be32 *stc_ix = wqe_ctrl->stc_ix;

	ct_rule_apply_modify_action(actions_data, tbl, queue,
				    wqe_ctrl, wqe_data, send_attr,
				    reverse, tbl_type);

	/* Assume dirty wqe_ctrl and clean wqe_data, and
	 * that bind function already moved the action stc offset to be32, that
	 * was done for actions tag counter and modify-header only.
	 */
	stc_ix[ACTION_STC_IDX_DW5] =
		actions_data->tag.action ?
		actions_data->tag.action->stc->offset :
		htobe32(tbl->common_res->default_stc->nop_dw5.offset);
	stc_ix[ACTION_STC_IDX_HIT] =
		reverse ? htobe32(actions_data->dest_table_reverse->stc->offset) :
			  htobe32(actions_data->dest_table->stc->offset);
	stc_ix[ACTION_STC_IDX_CTRL] =
		(actions_data->ctr.action ?
		actions_data->ctr.action->stc->offset :
		htobe32(tbl->common_res->default_stc->nop_ctr.offset)) |
			htobe32(stc_ix[ACTION_STC_IDX_DW7] ?
				ACTION_STC_IDX_LAST_COMBO2 << 29 :
				ACTION_STC_IDX_LAST_COMBO1 << 29);

	((uint32_t *)wqe_data)[ACTION_OFFSET_DW0] = htobe32(actions_data->ctr.offset);
	/*
	 * Assume tag value is 0 when no tag action.
	 * User provide tag value in BE32.
	 */
	((uint32_t *)wqe_data)[ACTION_OFFSET_DW5] = actions_data->tag.tag_value;
}

static void ct_rule_v6_send_actions(struct mlx5dv_hws_matcher *matcher,
				    struct send_engine *queue,
				    struct mlx5dv_hws_rule *rule,
				    struct mlx5dv_hws_ct_action_data *actions_data,
				    uint32_t cur_rule_idx,
				    bool is_reverse)
{
	struct send_engine_post_attr send_attr = {0};
	struct send_wqe_gta_data_seg_ste *wqe_data;
	struct send_wqe_gta_ctrl_seg *wqe_ctrl;
	struct send_engine_post_ctrl ctrl;
	size_t wqe_len_wr;

	ctrl = send_engine_post_start(queue);
	send_engine_post_req_wqe(&ctrl, (void *)&wqe_ctrl, &send_attr.len);
	send_engine_post_req_wqe(&ctrl, (void *)&wqe_data, &wqe_len_wr);
	memset(wqe_data, 0, wqe_len_wr);

	send_attr.opmod = SEND_WQE_GTA_OPMOD_STE;
	send_attr.opcode = SEND_WQE_OPCODE_TBL_ACCESS;
	send_attr.len += wqe_len_wr;
	send_attr.notify_hw = 0;
	send_attr.user_data = NULL;
	send_attr.id = matcher->action_ste.rtc->object_id;

	wqe_ctrl->op_dirix =
		htobe32(SEND_WQE_GTA_OP_ACTIVATE << 28 | (cur_rule_idx + !!is_reverse));

	/* Set actions */
	ct_rule_actions_apply(matcher, actions_data, queue, wqe_ctrl, wqe_data,
			      &send_attr, is_reverse);

	/* Send the wqe's now */
	send_engine_post_end(&ctrl, &send_attr);
}

static void
ct_rule_v6_set_ip(struct definer_fc **fc, uint8_t *addr, uint8_t *tag)
{
	CT_SET_BE32(tag, *(__be32 *)addr, (*fc)->bit_off);
	(*fc)++;
	CT_SET_BE32(tag, *(__be32 *)(addr + 4), (*fc)->bit_off);
	(*fc)++;
	CT_SET_BE32(tag, *(__be32 *)(addr + 8), (*fc)->bit_off);
	(*fc)++;
	CT_SET_BE32(tag, *(__be32 *)(addr + 12), (*fc)->bit_off);
}

static void
ct_rule_v6_set_tag(struct mlx5dv_hws_mt *mt,
		   struct mlx5dv_hws_ctv6_match *match_val,
		   struct send_wqe_gta_data_seg_ste *wqe_data,
		   uint8_t reverse)
{
	uint8_t *tag = (uint8_t *)wqe_data->action;
	struct definer_fc *fc = mt->fcm;

	CT_SET_BE32(tag, match_val->metadata, fc->bit_off);
	fc++;

	CT_SET(tag, CT_MATCH_STE_IPV6, fc->bit_off, fc->bit_sz);
	fc++;

	CT_SET_BYTE(tag, match_val->protocol, fc->bit_off);
	fc++;

	if (reverse) {
		ct_rule_v6_set_ip(&fc, match_val->dst_addr, tag);
		fc++;

		ct_rule_v6_set_ip(&fc, match_val->src_addr, tag);
		fc++;

		CT_SET_BE16(tag, match_val->dst_port, fc->bit_off);
		fc++;

		CT_SET_BE16(tag, match_val->src_port, fc->bit_off);
		fc++;
	} else {
		ct_rule_v6_set_ip(&fc, match_val->src_addr, tag);
		fc++;

		ct_rule_v6_set_ip(&fc, match_val->dst_addr, tag);
		fc++;

		CT_SET_BE16(tag, match_val->src_port, fc->bit_off);
		fc++;

		CT_SET_BE16(tag, match_val->dst_port, fc->bit_off);
	}
}

static void
ct_rule_v6_match_apply(struct mlx5dv_hws_matcher *matcher,
		       struct send_wqe_gta_ctrl_seg *wqe_ctrl,
		       struct send_wqe_gta_data_seg_ste *wqe_data,
		       uint32_t rule_idx)
{
	struct mlx5dv_hws_table *tbl = matcher->tbl;
	__be32 *stc_ix = wqe_ctrl->stc_ix;

	/* First jumbo match ste (assume dirty wqe_ctrl and clean wqe_data)
	 * STC for DW5-7 not set since this a jumbo STE, these DW hold match.
	 */
	stc_ix[ACTION_STC_IDX_DW5] = 0;
	stc_ix[ACTION_STC_IDX_DW6] = 0;
	stc_ix[ACTION_STC_IDX_DW7] = 0;
	stc_ix[ACTION_STC_IDX_HIT] = htobe32(matcher->action_ste.stc.offset);
	stc_ix[ACTION_STC_IDX_CTRL] = htobe32(tbl->common_res->default_stc->nop_ctr.offset) |
		htobe32(ACTION_STC_IDX_LAST_JUMBO_STE << 29);

	((uint32_t *)wqe_data)[ACTION_OFFSET_HIT_LSB] = htobe32(rule_idx << 6);
}

static int ct_rule_v6_create(struct mlx5dv_hws_matcher *matcher,
			     struct mlx5dv_hws_rule *rule,
			     struct mlx5dv_hws_ct_match *match_val,
			     struct mlx5dv_hws_ct_action_data *actions_data,
			     struct mlx5dv_hws_ct_rule_attr *ct_attr,
			     struct mlx5dv_hws_rule_attr *attr,
			     bool is_update)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	struct mlx5dv_hws_mt *mt = &rule->matcher->mt[0]; /* Only one mt */
	struct send_ste_attr ste_attr = {0};
	struct send_ring_dep_wqe *dep_wqe;
	struct send_engine *queue;
	uint32_t cur_rule_idx;

	cur_rule_idx = attr->rule_idx *
		((matcher->flags & MATCHER_FLAGS_CT_ONE_DIRECTION) ? 1 : 2);

	queue = &ctx->send_queue[attr->queue_id];
	if (unlikely(send_engine_err(queue))) {
		errno = EIO;
		return -errno;
	}

	if (unlikely(send_engine_full(queue))) {
		errno = EBUSY;
		return -errno;
	}

	/* Set and send actions for CT ipv6 */
	ct_rule_v6_send_actions(matcher, queue, rule, actions_data,
				cur_rule_idx, false);

	rule_create_init_ste_attr(rule, &ste_attr, is_update);
	dep_wqe = send_add_new_dep_wqe(queue);
	rule_init_dep_wqe(dep_wqe, rule, attr, is_update);
	memset(&dep_wqe->wqe_data, 0, SEND_WQE_SZ_GTA_DATA);

	ste_attr.wqe_ctrl = &dep_wqe->wqe_ctrl;
	ste_attr.wqe_data = &dep_wqe->wqe_data;

	ct_rule_v6_set_tag(mt, &match_val->ctv6, ste_attr.wqe_data, false);
	ct_rule_v6_match_apply(matcher, ste_attr.wqe_ctrl, ste_attr.wqe_data,
			       cur_rule_idx);

	send_engine_inc_entry(queue);

	if (ct_attr->bi_direction) {
		assert(!(matcher->flags & MATCHER_FLAGS_CT_ONE_DIRECTION));
		/* Set and send actions for CT ipv6 */
		ct_rule_v6_send_actions(matcher, queue, rule, actions_data,
					cur_rule_idx, true);

		/* Insert CTv6 rule B -> A */
		dep_wqe = send_add_new_dep_wqe(queue);
		rule_init_dep_wqe(dep_wqe, rule, attr, is_update);
		memset(&dep_wqe->wqe_data, 0, SEND_WQE_SZ_GTA_DATA);

		if (is_update)
			dep_wqe->rtc = rule->rtc_ext;

		dep_wqe->used_id = &rule->rtc_ext;

		ste_attr.wqe_ctrl = &dep_wqe->wqe_ctrl;
		ste_attr.wqe_data = &dep_wqe->wqe_data;

		ct_rule_v6_set_tag(mt, &match_val->ctv6, ste_attr.wqe_data, true);
		ct_rule_v6_match_apply(matcher, ste_attr.wqe_ctrl, ste_attr.wqe_data,
				       cur_rule_idx + 1);
	}

	/* Send dependent WQEs */
	if (!attr->burst)
		send_all_dep_wqe(queue);

	return 0;
}

static void
ct_rule_v4_set_tag(struct mlx5dv_hws_mt *mt,
		   struct mlx5dv_hws_ctv4_match *match_val,
		   struct send_wqe_gta_data_seg_ste *wqe_data,
		   uint8_t reverse)
{
	uint8_t *tag = (uint8_t *)wqe_data->action;
	struct definer_fc *fc = mt->fcm;

	CT_SET_BE32(tag, match_val->metadata, fc->bit_off);
	fc++;

	CT_SET(tag, CT_MATCH_STE_IPV4, fc->bit_off, fc->bit_sz);
	fc++;

	CT_SET_BYTE(tag, match_val->protocol, fc->bit_off);
	fc++;

	if (reverse) {
		CT_SET_BE32(tag, match_val->dst_addr, fc->bit_off);
		fc++;

		CT_SET_BE32(tag, match_val->src_addr, fc->bit_off);
		fc++;

		CT_SET_BE16(tag, match_val->dst_port, fc->bit_off);
		fc++;

		CT_SET_BE16(tag, match_val->src_port, fc->bit_off);
		fc++;
	} else {
		CT_SET_BE32(tag, match_val->src_addr, fc->bit_off);
		fc++;

		CT_SET_BE32(tag, match_val->dst_addr, fc->bit_off);
		fc++;

		CT_SET_BE16(tag, match_val->src_port, fc->bit_off);
		fc++;

		CT_SET_BE16(tag, match_val->dst_port, fc->bit_off);
	}
}

static int ct_rule_v4_create(struct mlx5dv_hws_matcher *matcher,
			     struct mlx5dv_hws_rule *rule,
			     struct mlx5dv_hws_ct_match *match_val,
			     struct mlx5dv_hws_ct_action_data *actions_data,
			     struct mlx5dv_hws_ct_rule_attr *ct_attr,
			     struct mlx5dv_hws_rule_attr *attr,
			     bool is_update)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	struct mlx5dv_hws_mt *mt = &rule->matcher->mt[0]; /* Only one mt */
	struct send_ste_attr ste_attr = {0};
	struct send_ring_dep_wqe *dep_wqe;
	struct send_engine *queue;
	bool need_dep_wqe;
	uint8_t notify_hw;

	need_dep_wqe = (actions_data->mh.mode == MLX5DV_HWS_CT_MH_MODE_DATA_INLINE);

	queue = &ctx->send_queue[attr->queue_id];
	if (unlikely(send_engine_err(queue))) {
		errno = EIO;
		return -errno;
	}

	if (unlikely(send_engine_full(queue))) {
		errno = EBUSY;
		return -errno;
	}

	rule_create_init_ste_attr(rule, &ste_attr, is_update);
	dep_wqe = send_add_new_dep_wqe(queue);
	rule_init_dep_wqe(dep_wqe, rule, attr, is_update);
	memset(&dep_wqe->wqe_data, 0, SEND_WQE_SZ_GTA_DATA);

	ste_attr.wqe_ctrl = &dep_wqe->wqe_ctrl;
	ste_attr.wqe_data = &dep_wqe->wqe_data;

	ct_rule_actions_apply(matcher, actions_data, queue,
			      ste_attr.wqe_ctrl, ste_attr.wqe_data,
			      &ste_attr.send_attr, false);

	ct_rule_v4_set_tag(mt, &match_val->ctv4, ste_attr.wqe_data, false);

	/* Notify if requested and no dep wqes to avoid extra doorbell */
	notify_hw = !attr->burst && !need_dep_wqe;

	if (!need_dep_wqe) {
		send_abort_new_dep_wqe(queue);
		ste_attr.send_attr.notify_hw = notify_hw && !(ct_attr->bi_direction);
		ste_attr.send_attr.user_data = dep_wqe->user_data;
		ste_attr.send_attr.rule = dep_wqe->rule;
		ste_attr.send_attr.used_id = &rule->rtc;
		ste_attr.send_attr.id = dep_wqe->rtc;
		ste_attr.send_attr.retry_id = dep_wqe->retry_rtc;
		send_ste(queue, &ste_attr);
	}

	send_engine_inc_entry(queue);

	if (ct_attr->bi_direction) {
		/* Insert CTv4 rule B -> A */
		dep_wqe = send_add_new_dep_wqe(queue);
		rule_init_dep_wqe(dep_wqe, rule, attr, is_update);
		memset(&dep_wqe->wqe_data, 0, SEND_WQE_SZ_GTA_DATA);

		if (is_update)
			dep_wqe->rtc = rule->rtc_ext;

		dep_wqe->used_id = &rule->rtc_ext;
		ste_attr.wqe_ctrl = &dep_wqe->wqe_ctrl;
		ste_attr.wqe_data = &dep_wqe->wqe_data;

		ct_rule_actions_apply(matcher, actions_data, queue,
				      ste_attr.wqe_ctrl, ste_attr.wqe_data,
				      &ste_attr.send_attr, true);

		ct_rule_v4_set_tag(mt, &match_val->ctv4, ste_attr.wqe_data, true);

		if (!need_dep_wqe) {
			send_abort_new_dep_wqe(queue);
			ste_attr.send_attr.notify_hw = notify_hw;
			ste_attr.send_attr.used_id = &rule->rtc_ext;
			send_ste(queue, &ste_attr);
		}
	}

	/* Send dependent WQEs */
	if (!attr->burst)
		send_all_dep_wqe(queue);

	return 0;
}

int mlx5dv_hws_ct_rule_create(struct mlx5dv_hws_matcher *matcher,
			      struct mlx5dv_hws_ct_match *match_val,
			      struct mlx5dv_hws_ct_action_data *actions_data,
			      struct mlx5dv_hws_ct_rule_attr *ct_attr,
			      struct mlx5dv_hws_rule *rule)
{
	struct mlx5dv_hws_rule_attr *attr = ct_attr->attr;

	rule->matcher = matcher;

	if (matcher->flags & MATCHER_FLAGS_CT_V4)
		return ct_rule_v4_create(matcher,
					 rule,
					 match_val,
					 actions_data,
					 ct_attr,
					 attr,
					 false);

	else if (matcher->flags & MATCHER_FLAGS_CT_V6)
		return ct_rule_v6_create(matcher,
					 rule,
					 match_val,
					 actions_data,
					 ct_attr,
					 attr,
					 false);

	return -ENOTSUP;
}

int mlx5dv_hws_ct_rule_update(struct mlx5dv_hws_matcher *matcher,
			      struct mlx5dv_hws_ct_match *match_val,
			      struct mlx5dv_hws_ct_action_data *actions_data,
			      struct mlx5dv_hws_ct_rule_attr *ct_attr,
			      struct mlx5dv_hws_rule *rule)
{
	struct mlx5dv_hws_rule_attr *attr = ct_attr->attr;

	if (unlikely(rule->status != RULE_STATUS_CREATED)) {
		HWS_ERR("Current rule status does not allow update");
		errno = EBUSY;
		return -errno;
	}

	rule->matcher = matcher;
	if (matcher->flags & MATCHER_FLAGS_CT_V4)
		return ct_rule_v4_create(matcher,
					 rule,
					 match_val,
					 actions_data,
					 ct_attr,
					 attr,
					 true);

	else if (matcher->flags & MATCHER_FLAGS_CT_V6)
		return ct_rule_v6_create(matcher,
					 rule,
					 match_val,
					 actions_data,
					 ct_attr,
					 attr,
					 true);

	return -ENOTSUP;
}

static void ct_rule_destroy_set_tag(struct mlx5dv_hws_rule *rule,
				    struct mlx5dv_hws_mt *mt,
				    struct mlx5dv_hws_ct_match *match_val,
				    struct send_wqe_gta_data_seg_ste *wqe_data,
				    bool is_revers)
{
	if (rule->matcher->flags & MATCHER_FLAGS_CT_V4)
		ct_rule_v4_set_tag(mt, &match_val->ctv4, wqe_data, is_revers);
	else
		ct_rule_v6_set_tag(mt, &match_val->ctv6, wqe_data, is_revers);
}

int mlx5dv_hws_ct_rule_destroy(struct mlx5dv_hws_rule *rule,
			       struct mlx5dv_hws_ct_match *match_val,
			       struct mlx5dv_hws_ct_rule_attr *ct_attr)
{
	struct mlx5dv_hws_context *ctx = rule->matcher->tbl->ctx;
	struct mlx5dv_hws_rule_attr *attr = ct_attr->attr;
	struct mlx5dv_hws_mt *mt = &rule->matcher->mt[0]; /* Only one mt */
	struct send_wqe_gta_data_seg_ste wqe_data = {0};
	struct send_wqe_gta_ctrl_seg wqe_ctrl = {0};
	struct send_ste_attr ste_attr = {0};
	struct send_engine *queue;

	queue = &ctx->send_queue[attr->queue_id];

	if (unlikely(send_engine_full(queue))) {
		errno = EBUSY;
		return -errno;
	}

	if (unlikely(send_engine_err(queue))) {
		rule_destroy_failed_hws(rule, attr);
		return 0;
	}

	/* Rule failed and doesn't require cleanup */
	if (rule->status == RULE_STATUS_FAILED) {
		rule_destroy_failed_hws(rule, attr);
		return 0;
	}

	/* Rule is not completed yet */
	if (rule->status != RULE_STATUS_CREATED &&
	    rule->status != RULE_STATUS_FAILING) {
		errno = EBUSY;
		return -errno;
	}

	send_engine_inc_entry(queue);

	/* Send dependent WQE */
	if (!attr->burst)
		send_all_dep_wqe(queue);

	rule->status = RULE_STATUS_DELETING;

	ste_attr.send_attr.opmod = SEND_WQE_GTA_OPMOD_STE;
	ste_attr.send_attr.opcode = SEND_WQE_OPCODE_TBL_ACCESS;
	ste_attr.send_attr.len = SEND_WQE_SZ_GTA_CTRL + SEND_WQE_SZ_GTA_DATA;

	ste_attr.send_attr.rule = rule;
	ste_attr.send_attr.notify_hw = !attr->burst && !ct_attr->bi_direction;
	ste_attr.send_attr.user_data = attr->user_data;
	ste_attr.send_attr.id = rule->rtc;
	ste_attr.send_attr.used_id = &rule->rtc;

	ste_attr.wqe_ctrl = &wqe_ctrl;
	ste_attr.wqe_data = &wqe_data;
	ste_attr.gta_opcode = SEND_WQE_GTA_OP_DEACTIVATE;

	ct_rule_destroy_set_tag(rule, mt, match_val, ste_attr.wqe_data, false);

	send_ste(queue, &ste_attr);

	if (ct_attr->bi_direction) {
		/* Remove CTv4 rule B -> A */
		memset(&wqe_data, 0, sizeof(struct send_wqe_gta_data_seg_ste));
		ct_rule_destroy_set_tag(rule, mt, match_val, ste_attr.wqe_data, true);
		ste_attr.send_attr.notify_hw = !attr->burst;
		ste_attr.send_attr.id = rule->rtc_ext;
		ste_attr.send_attr.used_id = &rule->rtc_ext;
		send_ste(queue, &ste_attr);
	}

	return 0;
}

static bool ct_matcher_is_metadata_field(struct definer_fc *fc)
{
	if (fc->ctrl.fname >= MLX5DV_HWS_FNAME_REG_C0 &&
	    fc->ctrl.fname <= MLX5DV_HWS_FNAME_REG_C11)
		return true;

	return false;
}

/* The function arrange the matching fields in certain order.
 * The first is the metadata field and then coming the rest.
 * When wew will insert rule, the values will be set according to that fixed order.
 */
static int
ct_matcher_verify_and_set_match(struct mlx5dv_hws_mt *mt, uint8_t num_of_fields)
{
	enum mlx5dv_hws_field_name ctv4_fields[CT_MATCHER_V4_FIELD_NUM] = {
						 MLX5DV_HWS_FNAME_NONE,
						 MLX5DV_HWS_FNAME_IP_VERSION,
						 MLX5DV_HWS_FNAME_IP_PROTOCOL,
						 MLX5DV_HWS_FNAME_IPV4_SRC,
						 MLX5DV_HWS_FNAME_IPV4_DST,
						 MLX5DV_HWS_FNAME_IP_SPORT,
						 MLX5DV_HWS_FNAME_IP_DPORT};

	enum mlx5dv_hws_field_name ctv6_fields[CT_MATCHER_V6_FIELD_NUM] = {
						 MLX5DV_HWS_FNAME_NONE,
						 MLX5DV_HWS_FNAME_IP_VERSION,
						 MLX5DV_HWS_FNAME_IP_PROTOCOL,
						 MLX5DV_HWS_FNAME_IPV6_SRC_127_96,
						 MLX5DV_HWS_FNAME_IPV6_SRC_95_64,
						 MLX5DV_HWS_FNAME_IPV6_SRC_63_32,
						 MLX5DV_HWS_FNAME_IPV6_SRC_31_0,
						 MLX5DV_HWS_FNAME_IPV6_DST_127_96,
						 MLX5DV_HWS_FNAME_IPV6_DST_95_64,
						 MLX5DV_HWS_FNAME_IPV6_DST_63_32,
						 MLX5DV_HWS_FNAME_IPV6_DST_31_0,
						 MLX5DV_HWS_FNAME_IP_SPORT,
						 MLX5DV_HWS_FNAME_IP_DPORT};
	enum mlx5dv_hws_field_name *fields;
	struct definer_fc tmp_fc;
	bool found = false;
	int i, j;

	if (num_of_fields == CT_MATCHER_V4_FIELD_NUM)
		fields = ctv4_fields;
	else
		fields = ctv6_fields;

	if (mt->fcm_sz != CT_MATCHER_V4_FIELD_NUM &&
	    mt->fcm_sz != CT_MATCHER_V6_FIELD_NUM) {
		HWS_ERR("The num of the fields (%d) is not fit CTV4 or CTV6",
			mt->fcm_sz);
		errno = ENOTSUP;
		return errno;
	}

	if (mt->fcc_sz || mt->fcr_sz) {
		HWS_ERR("Range (%d) or compare (%d) is not supported in CT",
			mt->fcr_sz, mt->fcc_sz);
		errno = ENOTSUP;
		return errno;
	}

	/* Put metadata in the first fc */
	for (i = 0; i < num_of_fields; i++) {
		if (ct_matcher_is_metadata_field(&mt->fcm[i])) {
			tmp_fc = mt->fcm[0];
			mt->fcm[0] = mt->fcm[i];
			mt->fcm[i] = tmp_fc;
			found = true;
			break;
		}
	}

	/* Verify we had one metadata */
	if (!found) {
		HWS_ERR("Metadata field does not exists in matcher contains: %d",
			mt->fcm_sz);
		errno = EINVAL;
		return errno;
	}

	found = false;

	/* Go over all the field and arrange them according to the special setter */
	for (j = 1; j < num_of_fields; j++) {
		for (i = 1; i < num_of_fields; i++) {
			if (mt->fcm[i].ctrl.fname == fields[j]) {
				tmp_fc = mt->fcm[j];
				mt->fcm[j] = mt->fcm[i];
				mt->fcm[i] = tmp_fc;
				found = true;
			}
		}
		if (!found) {
			HWS_ERR("Field (%d) is not exists in matcher ", fields[j]);
			errno = ENOTSUP;
			return errno;
		}
		found = false;
	}

	return 0;
}

static int ct_matcher_check_and_place_match(struct mlx5dv_hws_matcher *matcher,
					    enum mlx5dv_hws_ct_type ct_type)
{
	struct mlx5dv_hws_mt *mt = &matcher->mt[0];

	if (ct_type == MLX5DV_HWS_CT_TYPE_V4 &&
	    mt->fcm_sz == CT_MATCHER_V4_FIELD_NUM) {
		if (ct_matcher_verify_and_set_match(mt, mt->fcm_sz))
			return errno;

		matcher->flags |= MATCHER_FLAGS_CT_V4;
	} else if (ct_type == MLX5DV_HWS_CT_TYPE_V6 &&
		   mt->fcm_sz == CT_MATCHER_V6_FIELD_NUM) {
		if (ct_matcher_verify_and_set_match(mt, mt->fcm_sz))
			return errno;

		matcher->flags |= MATCHER_FLAGS_CT_V6;
	} else {
		HWS_ERR(" Num of fields (%d) does not fit CT", mt->fcm_sz);
		errno = ENOTSUP;
		return errno;
	}

	return 0;
}

static bool matcher_verify_ct_action_type(struct mlx5dv_hws_action *action)
{
	enum action_type action_type = action->type;

	if (action_type != ACTION_TYP_CTR &&
	    action_type != ACTION_TYP_TAG &&
	    action_type != ACTION_TYP_MODIFY_HDR &&
	    action_type != ACTION_TYP_DEST_TBL &&
	    action_type != ACTION_TYP_INLINE_SET &&
	    action_type != ACTION_TYP_INLINE_ADD &&
	    action_type != ACTION_TYP_INLINE_COPY)
		return false;

	if (action_type == ACTION_TYP_MODIFY_HDR) {
		if (action->modify_header.num_of_resources > 1) {
			HWS_ERR("Modify Action for CT should have one resource only");
			return false;
		}
	}

	return true;
}

int mlx5dv_hws_ct_actions_bind(struct mlx5dv_hws_matcher *matcher,
			       struct mlx5dv_hws_action *actions[],
			       size_t num_of_actions)
{
	struct mlx5dv_hws_at at = {0};
	int ret;
	int i;

	for (i = 0; i < num_of_actions; i++) {
		if (!matcher_verify_ct_action_type(actions[i]) ||
		    actions[i]->type == ACTION_TYP_DEST_TBL) {
			HWS_ERR("Action %d not supported in CT binding",
				actions[i]->type);
				errno = ENOTSUP;
				return errno;
		}

		at.actions = actions[i];
		at.num_actions = 1;
		ret = action_template_bind(matcher, &at);
		if (ret) {
			HWS_ERR("Failed to bind actions");
			goto err_out;
		}
		/* Copy the value, and change it to BE32 */
		actions[i]->stc->offset = htobe32(actions[i]->stc->offset);
	}

	return 0;

err_out:
	for(i = i - 1; i >= 0; i--) {
		/* Return it back to host to let the cache find it */
		actions[i]->stc->offset = be32toh(actions[i]->stc->offset);
		at.actions = actions[i];
		at.num_actions = 1;
		action_template_unbind(matcher, &at);
	}

	return ret;
}

int mlx5dv_hws_ct_actions_unbind(struct mlx5dv_hws_matcher *matcher,
				 struct mlx5dv_hws_action *actions[],
				 size_t num_of_actions)
{
	struct mlx5dv_hws_at at = {0};
	int i;

	for (i = 0; i < num_of_actions; i++) {
		if (!matcher_verify_ct_action_type(actions[i]) ||
		    actions[i]->type == ACTION_TYP_DEST_TBL) {
			HWS_ERR("Action %d not supported in CT unbinding",
				actions[i]->type);
				errno = ENOTSUP;
				return errno;
		}
		if (actions[i]->type == ACTION_TYP_DEST_TBL)
			continue;

		/* Return it back to host to let the cache find it */
		actions[i]->stc->offset = be32toh(actions[i]->stc->offset);
		at.actions = actions[i];
		at.num_actions = 1;
		action_template_unbind(matcher, &at);
	}

	return 0;
}

struct mlx5dv_hws_matcher *
mlx5dv_hws_ct_matcher_create(struct mlx5dv_hws_table *tbl,
			     struct mlx5dv_hws_mt *mt,
			     struct mlx5dv_hws_ct_matcher_attr *ct_attr)
{
	struct mlx5dv_hws_matcher_attr *attr = ct_attr->attr;
	struct mlx5dv_hws_at_attr at_attr = {0};
	struct mlx5dv_hws_matcher *matcher;
	struct mlx5dv_hws_at *at;
	int ret;

	if (table_is_root(tbl)) {
		HWS_ERR("CT not supported over root");
		errno = ENOTSUP;
		return NULL;
	}

	if (ct_attr->attr->insert_mode != MLX5DV_HWS_MATCHER_INSERT_BY_HASH) {
		HWS_ERR("CT matcher: Insert mode is required to be by hash");
		errno = ENOTSUP;
		return NULL;
	}

	if (ct_attr->attr->distribute_mode != MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH) {
		HWS_ERR("CT matcher: Distribution is required to be by hash");
		errno = ENOTSUP;
		return NULL;
	}

	if (ct_attr->ct_type == MLX5DV_HWS_CT_TYPE_V6 &&
	    !ct_attr->attr->optimize_using_rule_idx) {
		HWS_ERR("CT matcher: optimize_using_rule_idx is required for IPv6 match");
		errno = ENOTSUP;
		return NULL;
	}

	/* Creating an empty at */
	at = mlx5dv_hws_action_template_create(NULL, 0, &at_attr);
	if (!at) {
		HWS_ERR("Failed to create at for type: %d", ct_attr->ct_type);
		return NULL;
	}

	matcher = mlx5dv_hws_matcher_create(tbl, &mt, 1, &at, 1, attr);
	if (!matcher) {
		HWS_ERR("Failed to create matcher");
		goto at_destroy;
	}

	ret = ct_matcher_check_and_place_match(matcher, ct_attr->ct_type);
	if (ret) {
		HWS_ERR("Failed to create matcher");
		goto matcher_destroy;
	}

	if (!ct_attr->bi_direction)
		matcher->flags |= MATCHER_FLAGS_CT_ONE_DIRECTION;

	/* Make action STE if needed */
	if (ct_attr->ct_type == MLX5DV_HWS_CT_TYPE_V6) {
		matcher->action_ste.max_stes = ct_attr->bi_direction ? 2 : 1;
		ret = matcher_action_ste_create(matcher);
		if (ret) {
			matcher->action_ste.max_stes = 0;
			goto matcher_destroy;
		}
	}

	mlx5dv_hws_action_template_destroy(at);

	return matcher;

matcher_destroy:
	mlx5dv_hws_matcher_destroy(matcher);
at_destroy:
	mlx5dv_hws_action_template_destroy(at);
	return NULL;
}

int mlx5dv_hws_ct_matcher_destroy(struct mlx5dv_hws_matcher *matcher)
{
	return mlx5dv_hws_matcher_destroy(matcher);
}