/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"

static const char *debug_action_type_str[] = {
	[ACTION_TYP_REFORMAT_TNL_L2_TO_L2] = "TNL_L2_TO_L2",
	[ACTION_TYP_REFORMAT_L2_TO_TNL_L2] = "L2_TO_TNL_L2",
	[ACTION_TYP_REFORMAT_TNL_L3_TO_L2] = "TNL_L3_TO_L2",
	[ACTION_TYP_REFORMAT_L2_TO_TNL_L3] = "L2_TO_TNL_L3",
	[ACTION_TYP_INSERT_HEADER] = "INSERT_HEADER",
	[ACTION_TYP_REMOVE_HEADER] = "REMOVE_HEADER",
	[ACTION_TYP_CTR] = "CTR",
	[ACTION_TYP_TAG] = "TAG",
	[ACTION_TYP_MODIFY_HDR] = "MODIFY_HDR",
	[ACTION_TYP_POP_VLAN] = "POP_VLAN",
	[ACTION_TYP_PUSH_VLAN] = "PUSH_VLAN",
	[ACTION_TYP_ASO_METER] = "ASO_METER",
	[ACTION_TYP_ASO_CT] = "ASO_CT",
	[ACTION_TYP_ASO_IPSEC] = "ASO_IPSEC",
	[ACTION_TYP_ASO_FIRST_HIT] = "ASO_FIRST_HIT",
	[ACTION_TYP_NAT64] = "NAT64",
	[ACTION_TYP_CRYPTO_ENCRYPT] = "CRYPTO_ENCRYPT",
	[ACTION_TYP_CRYPTO_DECRYPT] = "CRYPTO_DECRYPT",
	[ACTION_TYP_REFORMAT_TRAILER] = "REFORMAT_TRAILER",
	[ACTION_TYP_DEST_ROOT] = "DEST_ROOT",
	[ACTION_TYP_DEST_ARRAY] = "DEST_ARRAY",
	[ACTION_TYP_DEST_DROP] = "DROP",
	[ACTION_TYP_DEST_TIR] = "TIR",
	[ACTION_TYP_DEST_TBL] = "TBL",
	[ACTION_TYP_DEST_VPORT] = "VPORT",
	[ACTION_TYP_DEST_MISS] = "DEFAULT_MISS",
	[ACTION_TYP_DEST_MATCHER] = "DEST_MATCHER",
	[ACTION_TYP_INLINE_SET] = "ACTION_TYP_INLINE_SET",
	[ACTION_TYP_INLINE_ADD] = "ACTION_TYP_INLINE_ADD",
	[ACTION_TYP_INLINE_ADD_FIELD] = "ACTION_TYP_INLINE_ADD_FIELD",
	[ACTION_TYP_INLINE_COPY] = "ACTION_TYP_INLINE_COPY",
	[ACTION_TYP_INLINE_INSERT] = "ACTION_TYP_INLINE_INSERT",
};

enum debug_dump_tool_table_type {
	DEBUG_DUMP_TABLE_TYPE_NIC_RX,
	DEBUG_DUMP_TABLE_TYPE_NIC_TX,
	DEBUG_DUMP_TABLE_TYPE_FDB,
	DEBUG_DUMP_TABLE_TYPE_FDB_RX,
	DEBUG_DUMP_TABLE_TYPE_FDB_TX,
	DEBUG_DUMP_TABLE_TYPE_FDB_UNIFIED,
	DEBUG_DUMP_TABLE_TYPE_RDMA_TRANSPORT_RX,
	DEBUG_DUMP_TABLE_TYPE_RDMA_TRANSPORT_TX,
};

/* Table type translation to stay compatiblity to mlx_steering_dump tool */
static const enum debug_dump_tool_table_type debug_table_type_to_dump_tool[] = {
	[MLX5DV_HWS_TABLE_TYPE_NIC_RX] = DEBUG_DUMP_TABLE_TYPE_NIC_RX,
	[MLX5DV_HWS_TABLE_TYPE_NIC_TX] = DEBUG_DUMP_TABLE_TYPE_NIC_TX,
	[MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX] = DEBUG_DUMP_TABLE_TYPE_RDMA_TRANSPORT_RX,
	[MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX] = DEBUG_DUMP_TABLE_TYPE_RDMA_TRANSPORT_TX,
	[MLX5DV_HWS_TABLE_TYPE_FDB_RX] = DEBUG_DUMP_TABLE_TYPE_FDB_RX,
	[MLX5DV_HWS_TABLE_TYPE_FDB_TX] = DEBUG_DUMP_TABLE_TYPE_FDB_TX,
	[MLX5DV_HWS_TABLE_TYPE_FDB] = DEBUG_DUMP_TABLE_TYPE_FDB_UNIFIED,
};

static_assert(ARRAY_SIZE(debug_action_type_str) == ACTION_TYP_MAX,
	      "Missing debug_action_type_str");

const char *
debug_action_type_to_str(enum action_type type)
{
	return debug_action_type_str[type];
}

static int
debug_dump_matcher_template_definer(FILE *f,
				    void *parent_obj,
				    struct definer *definer,
				    enum debug_res_type type)
{
	int i, ret;

	if (!definer)
		return 0;

	ret = fprintf(f, "%d,0x%" PRIx64 ",0x%" PRIx64 ",%d,%d,",
		      type,
		      (uint64_t)(uintptr_t)definer,
		      (uint64_t)(uintptr_t)parent_obj,
		      definer->obj->object_id,
		      definer->type);
	if (ret < 0) {
		errno = EINVAL;
		return -errno;
	}

	for (i = 0; i < DW_SELECTORS; i++) {
		ret = fprintf(f, "0x%x%s", definer->dw_selector[i],
			      (i == DW_SELECTORS - 1) ? "," : "-");
		if (ret < 0) {
			errno = EINVAL;
			return -errno;
		}
	}

	for (i = 0; i < BYTE_SELECTORS; i++) {
		ret = fprintf(f, "0x%x%s", definer->byte_selector[i],
			      (i == BYTE_SELECTORS - 1) ? "," : "-");
		if (ret < 0) {
			errno = EINVAL;
			return -errno;
		}
	}

	for (i = 0; i < STE_JUMBO_TAG_SZ; i++) {
		ret = fprintf(f, "%02x", definer->mask->jumbo[i]);
		if (ret < 0) {
			errno = EINVAL;
			return -errno;
		}
	}

	ret = fprintf(f, "\n");
	if (ret < 0) {
		errno = EINVAL;
		return -errno;
	}

	return 0;
}

static int
debug_dump_matcher_match_template(FILE *f, struct mlx5dv_hws_matcher *matcher)
{
	bool is_root = matcher->tbl->level == TABLE_ROOT_LEVEL;
	enum debug_res_type type;
	int i, ret;

	for (i = 0; i < matcher->num_of_mt; i++) {
		struct mlx5dv_hws_mt *mt = &matcher->mt[i];

		ret = fprintf(f, "%d,0x%" PRIx64 ",0x%" PRIx64 ",%d,%d,%d,%d\n",
			      DEBUG_RES_TYPE_MATCHER_MATCH_TEMPLATE,
			      (uint64_t)(uintptr_t)mt,
			      (uint64_t)(uintptr_t)matcher,
			      is_root ? 0 : mt->fcm_sz,
			      0,//For legacy usage of dump tool also for HWS
			      is_root ? 0 : mt->fcr_sz,
			      is_root ? 0 : mt->fcc_sz);
		if (ret < 0) {
			errno = EINVAL;
			return -errno;
		}

		type = DEBUG_RES_TYPE_MATCHER_TEMPLATE_MATCH_DEFINER;
		ret = debug_dump_matcher_template_definer(f, mt,
							  mt->definer,
							  type);
		if (ret)
			return ret;

		type = DEBUG_RES_TYPE_MATCHER_TEMPLATE_RANGE_DEFINER;
		ret = debug_dump_matcher_template_definer(f, mt,
							  mt->range_definer,
							  type);
		if (ret)
			return ret;

		type = DEBUG_RES_TYPE_MATCHER_TEMPLATE_COMPARE_MATCH_DEFINER;
		ret = debug_dump_matcher_template_definer(f, mt,
							  mt->compare_definer,
							  type);
		if (ret)
			return ret;

	}

	type = DEBUG_RES_TYPE_MATCHER_TEMPLATE_HASH_DEFINER;
	ret = debug_dump_matcher_template_definer(f, matcher,
						  matcher->hash_definer, type);
	if (ret)
		return ret;

	return 0;
}

static int
debug_dump_matcher_action_template(FILE *f, struct mlx5dv_hws_matcher *matcher)
{
	bool is_root = matcher->tbl->level == TABLE_ROOT_LEVEL;
	enum action_type action_type;
	int i, j, ret;

	for (i = 0; i < matcher->num_of_at; i++) {
		struct mlx5dv_hws_at *at = &matcher->at[i];

		ret = fprintf(f, "%d,0x%" PRIx64 ",0x%" PRIx64 ",%d,%d,%d",
			      DEBUG_RES_TYPE_MATCHER_ACTION_TEMPLATE,
			      (uint64_t)(uintptr_t)at,
			      (uint64_t)(uintptr_t)matcher,
			      at->only_term,
			      is_root ? 0 : at->num_of_action_stes,
			      at->num_actions);
		if (ret < 0) {
			errno = EINVAL;
			return -errno;
		}

		for (j = 0; j < at->num_actions; j++) {
			action_type = at->actions[j].type;
			ret = fprintf(f, ",%s", debug_action_type_to_str(action_type));
			if (ret < 0) {
				errno = EINVAL;
				return -errno;
			}
		}

		fprintf(f, "\n");
	}

	return 0;
}

static int
debug_dump_matcher_attr(FILE *f, struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_matcher_attr *attr = &matcher->attr;
	int ret;

	ret = fprintf(f, "%d,0x%" PRIx64 ",%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
		      DEBUG_RES_TYPE_MATCHER_ATTR,
		      (uint64_t)(uintptr_t)matcher,
		      attr->priority,
		      attr->resource_mode,
		      attr->table.sz_row_log,
		      attr->table.sz_col_log,
		      attr->optimize_using_rule_idx,
		      0,//For legacy usage of dump tool also for HWS
		      attr->insert_mode,
		      attr->distribute_mode,
		      attr->always_hit,
		      attr->isolated);
	if (ret < 0) {
		errno = EINVAL;
		return -errno;
	}

	return 0;
}

static int debug_dump_matcher(FILE *f, struct mlx5dv_hws_matcher *matcher)
{
	bool is_shared = context_shared_gvmi_used(matcher->tbl->ctx);
	bool is_root = matcher->tbl->level == TABLE_ROOT_LEVEL;
	enum mlx5dv_hws_table_type tbl_type = matcher->tbl->type;
	struct matcher_resize_data *resize_data;
	struct cmd_ft_query_attr ft_attr = {0};
	struct mlx5dv_devx_obj *ste_0, *ste_1;
	struct mlx5dv_devx_obj *rtc_0, *rtc_1;
	uint64_t icm_addr_0 = 0;
	uint64_t icm_addr_1 = 0;
	int ret;

	ret = fprintf(f, "%d,0x%" PRIx64 ",0x%" PRIx64 ",%d,%d,0x%" PRIx64,
			DEBUG_RES_TYPE_MATCHER,
			(uint64_t)(uintptr_t)matcher,
			(uint64_t)(uintptr_t)matcher->tbl,
			matcher->num_of_mt,
			is_root ? 0 : matcher->end_ft->object_id,
			matcher->col_matcher ?
				(uint64_t)(uintptr_t)matcher->col_matcher : 0);
	if (ret < 0)
		goto out_err;

	ste_0 = matcher->match_ste.ste_0;
	ste_1 = matcher->match_ste.ste_1;
	ret = fprintf(f, ",%d,%d,%d,%d",
			matcher->match_ste.rtc_0 ?
				matcher->match_ste.rtc_0->object_id : 0,
			ste_0 ? (int)ste_0->object_id : -1,
			matcher->match_ste.rtc_1 ?
				matcher->match_ste.rtc_1->object_id : 0,
			ste_1 ? (int)ste_1->object_id : -1);
	if (ret < 0)
		goto out_err;

	ste_0 = NULL;
	ste_1 = NULL;
	rtc_0 = NULL;
	rtc_1 = NULL;
	if (tbl_type == MLX5DV_HWS_TABLE_TYPE_FDB) {
		ste_0 = matcher->action_ste.ste;
		ste_1 = matcher->action_ste.ste;
		rtc_0 = matcher->action_ste.rtc;
		rtc_1 = matcher->action_ste.rtc;
	} else if (tbl_type == MLX5DV_HWS_TABLE_TYPE_FDB_TX) {
			ste_1 = matcher->action_ste.ste;
			rtc_1 = matcher->action_ste.rtc;
	} else {
		ste_0 = matcher->action_ste.ste;
		rtc_0 = matcher->action_ste.rtc;
	}

	if (!is_root) {
		ft_attr.type = matcher->tbl->fw_ft_type;
		ret = cmd_flow_table_query(matcher->end_ft,
					   &ft_attr,
					   &icm_addr_0,
					   &icm_addr_1);
		if (ret)
			return ret;
	}

	ret = fprintf(f, ",%d,%d,%d,%d,%d,0x%" PRIx64 ",0x%" PRIx64 "\n",
		      rtc_0 ? rtc_0->object_id : 0,
		      ste_0 ? (int)ste_0->object_id : -1,
		      rtc_1 ? rtc_1->object_id : 0,
		      ste_1 ? (int)ste_1->object_id : -1,
		      is_shared && !is_root ?
		      matcher->match_ste.aliased_rtc_0->object_id : 0,
		      debug_icm_to_idx(icm_addr_0),
		      debug_icm_to_idx(icm_addr_1));
	if (ret < 0)
		goto out_err;

	ret = debug_dump_matcher_attr(f, matcher);
	if (ret)
		return ret;

	ret = debug_dump_matcher_match_template(f, matcher);
	if (ret)
		return ret;

	ret = debug_dump_matcher_action_template(f, matcher);
	if (ret)
		return ret;

	list_for_each(&matcher->resize_data, resize_data, next) {
		ste_0 = NULL;
		ste_1 = NULL;
		rtc_0 = NULL;
		rtc_1 = NULL;
		if (tbl_type == MLX5DV_HWS_TABLE_TYPE_FDB) {
			ste_0 = resize_data->action_ste;
			ste_1 = resize_data->action_ste;
			rtc_0 = resize_data->action_rtc;
			rtc_1 = resize_data->action_rtc;
		} else if (tbl_type == MLX5DV_HWS_TABLE_TYPE_FDB_TX) {
			ste_1 = resize_data->action_ste;
			rtc_1 = resize_data->action_rtc;
		} else {
			ste_0 = resize_data->action_ste;
			rtc_0 = resize_data->action_rtc;
		}

		ret = fprintf(f, "%d,0x%" PRIx64 ",%d,%d,%d,%d\n",
			      DEBUG_RES_TYPE_MATCHER_RESIZABLE_ACTION_ARRAY,
			      (uint64_t)(uintptr_t)matcher,
			      rtc_0 ? rtc_0->object_id : 0,
			      ste_0 ? (int)ste_0->object_id : -1,
			      rtc_1 ? rtc_1->object_id : 0,
			      ste_1 ? (int)ste_1->object_id : -1);
		if (ret < 0)
			return ret;
	}

	return 0;

out_err:
	errno = EINVAL;
	return -errno;
}

static int debug_dump_table(FILE *f, struct mlx5dv_hws_table *tbl)
{
	bool is_shared = context_shared_gvmi_used(tbl->ctx);
	bool is_root = tbl->level == TABLE_ROOT_LEVEL;
	struct cmd_ft_query_attr ft_attr = {0};
	struct mlx5dv_hws_matcher *matcher;
	uint64_t local_icm_addr_0 = 0;
	uint64_t local_icm_addr_1 = 0;
	uint64_t icm_addr_0 = 0;
	uint64_t icm_addr_1 = 0;
	int ret;

	ret = fprintf(f, "%d,0x%" PRIx64 ",0x%" PRIx64 ",%d,%d,%d,%d,%d",
		      DEBUG_RES_TYPE_TABLE,
		      (uint64_t)(uintptr_t)tbl,
		      (uint64_t)(uintptr_t)tbl->ctx,
		      is_root ? 0 : tbl->ft->object_id,
		      debug_table_type_to_dump_tool[tbl->type],
		      is_root ? 0 : tbl->fw_ft_type,
		      tbl->level,
		      is_shared && !is_root ? tbl->local_ft->object_id : 0);
	if (ret < 0)
		goto out_err;

	if (!is_root) {
		ft_attr.type = tbl->fw_ft_type;
		ret = cmd_flow_table_query(tbl->ft,
					   &ft_attr,
					   &icm_addr_0,
					   &icm_addr_1);
		if (ret)
			return ret;

		if (is_shared) {
			ft_attr.type = tbl->fw_ft_type;
			ret = cmd_flow_table_query(tbl->local_ft,
						   &ft_attr,
						   &local_icm_addr_0,
						   &local_icm_addr_1);
			if (ret)
				return ret;
		}
	}

	ret = fprintf(f, ",0x%" PRIx64 ",0x%" PRIx64 ",0x%" PRIx64 ",0x%" PRIx64 ",0x%" PRIx64 "",
		      debug_icm_to_idx(icm_addr_0),
		      debug_icm_to_idx(icm_addr_1),
		      debug_icm_to_idx(local_icm_addr_0),
		      debug_icm_to_idx(local_icm_addr_1),
		      (uint64_t)(uintptr_t)tbl->default_miss.miss_tbl);
	if (ret < 0)
		goto out_err;

	ret = fprintf(f, ",%d,%d,%d\n", tbl->ib_port, tbl->vport, tbl->other_vport);
	if (ret < 0)
		goto out_err;

	list_for_each(&tbl->head, matcher, next) {
		ret = debug_dump_matcher(f, matcher);
		if (ret)
			return ret;
	}

	list_for_each(&tbl->isolated_matchers, matcher, next) {
		ret = debug_dump_matcher(f, matcher);
		if (ret)
			return ret;
	}

	return 0;

out_err:
	errno = EINVAL;
	return -errno;
}

static int
debug_dump_context_send_engine(FILE *f, struct mlx5dv_hws_context *ctx)
{
	struct send_engine *send_queue;
	int ret, i, j;

	for (i = 0; i < (int)ctx->queues; i++) {
		send_queue = &ctx->send_queue[i];
		ret = fprintf(f, "%d,0x%" PRIx64 ",%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
			      DEBUG_RES_TYPE_CONTEXT_SEND_ENGINE,
			      (uint64_t)(uintptr_t)ctx,
			      i,
			      send_queue->used_entries,
			      send_queue->num_entries,
			      send_queue->rings,
			      send_queue->num_entries,
			      send_queue->err,
			      send_queue->completed.ci,
			      send_queue->completed.pi,
			      send_queue->completed.mask);
		if (ret < 0) {
			errno = EINVAL;
			return -errno;
		}

		for (j = 0; j < SEND_NUM_SEND_RINGS; j++) {
			struct send_ring *send_ring = &send_queue->send_ring[j];
			struct send_ring_cq *cq = &send_ring->send_cq;
			struct send_ring_sq *sq = &send_ring->send_sq;

			ret = fprintf(f, "%d,0x%" PRIx64 ",%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
				      DEBUG_RES_TYPE_CONTEXT_SEND_RING,
				      (uint64_t)(uintptr_t)ctx,
				      j,
				      i,
				      cq->cqn,
				      cq->cons_index,
				      cq->ncqe_mask,
				      cq->buf_sz,
				      cq->ncqe,
				      cq->cqe_log_sz,
				      cq->poll_wqe,
				      cq->cqe_sz,
				      sq->sqn,
				      sq->obj->object_id,
				      sq->cur_post,
				      sq->buf_mask);
			if (ret < 0) {
				errno = EINVAL;
				return -errno;
			}
		}
	}

	return 0;
}

static int debug_dump_context_caps(FILE *f, struct mlx5dv_hws_context *ctx)
{
	struct cmd_query_caps *caps = ctx->caps;
	int ret;

	ret = fprintf(f, "%d,0x%" PRIx64 ",%s,%d,%d,%d,%d,",
		      DEBUG_RES_TYPE_CONTEXT_CAPS,
		      (uint64_t)(uintptr_t)ctx,
		      caps->fw_ver,
		      caps->wqe_based_update,
		      caps->ste_format,
		      caps->ste_alloc_log_max,
		      caps->log_header_modify_argument_max_alloc);
	if (ret < 0) {
		errno = EINVAL;
		return -errno;
	}

	ret = fprintf(f, "%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%s\n",
		      caps->flex_protocols,
		      caps->rtc_reparse_mode,
		      caps->rtc_index_mode,
		      caps->ste_alloc_log_gran,
		      caps->stc_alloc_log_max,
		      caps->stc_alloc_log_gran,
		      caps->rtc_log_depth_max,
		      caps->format_select_gtpu_dw_0,
		      caps->format_select_gtpu_dw_1,
		      caps->format_select_gtpu_dw_2,
		      caps->format_select_gtpu_ext_dw_0,
		      caps->nic_ft.max_level,
		      caps->nic_ft.reparse,
		      caps->fdb_ft.max_level,
		      caps->fdb_ft.reparse,
		      caps->log_header_modify_argument_granularity,
		      caps->linear_match_definer,
		      "regc_3");
	if (ret < 0) {
		errno = EINVAL;
		return -errno;
	}

	return 0;
}

static int debug_dump_context_attr(FILE *f, struct mlx5dv_hws_context *ctx)
{
	int ret;

	ret = fprintf(f, "%u,0x%" PRIx64 ",%d,%zu,%d,%s,%d,%d\n",
		      DEBUG_RES_TYPE_CONTEXT_ATTR,
		      (uint64_t)(uintptr_t)ctx,
		      ctx->pd_num,
		      ctx->queues,
		      ctx->send_queue->num_entries,
		      context_shared_gvmi_used(ctx) ?
		      ibv_get_device_name(ctx->ibv_ctx->device) : "None",
		      ctx->caps->vhca_id,
		      context_shared_gvmi_used(ctx) ?
		      ctx->caps->shared_vhca_id : 0xffff);
	if (ret < 0) {
		errno = EINVAL;
		return -errno;
	}

	return 0;
}

static int debug_dump_context_info(FILE *f, struct mlx5dv_hws_context *ctx)
{
	int ret;

	ret = fprintf(f, "%d,0x%" PRIx64 ",%d,%s,%s\n",
			DEBUG_RES_TYPE_CONTEXT,
			(uint64_t)(uintptr_t)ctx,
			ctx->flags & CONTEXT_FLAG_HWS_SUPPORT,
			ibv_get_device_name(context_get_local_ibv(ctx)->device),
			DEBUG_VERSION);
	if (ret < 0) {
		errno = EINVAL;
		return -errno;
	}

	ret = debug_dump_context_attr(f, ctx);
	if (ret)
		return ret;

	ret = debug_dump_context_caps(f, ctx);
	if (ret)
		return ret;

	return 0;
}

static int debug_dump_context_stc_resource(FILE *f,
					   struct mlx5dv_hws_context *ctx,
					   uint32_t tbl_type,
					   uint32_t ib_port,
					   struct pool_resource *resource)
{
	uint32_t id = resource->devx_obj ? resource->devx_obj->object_id : 0;
	int ret;

	ret = fprintf(f, "%d,0x%" PRIx64 ",%u,%u,%u\n",
		      DEBUG_RES_TYPE_CONTEXT_STC,
		      (uint64_t)(uintptr_t)ctx,
		      debug_table_type_to_dump_tool[tbl_type],
		      id,
		      ib_port);
	if (ret < 0) {
		errno = EINVAL;
		return -errno;
	}

	return 0;
}

static int debug_dump_context_stc(FILE *f, struct mlx5dv_hws_context *ctx)
{
	struct context_obj *ctx_stc_pool;
	struct pool *stc_pool;
	int ret;

	list_for_each(&ctx->stc_pool, ctx_stc_pool, next) {
		stc_pool = ctx_stc_pool->obj;
		if (stc_pool->resource[0] != NULL) {
			ret = debug_dump_context_stc_resource(f, ctx,
							      ctx_stc_pool->tbl_type,
							      ctx_stc_pool->ib_port,
							      stc_pool->resource[0]);
			if (ret)
				return ret;
		}
	}

	return 0;
}

static int debug_dump_context(FILE *f, struct mlx5dv_hws_context *ctx)
{
	struct mlx5dv_hws_table *tbl;
	int ret;

	ret = debug_dump_context_info(f, ctx);
	if (ret)
		return ret;

	ret = debug_dump_context_send_engine(f, ctx);
	if (ret)
		return ret;

	ret = debug_dump_context_stc(f, ctx);
	if (ret)
		return ret;

	list_for_each(&ctx->head, tbl, next) {
		ret = debug_dump_table(f, tbl);
		if (ret)
			return ret;
	}

	return 0;
}

int mlx5dv_hws_debug_dump(struct mlx5dv_hws_context *ctx, FILE *f)
{
	int ret;

	if (!f || !ctx) {
		errno = EINVAL;
		return -errno;
	}

	pthread_spin_lock(&ctx->ctrl_lock);
	ret = debug_dump_context(f, ctx);
	fflush(f);
	pthread_spin_unlock(&ctx->ctrl_lock);

	return ret;
}
