/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"

#define DEFINER_MAX_ROW_LOG 32
#define IP_VERSION_IPV4 0x4
#define IP_VERSION_IPV6 0x6

/* Calculate BYTE offset for DW MSB unaligned */
#define UNALGN_DW_MSB(bit_off, bit_sz) \
	((bit_off + bit_sz + BITS_IN_BYTE - 1) / BITS_IN_BYTE - DW_SIZE)

/* Get field 32 in case field is not DW aligned */
#define HWS_GET_UNALGN(p, bit_off, bit_sz) \
	(be32toh(*(const __be32 *)((uint8_t *)(p) + UNALGN_DW_MSB(bit_off, bit_sz))) >> \
		((BITS_IN_BYTE - (bit_off + bit_sz) % BITS_IN_BYTE) % BITS_IN_BYTE)) & \
		__devx_mask(bit_sz)

/* Get field 32 in case field is DW aligned (same as devx_set) */
#define HWS_GET_ALGN(p, bit_off, bit_sz) \
	(be32toh(*(const __be32 *)((__be32 *)(p) + (bit_off) / BITS_IN_DW)) >> \
		__devx_dw_bit_off(bit_sz, bit_off)) & \
		__devx_mask(bit_sz)

/* Simple field 32 GET */
#define HWS_GET(p, bit_off, bit_sz) \
	(((bit_off) % BITS_IN_DW + bit_sz) <= BITS_IN_DW) ? \
		HWS_GET_ALGN(p, bit_off, bit_sz) : \
		HWS_GET_UNALGN(p, bit_off, bit_sz)

/* Setter function based on bit offset and bit_sz, for unaligned 32bit value */
#define HWS_SET_UNALGN(p, v, bit_off, bit_sz) \
do { \
	uint32_t _v = v; \
	uint32_t _m = __devx_mask(bit_sz); \
	uint32_t _dw_bit_off = ((BITS_IN_BYTE - (bit_off + bit_sz) % BITS_IN_BYTE) % BITS_IN_BYTE); \
	__be32 *_p = (__be32 *)((uint8_t *)(p) + UNALGN_DW_MSB(bit_off, bit_sz)); \
	*_p = htobe32((be32toh(*(_p)) & (~((_m) << (_dw_bit_off)))) | ((_v & _m) << _dw_bit_off)); \
} while (0)

/* Setter function based on bit offset and bit_sz, for aligned 32bit value */
#define HWS_SET_ALGN(p, v, bit_off, bit_sz) \
do { \
	uint32_t _v = v; \
	uint32_t _m = __devx_mask(bit_sz); \
	uint32_t _dw_bit_off = __devx_dw_bit_off(bit_sz, bit_off); \
	__be32 *_p = ((__be32 *)(p) + ((bit_off) / BITS_IN_DW)); \
	*_p = htobe32((be32toh(*(_p)) & (~((_m) << (_dw_bit_off)))) | ((_v & _m) << _dw_bit_off)); \
} while (0)

/* Simple field 32 SET */
#define HWS_SET(p, v, bit_off, bit_sz) \
do { \
	if (((bit_off) % BITS_IN_DW + bit_sz) <= BITS_IN_DW) { \
		HWS_SET_ALGN(p, v, bit_off, bit_sz); \
	} else { \
		HWS_SET_UNALGN(p, v, bit_off, bit_sz); \
	} \
} while (0)

/* Calculate number of bits in field */
#define DEFINER_CALC_BIT_SZ(typ, fld) (sizeof(__devx_nullp(typ)->fld))
/* Calculate total bit offset to MSB of field */
#define DEFINER_CALC_BIT_OFF(typ, fld) ((unsigned int)(uintptr_t) (&(__devx_nullp(typ)->fld)))
/* Calculate total byte offset to MSB of field */
#define DEFINER_CALC_BYTE_OFF(typ, fld) (DEFINER_CALC_BIT_OFF(typ, fld) / 8)
/* Calculate absolute bit mask */
#define DEFINER_CALC_BIT_MASK(typ, fld) ((1 << DEFINER_CALC_BIT_SZ(typ, fld)) - 1)
/* Calculate DW bit offset from MSB */
#define DEFINER_CALC_DW_BIT_OFF(typ, fld) (DEFINER_CALC_BIT_OFF(typ, fld) & 0x1f)

/* Helper to initialize the fc_arr with less code */
#define DEFINER_OFFSET(hdr, fld) \
{ \
	.bit_off = DEFINER_CALC_BIT_OFF(hdr, fld), \
	.bit_sz = DEFINER_CALC_BIT_SZ(hdr, fld), \
}

/* Helper to initialize the fc_arr with less code */
#define DEFINER_HL_OFFSET(fld) DEFINER_OFFSET(definer_hl, fld)
#define DEFINER_MP_OFFSET(fld) DEFINER_OFFSET(dr_match_param, fld)

/* Helper to initialize the fc_arr members whose HL offsets are
 * dynamically fetched during the run time.
 */
#define DEFINER_DYNAMIC_OFFSET(typ, fld) \
{ \
	.bit_off = DEFINER_CALC_DW_BIT_OFF(typ, fld), \
	.bit_sz = DEFINER_CALC_BIT_SZ(typ, fld), \
}

struct definer_sel_ctrl {
	uint8_t allowed_full_dw; /* Full DW selectors cover all offsets */
	uint8_t allowed_lim_dw;  /* Limited DW selectors cover offset < 64 */
	uint8_t allowed_bytes;   /* Bytes selectors, up to offset 255 */
	uint8_t used_full_dw;
	uint8_t used_lim_dw;
	uint8_t used_bytes;
	uint8_t full_dw_selector[DW_SELECTORS];
	uint8_t lim_dw_selector[DW_SELECTORS_LIMITED];
	uint8_t byte_selector[BYTE_SELECTORS];
};

struct definer_field_conv_offset {
	uint32_t bit_off;
	uint32_t bit_sz;
};

static const struct definer_field_conv_offset
fco_hws_arr[MLX5DV_HWS_HEADER_MAX][MLX5DV_HWS_FNAME_MAX] = {
	[MLX5DV_HWS_HEADER_NONE] = {
		[MLX5DV_HWS_FNAME_REG_C0] = DEFINER_HL_OFFSET(registers.register_c_0),
		[MLX5DV_HWS_FNAME_REG_C1] = DEFINER_HL_OFFSET(registers.register_c_1),
		[MLX5DV_HWS_FNAME_REG_C2] = DEFINER_HL_OFFSET(registers.register_c_2),
		[MLX5DV_HWS_FNAME_REG_C3] = DEFINER_HL_OFFSET(registers.register_c_3),
		[MLX5DV_HWS_FNAME_REG_C4] = DEFINER_HL_OFFSET(registers.register_c_4),
		[MLX5DV_HWS_FNAME_REG_C5] = DEFINER_HL_OFFSET(registers.register_c_5),
		[MLX5DV_HWS_FNAME_REG_C6] = DEFINER_HL_OFFSET(registers.register_c_6),
		[MLX5DV_HWS_FNAME_REG_C7] = DEFINER_HL_OFFSET(registers.register_c_7),
		[MLX5DV_HWS_FNAME_REG_C8] = DEFINER_HL_OFFSET(registers.register_c_8),
		[MLX5DV_HWS_FNAME_REG_C9] = DEFINER_HL_OFFSET(registers.register_c_9),
		[MLX5DV_HWS_FNAME_REG_C10] = DEFINER_HL_OFFSET(registers.register_c_10),
		[MLX5DV_HWS_FNAME_REG_C11] = DEFINER_HL_OFFSET(registers.register_c_11),
		[MLX5DV_HWS_FNAME_REG_A] = DEFINER_HL_OFFSET(metadata.general_purpose),
		[MLX5DV_HWS_FNAME_REG_B] = DEFINER_HL_OFFSET(metadata.metadata_to_cqe),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_0] = DEFINER_HL_OFFSET(flex_parser.flex_parser_0),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_1] = DEFINER_HL_OFFSET(flex_parser.flex_parser_1),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_2] = DEFINER_HL_OFFSET(flex_parser.flex_parser_2),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_3] = DEFINER_HL_OFFSET(flex_parser.flex_parser_3),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_4] = DEFINER_HL_OFFSET(flex_parser.flex_parser_4),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_5] = DEFINER_HL_OFFSET(flex_parser.flex_parser_5),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_6] = DEFINER_HL_OFFSET(flex_parser.flex_parser_6),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_7] = DEFINER_HL_OFFSET(flex_parser.flex_parser_7),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_0_OK] = DEFINER_HL_OFFSET(oks1.flex_parser0_steering_ok),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_1_OK] = DEFINER_HL_OFFSET(oks1.flex_parser1_steering_ok),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_2_OK] = DEFINER_HL_OFFSET(oks1.flex_parser2_steering_ok),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_3_OK] = DEFINER_HL_OFFSET(oks1.flex_parser3_steering_ok),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_4_OK] = DEFINER_HL_OFFSET(oks1.flex_parser4_steering_ok),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_5_OK] = DEFINER_HL_OFFSET(oks1.flex_parser5_steering_ok),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_6_OK] = DEFINER_HL_OFFSET(oks1.flex_parser6_steering_ok),
		[MLX5DV_HWS_FNAME_FLEX_PARSER_7_OK] = DEFINER_HL_OFFSET(oks1.flex_parser7_steering_ok),
		[MLX5DV_HWS_FNAME_RANDOM_NUM] = DEFINER_HL_OFFSET(random_number.random_number),
		[MLX5DV_HWS_FNAME_SOURCE_QP] = DEFINER_HL_OFFSET(source_qp_gvmi.source_qp),
		[MLX5DV_HWS_FNAME_HASH_RESULT] = DEFINER_HL_OFFSET(metadata.acomulated_hash),
		[MLX5DV_HWS_FNAME_UTC_63_32] = DEFINER_HL_OFFSET(utc.utc_timestamp_h),
		[MLX5DV_HWS_FNAME_UTC_31_0] = DEFINER_HL_OFFSET(utc.utc_timestamp_l),
		[MLX5DV_HWS_FNAME_FRC_63_32] = DEFINER_HL_OFFSET(frc.frc_timestamp_h),
		[MLX5DV_HWS_FNAME_FRC_31_0] = DEFINER_HL_OFFSET(frc.frc_timestamp_l),
		[MLX5DV_HWS_FNAME_IPSEC_SYNDROME] = DEFINER_HL_OFFSET(ipsec.ipsec_syndrome),
		[MLX5DV_HWS_FNAME_PSP_SYNDROME] = DEFINER_HL_OFFSET(psp.psp_syndrome),
	},
	[MLX5DV_HWS_HEADER_OUTER] = {
		[MLX5DV_HWS_FNAME_L2_TYPE] = DEFINER_HL_OFFSET(eth_l2_outer.l2_type),
		[MLX5DV_HWS_FNAME_L3_TYPE] = DEFINER_HL_OFFSET(eth_l2_outer.l3_type),
		[MLX5DV_HWS_FNAME_L4_TYPE] = DEFINER_HL_OFFSET(eth_l2_outer.l4_type),
		[MLX5DV_HWS_FNAME_L4_TYPE_BWC] = DEFINER_HL_OFFSET(eth_l2_outer.l4_type_bwc),
		[MLX5DV_HWS_FNAME_ETH_TYPE] = DEFINER_HL_OFFSET(eth_l2_outer.l3_ethertype),
		[MLX5DV_HWS_FNAME_ETH_PKT_LENGTH] = DEFINER_HL_OFFSET(eth_l3_outer.eth_packet_length),
		[MLX5DV_HWS_FNAME_ENCAP_TYPE] = DEFINER_HL_OFFSET(eth_l2_outer.encap_type),
		[MLX5DV_HWS_FNAME_SMAC_47_16] = DEFINER_HL_OFFSET(eth_l2_src_outer.smac_47_16),
		[MLX5DV_HWS_FNAME_SMAC_15_0] = DEFINER_HL_OFFSET(eth_l2_src_outer.smac_15_0),
		[MLX5DV_HWS_FNAME_DMAC_47_16] = DEFINER_HL_OFFSET(eth_l2_outer.dmac_47_16),
		[MLX5DV_HWS_FNAME_DMAC_15_0] = DEFINER_HL_OFFSET(eth_l2_outer.dmac_15_0),
		[MLX5DV_HWS_FNAME_FIRST_PRIO] = DEFINER_HL_OFFSET(eth_l2_outer.first_priority),
		[MLX5DV_HWS_FNAME_FIRST_CFI] = DEFINER_HL_OFFSET(eth_l2_outer.first_cfi),
		[MLX5DV_HWS_FNAME_FIRST_VID] = DEFINER_HL_OFFSET(eth_l2_outer.first_vlan_id),
		[MLX5DV_HWS_FNAME_FIRST_CVLAN] = DEFINER_HL_OFFSET(eth_l2_outer.first_cvlan_tag),
		[MLX5DV_HWS_FNAME_FIRST_SVLAN] = DEFINER_HL_OFFSET(eth_l2_outer.first_svlan_tag),
		[MLX5DV_HWS_FNAME_FIRST_VLAN_TYPE] = DEFINER_HL_OFFSET(eth_l2_outer.first_vlan_qualifier),
		[MLX5DV_HWS_FNAME_SECOND_PRIO] = DEFINER_HL_OFFSET(eth_l2_outer.second_priority),
		[MLX5DV_HWS_FNAME_SECOND_CFI] = DEFINER_HL_OFFSET(eth_l2_outer.second_cfi),
		[MLX5DV_HWS_FNAME_SECOND_VID] = DEFINER_HL_OFFSET(eth_l2_outer.second_vlan_id),
		[MLX5DV_HWS_FNAME_SECOND_CVLAN] = DEFINER_HL_OFFSET(eth_l2_outer.second_cvlan_tag),
		[MLX5DV_HWS_FNAME_SECOND_SVLAN] = DEFINER_HL_OFFSET(eth_l2_outer.second_svlan_tag),
		[MLX5DV_HWS_FNAME_SECOND_VLAN_TYPE] = DEFINER_HL_OFFSET(eth_l2_outer.second_vlan_qualifier),
		[MLX5DV_HWS_FNAME_IP_VERSION] = DEFINER_HL_OFFSET(eth_l3_outer.ip_version),
		[MLX5DV_HWS_FNAME_IP_FRAG] = DEFINER_HL_OFFSET(eth_l2_outer.ip_fragmented),
		[MLX5DV_HWS_FNAME_IP_TTL] = DEFINER_HL_OFFSET(eth_l3_outer.time_to_live_hop_limit),
		[MLX5DV_HWS_FNAME_IP_SPORT] = DEFINER_HL_OFFSET(eth_l4_outer.source_port),
		[MLX5DV_HWS_FNAME_IP_DPORT] = DEFINER_HL_OFFSET(eth_l4_outer.destination_port),
		[MLX5DV_HWS_FNAME_IP_PROTOCOL] = DEFINER_HL_OFFSET(eth_l3_outer.protocol_next_header),
		[MLX5DV_HWS_FNAME_IP_TOS] = DEFINER_HL_OFFSET(eth_l3_outer.tos),
		[MLX5DV_HWS_FNAME_IP_TCP_SPORT] = DEFINER_HL_OFFSET(eth_l4_outer.source_port),
		[MLX5DV_HWS_FNAME_IP_TCP_DPORT] = DEFINER_HL_OFFSET(eth_l4_outer.destination_port),
		[MLX5DV_HWS_FNAME_IP_TCP_FLAGS] = DEFINER_HL_OFFSET(eth_l4_outer.tcp_flags),
		[MLX5DV_HWS_FNAME_IP_TCP_DATA_OFFSET] = DEFINER_HL_OFFSET(eth_l4_outer.data_offset),
		[MLX5DV_HWS_FNAME_IP_TCP_NS] = DEFINER_HL_OFFSET(eth_l4_outer.tcp_ns),
		[MLX5DV_HWS_FNAME_IP_TCP_SEQ] = DEFINER_HL_OFFSET(tcp_icmp.tcp_seq),
		[MLX5DV_HWS_FNAME_IP_TCP_ACK] = DEFINER_HL_OFFSET(tcp_icmp.tcp_ack),
		[MLX5DV_HWS_FNAME_IP_TCP_URG] = DEFINER_HL_OFFSET(tcp_icmp.tcp_urg),
		[MLX5DV_HWS_FNAME_IP_TCP_WIN] = DEFINER_HL_OFFSET(tcp_icmp.tcp_win),
		[MLX5DV_HWS_FNAME_IP_TCP_UDP_CSUM] = DEFINER_HL_OFFSET(l4_misc_outer.checksum),
		[MLX5DV_HWS_FNAME_IP_UDP_LEN] = DEFINER_HL_OFFSET(l4_misc_outer.length),
		[MLX5DV_HWS_FNAME_IP_UDP_SPORT] = DEFINER_HL_OFFSET(eth_l4_outer.source_port),
		[MLX5DV_HWS_FNAME_IP_UDP_DPORT] = DEFINER_HL_OFFSET(eth_l4_outer.destination_port),
		[MLX5DV_HWS_FNAME_IPV4_ID] = DEFINER_HL_OFFSET(eth_l3_outer.identification),
		[MLX5DV_HWS_FNAME_IPV4_IHL] = DEFINER_HL_OFFSET(eth_l3_outer.ihl),
		[MLX5DV_HWS_FNAME_IPV4_LEN] = DEFINER_HL_OFFSET(eth_l3_outer.ipv4_total_length),
		[MLX5DV_HWS_FNAME_IPV4_CHECKSUM] = DEFINER_HL_OFFSET(eth_l3_outer.checksum),
		[MLX5DV_HWS_FNAME_IPV4_FLAGS] = DEFINER_HL_OFFSET(eth_l3_outer.flags),
		[MLX5DV_HWS_FNAME_IPV4_FRAGMENT_OFFSET] = DEFINER_HL_OFFSET(eth_l3_outer.fragment_offset),
		[MLX5DV_HWS_FNAME_IPV4_SRC] = DEFINER_HL_OFFSET(ipv4_src_dest_outer.source_address),
		[MLX5DV_HWS_FNAME_IPV4_DST] = DEFINER_HL_OFFSET(ipv4_src_dest_outer.destination_address),
		[MLX5DV_HWS_FNAME_IPV6_FLOW_LABEL] = DEFINER_HL_OFFSET(eth_l3_outer.flow_label),
		[MLX5DV_HWS_FNAME_IPV6_PAYLOAD_LEN] = DEFINER_HL_OFFSET(eth_l3_outer.ipv6_payload_length),
		[MLX5DV_HWS_FNAME_IPV6_DST_127_96] = DEFINER_HL_OFFSET(ipv6_dst_outer.ipv6_address_127_96),
		[MLX5DV_HWS_FNAME_IPV6_DST_95_64] = DEFINER_HL_OFFSET(ipv6_dst_outer.ipv6_address_95_64),
		[MLX5DV_HWS_FNAME_IPV6_DST_63_32] = DEFINER_HL_OFFSET(ipv6_dst_outer.ipv6_address_63_32),
		[MLX5DV_HWS_FNAME_IPV6_DST_31_0] = DEFINER_HL_OFFSET(ipv6_dst_outer.ipv6_address_31_0),
		[MLX5DV_HWS_FNAME_IPV6_SRC_127_96] = DEFINER_HL_OFFSET(ipv6_src_outer.ipv6_address_127_96),
		[MLX5DV_HWS_FNAME_IPV6_SRC_95_64] = DEFINER_HL_OFFSET(ipv6_src_outer.ipv6_address_95_64),
		[MLX5DV_HWS_FNAME_IPV6_SRC_63_32] = DEFINER_HL_OFFSET(ipv6_src_outer.ipv6_address_63_32),
		[MLX5DV_HWS_FNAME_IPV6_SRC_31_0] = DEFINER_HL_OFFSET(ipv6_src_outer.ipv6_address_31_0),
		[MLX5DV_HWS_FNAME_TNL_HDR_0] = DEFINER_HL_OFFSET(tunnel_header.tunnel_header_0),
		[MLX5DV_HWS_FNAME_TNL_HDR_1] = DEFINER_HL_OFFSET(tunnel_header.tunnel_header_1),
		[MLX5DV_HWS_FNAME_TNL_HDR_2] = DEFINER_HL_OFFSET(tunnel_header.tunnel_header_2),
		[MLX5DV_HWS_FNAME_TNL_HDR_3] = DEFINER_HL_OFFSET(tunnel_header.tunnel_header_3),
		[MLX5DV_HWS_FNAME_IB_L4_OPCODE] = DEFINER_HL_OFFSET(ib_l4.opcode),
		[MLX5DV_HWS_FNAME_IB_L4_QPN] = DEFINER_HL_OFFSET(ib_l4.qp),
		[MLX5DV_HWS_FNAME_IB_L4_ACK_REQ] = DEFINER_HL_OFFSET(ib_l4.ackreq),
		[MLX5DV_HWS_FNAME_ICMP_DW0] = DEFINER_HL_OFFSET(tcp_icmp.icmp_dw0),
		[MLX5DV_HWS_FNAME_ICMP_DW1] = DEFINER_HL_OFFSET(tcp_icmp.icmp_dw1),
		[MLX5DV_HWS_FNAME_ICMP_DW2] = DEFINER_HL_OFFSET(tcp_icmp.icmp_dw2),
		[MLX5DV_HWS_FNAME_ICMP_TYPE] = DEFINER_HL_OFFSET(tcp_icmp.icmp_type),
		[MLX5DV_HWS_FNAME_ICMP_CODE] = DEFINER_HL_OFFSET(tcp_icmp.icmp_code),
		[MLX5DV_HWS_FNAME_IPSEC_LAYER] = DEFINER_HL_OFFSET(eth_l2_outer.ipsec_layer),
		[MLX5DV_HWS_FNAME_IPSEC_SPI] = DEFINER_HL_OFFSET(ipsec.spi),
		[MLX5DV_HWS_FNAME_IPSEC_NEXT_HEADER] = DEFINER_HL_OFFSET(ipsec.next_header),
		[MLX5DV_HWS_FNAME_IPSEC_SEQ_NUM] = DEFINER_HL_OFFSET(ipsec.sequence_number),
		[MLX5DV_HWS_FNAME_OK_L2] = DEFINER_HL_OFFSET(oks1.first_l2_ok),
		[MLX5DV_HWS_FNAME_OK_L3] = DEFINER_HL_OFFSET(oks1.first_l3_ok),
		[MLX5DV_HWS_FNAME_OK_L4] = DEFINER_HL_OFFSET(oks1.first_l4_ok),
		[MLX5DV_HWS_FNAME_OK_IPV4_CSUM] = DEFINER_HL_OFFSET(oks1.first_ipv4_checksum_ok),
		[MLX5DV_HWS_FNAME_OK_L4_CSUM] = DEFINER_HL_OFFSET(oks1.first_l4_checksum_ok),
		[MLX5DV_HWS_FNAME_PSP_DW0] = DEFINER_HL_OFFSET(psp.psp_dw0),
		[MLX5DV_HWS_FNAME_PSP_DW1] = DEFINER_HL_OFFSET(psp.psp_dw1),
		[MLX5DV_HWS_FNAME_PSP_DW2] = DEFINER_HL_OFFSET(psp.psp_dw2),
		[MLX5DV_HWS_FNAME_PSP_DW3] = DEFINER_HL_OFFSET(psp.psp_dw3),
		[MLX5DV_HWS_FNAME_PSP_DW4] = DEFINER_HL_OFFSET(psp.psp_dw4),
		[MLX5DV_HWS_FNAME_PSP_DW5] = DEFINER_HL_OFFSET(psp.psp_dw5),
		[MLX5DV_HWS_FNAME_GTP_TEID] = DEFINER_DYNAMIC_OFFSET(header_gtp, teid),
		[MLX5DV_HWS_FNAME_GTP_MSG_TYPE] = DEFINER_DYNAMIC_OFFSET(header_gtp, msg_type),
		[MLX5DV_HWS_FNAME_GTP_EXT_FLAG] = DEFINER_DYNAMIC_OFFSET(header_gtp, ext_hdr_flag),
		[MLX5DV_HWS_FNAME_GTP_NEXT_EXT_HDR] = DEFINER_DYNAMIC_OFFSET(header_opt_gtp, next_ext_hdr_type),
		[MLX5DV_HWS_FNAME_GTP_EXT_HDR_PDU] = DEFINER_DYNAMIC_OFFSET(header_gtp_psc, pdu_type),
		[MLX5DV_HWS_FNAME_GTP_EXT_HDR_QFI] = DEFINER_DYNAMIC_OFFSET(header_gtp_psc, qfi),
		[MLX5DV_HWS_FNAME_GRE_C_PRESENT] = DEFINER_HL_OFFSET(tunnel_header.gre.gre_c_present),
		[MLX5DV_HWS_FNAME_GRE_K_PRESENT] = DEFINER_HL_OFFSET(tunnel_header.gre.gre_k_present),
		[MLX5DV_HWS_FNAME_GRE_S_PRESENT] = DEFINER_HL_OFFSET(tunnel_header.gre.gre_s_present),
		[MLX5DV_HWS_FNAME_GRE_VERSION] = DEFINER_HL_OFFSET(tunnel_header.gre.version),
		[MLX5DV_HWS_FNAME_GRE_PROTOCOL] = DEFINER_HL_OFFSET(tunnel_header.gre.protocol),
		[MLX5DV_HWS_FNAME_GRE_CSUM] = DEFINER_HL_OFFSET(tunnel_header.gre.checksum),
		[MLX5DV_HWS_FNAME_GRE_KEY] = DEFINER_HL_OFFSET(tunnel_header.gre.key),
		[MLX5DV_HWS_FNAME_GRE_SEQ_NUM] = DEFINER_HL_OFFSET(tunnel_header.gre.sequence_num),
		[MLX5DV_HWS_FNAME_NVGRE_PROTOCOL] = DEFINER_HL_OFFSET(tunnel_header.nvgre.protocol),
		[MLX5DV_HWS_FNAME_NVGRE_VSID] = DEFINER_HL_OFFSET(tunnel_header.nvgre.vsid),
		[MLX5DV_HWS_FNAME_NVGRE_FLOW_ID] = DEFINER_HL_OFFSET(tunnel_header.nvgre.flow_id),
		[MLX5DV_HWS_FNAME_VXLAN_GPE_FLAGS] = DEFINER_HL_OFFSET(tunnel_header.vxlan_gpe.flags),
		[MLX5DV_HWS_FNAME_VXLAN_GPE_NEXT_PROTOCOL] = DEFINER_HL_OFFSET(tunnel_header.vxlan_gpe.protocol),
		[MLX5DV_HWS_FNAME_VXLAN_GPE_VNI] = DEFINER_HL_OFFSET(tunnel_header.vxlan_gpe.vni),
		[MLX5DV_HWS_FNAME_VXLAN_GPE_VNI_RSVD_DW1] = DEFINER_HL_OFFSET(tunnel_header.vxlan_gpe.rsvd1),
		[MLX5DV_HWS_FNAME_VXLAN_FLAGS] = DEFINER_HL_OFFSET(tunnel_header.vxlan.flags),
		[MLX5DV_HWS_FNAME_VXLAN_VNI] = DEFINER_HL_OFFSET(tunnel_header.vxlan.vni),
		[MLX5DV_HWS_FNAME_VXLAN_VNI_RSVD_DW1] = DEFINER_HL_OFFSET(tunnel_header.vxlan.reserved2),
		[MLX5DV_HWS_FNAME_VXLAN_GBP_GROUP_POLICY_ID] = DEFINER_HL_OFFSET(tunnel_header.vxlan_gbp.group_id),
		[MLX5DV_HWS_FNAME_VXLAN_GBP_VNI] = DEFINER_HL_OFFSET(tunnel_header.vxlan_gbp.vni),
		[MLX5DV_HWS_FNAME_MPLS0] = DEFINER_HL_OFFSET(mpls_outer.mpls0_label),
		[MLX5DV_HWS_FNAME_MPLS1] = DEFINER_HL_OFFSET(mpls_outer.mpls1_label),
		[MLX5DV_HWS_FNAME_MPLS2] = DEFINER_HL_OFFSET(mpls_outer.mpls2_label),
		[MLX5DV_HWS_FNAME_MPLS3] = DEFINER_HL_OFFSET(mpls_outer.mpls3_label),
		[MLX5DV_HWS_FNAME_MPLS4] = DEFINER_HL_OFFSET(mpls_outer.mpls4_label),
		[MLX5DV_HWS_FNAME_MPLS0_OK] = DEFINER_HL_OFFSET(oks2.first_mpls0_qualifier),
		[MLX5DV_HWS_FNAME_MPLS1_OK] = DEFINER_HL_OFFSET(oks2.first_mpls1_qualifier),
		[MLX5DV_HWS_FNAME_MPLS2_OK] = DEFINER_HL_OFFSET(oks2.first_mpls2_qualifier),
		[MLX5DV_HWS_FNAME_MPLS3_OK] = DEFINER_HL_OFFSET(oks2.first_mpls3_qualifier),
		[MLX5DV_HWS_FNAME_MPLS4_OK] = DEFINER_HL_OFFSET(oks2.first_mpls4_qualifier),
	},
	[MLX5DV_HWS_HEADER_INNER] = {
		[MLX5DV_HWS_FNAME_L2_TYPE] = DEFINER_HL_OFFSET(eth_l2_inner.l2_type),
		[MLX5DV_HWS_FNAME_L3_TYPE] = DEFINER_HL_OFFSET(eth_l2_inner.l3_type),
		[MLX5DV_HWS_FNAME_L4_TYPE] = DEFINER_HL_OFFSET(eth_l2_inner.l4_type),
		[MLX5DV_HWS_FNAME_L4_TYPE_BWC] = DEFINER_HL_OFFSET(eth_l2_inner.l4_type_bwc),
		[MLX5DV_HWS_FNAME_ETH_TYPE] = DEFINER_HL_OFFSET(eth_l2_inner.l3_ethertype),
		[MLX5DV_HWS_FNAME_SMAC_47_16] = DEFINER_HL_OFFSET(eth_l2_src_inner.smac_47_16),
		[MLX5DV_HWS_FNAME_SMAC_15_0] = DEFINER_HL_OFFSET(eth_l2_src_inner.smac_15_0),
		[MLX5DV_HWS_FNAME_DMAC_47_16] = DEFINER_HL_OFFSET(eth_l2_inner.dmac_47_16),
		[MLX5DV_HWS_FNAME_DMAC_15_0] = DEFINER_HL_OFFSET(eth_l2_inner.dmac_15_0),
		[MLX5DV_HWS_FNAME_FIRST_PRIO] = DEFINER_HL_OFFSET(eth_l2_inner.first_priority),
		[MLX5DV_HWS_FNAME_FIRST_CFI] = DEFINER_HL_OFFSET(eth_l2_inner.first_cfi),
		[MLX5DV_HWS_FNAME_FIRST_VID] = DEFINER_HL_OFFSET(eth_l2_inner.first_vlan_id),
		[MLX5DV_HWS_FNAME_FIRST_CVLAN] = DEFINER_HL_OFFSET(eth_l2_inner.first_cvlan_tag),
		[MLX5DV_HWS_FNAME_FIRST_SVLAN] = DEFINER_HL_OFFSET(eth_l2_inner.first_svlan_tag),
		[MLX5DV_HWS_FNAME_FIRST_VLAN_TYPE] = DEFINER_HL_OFFSET(eth_l2_inner.first_vlan_qualifier),
		[MLX5DV_HWS_FNAME_SECOND_PRIO] = DEFINER_HL_OFFSET(eth_l2_inner.second_priority),
		[MLX5DV_HWS_FNAME_SECOND_CFI] = DEFINER_HL_OFFSET(eth_l2_inner.second_cfi),
		[MLX5DV_HWS_FNAME_SECOND_VID] = DEFINER_HL_OFFSET(eth_l2_inner.second_vlan_id),
		[MLX5DV_HWS_FNAME_SECOND_CVLAN] = DEFINER_HL_OFFSET(eth_l2_inner.second_cvlan_tag),
		[MLX5DV_HWS_FNAME_SECOND_SVLAN] = DEFINER_HL_OFFSET(eth_l2_inner.second_svlan_tag),
		[MLX5DV_HWS_FNAME_SECOND_VLAN_TYPE] = DEFINER_HL_OFFSET(eth_l2_inner.second_vlan_qualifier),
		[MLX5DV_HWS_FNAME_IP_VERSION] = DEFINER_HL_OFFSET(eth_l3_inner.ip_version),
		[MLX5DV_HWS_FNAME_IP_FRAG] = DEFINER_HL_OFFSET(eth_l2_inner.ip_fragmented),
		[MLX5DV_HWS_FNAME_IP_TTL] = DEFINER_HL_OFFSET(eth_l3_inner.time_to_live_hop_limit),
		[MLX5DV_HWS_FNAME_IP_SPORT] = DEFINER_HL_OFFSET(eth_l4_inner.source_port),
		[MLX5DV_HWS_FNAME_IP_DPORT] = DEFINER_HL_OFFSET(eth_l4_inner.destination_port),
		[MLX5DV_HWS_FNAME_IP_PROTOCOL] = DEFINER_HL_OFFSET(eth_l3_inner.protocol_next_header),
		[MLX5DV_HWS_FNAME_IP_TOS] = DEFINER_HL_OFFSET(eth_l3_inner.tos),
		[MLX5DV_HWS_FNAME_IP_TCP_SPORT] = DEFINER_HL_OFFSET(eth_l4_inner.source_port),
		[MLX5DV_HWS_FNAME_IP_TCP_DPORT] = DEFINER_HL_OFFSET(eth_l4_inner.destination_port),
		[MLX5DV_HWS_FNAME_IP_TCP_FLAGS] = DEFINER_HL_OFFSET(eth_l4_inner.tcp_flags),
		[MLX5DV_HWS_FNAME_IP_TCP_DATA_OFFSET] = DEFINER_HL_OFFSET(eth_l4_inner.data_offset),
		[MLX5DV_HWS_FNAME_IP_TCP_NS] = DEFINER_HL_OFFSET(eth_l4_inner.tcp_ns),
		[MLX5DV_HWS_FNAME_IP_TCP_SEQ] = DEFINER_HL_OFFSET(tcp_icmp.tcp_seq),
		[MLX5DV_HWS_FNAME_IP_TCP_ACK] = DEFINER_HL_OFFSET(tcp_icmp.tcp_ack),
		[MLX5DV_HWS_FNAME_IP_TCP_URG] = DEFINER_HL_OFFSET(tcp_icmp.tcp_urg),
		[MLX5DV_HWS_FNAME_IP_TCP_WIN] = DEFINER_HL_OFFSET(tcp_icmp.tcp_win),
		[MLX5DV_HWS_FNAME_IP_TCP_UDP_CSUM] = DEFINER_HL_OFFSET(l4_misc_inner.checksum),
		[MLX5DV_HWS_FNAME_IP_UDP_LEN] = DEFINER_HL_OFFSET(l4_misc_inner.length),
		[MLX5DV_HWS_FNAME_IP_UDP_SPORT] = DEFINER_HL_OFFSET(eth_l4_inner.source_port),
		[MLX5DV_HWS_FNAME_IP_UDP_DPORT] = DEFINER_HL_OFFSET(eth_l4_inner.destination_port),
		[MLX5DV_HWS_FNAME_IPV4_ID] = DEFINER_HL_OFFSET(eth_l3_inner.identification),
		[MLX5DV_HWS_FNAME_IPV4_IHL] = DEFINER_HL_OFFSET(eth_l3_inner.ihl),
		[MLX5DV_HWS_FNAME_IPV4_LEN] = DEFINER_HL_OFFSET(eth_l3_inner.ipv4_total_length),
		[MLX5DV_HWS_FNAME_IPV4_CHECKSUM] = DEFINER_HL_OFFSET(eth_l3_inner.checksum),
		[MLX5DV_HWS_FNAME_IPV4_FLAGS] = DEFINER_HL_OFFSET(eth_l3_inner.flags),
		[MLX5DV_HWS_FNAME_IPV4_FRAGMENT_OFFSET] = DEFINER_HL_OFFSET(eth_l3_inner.fragment_offset),
		[MLX5DV_HWS_FNAME_IPV4_SRC] = DEFINER_HL_OFFSET(ipv4_src_dest_inner.source_address),
		[MLX5DV_HWS_FNAME_IPV4_DST] = DEFINER_HL_OFFSET(ipv4_src_dest_inner.destination_address),
		[MLX5DV_HWS_FNAME_IPV6_FLOW_LABEL] = DEFINER_HL_OFFSET(eth_l3_inner.flow_label),
		[MLX5DV_HWS_FNAME_IPV6_PAYLOAD_LEN] = DEFINER_HL_OFFSET(eth_l3_inner.ipv6_payload_length),
		[MLX5DV_HWS_FNAME_IPV6_DST_127_96] = DEFINER_HL_OFFSET(ipv6_dst_inner.ipv6_address_127_96),
		[MLX5DV_HWS_FNAME_IPV6_DST_95_64] = DEFINER_HL_OFFSET(ipv6_dst_inner.ipv6_address_95_64),
		[MLX5DV_HWS_FNAME_IPV6_DST_63_32] = DEFINER_HL_OFFSET(ipv6_dst_inner.ipv6_address_63_32),
		[MLX5DV_HWS_FNAME_IPV6_DST_31_0] = DEFINER_HL_OFFSET(ipv6_dst_inner.ipv6_address_31_0),
		[MLX5DV_HWS_FNAME_IPV6_SRC_127_96] = DEFINER_HL_OFFSET(ipv6_src_inner.ipv6_address_127_96),
		[MLX5DV_HWS_FNAME_IPV6_SRC_95_64] = DEFINER_HL_OFFSET(ipv6_src_inner.ipv6_address_95_64),
		[MLX5DV_HWS_FNAME_IPV6_SRC_63_32] = DEFINER_HL_OFFSET(ipv6_src_inner.ipv6_address_63_32),
		[MLX5DV_HWS_FNAME_IPV6_SRC_31_0] = DEFINER_HL_OFFSET(ipv6_src_inner.ipv6_address_31_0),
		[MLX5DV_HWS_FNAME_ICMP_DW0] = DEFINER_HL_OFFSET(tcp_icmp.icmp_dw0),
		[MLX5DV_HWS_FNAME_ICMP_DW1] = DEFINER_HL_OFFSET(tcp_icmp.icmp_dw1),
		[MLX5DV_HWS_FNAME_ICMP_DW2] = DEFINER_HL_OFFSET(tcp_icmp.icmp_dw2),
		[MLX5DV_HWS_FNAME_ICMP_TYPE] = DEFINER_HL_OFFSET(tcp_icmp.icmp_type),
		[MLX5DV_HWS_FNAME_ICMP_CODE] = DEFINER_HL_OFFSET(tcp_icmp.icmp_code),
		[MLX5DV_HWS_FNAME_IPSEC_LAYER] = DEFINER_HL_OFFSET(eth_l2_inner.ipsec_layer),
		[MLX5DV_HWS_FNAME_OK_L2] = DEFINER_HL_OFFSET(oks1.second_l2_ok),
		[MLX5DV_HWS_FNAME_OK_L3] = DEFINER_HL_OFFSET(oks1.second_l3_ok),
		[MLX5DV_HWS_FNAME_OK_L4] = DEFINER_HL_OFFSET(oks1.second_l4_ok),
		[MLX5DV_HWS_FNAME_OK_IPV4_CSUM] = DEFINER_HL_OFFSET(oks1.second_ipv4_checksum_ok),
		[MLX5DV_HWS_FNAME_OK_L4_CSUM] = DEFINER_HL_OFFSET(oks1.second_l4_checksum_ok),
		[MLX5DV_HWS_FNAME_MPLS0] = DEFINER_HL_OFFSET(mpls_inner.mpls0_label),
		[MLX5DV_HWS_FNAME_MPLS1] = DEFINER_HL_OFFSET(mpls_inner.mpls1_label),
		[MLX5DV_HWS_FNAME_MPLS2] = DEFINER_HL_OFFSET(mpls_inner.mpls2_label),
		[MLX5DV_HWS_FNAME_MPLS3] = DEFINER_HL_OFFSET(mpls_inner.mpls3_label),
		[MLX5DV_HWS_FNAME_MPLS4] = DEFINER_HL_OFFSET(mpls_inner.mpls4_label),
		[MLX5DV_HWS_FNAME_MPLS0_OK] = DEFINER_HL_OFFSET(oks2.second_mpls0_qualifier),
		[MLX5DV_HWS_FNAME_MPLS1_OK] = DEFINER_HL_OFFSET(oks2.second_mpls1_qualifier),
		[MLX5DV_HWS_FNAME_MPLS2_OK] = DEFINER_HL_OFFSET(oks2.second_mpls2_qualifier),
		[MLX5DV_HWS_FNAME_MPLS3_OK] = DEFINER_HL_OFFSET(oks2.second_mpls3_qualifier),
		[MLX5DV_HWS_FNAME_MPLS4_OK] = DEFINER_HL_OFFSET(oks2.second_mpls4_qualifier),
	},
};

static const struct definer_field_conv_offset
fco_root_arr[MLX5DV_HWS_HEADER_MAX][MLX5DV_HWS_FNAME_MAX] = {
	[MLX5DV_HWS_HEADER_NONE] = {
		[MLX5DV_HWS_FNAME_REG_C0] = DEFINER_MP_OFFSET(misc2.metadata_reg_c_0),
		[MLX5DV_HWS_FNAME_REG_C1] = DEFINER_MP_OFFSET(misc2.metadata_reg_c_1),
		[MLX5DV_HWS_FNAME_REG_C2] = DEFINER_MP_OFFSET(misc2.metadata_reg_c_2),
		[MLX5DV_HWS_FNAME_REG_C3] = DEFINER_MP_OFFSET(misc2.metadata_reg_c_3),
		[MLX5DV_HWS_FNAME_REG_C4] = DEFINER_MP_OFFSET(misc2.metadata_reg_c_4),
		[MLX5DV_HWS_FNAME_REG_C5] = DEFINER_MP_OFFSET(misc2.metadata_reg_c_5),
		[MLX5DV_HWS_FNAME_REG_C6] = DEFINER_MP_OFFSET(misc2.metadata_reg_c_6),
		[MLX5DV_HWS_FNAME_REG_C7] = DEFINER_MP_OFFSET(misc2.metadata_reg_c_7),
		[MLX5DV_HWS_FNAME_REG_A] = DEFINER_MP_OFFSET(misc2.metadata_reg_a),
		[MLX5DV_HWS_FNAME_SOURCE_QP] = DEFINER_MP_OFFSET(misc.source_sqn),
		[MLX5DV_HWS_FNAME_IPSEC_SYNDROME] = DEFINER_MP_OFFSET(misc2.ipsec_syndrome),
		[MLX5DV_HWS_FNAME_IPSEC_NEXT_HEADER] = DEFINER_MP_OFFSET(misc2.ipsec_next_header),
		[MLX5DV_HWS_FNAME_PSP_DW0] = DEFINER_MP_OFFSET(misc6.psp_header_0),
		[MLX5DV_HWS_FNAME_PSP_DW1] = DEFINER_MP_OFFSET(misc6.psp_header_1),
		[MLX5DV_HWS_FNAME_PSP_DW2] = DEFINER_MP_OFFSET(misc6.psp_header_2),
		[MLX5DV_HWS_FNAME_PSP_DW3] = DEFINER_MP_OFFSET(misc6.psp_header_3),
		[MLX5DV_HWS_FNAME_PSP_DW4] = DEFINER_MP_OFFSET(misc6.psp_header_4),
		[MLX5DV_HWS_FNAME_PSP_DW5] = DEFINER_MP_OFFSET(misc6.psp_header_5),
		[MLX5DV_HWS_FNAME_PSP_SYNDROME] = DEFINER_MP_OFFSET(misc2.psp_syndrome),
	},
	[MLX5DV_HWS_HEADER_OUTER] = {
		[MLX5DV_HWS_FNAME_L4_TYPE_BWC] = DEFINER_MP_OFFSET(outer.l4_type),
		[MLX5DV_HWS_FNAME_ETH_TYPE] = DEFINER_MP_OFFSET(outer.ethertype),
		[MLX5DV_HWS_FNAME_SMAC_47_16] = DEFINER_MP_OFFSET(outer.smac_47_16),
		[MLX5DV_HWS_FNAME_SMAC_15_0] = DEFINER_MP_OFFSET(outer.smac_15_0),
		[MLX5DV_HWS_FNAME_DMAC_47_16] = DEFINER_MP_OFFSET(outer.dmac_47_16),
		[MLX5DV_HWS_FNAME_DMAC_15_0] = DEFINER_MP_OFFSET(outer.dmac_15_0),
		[MLX5DV_HWS_FNAME_FIRST_PRIO] = DEFINER_MP_OFFSET(outer.first_prio),
		[MLX5DV_HWS_FNAME_FIRST_CFI] = DEFINER_MP_OFFSET(outer.first_cfi),
		[MLX5DV_HWS_FNAME_FIRST_VID] = DEFINER_MP_OFFSET(outer.first_vid),
		[MLX5DV_HWS_FNAME_FIRST_CVLAN] = DEFINER_MP_OFFSET(outer.cvlan_tag),
		[MLX5DV_HWS_FNAME_FIRST_SVLAN] = DEFINER_MP_OFFSET(outer.svlan_tag),
		[MLX5DV_HWS_FNAME_SECOND_PRIO] = DEFINER_MP_OFFSET(misc.outer_second_prio),
		[MLX5DV_HWS_FNAME_SECOND_CFI] = DEFINER_MP_OFFSET(misc.outer_second_cfi),
		[MLX5DV_HWS_FNAME_SECOND_VID] = DEFINER_MP_OFFSET(misc.outer_second_vid),
		[MLX5DV_HWS_FNAME_SECOND_CVLAN] = DEFINER_MP_OFFSET(misc.outer_second_cvlan_tag),
		[MLX5DV_HWS_FNAME_SECOND_SVLAN] = DEFINER_MP_OFFSET(misc.outer_second_svlan_tag),
		[MLX5DV_HWS_FNAME_IP_VERSION] = DEFINER_MP_OFFSET(outer.ip_version),
		[MLX5DV_HWS_FNAME_IP_FRAG] = DEFINER_MP_OFFSET(outer.frag),
		[MLX5DV_HWS_FNAME_IP_TTL] = DEFINER_MP_OFFSET(outer.ip_ttl_hoplimit),
		[MLX5DV_HWS_FNAME_IP_PROTOCOL] = DEFINER_MP_OFFSET(outer.ip_protocol),
		[MLX5DV_HWS_FNAME_IP_TOS] = DEFINER_MP_OFFSET(outer.tos),
		[MLX5DV_HWS_FNAME_IP_TCP_SPORT] = DEFINER_MP_OFFSET(outer.tcp_sport),
		[MLX5DV_HWS_FNAME_IP_TCP_DPORT] = DEFINER_MP_OFFSET(outer.tcp_dport),
		[MLX5DV_HWS_FNAME_IP_TCP_FLAGS] = DEFINER_MP_OFFSET(outer.tcp_flags),
		[MLX5DV_HWS_FNAME_IP_TCP_SEQ] = DEFINER_MP_OFFSET(misc3.outer_tcp_seq_num),
		[MLX5DV_HWS_FNAME_IP_TCP_ACK] = DEFINER_MP_OFFSET(misc3.outer_tcp_ack_num),
		[MLX5DV_HWS_FNAME_IP_UDP_SPORT] = DEFINER_MP_OFFSET(outer.udp_sport),
		[MLX5DV_HWS_FNAME_IP_UDP_DPORT] = DEFINER_MP_OFFSET(outer.udp_dport),
		[MLX5DV_HWS_FNAME_IPV4_IHL] = DEFINER_MP_OFFSET(outer.ipv4_ihl),
		[MLX5DV_HWS_FNAME_IPV4_SRC] = DEFINER_MP_OFFSET(outer.src_ip_31_0),
		[MLX5DV_HWS_FNAME_IPV4_DST] = DEFINER_MP_OFFSET(outer.dst_ip_31_0),
		[MLX5DV_HWS_FNAME_IPV6_FLOW_LABEL] = DEFINER_MP_OFFSET(misc.outer_ipv6_flow_label),
		[MLX5DV_HWS_FNAME_IPV6_DST_127_96] = DEFINER_MP_OFFSET(outer.dst_ip_127_96),
		[MLX5DV_HWS_FNAME_IPV6_DST_95_64] = DEFINER_MP_OFFSET(outer.dst_ip_95_64),
		[MLX5DV_HWS_FNAME_IPV6_DST_63_32] = DEFINER_MP_OFFSET(outer.dst_ip_63_32),
		[MLX5DV_HWS_FNAME_IPV6_DST_31_0] = DEFINER_MP_OFFSET(outer.dst_ip_31_0),
		[MLX5DV_HWS_FNAME_IPV6_SRC_127_96] = DEFINER_MP_OFFSET(outer.src_ip_127_96),
		[MLX5DV_HWS_FNAME_IPV6_SRC_95_64] = DEFINER_MP_OFFSET(outer.src_ip_95_64),
		[MLX5DV_HWS_FNAME_IPV6_SRC_63_32] = DEFINER_MP_OFFSET(outer.src_ip_63_32),
		[MLX5DV_HWS_FNAME_IPV6_SRC_31_0] = DEFINER_MP_OFFSET(outer.src_ip_31_0),
		[MLX5DV_HWS_FNAME_TNL_HDR_0] = DEFINER_MP_OFFSET(misc5.tunnel_header_0),
		[MLX5DV_HWS_FNAME_TNL_HDR_1] = DEFINER_MP_OFFSET(misc5.tunnel_header_1),
		[MLX5DV_HWS_FNAME_TNL_HDR_2] = DEFINER_MP_OFFSET(misc5.tunnel_header_2),
		[MLX5DV_HWS_FNAME_TNL_HDR_3] = DEFINER_MP_OFFSET(misc5.tunnel_header_3),
		[MLX5DV_HWS_FNAME_ICMP_TYPE] = DEFINER_MP_OFFSET(misc3.icmp_type),
		[MLX5DV_HWS_FNAME_ICMP_CODE] = DEFINER_MP_OFFSET(misc3.icmp_code),
		[MLX5DV_HWS_FNAME_IPSEC_SPI] = DEFINER_MP_OFFSET(misc.outer_esp_spi),
		[MLX5DV_HWS_FNAME_OK_L3] = DEFINER_MP_OFFSET(outer.l3_ok),
		[MLX5DV_HWS_FNAME_OK_L4] = DEFINER_MP_OFFSET(outer.l4_ok),
		[MLX5DV_HWS_FNAME_OK_IPV4_CSUM] = DEFINER_MP_OFFSET(outer.ipv4_checksum_ok),
		[MLX5DV_HWS_FNAME_OK_L4_CSUM] = DEFINER_MP_OFFSET(outer.l4_checksum_ok),
		[MLX5DV_HWS_FNAME_MPLS0] = DEFINER_MP_OFFSET(misc2.outer_first_mpls),
		[MLX5DV_HWS_FNAME_GTP_TEID] = DEFINER_MP_OFFSET(misc3.gtpu_teid),
		[MLX5DV_HWS_FNAME_GTP_MSG_TYPE] = DEFINER_MP_OFFSET(misc3.gtpu_dw_0_msg_type),
		[MLX5DV_HWS_FNAME_GTP_EXT_FLAG] = DEFINER_MP_OFFSET(misc3.gtpu_dw_0_ext_hdr_flag),
		[MLX5DV_HWS_FNAME_GTP_NEXT_EXT_HDR] = DEFINER_MP_OFFSET(misc3.gtpu_dw_2_next_ext_hdr_type),
		[MLX5DV_HWS_FNAME_GTP_EXT_HDR_PDU] = DEFINER_MP_OFFSET(misc3.gtpu_first_ext_dw_0_pdu_type),
		[MLX5DV_HWS_FNAME_GTP_EXT_HDR_QFI] = DEFINER_MP_OFFSET(misc3.gtpu_first_ext_dw_0_qfi),
	},
	[MLX5DV_HWS_HEADER_INNER] = {
		[MLX5DV_HWS_FNAME_L4_TYPE_BWC] = DEFINER_MP_OFFSET(inner.l4_type),
		[MLX5DV_HWS_FNAME_ETH_TYPE] = DEFINER_MP_OFFSET(inner.ethertype),
		[MLX5DV_HWS_FNAME_SMAC_47_16] = DEFINER_MP_OFFSET(inner.smac_47_16),
		[MLX5DV_HWS_FNAME_SMAC_15_0] = DEFINER_MP_OFFSET(inner.smac_15_0),
		[MLX5DV_HWS_FNAME_DMAC_47_16] = DEFINER_MP_OFFSET(inner.dmac_47_16),
		[MLX5DV_HWS_FNAME_DMAC_15_0] = DEFINER_MP_OFFSET(inner.dmac_15_0),
		[MLX5DV_HWS_FNAME_FIRST_PRIO] = DEFINER_MP_OFFSET(inner.first_prio),
		[MLX5DV_HWS_FNAME_FIRST_CFI] = DEFINER_MP_OFFSET(inner.first_cfi),
		[MLX5DV_HWS_FNAME_FIRST_VID] = DEFINER_MP_OFFSET(inner.first_vid),
		[MLX5DV_HWS_FNAME_FIRST_CVLAN] = DEFINER_MP_OFFSET(inner.cvlan_tag),
		[MLX5DV_HWS_FNAME_FIRST_SVLAN] = DEFINER_MP_OFFSET(inner.svlan_tag),
		[MLX5DV_HWS_FNAME_SECOND_PRIO] = DEFINER_MP_OFFSET(misc.inner_second_prio),
		[MLX5DV_HWS_FNAME_SECOND_CFI] = DEFINER_MP_OFFSET(misc.inner_second_cfi),
		[MLX5DV_HWS_FNAME_SECOND_VID] = DEFINER_MP_OFFSET(misc.inner_second_vid),
		[MLX5DV_HWS_FNAME_SECOND_CVLAN] = DEFINER_MP_OFFSET(misc.inner_second_cvlan_tag),
		[MLX5DV_HWS_FNAME_SECOND_SVLAN] = DEFINER_MP_OFFSET(misc.inner_second_svlan_tag),
		[MLX5DV_HWS_FNAME_IP_VERSION] = DEFINER_MP_OFFSET(inner.ip_version),
		[MLX5DV_HWS_FNAME_IP_FRAG] = DEFINER_MP_OFFSET(inner.frag),
		[MLX5DV_HWS_FNAME_IP_TTL] = DEFINER_MP_OFFSET(inner.ip_ttl_hoplimit),
		[MLX5DV_HWS_FNAME_IP_PROTOCOL] = DEFINER_MP_OFFSET(inner.ip_protocol),
		[MLX5DV_HWS_FNAME_IP_TOS] = DEFINER_MP_OFFSET(inner.tos),
		[MLX5DV_HWS_FNAME_IP_TCP_SPORT] = DEFINER_MP_OFFSET(inner.tcp_sport),
		[MLX5DV_HWS_FNAME_IP_TCP_DPORT] = DEFINER_MP_OFFSET(inner.tcp_dport),
		[MLX5DV_HWS_FNAME_IP_TCP_FLAGS] = DEFINER_MP_OFFSET(inner.tcp_flags),
		[MLX5DV_HWS_FNAME_IP_UDP_SPORT] = DEFINER_MP_OFFSET(inner.udp_sport),
		[MLX5DV_HWS_FNAME_IP_UDP_DPORT] = DEFINER_MP_OFFSET(inner.udp_dport),
		[MLX5DV_HWS_FNAME_IPV4_IHL] = DEFINER_MP_OFFSET(inner.ipv4_ihl),
		[MLX5DV_HWS_FNAME_IPV4_SRC] = DEFINER_MP_OFFSET(inner.src_ip_31_0),
		[MLX5DV_HWS_FNAME_IPV4_DST] = DEFINER_MP_OFFSET(inner.dst_ip_31_0),
		[MLX5DV_HWS_FNAME_IPV6_FLOW_LABEL] = DEFINER_MP_OFFSET(misc.inner_ipv6_flow_label),
		[MLX5DV_HWS_FNAME_IPV6_DST_127_96] = DEFINER_MP_OFFSET(inner.dst_ip_127_96),
		[MLX5DV_HWS_FNAME_IPV6_DST_95_64] = DEFINER_MP_OFFSET(inner.dst_ip_95_64),
		[MLX5DV_HWS_FNAME_IPV6_DST_63_32] = DEFINER_MP_OFFSET(inner.dst_ip_63_32),
		[MLX5DV_HWS_FNAME_IPV6_DST_31_0] = DEFINER_MP_OFFSET(inner.dst_ip_31_0),
		[MLX5DV_HWS_FNAME_IPV6_SRC_127_96] = DEFINER_MP_OFFSET(inner.src_ip_127_96),
		[MLX5DV_HWS_FNAME_IPV6_SRC_95_64] = DEFINER_MP_OFFSET(inner.src_ip_95_64),
		[MLX5DV_HWS_FNAME_IPV6_SRC_63_32] = DEFINER_MP_OFFSET(inner.src_ip_63_32),
		[MLX5DV_HWS_FNAME_IPV6_SRC_31_0] = DEFINER_MP_OFFSET(inner.src_ip_31_0),
		[MLX5DV_HWS_FNAME_IPSEC_SPI] = DEFINER_MP_OFFSET(misc.inner_esp_spi),
		[MLX5DV_HWS_FNAME_OK_L3] = DEFINER_MP_OFFSET(inner.l3_ok),
		[MLX5DV_HWS_FNAME_OK_L4] = DEFINER_MP_OFFSET(inner.l4_ok),
		[MLX5DV_HWS_FNAME_OK_IPV4_CSUM] = DEFINER_MP_OFFSET(inner.ipv4_checksum_ok),
		[MLX5DV_HWS_FNAME_OK_L4_CSUM] = DEFINER_MP_OFFSET(inner.l4_checksum_ok),
		[MLX5DV_HWS_FNAME_MPLS0] = DEFINER_MP_OFFSET(misc2.inner_first_mpls),
	},
};

/* Each row (i) indicates a different matcher size, and each column (j)
 * represents {DW5, DW4, DW3, DW2, DW1, DW0}.
 * For values 0,..,2^i, and j (DW) 0,..,5: optimal_dist_dw[i][j] is 1 if the
 * number of different hash results on these values equals 2^i, meaning this
 * DW hash distribution is complete.
 */
static int optimal_dist_dw[DEFINER_MAX_ROW_LOG][DW_SELECTORS_MATCH] = {
	{0, 0, 0, 0, 0, 0}, {1, 1, 1, 1, 1, 1}, {0, 1, 1, 0, 1, 0},
	{0, 1, 1, 0, 1, 0}, {1, 0, 1, 0, 1, 0}, {0, 0, 0, 1, 1, 0},
	{0, 1, 1, 0, 1, 0}, {0, 0, 0, 0, 1, 0}, {0, 1, 1, 0, 1, 0},
	{0, 0, 0, 0, 0, 0}, {1, 0, 1, 0, 0, 0}, {0, 0, 0, 0, 0, 0},
	{0, 1, 0, 1, 0, 0}, {1, 0, 0, 0, 0, 0}, {0, 0, 1, 0, 0, 1},
	{1, 1, 1, 0, 0, 0}, {1, 1, 1, 0, 1, 0}, {0, 0, 1, 1, 0, 0},
	{0, 1, 1, 0, 0, 1}, {0, 0, 1, 0, 0, 1}, {0, 0, 1, 0, 0, 0},
	{1, 0, 1, 1, 0, 0}, {1, 0, 1, 0, 0, 1}, {0, 0, 1, 1, 0, 1},
	{1, 1, 1, 0, 0, 0}, {0, 1, 0, 1, 0, 1}, {0, 0, 0, 0, 0, 1},
	{0, 0, 0, 1, 1, 1}, {0, 0, 1, 0, 0, 1}, {1, 1, 0, 1, 1, 0},
	{0, 0, 0, 0, 1, 0}, {0, 0, 0, 1, 1, 0}};

static uint8_t
definer_dw_selector_to_tag_dw(uint8_t dw_selector)
{
	/* DW8 -> Match_DW0, DW7 -> Match_DW1, ..., DW0 -> Match_DW8 */
	return (DW_SELECTORS - dw_selector - 1);
}

static int
definer_find_byte_in_tag(struct definer *definer,
			 uint32_t hl_byte_off,
			 uint32_t *tag_byte_off)
{
	uint8_t byte_offset;
	int i, dw_to_scan;

	/* Avoid accessing unused DW selectors */
	dw_to_scan = definer_is_jumbo(definer) ?
		DW_SELECTORS : DW_SELECTORS_MATCH;

	/* Add offset since each DW covers multiple BYTEs */
	byte_offset = hl_byte_off % DW_SIZE;
	for (i = 0; i < dw_to_scan; i++) {
		if (definer->dw_selector[i] == hl_byte_off / DW_SIZE) {
			*tag_byte_off = byte_offset + DW_SIZE * (DW_SELECTORS - i - 1);
			return 0;
		}
	}

	/* Add offset to skip DWs in definer */
	byte_offset = DW_SIZE * DW_SELECTORS;
	/* Iterate in reverse since the code uses bytes from 7 -> 0 */
	for (i = BYTE_SELECTORS; i-- > 0 ;) {
		if (definer->byte_selector[i] == hl_byte_off) {
			*tag_byte_off = byte_offset + (BYTE_SELECTORS - i - 1);
			return 0;
		}
	}

	/* The hl byte offset must be part of the definer */
	HWS_DBG("Failed to map to definer, HL byte [%d] not found", byte_offset);
	errno = EINVAL;
	return -errno;
}

static int definer_fc_bind(struct definer *definer,
			   struct definer_fc *fc,
			   uint32_t fc_sz)
{
	uint32_t byte_off, i, tag_offset = 0;
	int ret;

	for (i = 0; i < fc_sz; i++) {
		/* Map header layout byte offset to byte offset in tag */
		byte_off = fc->bit_off / BITS_IN_BYTE;
		ret = definer_find_byte_in_tag(definer, byte_off, &tag_offset);
		if (ret)
			return ret;

		/* Update offset in headers layout to offset in tag */
		fc->bit_off = fc->bit_off % BITS_IN_BYTE + tag_offset * BITS_IN_BYTE;
		fc++;
	}

	return 0;
}

static bool
definer_best_hl_fit_recu(struct definer_sel_ctrl *ctrl,
			 uint32_t cur_dw,
			 uint32_t *data)
{
	uint8_t bytes_set;
	int byte_idx;
	bool ret;
	int i;

	/* Reached end, nothing left to do */
	if (cur_dw == DEVX_ST_SZ_DW(definer_hl))
		return true;

	/* No data set, can skip to next DW */
	while (!*data) {
		cur_dw++;
		data++;

		/* Reached end, nothing left to do */
		if (cur_dw == DEVX_ST_SZ_DW(definer_hl))
			return true;
	}

	/* Used all DW selectors and Byte selectors, no possible solution */
	if (ctrl->allowed_full_dw == ctrl->used_full_dw &&
	    ctrl->allowed_lim_dw == ctrl->used_lim_dw &&
	    ctrl->allowed_bytes == ctrl->used_bytes)
		return false;

	/* Try to use limited DW selectors */
	if (ctrl->allowed_lim_dw > ctrl->used_lim_dw && cur_dw < 64) {
		ctrl->lim_dw_selector[ctrl->used_lim_dw++] = cur_dw;

		ret = definer_best_hl_fit_recu(ctrl, cur_dw + 1, data + 1);
		if (ret)
			return ret;

		ctrl->lim_dw_selector[--ctrl->used_lim_dw] = 0;
	}

	/* Try to use DW selectors */
	if (ctrl->allowed_full_dw > ctrl->used_full_dw) {
		ctrl->full_dw_selector[ctrl->used_full_dw++] = cur_dw;

		ret = definer_best_hl_fit_recu(ctrl, cur_dw + 1, data + 1);
		if (ret)
			return ret;

		ctrl->full_dw_selector[--ctrl->used_full_dw] = 0;
	}

	/* No byte selector for offset bigger than 255 */
	if (cur_dw * DW_SIZE > 255)
		return false;

	bytes_set = !!(0x000000ff & *data) +
		    !!(0x0000ff00 & *data) +
		    !!(0x00ff0000 & *data) +
		    !!(0xff000000 & *data);

	/* Check if there are enough byte selectors left */
	if (bytes_set + ctrl->used_bytes > ctrl->allowed_bytes)
		return false;

	/* Try to use Byte selectors */
	for (i = 0; i < DW_SIZE; i++)
		if ((0xff000000 >> (i * BITS_IN_BYTE)) & be32toh((__force __be32)*data)) {
			/* Use byte selectors high to low */
			byte_idx = ctrl->allowed_bytes - ctrl->used_bytes - 1;
			ctrl->byte_selector[byte_idx] = cur_dw * DW_SIZE + i;
			ctrl->used_bytes++;
		}

	ret = definer_best_hl_fit_recu(ctrl, cur_dw + 1, data + 1);
	if (ret)
		return ret;

	for (i = 0; i < DW_SIZE; i++)
		if ((0xff << (i * BITS_IN_BYTE)) & be32toh((__force __be32)*data)) {
			ctrl->used_bytes--;
			byte_idx = ctrl->allowed_bytes - ctrl->used_bytes - 1;
			ctrl->byte_selector[byte_idx] = 0;
		}

	return false;
}

static void
definer_copy_sel_ctrl(struct definer_sel_ctrl *ctrl,
		      struct definer *definer)
{
	memcpy(definer->byte_selector, ctrl->byte_selector, ctrl->allowed_bytes);
	memcpy(definer->dw_selector, ctrl->full_dw_selector, ctrl->allowed_full_dw);
	memcpy(definer->dw_selector + ctrl->allowed_full_dw,
	       ctrl->lim_dw_selector, ctrl->allowed_lim_dw);
}

static int
definer_find_best_range_fit(struct definer *definer,
			    struct mlx5dv_hws_matcher *matcher)
{
	uint32_t tag_bit_offset[MLX5DV_HWS_FNAME_MAX] = {0};
	uint8_t field_select[MLX5DV_HWS_FNAME_MAX] = {0};
	struct definer_sel_ctrl ctrl = {0};
	uint32_t algn_byte_off;
	struct definer_fc *fcr;
	uint32_t bit_offset;
	bool require_dw;
	int idx, i, j;

	/* Try to create a range definer */
	ctrl.allowed_full_dw = DW_SELECTORS_RANGE;
	ctrl.allowed_bytes = BYTE_SELECTORS_RANGE;

	/* Multiple fields cannot share the same DW for range match.
	 * The HW doesn't recognize each field but compares the full dw.
	 * For example definer DW consists of FieldA_FieldB
	 * FieldA: Mask 0xFFFF range 0x1 to 0x2
	 * FieldB: Mask 0xFFFF range 0x3 to 0x4
	 * STE DW range will be 0x00010003 - 0x00020004
	 * This will cause invalid match for FieldB if FieldA=1 and FieldB=8
	 * Since 0x10003 < 0x10008 < 0x20004
	 */
	for (i = 0; i < matcher->num_of_mt; i++) {
		for (j = 0; j < matcher->mt[i].fcr_sz; j++) {
			fcr = &(matcher->mt[i].fcr[j]);

			/* Found - Reuse previous mt binding */
			if (field_select[fcr->ctrl.fname]) {
				fcr->bit_off = tag_bit_offset[fcr->ctrl.fname];
				continue;
			}

			/* Not found */
			require_dw = fcr->bit_off >= (64 * BITS_IN_DW);
			if (require_dw || ctrl.used_bytes == ctrl.allowed_bytes) {
				/* Try to cover using DW selector */
				if (ctrl.used_full_dw == ctrl.allowed_full_dw)
					goto not_supported;

				ctrl.full_dw_selector[ctrl.used_full_dw++] =
					fcr->bit_off / BITS_IN_DW;

				/* Bind DW */
				idx = ctrl.used_full_dw - 1;
				bit_offset = fcr->bit_off % BITS_IN_DW;
				bit_offset += BITS_IN_DW * (DW_SELECTORS - idx - 1);
			} else {
				/* Try to cover using Bytes selectors */
				if (ctrl.used_bytes == ctrl.allowed_bytes)
					goto not_supported;

				algn_byte_off = DW_SIZE * (fcr->bit_off / BITS_IN_DW);
				ctrl.byte_selector[ctrl.used_bytes++] = algn_byte_off + 3;
				ctrl.byte_selector[ctrl.used_bytes++] = algn_byte_off + 2;
				ctrl.byte_selector[ctrl.used_bytes++] = algn_byte_off + 1;
				ctrl.byte_selector[ctrl.used_bytes++] = algn_byte_off;

				/* Bind BYTE */
				bit_offset = BITS_IN_DW * DW_SELECTORS;
				bit_offset += BITS_IN_BYTE * (BYTE_SELECTORS - ctrl.used_bytes);
				bit_offset += fcr->bit_off % BITS_IN_DW;
				/* No need to bind BIT since we always take full DW */
			}

			fcr->bit_off = bit_offset;
			tag_bit_offset[fcr->ctrl.fname] = bit_offset;
			field_select[fcr->ctrl.fname] = 1;
		}
	}

	definer_copy_sel_ctrl(&ctrl, definer);
	definer->type = DEFINER_TYPE_RANGE;

	return 0;

not_supported:
	HWS_ERR("Unable to find supporting range definer combination");
	errno = ENOTSUP;
	return -errno;
}

static void definer_optimize_order(struct definer *definer, int num_log)
{
	uint8_t hl_prio[] = {
		DEVX_BYTE_OFF(definer_hl, ipv4_src_dest_outer.source_address) / DW_SIZE,
		DEVX_BYTE_OFF(definer_hl, ipv4_src_dest_outer.destination_address) / DW_SIZE};
	size_t hl_prio_sz = sizeof(hl_prio)/sizeof(hl_prio[0]);
	int dw = 0, i = 0, j;
	int *dw_flag;
	uint8_t tmp;

	dw_flag = optimal_dist_dw[num_log];

	while (i < hl_prio_sz) {
		j = 0;
		/* Finding a candidate to improve its hash distribution */
		while (j < DW_SELECTORS_MATCH && (hl_prio[i] != definer->dw_selector[j]))
			j++;

		/* Finding a DW location with good hash distribution */
		while (dw < DW_SELECTORS_MATCH && dw_flag[dw] == 0)
			dw++;

		if (dw < DW_SELECTORS_MATCH && j < DW_SELECTORS_MATCH) {
			tmp = definer->dw_selector[dw];
			definer->dw_selector[dw] = definer->dw_selector[j];
			definer->dw_selector[j] = tmp;
			dw++;
		}
		i++;
	}
}

static int
definer_find_best_match_fit(struct mlx5dv_hws_matcher *matcher,
			    struct definer *definer)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	struct definer_sel_ctrl ctrl = {0};
	uint32_t *hl;
	bool found;
	int i, j;

	/* Union header-layout (hl) is used for creating a single definer
	 * field layout used with different bitmasks for hash and match.
	 */
	hl = calloc(1, DEVX_ST_SZ_BYTES(definer_hl));
	if (!hl) {
		HWS_ERR("Failed to allocate memory for header layout");
		errno = ENOMEM;
		return -errno;
	}

	for (i = 0; i < matcher->num_of_mt; i++)
		for (j = 0; j < matcher->mt[i].fcm_sz; j++)
			HWS_SET(hl, -1,
				matcher->mt[i].fcm[j].bit_off,
				matcher->mt[i].fcm[j].bit_sz);

	/* Try to create a match definer */
	ctrl.allowed_full_dw = DW_SELECTORS_MATCH;
	ctrl.allowed_lim_dw = 0;
	ctrl.allowed_bytes = BYTE_SELECTORS;

	found = definer_best_hl_fit_recu(&ctrl, 0, hl);
	if (found) {
		definer_copy_sel_ctrl(&ctrl, definer);
		definer->type = DEFINER_TYPE_MATCH;
		free(hl);
		return 0;
	}

	/* Try to create a full/limited jumbo definer */
	ctrl.allowed_full_dw = ctx->caps->full_dw_jumbo_support ? DW_SELECTORS :
								  DW_SELECTORS_MATCH;
	ctrl.allowed_lim_dw = ctx->caps->full_dw_jumbo_support ? 0 :
								 DW_SELECTORS_LIMITED;
	ctrl.allowed_bytes = BYTE_SELECTORS;

	found = definer_best_hl_fit_recu(&ctrl, 0, hl);
	if (found) {
		definer_copy_sel_ctrl(&ctrl, definer);
		definer->type = DEFINER_TYPE_JUMBO;
		free(hl);
		return 0;
	}

	HWS_DBG("Unable to find supporting match/jumbo definer combination");
	errno = E2BIG;
	free(hl);
	return -errno;
}

static uint32_t
definer_fc_read_field(struct definer_fc *fc,
		      const struct mlx5dv_hws_item_data *data,
		      const void *match,
		      enum mlx5dv_hws_field_user_conv_flag flag)
{
	if (unlikely(fc->user_conv_func)) {
		uint32_t val;

		fc->user_conv_func(data, fc->user_conv_data, flag, &val);
		return val;
	}

	return HWS_GET(match, fc->src_bit_off, fc->bit_sz);
}

static uint32_t
definer_fc_read_field_mask(struct definer_fc *fc,
			   const struct mlx5dv_hws_item_data *data)
{
	return definer_fc_read_field(fc,
				     data,
				     data->mask,
				     MLX5DV_HWS_FIELD_USER_CONV_FLAG_MASK);
}

void definer_create_tag_mask(struct definer_fc *fc,
			     uint32_t fc_sz,
			     uint8_t *tag)
{
	uint32_t i;

	for (i = 0; i < fc_sz; i++) {
		HWS_SET(tag, fc->ctrl.mask, fc->bit_off, fc->bit_sz);
		fc++;
	}
}

void definer_create_tag(struct mlx5dv_hws_item_data *data,
			struct definer_fc *fc,
			uint32_t fc_sz,
			uint8_t *tag)
{
	uint32_t i, val;

	for (i = 0; i < fc_sz; i++) {
		val = definer_fc_read_field(fc,
					    &data[fc->item_idx],
					    data[fc->item_idx].match,
					    MLX5DV_HWS_FIELD_USER_CONV_FLAG_MATCH);

		HWS_SET(tag, val, fc->bit_off, fc->bit_sz);
		fc++;
	}
}

static uint32_t definer_get_range_bit_off(uint32_t match_bit_off)
{
	uint8_t curr_dw_idx = match_bit_off / BITS_IN_DW;
	uint8_t new_dw_idx;

	/* Range DW can have the following values 7,8,9,10
	 * -DW7 is mapped to DW9
	 * -DW8 is mapped to DW7
	 * -DW9 is mapped to DW5
	 * -DW10 is mapped to DW3
	 * To reduce calculation the following formula is used:
	 */
	new_dw_idx = curr_dw_idx * (-2) + 23;

	return new_dw_idx * BITS_IN_DW + match_bit_off % BITS_IN_DW;
}

void definer_create_tag_range(struct mlx5dv_hws_item_data *data,
			      struct definer_fc *fcr,
			      uint32_t fcr_sz,
			      uint8_t *tag)
{
	uint32_t i, bit_off, max, min;

	for (i = 0; i < fcr_sz; i++) {
		max = definer_fc_read_field(fcr,
					    &data[fcr->item_idx],
					    data[fcr->item_idx].last,
					    MLX5DV_HWS_FIELD_USER_CONV_FLAG_LAST);

		min = definer_fc_read_field(fcr,
					    &data[fcr->item_idx],
					    data[fcr->item_idx].match,
					    MLX5DV_HWS_FIELD_USER_CONV_FLAG_MATCH);

		/* Set MAX and MIN value */
		bit_off = definer_get_range_bit_off(fcr->bit_off);
		HWS_SET(tag, max, bit_off, fcr->bit_sz);
		HWS_SET(tag, min, bit_off + BITS_IN_DW, fcr->bit_sz);
		fcr++;
	}
}

static void
definer_translate_compare_op(enum mlx5dv_hws_field_match_op op,
			     uint8_t *operator,
			     bool *inverse)
{
	switch (op) {
	case MLX5DV_HWS_FIELD_MATCH_OP_EQ:
		*operator = MLX5_IFC_STE_MATCH_4DW_RANGE_MATCH_OP_BETWEEN;
		*inverse = false;
		break;
	case MLX5DV_HWS_FIELD_MATCH_OP_NE:
		*operator = MLX5_IFC_STE_MATCH_4DW_RANGE_MATCH_OP_BETWEEN;
		*inverse = true;
		break;
	case MLX5DV_HWS_FIELD_MATCH_OP_LT:
		*operator = MLX5_IFC_STE_MATCH_4DW_RANGE_MATCH_OP_BIGGER;
		*inverse = true;
		break;
	case MLX5DV_HWS_FIELD_MATCH_OP_LE:
		*operator = MLX5_IFC_STE_MATCH_4DW_RANGE_MATCH_OP_SMALLER;
		*inverse = false;
		break;
	case MLX5DV_HWS_FIELD_MATCH_OP_GT:
		*operator = MLX5_IFC_STE_MATCH_4DW_RANGE_MATCH_OP_SMALLER;
		*inverse = true;
		break;
	case MLX5DV_HWS_FIELD_MATCH_OP_GE:
		*operator = MLX5_IFC_STE_MATCH_4DW_RANGE_MATCH_OP_BIGGER;
		*inverse = false;
		break;
	default:
		HWS_ERR("Invalid match op [%d]", op);
		assert(false);
	}
}

void definer_create_tag_compare(struct mlx5dv_hws_item_data *data,
				struct definer_fc *fcc,
				uint32_t fcc_sz,
				uint8_t *tag)
{
	uint8_t op, bit_shift;
	bool compare_val;
	uint32_t val;
	bool inverse;

	/* Set number of operations 0-One match, 1-Two matches */
	DEVX_SET(ste_match_4dw_range_ctrl_dw, tag, num_cond, fcc_sz - 1);

	if (fcc_sz >= 1) {
		compare_val = fcc[0].ctrl.type == MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
		definer_translate_compare_op(fcc[0].ctrl.op, &op, &inverse);
		DEVX_SET(ste_match_4dw_range_ctrl_dw, tag, cond0_base_src, compare_val);
		DEVX_SET(ste_match_4dw_range_ctrl_dw, tag, cond0_operator, op);
		DEVX_SET(ste_match_4dw_range_ctrl_dw, tag, cond0_inverse, inverse);

		/* Compare to DW4 in definer to value or to DW2 */
		if (compare_val) {
			val = definer_fc_read_field(&fcc[0],
						    &data[fcc[0].item_idx],
						    data[fcc[0].item_idx].match,
						    MLX5DV_HWS_FIELD_USER_CONV_FLAG_MATCH);
			bit_shift = BITS_IN_DW - fcc[0].bit_sz - fcc[0].bit_off % BITS_IN_DW;
			DEVX_SET(ste_match_4dw_range_ctrl_dw, tag, match_base_0, val << bit_shift);
		}
	}

	if (fcc_sz == 2) {
		compare_val = fcc[1].ctrl.type == MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE;
		definer_translate_compare_op(fcc[1].ctrl.op, &op, &inverse);
		DEVX_SET(ste_match_4dw_range_ctrl_dw, tag, cond1_base_src, compare_val);
		DEVX_SET(ste_match_4dw_range_ctrl_dw, tag, cond1_operator, op);
		DEVX_SET(ste_match_4dw_range_ctrl_dw, tag, cond1_inverse, inverse);

		/* Compare to DW5 in definer to value or to DW3 */
		if (compare_val) {
			val = definer_fc_read_field(&fcc[1],
						    &data[fcc[1].item_idx],
						    data[fcc[1].item_idx].match,
						    MLX5DV_HWS_FIELD_USER_CONV_FLAG_MATCH);
			bit_shift = BITS_IN_DW - fcc[1].bit_sz - fcc[1].bit_off % BITS_IN_DW;
			DEVX_SET(ste_match_4dw_range_ctrl_dw, tag, match_base_1, val << bit_shift);
		}
	}
}

int definer_get_id(struct definer *definer)
{
	return definer->obj->object_id;
}

int definer_compare(struct definer *definer_a,
		    struct definer *definer_b)
{
	int i;

	/* Future: Optimize by comparing selectors with valid mask only */
	for (i = 0; i < BYTE_SELECTORS; i++)
		if (definer_a->byte_selector[i] != definer_b->byte_selector[i])
			return 1;

	for (i = 0; i < DW_SELECTORS; i++)
		if (definer_a->dw_selector[i] != definer_b->dw_selector[i])
			return 1;

	for (i = 0; i < STE_JUMBO_TAG_SZ; i++)
		if (definer_a->mask->jumbo[i] != definer_b->mask->jumbo[i])
			return 1;

	return 0;
}

static int
definer_optimize_order_supported(struct definer *match_definer,
				 struct mlx5dv_hws_matcher *matcher)
{
	return !definer_is_jumbo(match_definer) &&
	       !matcher_req_fw_wqe(matcher) &&
	       !matcher_is_resizable(matcher) &&
	       !matcher_is_insert_by_idx(matcher);
}

static int
definer_calc_layout(struct mlx5dv_hws_matcher *matcher,
		    struct definer *match_definer,
		    struct definer *range_definer)
{
	int ret;

	/* Find the match definer layout for match templates fcs */
	ret = definer_find_best_match_fit(matcher, match_definer);
	if (ret) {
		HWS_DBG("Failed to create match definer from header layout");
		return ret;
	}

	if (definer_optimize_order_supported(match_definer, matcher))
		definer_optimize_order(match_definer, matcher->attr.rule.num_log);

	/* Find the range definer layout for match templates fcrs */
	ret = definer_find_best_range_fit(range_definer, matcher);
	if (ret) {
		HWS_ERR("Failed to create range definer from header layout");
		return ret;
	}

	return 0;
}

int definer_cache_init(struct definer_cache **cache)
{
	struct definer_cache *new_cache;

	new_cache = calloc(1, sizeof(*new_cache));
	if (!new_cache) {
		errno = ENOMEM;
		return -errno;
	}
	list_head_init(&new_cache->head);
	*cache = new_cache;

	return 0;
}

void definer_cache_uninit(struct definer_cache *cache)
{
	free(cache);
}

static struct mlx5dv_devx_obj *definer_get_obj(struct mlx5dv_hws_context *ctx,
					       struct definer *definer)
{
	struct definer_cache *cache = ctx->definer_cache;
	struct cmd_definer_create_attr def_attr = {0};
	struct definer_cache_item *cached_definer;
	struct mlx5dv_devx_obj *obj;

	/* Search definer cache for requested definer */
	list_for_each(&cache->head, cached_definer, next) {
		if (definer_compare(&cached_definer->definer, definer))
			continue;

		/* Reuse definer and set LRU (move to be first in the list) */
		list_del(&cached_definer->next);
		list_add(&cache->head, &cached_definer->next);
		cached_definer->refcount++;
		return cached_definer->definer.obj;
	}

	/* Allocate and create definer based on the bitmask tag */
	def_attr.match_mask = definer->mask->jumbo;
	def_attr.dw_selector = definer->dw_selector;
	def_attr.byte_selector = definer->byte_selector;

	obj = cmd_definer_create(ctx->ibv_ctx, &def_attr);
	if (!obj)
		return NULL;

	cached_definer = calloc(1, sizeof(*cached_definer));
	if (!cached_definer) {
		errno = ENOMEM;
		goto free_definer_obj;
	}

	cached_definer->definer.mask = calloc(1, sizeof(struct rule_match_tag));
	if (!cached_definer->definer.mask) {
		HWS_ERR("Failed to allocate mask for cached definer");
		errno = ENOMEM;
		goto free_cached_definer;
	}

	cached_definer->definer.type = definer->type;
	memcpy(cached_definer->definer.byte_selector, definer->byte_selector,
	       BYTE_SELECTORS);
	memcpy(cached_definer->definer.dw_selector, definer->dw_selector,
	       DW_SELECTORS);
	memcpy(cached_definer->definer.mask, definer->mask,
	       sizeof(struct rule_match_tag));
	cached_definer->definer.obj = obj;
	cached_definer->refcount = 1;
	list_add(&cache->head, &cached_definer->next);

	return obj;

free_cached_definer:
	free(cached_definer);
free_definer_obj:
	cmd_destroy_obj(obj);
	return NULL;
}

static void definer_put_obj(struct mlx5dv_hws_context *ctx,
			    struct mlx5dv_devx_obj *obj)
{
	struct definer_cache_item *cached_definer;

	list_for_each(&ctx->definer_cache->head, cached_definer, next) {
		if (cached_definer->definer.obj != obj)
			continue;

		/* Object found */
		if (--cached_definer->refcount)
			return;

		list_del(&cached_definer->next);
		cmd_destroy_obj(cached_definer->definer.obj);
		free(cached_definer->definer.mask);
		free(cached_definer);
		return;
	}

	/* Programming error, object must be part of cache */
	assert(false);
}

static struct definer *definer_alloc(struct mlx5dv_hws_context *ctx,
				     struct definer_fc *fc,
				     int fc_sz,
				     struct definer *layout,
				     bool bind_fc)
{
	struct definer *definer;
	int ret;

	definer = calloc(1, sizeof(*definer));
	if (!definer) {
		HWS_ERR("Failed to allocate memory for definer");
		errno = ENOMEM;
		return NULL;
	}

	memcpy(definer, layout, sizeof(*definer));

	definer->mask = calloc(1, sizeof(struct rule_match_tag));
	if (!definer->mask) {
		HWS_ERR("Failed to allocate mask for definer");
		errno = ENOMEM;
		goto free_definer;
	}

	/* Align field copy array based on given layout */
	if (bind_fc) {
		ret = definer_fc_bind(definer, fc, fc_sz);
		if (ret) {
			HWS_ERR("Failed to bind field copy to definer");
			goto free_mask;
		}
	}

	if (layout->mask) {
		/* Use layout mask if provided */
		memcpy(definer->mask, layout->mask, sizeof(struct rule_match_tag));
	} else {
		/* Create the tag mask used for definer creation */
		definer_create_tag_mask(fc, fc_sz, definer->mask->jumbo);
	}

	definer->obj = definer_get_obj(ctx, definer);
	if (!definer->obj)
		goto free_mask;

	return definer;

free_mask:
	free(definer->mask);
free_definer:
	free(definer);
	return NULL;
}

static void definer_free(struct mlx5dv_hws_context *ctx,
			 struct definer *definer)
{
	definer_put_obj(ctx, definer->obj);
	free(definer->mask);
	free(definer);
}

static int
definer_matcher_match_init(struct mlx5dv_hws_context *ctx,
			   struct mlx5dv_hws_matcher *matcher,
			   struct definer *match_layout)
{
	struct mlx5dv_hws_mt *mt = matcher->mt;
	int i;

	/* Create mandatory match definer */
	for (i = 0; i < matcher->num_of_mt; i++) {
		mt[i].definer = definer_alloc(ctx,
					      mt[i].fcm,
					      mt[i].fcm_sz,
					      match_layout,
					      true);
		if (!mt[i].definer) {
			HWS_ERR("Failed to create match definer");
			goto free_definers;
		}
	}
	return 0;

free_definers:
	while (i--)
		definer_free(ctx, mt[i].definer);

	return -errno;
}

static void
definer_matcher_match_uninit(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	int i;

	for (i = 0; i < matcher->num_of_mt; i++)
		definer_free(ctx, matcher->mt[i].definer);
}

static int
definer_matcher_range_init(struct mlx5dv_hws_context *ctx,
			   struct mlx5dv_hws_matcher *matcher,
			   struct definer *range_layout)
{
	struct mlx5dv_hws_mt *mt = matcher->mt;
	int i;

	/* Create optional range definers */
	for (i = 0; i < matcher->num_of_mt; i++) {
		bool is_range = !!mt[i].fcr_sz;
		bool has_range = matcher->flags & MATCHER_FLAGS_RANGE_DEFINER;

		/* All must use range if requested */
		if (i && ((is_range && !has_range) || (!is_range && has_range))) {
			HWS_ERR("Using range and non range templates is not allowed");
			goto free_definers;
		}

		if (!mt[i].fcr_sz)
			continue;

		matcher->flags |= MATCHER_FLAGS_RANGE_DEFINER;
		/* Create definer without fcr binding, already binded */
		mt[i].range_definer = definer_alloc(ctx,
						    mt[i].fcr,
						    mt[i].fcr_sz,
						    range_layout,
						    false);
		if (!mt[i].range_definer) {
			HWS_ERR("Failed to create range definer");
			goto free_definers;
		}
	}
	return 0;

free_definers:
	while (i--)
		if (mt[i].range_definer)
			definer_free(ctx, mt[i].range_definer);

	return -errno;
}

static void
definer_matcher_range_uninit(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	int i;

	for (i = 0; i < matcher->num_of_mt; i++)
		if (matcher->mt[i].range_definer)
			definer_free(ctx, matcher->mt[i].range_definer);
}

static int
definer_matcher_compare_init(struct mlx5dv_hws_context *ctx,
			     struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_mt *mt = matcher->mt;
	uint8_t *dw2, *dw3, *dw4, *dw5;
	int i, j;

	/* Create optional compare definers */
	for (i = 0; i < matcher->num_of_mt; i++) {
		bool has_compare = matcher->flags & MATCHER_FLAGS_COMPARE_DEFINER;
		struct definer compare_layout = {0};
		bool is_compare = !!mt[i].fcc_sz;
		struct rule_match_tag mask = {0};
		uint32_t *tag;

		/* All must use compare if requested */
		if (i && ((is_compare && !has_compare) || (!is_compare && has_compare))) {
			HWS_ERR("Using compare and non compare templates is not allowed");
			goto free_definers;
		}

		if (!mt[i].fcc_sz)
			continue;

		matcher->flags |= MATCHER_FLAGS_COMPARE_DEFINER;

		/* Due some HW limitation need to fill unused
		 * DW's 0-5 and byte selectors with 0xff.
		 */
		for (j = 0; j < DW_SELECTORS_MATCH; j++)
			compare_layout.dw_selector[j] = 0xff;

		for (j = 0; j < BYTE_SELECTORS; j++)
			compare_layout.byte_selector[j] = 0xff;

		dw2 = &compare_layout.dw_selector[DEFINER_SELECTOR_DW2];
		dw3 = &compare_layout.dw_selector[DEFINER_SELECTOR_DW3];
		dw4 = &compare_layout.dw_selector[DEFINER_SELECTOR_DW4];
		dw5 = &compare_layout.dw_selector[DEFINER_SELECTOR_DW5];

		tag = (uint32_t *)mask.jumbo;

		/* Compare DW4 to DW2/Value */
		if (mt[i].fcc_sz >= 1) {
			*dw4 = mt[i].fcc[0].bit_off / BITS_IN_DW;
			tag[definer_dw_selector_to_tag_dw(DEFINER_SELECTOR_DW4)] = -1;

			if (mt[i].fcc[0].ctrl.type ==
			    MLX5DV_HWS_FIELD_MATCH_TYPE_TO_FIELD) {
				*dw2 = mt[i].fcc[0].ctrl.compare_dw;
				tag[definer_dw_selector_to_tag_dw(DEFINER_SELECTOR_DW2)] = -1;
			}
		}

		/* Compare DW5 to DW3/Value */
		if (mt[i].fcc_sz == 2) {
			*dw5 = mt[i].fcc[1].bit_off / BITS_IN_DW;
			tag[definer_dw_selector_to_tag_dw(DEFINER_SELECTOR_DW5)] = -1;
			if (mt[i].fcc[1].ctrl.type ==
			    MLX5DV_HWS_FIELD_MATCH_TYPE_TO_FIELD) {
				*dw3 = mt[i].fcc[1].ctrl.compare_dw;
				tag[definer_dw_selector_to_tag_dw(DEFINER_SELECTOR_DW3)] = -1;
			}
		}

		/* Use custom definer mask */
		compare_layout.mask = &mask;
		compare_layout.type = DEFINER_TYPE_COMPARE;

		mt[i].compare_definer = definer_alloc(ctx,
						      mt[i].fcc,
						      mt[i].fcc_sz,
						      &compare_layout,
						      false);
		if (!mt[i].compare_definer) {
			HWS_ERR("Failed to create compare definer");
			goto free_definers;
		}
	}

	return 0;

free_definers:
	while (i--)
		if (mt[i].compare_definer)
			definer_free(ctx, mt[i].compare_definer);

	return -errno;
}

static void
definer_matcher_compare_uninit(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	int i;

	for (i = 0; i < matcher->num_of_mt; i++)
		if (matcher->mt[i].compare_definer)
			definer_free(ctx, matcher->mt[i].compare_definer);
}

static int
definer_matcher_hash_init(struct mlx5dv_hws_context *ctx,
			  struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_mt *mt = matcher->mt;
	struct rule_match_tag *mask;
	uint8_t *bit_mask;
	int i, j;

	for (i = 1; i < matcher->num_of_mt; i++)
		if (definer_compare(mt[i].definer, mt[i - 1].definer))
			matcher->flags |= MATCHER_FLAGS_HASH_DEFINER;

	if (!(matcher->flags & MATCHER_FLAGS_HASH_DEFINER))
		return 0;

	/* Insert by index requires all MT using the same definer */
	if (matcher->attr.insert_mode == MLX5DV_HWS_MATCHER_INSERT_BY_INDEX) {
		HWS_ERR("Insert by index not supported with MT combination");
		errno = EOPNOTSUPP;
		return -errno;
	}

	matcher->hash_definer = calloc(1, sizeof(*matcher->hash_definer));
	if (!matcher->hash_definer) {
		HWS_ERR("Failed to allocate memory for hash definer");
		errno = ENOMEM;
		return -errno;
	}

	mask = calloc(1, sizeof(struct rule_match_tag));
	if (!mask) {
		HWS_ERR("Failed to allocate mask for hash definer");
		errno = ENOMEM;
		goto free_hash_definer;
	}

	/* Calculate intersection between all match templates bitmasks.
	 * We will use mt[0] as reference and intersect it with mt[1..n].
	 * From this we will get:
	 * hash_definer.selectors = mt[0].selecotrs
	 * hash_definer.mask =  mt[0].mask & mt[0].mask & ... & mt[n].mask
	 */

	/* Use first definer which should also contain intersection fields */
	memcpy(matcher->hash_definer, mt->definer, sizeof(struct definer));
	memcpy(mask, mt->definer->mask, sizeof(struct rule_match_tag));

	matcher->hash_definer->mask = mask;

	/* Calculate intersection between first to all match templates bitmasks */
	for (i = 1; i < matcher->num_of_mt; i++) {
		bit_mask = mt[i].definer->mask->jumbo;
		for (j = 0; j < STE_JUMBO_TAG_SZ; j++)
			matcher->hash_definer->mask->jumbo[j] &= bit_mask[j];
	}

	matcher->hash_definer->obj = definer_get_obj(ctx, matcher->hash_definer);
	if (!matcher->hash_definer->obj)
		goto free_mask;

	return 0;

free_mask:
	free(mask);
free_hash_definer:
	free(matcher->hash_definer);
	return -errno;
}

static void
definer_matcher_hash_uninit(struct mlx5dv_hws_matcher *matcher)
{
	if (!matcher->hash_definer)
		return;

	definer_put_obj(matcher->tbl->ctx, matcher->hash_definer->obj);
	free(matcher->hash_definer->mask);
	free(matcher->hash_definer);
}

int definer_matcher_init(struct mlx5dv_hws_context *ctx,
			 struct mlx5dv_hws_matcher *matcher)
{
	struct definer match_layout = {0};
	struct definer range_layout = {0};
	int ret;

	if (matcher->flags & MATCHER_FLAGS_COLLISION)
		return 0;

	ret = definer_calc_layout(matcher, &match_layout, &range_layout);
	if (ret) {
		if (errno == E2BIG)
			HWS_DBG("Failed to calculate matcher definer layout");
		else
			HWS_ERR("Failed to calculate matcher definer layout");

		return ret;
	}

	/* Calculate definers needed for exact match */
	ret = definer_matcher_match_init(ctx, matcher, &match_layout);
	if (ret) {
		HWS_ERR("Failed to init match definers");
		return ret;
	}

	/* Calculate definers needed for range */
	ret = definer_matcher_range_init(ctx, matcher, &range_layout);
	if (ret) {
		HWS_ERR("Failed to init range definers");
		goto uninit_match_definer;
	}

	/* Calculate definers needed for compare */
	ret = definer_matcher_compare_init(ctx, matcher);
	if (ret) {
		HWS_ERR("Failed to init compare definer");
		goto uninit_range_definer;
	}

	/* Calculate partial hash definer */
	ret = definer_matcher_hash_init(ctx, matcher);
	if (ret) {
		HWS_ERR("Failed to init hash definer");
		goto uninit_compare_definer;
	}

	return 0;

uninit_compare_definer:
	definer_matcher_compare_uninit(matcher);
uninit_range_definer:
	definer_matcher_range_uninit(matcher);
uninit_match_definer:
	definer_matcher_match_uninit(matcher);
	return ret;
}

void definer_matcher_uninit(struct mlx5dv_hws_matcher *matcher)
{
	if (matcher->flags & MATCHER_FLAGS_COLLISION)
		return;

	definer_matcher_hash_uninit(matcher);
	definer_matcher_compare_uninit(matcher);
	definer_matcher_range_uninit(matcher);
	definer_matcher_match_uninit(matcher);
}

int definer_mt_check_root(struct mlx5dv_hws_mt *mt)
{
	bool has_ipv4[MLX5DV_HWS_HEADER_MAX] = {0};
	bool has_ipv6[MLX5DV_HWS_HEADER_MAX] = {0};
	enum mlx5dv_hws_field_header hdr;
	struct definer_fc *fcm = mt->fcm;
	int i;

	for (i = 0; i < mt->fcm_sz; i++) {
		switch (fcm->ctrl.fname) {
		case MLX5DV_HWS_FNAME_IPV6_DST_127_96:
		case MLX5DV_HWS_FNAME_IPV6_DST_95_64:
		case MLX5DV_HWS_FNAME_IPV6_DST_63_32:
		case MLX5DV_HWS_FNAME_IPV6_DST_31_0:
		case MLX5DV_HWS_FNAME_IPV6_SRC_127_96:
		case MLX5DV_HWS_FNAME_IPV6_SRC_95_64:
		case MLX5DV_HWS_FNAME_IPV6_SRC_63_32:
		case MLX5DV_HWS_FNAME_IPV6_SRC_31_0:
			has_ipv6[fcm->ctrl.header] = 1;
			break;
		case MLX5DV_HWS_FNAME_IPV4_SRC:
		case MLX5DV_HWS_FNAME_IPV4_DST:
			has_ipv4[fcm->ctrl.header] = 1;
			break;
		default:
			assert(1);
			break;
		}
	}

	for (hdr = MLX5DV_HWS_HEADER_NONE; hdr < MLX5DV_HWS_HEADER_MAX; hdr++) {
		if (has_ipv4[hdr] && has_ipv6[hdr]) {
			HWS_ERR("IPv4 address and IPv6 address cannot be used together");
			errno = ENOTSUP;
			return -errno;
		}
	}

	return 0;
}

static int definer_dynamic_bit_off_fetch(struct mlx5dv_hws_context *ctx,
					 struct definer_fc *fc,
					 bool is_root,
					 uint32_t *bit_off)
{
	struct cmd_query_caps *caps = ctx->caps;
	enum mlx5dv_hws_field_name fname;

	if (is_root)
		return 0;

	fname = fc->ctrl.fname;

	switch (fname) {
	case MLX5DV_HWS_FNAME_GTP_TEID:
		if (!(caps->flex_protocols & MLX5_HCA_FLEX_GTPU_TEID_ENABLED))
			goto out_enotsup;

		*bit_off += caps->format_select_gtpu_dw_1 * BITS_IN_DW;
		break;
	case MLX5DV_HWS_FNAME_GTP_MSG_TYPE:
		if (!(caps->flex_protocols & MLX5_HCA_FLEX_GTPU_DW_0_ENABLED))
			goto out_enotsup;

		*bit_off += caps->format_select_gtpu_dw_0 * BITS_IN_DW;
		break;
	case MLX5DV_HWS_FNAME_GTP_EXT_FLAG:
		if (!(caps->flex_protocols & MLX5_HCA_FLEX_GTPU_DW_0_ENABLED))
			goto out_enotsup;

		*bit_off += caps->format_select_gtpu_dw_0 * BITS_IN_DW;
		break;
	case MLX5DV_HWS_FNAME_GTP_NEXT_EXT_HDR:
		if (!(caps->flex_protocols & MLX5_HCA_FLEX_GTPU_DW_2_ENABLED))
			goto out_enotsup;

		*bit_off += caps->format_select_gtpu_dw_2 * BITS_IN_DW;
		break;
	case MLX5DV_HWS_FNAME_GTP_EXT_HDR_PDU:
		if (!(caps->flex_protocols & MLX5_HCA_FLEX_GTPU_FIRST_EXT_DW_0_ENABLED))
			goto out_enotsup;

		*bit_off += caps->format_select_gtpu_ext_dw_0 * BITS_IN_DW;
		break;
	case MLX5DV_HWS_FNAME_GTP_EXT_HDR_QFI:
		if (!(caps->flex_protocols & MLX5_HCA_FLEX_GTPU_FIRST_EXT_DW_0_ENABLED))
			goto out_enotsup;

		*bit_off += caps->format_select_gtpu_ext_dw_0 * BITS_IN_DW;
		break;
	default:
		break;
	}

	return 0;

out_enotsup:
	HWS_ERR("Field [%d] header [%d] not supported over %s, missing capability support",
		fname,
		fc->ctrl.header,
		is_root ? "ROOT" : "HWS");

	errno = ENOTSUP;
	return -errno;
}

int definer_mt_align(struct mlx5dv_hws_context *ctx,
		     struct mlx5dv_hws_mt *mt,
		     bool is_root)
{
	const struct definer_field_conv_offset *fco;
	struct definer_fc *fc = mt->fcm;
	uint32_t bit_off;
	int i;

	for (i = 0; i < mt->fcm_sz + mt->fcr_sz + mt->fcc_sz; i++) {
		fco = is_root ? &fco_root_arr[fc->ctrl.header][fc->ctrl.fname] :
				&fco_hws_arr[fc->ctrl.header][fc->ctrl.fname];

		if (!fco->bit_sz) {
			HWS_ERR("Field [%d] header [%d] not supported over %s matcher",
				fc->ctrl.fname,
				fc->ctrl.header,
				is_root ? "ROOT" : "HWS");
			errno = ENOTSUP;
			return -errno;
		}

		/* Shift bit offset based on match param layout */
		bit_off = fco->bit_off;

		/* Update dynamic bit_off for flex fields */
		if (definer_dynamic_bit_off_fetch(ctx, fc, is_root, &bit_off))
			return -errno;

		fc->bit_off = bit_off;
		fc++;
	}

	return 0;
}

static bool definer_field_is_range(const struct mlx5dv_hws_item *item,
				   struct mlx5dv_hws_field *field)
{
	uint8_t bit_sz;
	uint32_t val;

	if (field->type != MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE)
		return false;

	if (!item->data.last)
		return false;

	if (unlikely(field->value.user_conv_func)) {
		field->value.user_conv_func(&item->data,
					    field->value.user_conv_data,
					    MLX5DV_HWS_FIELD_USER_CONV_FLAG_LAST,
					    &val);
		return !!val;
	}

	bit_sz = fco_hws_arr[field->header][field->fname].bit_sz |
		 fco_root_arr[field->header][field->fname].bit_sz;

	return !!HWS_GET(item->data.last,
			 field->value.bit_off,
			 bit_sz);
}

static bool definer_field_is_compare(struct mlx5dv_hws_field *field)
{
	return (field->type == MLX5DV_HWS_FIELD_MATCH_TYPE_TO_FIELD ||
		field->op != MLX5DV_HWS_FIELD_MATCH_OP_EQ);
}

static int definer_mt_bind_field(struct mlx5dv_hws_context *ctx,
				 struct mlx5dv_hws_mt *mt,
				 const struct mlx5dv_hws_item *item,
				 int item_idx,
				 int field_idx)
{
	struct mlx5dv_hws_field *field = &item->fields[field_idx];
	const struct definer_field_conv_offset *fco, *fco_b;
	bool is_range, is_compare;
	struct definer_fc *fc;
	uint8_t bit_sz;

	if (field->fname == MLX5DV_HWS_FNAME_NONE)
		return 0;

	/* Check field exists over root or HWS based on field size */
	bit_sz = fco_hws_arr[field->header][field->fname].bit_sz |
		 fco_root_arr[field->header][field->fname].bit_sz;

	if (!bit_sz) {
		HWS_ERR("Unsupported header[%d] fname[%d]",
			field->header,
			field->fname);
		errno = ENOTSUP;
		return -errno;
	}

	is_range = definer_field_is_range(item, field);
	is_compare = definer_field_is_compare(field);

	/* Get field copy (fc) based on type compare/range/match */
	if (is_range)
		fc = &mt->fcr[mt->fcr_sz++];
	else if (is_compare)
		fc = &mt->fcc[mt->fcc_sz++];
	else
		fc = &mt->fcm[mt->fcm_sz++];

	fc->bit_sz = bit_sz;
	fc->item_idx = item_idx;
	fc->ctrl.op = field->op;
	fc->ctrl.type = field->type;
	fc->ctrl.fname = field->fname;
	fc->ctrl.header = field->header;

	switch (field->type) {
	case MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE:
		fc->src_bit_off = field->value.bit_off;
		fc->user_conv_data = field->value.user_conv_data;
		fc->user_conv_func = field->value.user_conv_func;

		if (bit_sz + field->value.bit_off % BITS_IN_BYTE > BITS_IN_DW) {
			HWS_ERR("Limitation: Invalid field memory alignment");
			goto out_no_supp;
		}

		/* Save mask for definer build without saving item */
		fc->ctrl.mask = definer_fc_read_field_mask(fc, &item->data);

		fco = &fco_hws_arr[field->header][field->fname];

		if (is_compare) {
			/* HW bug ignores definer mask, WA is using a single field
			 * per DW. Exception is random field (random + reserved).
			 */
			if (!fco->bit_sz ||
			    (fco->bit_sz != BITS_IN_DW &&
			     field->fname != MLX5DV_HWS_FNAME_RANDOM_NUM)) {
				HWS_ERR("Compare field [%d] not supported", field->fname);
				goto out_no_supp;
			}

			/* Compare operation is limited to full mask and HWS only */
			if (fc->ctrl.mask != __devx_mask(fco->bit_sz)) {
				HWS_ERR("Compare requires a full mask over the field");
				goto out_no_supp;
			}
		}

		/* Range operation is limited to HWS only */
		if (is_range && !fco->bit_sz) {
			HWS_ERR("Range field [%d] not supported", field->fname);
			goto out_no_supp;
		}
		break;

	case MLX5DV_HWS_FIELD_MATCH_TYPE_TO_FIELD:
		fco = &fco_hws_arr[field->header][field->fname];
		fco_b = &fco_hws_arr[field->field.header_b][field->field.fname_b];

		if (!fco->bit_sz || !fco_b->bit_sz) {
			HWS_ERR("Compare field/field_b not supported");
			goto out_no_supp;
		}

		if (fco->bit_sz != BITS_IN_DW || fco_b->bit_sz != BITS_IN_DW) {
			HWS_ERR("Compare field/field_b size must be a full DW");
			goto out_no_supp;
		}

		fc->ctrl.mask = -1;
		fc->ctrl.compare_dw = fco_b->bit_off / BITS_IN_DW;
		break;

	default:
		HWS_ERR("Invalid field type [%d]", field->type);
		goto out_no_supp;
	}

	return 0;

out_no_supp:
	errno = ENOTSUP;
	return -errno;
}

int definer_mt_bind_items(struct mlx5dv_hws_context *ctx,
			  struct mlx5dv_hws_mt *mt,
			  const struct mlx5dv_hws_item items[],
			  uint8_t num_of_items)
{
	struct definer_fc *fc_arr, *fc_arr_packed;
	int i, j, ret, total_fc_sz = 0;

	for (i = 0; i < num_of_items; i++)
		total_fc_sz += items[i].num_of_fields;

	/* Allocate unpacked arrays for FCs */
	fc_arr = calloc(3 * total_fc_sz, sizeof(*fc_arr));
	if (!fc_arr) {
		HWS_ERR("Failed allocating field copy (fc) array");
		errno = ENOMEM;
		return -errno;
	}

	mt->fcm = &fc_arr[0 * total_fc_sz];
	mt->fcr = &fc_arr[1 * total_fc_sz];
	mt->fcc = &fc_arr[2 * total_fc_sz];

	for (i = 0; i < num_of_items; i++) {
		for (j = 0; j < items[i].num_of_fields; j++) {
			ret = definer_mt_bind_field(ctx, mt, &items[i], i, j);
			if (ret) {
				HWS_ERR("Failed to bind item[%d] field[%d]", i, j);
				goto free_fc_arr;
			}
		}
	}

	if (mt->fcc_sz && mt->fcr_sz) {
		HWS_ERR("Compare and range are not supported over the same mt");
		errno = ENOTSUP;
		goto free_fc_arr;
	}

	if (mt->fcc_sz > 2) {
		HWS_ERR("Requested [%d] compare fields, HW limited to 2", mt->fcc_sz);
		errno = ENOTSUP;
		goto free_fc_arr;
	}

	/* Pack FCs to a single compact array for performance */
	fc_arr_packed = calloc(mt->fcc_sz + mt->fcr_sz + mt->fcm_sz, sizeof(*fc_arr));
	if (!fc_arr_packed) {
		HWS_ERR("Failed allocating packed field copy (fc) array");
		errno = ENOMEM;
		goto free_fc_arr;
	}

	mt->fcm = fc_arr_packed;
	mt->fcr = &fc_arr_packed[mt->fcm_sz];
	mt->fcc = &fc_arr_packed[mt->fcm_sz + mt->fcr_sz];

	memcpy(mt->fcm, &fc_arr[0 * total_fc_sz], mt->fcm_sz * sizeof(*fc_arr));
	memcpy(mt->fcr, &fc_arr[1 * total_fc_sz], mt->fcr_sz * sizeof(*fc_arr));
	memcpy(mt->fcc, &fc_arr[2 * total_fc_sz], mt->fcc_sz * sizeof(*fc_arr));
	free(fc_arr);

	return 0;

free_fc_arr:
	free(fc_arr);
	return -errno;
}

void definer_mt_unbind_items(struct mlx5dv_hws_mt *mt)
{
	free(mt->fcm);
}

enum mlx5dv_hws_field_name
definer_hl_to_fname(enum mlx5dv_hws_field_header hws_header,
		    uint32_t bit_off,
		    uint32_t bit_sz)
{
	int i;

	/* The header layout value is a unique identifier that defines the matching
	 * criteria for packet headers.
	 * Each field name (fname) is ultimately mapped to a header layout value, ensuring
	 * a one-to-one correspondence between bit offset (bit_off), bit size (bit_sz),
	 * and field name (fname).
	 */
	for (i = 0; i < ARRAY_SIZE(fco_hws_arr[hws_header]); i++) {
		if (bit_off == fco_hws_arr[hws_header][i].bit_off &&
		    bit_sz == fco_hws_arr[hws_header][i].bit_sz)
			return i;
	}

	return MLX5DV_HWS_FNAME_NONE;
}

uint32_t definer_get_field_width(enum mlx5dv_hws_field_name fname)
{
	return fco_hws_arr[MLX5DV_HWS_HEADER_NONE][fname].bit_sz |
	       fco_hws_arr[MLX5DV_HWS_HEADER_OUTER][fname].bit_sz |
	       fco_hws_arr[MLX5DV_HWS_HEADER_INNER][fname].bit_sz |
	       fco_root_arr[MLX5DV_HWS_HEADER_OUTER][fname].bit_sz |
	       fco_root_arr[MLX5DV_HWS_HEADER_INNER][fname].bit_sz;
}

static bool definer_fname_is_ipv4(enum mlx5dv_hws_field_name fname)
{
	return fname == MLX5DV_HWS_FNAME_IPV4_ID ||
	       fname == MLX5DV_HWS_FNAME_IPV4_IHL ||
	       fname == MLX5DV_HWS_FNAME_IPV4_LEN ||
	       fname == MLX5DV_HWS_FNAME_IPV4_CHECKSUM ||
	       fname == MLX5DV_HWS_FNAME_IPV4_FLAGS ||
	       fname == MLX5DV_HWS_FNAME_IPV4_FRAGMENT_OFFSET ||
	       fname == MLX5DV_HWS_FNAME_IPV4_SRC ||
	       fname == MLX5DV_HWS_FNAME_IPV4_DST;
}

static bool definer_fname_is_ipv6(enum mlx5dv_hws_field_name fname)
{
	return fname == MLX5DV_HWS_FNAME_IPV6_FLOW_LABEL ||
	       fname == MLX5DV_HWS_FNAME_IPV6_PAYLOAD_LEN ||
	       fname == MLX5DV_HWS_FNAME_IPV6_DST_127_96 ||
	       fname == MLX5DV_HWS_FNAME_IPV6_DST_95_64 ||
	       fname == MLX5DV_HWS_FNAME_IPV6_DST_63_32 ||
	       fname == MLX5DV_HWS_FNAME_IPV6_DST_31_0 ||
	       fname == MLX5DV_HWS_FNAME_IPV6_SRC_127_96 ||
	       fname == MLX5DV_HWS_FNAME_IPV6_SRC_95_64 ||
	       fname == MLX5DV_HWS_FNAME_IPV6_SRC_63_32 ||
	       fname == MLX5DV_HWS_FNAME_IPV6_SRC_31_0;
}

static void definer_fname_root_set_ip_version(enum mlx5dv_hws_field_header header,
					      uint8_t *match_param,
					      uint32_t val,
					      bool is_mask)
{
	val = is_mask ? -1 : val;
	HWS_SET(match_param, val,
		fco_root_arr[header][MLX5DV_HWS_FNAME_IP_VERSION].bit_off,
		fco_root_arr[header][MLX5DV_HWS_FNAME_IP_VERSION].bit_sz);
}

void definer_set_dependent_fields_root(struct definer_fc *fc,
				       uint32_t fc_sz,
				       uint8_t *match_param,
				       bool is_mask)
{
	bool has_ethtype[MLX5DV_HWS_HEADER_MAX] = {0};
	bool has_ipv4[MLX5DV_HWS_HEADER_MAX] = {0};
	bool has_ipv6[MLX5DV_HWS_HEADER_MAX] = {0};
	bool has_ip_v[MLX5DV_HWS_HEADER_MAX] = {0};
	uint32_t i, hdr, ip_version;

	for (i = 0; i < fc_sz; i++) {
		has_ipv4[fc[i].ctrl.header] |= definer_fname_is_ipv4(fc[i].ctrl.fname);
		has_ipv6[fc[i].ctrl.header] |= definer_fname_is_ipv6(fc[i].ctrl.fname);
		has_ip_v[fc[i].ctrl.header] |= fc[i].ctrl.fname == MLX5DV_HWS_FNAME_IP_VERSION;
		has_ethtype[fc[i].ctrl.header] |= fc[i].ctrl.fname == MLX5DV_HWS_FNAME_ETH_TYPE;

		/* Check for invalid case - let lower level fail this match */
		if (has_ipv4[i] && has_ipv6[i])
			return;
	}

	/* Set IP version on missing headers when IP related fields are set */
	for (hdr = 0; hdr < MLX5DV_HWS_HEADER_MAX; hdr++) {
		if ((has_ipv4[hdr] || has_ipv6[hdr]) &&
		    !(has_ip_v[hdr] || has_ethtype[hdr])) {
			ip_version = has_ipv4[hdr] ? IP_VERSION_IPV4 : IP_VERSION_IPV6;
			definer_fname_root_set_ip_version(hdr, match_param, ip_version, is_mask);
		}
	}
}
