/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"

static bool matcher_requires_col_tbl(uint8_t log_num_of_rules)
{
	/* Collision table concatenation is done only for large rule tables */
	return log_num_of_rules > MATCHER_ASSURED_RULES_TH;
}

static uint8_t matcher_rules_to_tbl_depth(uint8_t log_num_of_rules)
{
	if (matcher_requires_col_tbl(log_num_of_rules))
		return MATCHER_ASSURED_MAIN_TBL_DEPTH;

	/* For small rule tables we use a single deep table to assure insertion */
	return min_t(uint8_t, log_num_of_rules, MATCHER_ASSURED_COL_TBL_DEPTH);
}

static void matcher_destroy_end_ft(struct mlx5dv_hws_matcher *matcher)
{
	table_destroy_default_ft(matcher->tbl, matcher->end_ft);
}

int matcher_free_rtc_pointing(struct mlx5dv_hws_table *tbl,
			      struct mlx5dv_devx_obj *ft)
{
	int ret;

	if (!table_is_fdb_any(tbl->type) && !context_shared_gvmi_used(tbl->ctx))
		return 0;

	ret = cmd_flow_table_modify_rtc(ft,
					tbl->fw_ft_type,
					tbl->other_vport,
					tbl->vport,
					NULL, NULL);
	if (ret)
		HWS_ERR("Failed to disconnect previous RTC");

	return ret;
}

static int matcher_shared_point_end_ft(struct mlx5dv_hws_matcher *matcher)
{
	struct cmd_ft_modify_attr ft_attr = {0};
	int ret;

	cmd_set_attr_connect_miss_tbl(matcher->tbl->ctx,
				      matcher->tbl->fw_ft_type,
				      matcher->tbl->type,
				      matcher->tbl->other_vport,
				      matcher->tbl->vport,
				      &ft_attr);

	ret = cmd_flow_table_modify(matcher->end_ft, &ft_attr);
	if (ret) {
		HWS_ERR("Failed to connect new matcher to default miss alias RTC");
		return ret;
	}

	ret = matcher_free_rtc_pointing(matcher->tbl, matcher->end_ft);

	return ret;
}

static int matcher_shared_create_alias_rtc(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	struct mlx5dv_devx_obj *obj;

	obj = cmd_allow_and_create_aliased_obj(ctx->ibv_ctx,
					       ctx->local_ibv_ctx,
					       ctx->caps->shared_vhca_id,
					       matcher->match_ste.rtc_0->object_id,
					       MLX5_GENERAL_OBJ_TYPE_RTC);
	if (!obj) {
		HWS_ERR("Failed to allocate alias RTC");
		return -errno;

	}

	matcher->match_ste.aliased_rtc_0 = obj;

	return 0;
}

static int matcher_create_init_shared(struct mlx5dv_hws_matcher *matcher)
{
	if (!context_shared_gvmi_used(matcher->tbl->ctx))
		return 0;

	if (matcher_shared_point_end_ft(matcher)) {
		HWS_ERR("Failed to point shared matcher end flow table");
		return -errno;
	}

	if (matcher_shared_create_alias_rtc(matcher)) {
		HWS_ERR("Failed to create alias RTC");
		return -errno;
	}

	return 0;
}

static void matcher_create_uninit_shared(struct mlx5dv_hws_matcher *matcher)
{
	if (!context_shared_gvmi_used(matcher->tbl->ctx))
		return;

	if (matcher->match_ste.aliased_rtc_0) {
		cmd_destroy_obj(matcher->match_ste.aliased_rtc_0);
		matcher->match_ste.aliased_rtc_0 = NULL;
	}
}

static int matcher_create_end_ft(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_table *tbl = matcher->tbl;

	matcher->end_ft = table_create_default_ft(tbl->ctx->ibv_ctx, tbl);
	if (!matcher->end_ft) {
		HWS_ERR("Failed to create matcher end flow table");
		return -errno;
	}
	return 0;
}

static uint32_t
matcher_connect_get_rtc0(struct mlx5dv_hws_matcher *matcher)
{
	if (!matcher->match_ste.aliased_rtc_0)
		return matcher->match_ste.rtc_0->object_id;
	else
		return matcher->match_ste.aliased_rtc_0->object_id;
}

/* Updates tbl->local_ft to the first RTC or 0 if no more matchers */
static int matcher_shared_update_local_ft(struct mlx5dv_hws_table *tbl)
{
	struct cmd_ft_modify_attr cur_ft_attr = {0};
	struct mlx5dv_hws_matcher *first_matcher;
	int ret;

	if (!context_shared_gvmi_used(tbl->ctx))
		return 0;

	first_matcher = list_top(&tbl->head, struct mlx5dv_hws_matcher, next);
	if (!first_matcher) {
		/* local ft no longer points to any RTC, drop refcount */
		ret = matcher_free_rtc_pointing(tbl, tbl->local_ft);
		if (ret)
			HWS_ERR("Failed to clear local FT to prev alias RTC");

		return ret;
	}

	/* point local_ft to the first RTC */
	cur_ft_attr.modify_fs = MLX5_IFC_MODIFY_FLOW_TABLE_RTC_ID;
	cur_ft_attr.type = tbl->fw_ft_type;
	cur_ft_attr.vport = tbl->vport;
	cur_ft_attr.other_vport = tbl->other_vport;
	cur_ft_attr.rtc_id_0 = matcher_connect_get_rtc0(first_matcher);

	ret = cmd_flow_table_modify(tbl->local_ft, &cur_ft_attr);
	if (ret) {
		HWS_ERR("Failed to point local FT to alias RTC");
		return ret;
	}

	return 0;
}

static int matcher_connect(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_table *tbl = matcher->tbl;
	struct mlx5dv_hws_matcher *prev = NULL;
	struct mlx5dv_hws_matcher *next = NULL;
	struct mlx5dv_hws_matcher *tmp_matcher;
	struct mlx5dv_devx_obj *last_ft;
	int ret;

	if (matcher->attr.isolated) {
		list_add(&tbl->isolated_matchers, &matcher->next);
		ret = table_connect_src_ft_to_miss_table(tbl,
							 matcher->end_ft,
							 tbl->default_miss.miss_tbl);
		if (ret) {
			HWS_ERR("Failed to connect the new matcher to the miss_tbl");
			goto remove_from_list;
		}

		return 0;
	}

	/* First matcher */
	if (list_empty(&tbl->head)) {
		list_add(&tbl->head, &matcher->next);
		goto connect;
	}

	/* Find location in matcher list */
	list_for_each(&tbl->head, tmp_matcher, next) {
		if (tmp_matcher->attr.priority > matcher->attr.priority) {
			next = tmp_matcher;
			break;
		}
		prev = tmp_matcher;
	}

	if (next)
		list_add_before(&tbl->head, &next->next, &matcher->next);
	else
		list_add_after(&tbl->head, &prev->next, &matcher->next);

connect:
	if (next) {
		/* Connect to next RTC */
		ret = cmd_flow_table_modify_rtc(matcher->end_ft,
						tbl->fw_ft_type,
						tbl->other_vport,
						tbl->vport,
						next->match_ste.rtc_0,
						next->match_ste.rtc_1);
		if (ret) {
			HWS_ERR("Failed to connect new matcher to next RTC");
			goto remove_from_list;
		}
	} else {
		/* Connect last matcher to next miss_tbl if exists */
		last_ft = table_get_last_ft(tbl);
		ret = table_connect_src_ft_to_miss_table(tbl,
							 last_ft,
							 tbl->default_miss.miss_tbl);
		if (ret) {
			HWS_ERR("Failed connect new matcher to miss_tbl");
			goto remove_from_list;
		}
	}

	/* Connect to previous FT */
	ret = cmd_flow_table_modify_rtc(prev ? prev->end_ft : tbl->ft,
					tbl->fw_ft_type,
					tbl->other_vport,
					tbl->vport,
					matcher->match_ste.rtc_0,
					matcher->match_ste.rtc_1);
	if (ret) {
		HWS_ERR("Failed to connect new matcher to previous FT");
		goto remove_from_list;
	}

	ret = matcher_shared_update_local_ft(tbl);
	if (ret) {
		HWS_ERR("Failed to update local_ft anchor in shared table");
		goto remove_from_list;
	}

	/* Reset next miss FT to default (drop refcount) */
	ret = table_ft_set_default_next_ft(tbl, prev ? prev->end_ft : tbl->ft);
	if (ret) {
		HWS_ERR("Failed to reset matcher ft default miss");
		goto remove_from_list;
	}

	if (!prev) {
		/* Update tables missing to current matcher in the table */
		ret = table_update_connected_miss_tables(tbl);
		if (ret) {
			HWS_ERR("Fatal error updating connected miss tables");
			goto remove_from_list;
		}
	}

	return 0;

remove_from_list:
	list_del(&matcher->next);
	return ret;
}

static int matcher_disconnect(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_table *tbl = matcher->tbl;
	struct mlx5dv_devx_obj *prev_ft, *last_ft;
	struct mlx5dv_hws_matcher *tmp_matcher;
	struct mlx5dv_hws_matcher *next;
	int ret;

	if (matcher->attr.isolated) {
		list_del(&matcher->next);
		return 0;
	}

	prev_ft = tbl->ft;

	list_for_each(&tbl->head, tmp_matcher, next) {
		if (tmp_matcher == matcher)
			break;

		prev_ft = tmp_matcher->end_ft;
	}

	next = list_next(&tbl->head, matcher, next);
	list_del(&matcher->next);

	if (next) {
		/* Connect previous end FT to next RTC */
		ret = cmd_flow_table_modify_rtc(prev_ft,
						tbl->fw_ft_type,
						tbl->other_vport,
						tbl->vport,
						next->match_ste.rtc_0,
						next->match_ste.rtc_1);
		if (ret) {
			HWS_ERR("Fatal: failed to disconnect matcher");
			return ret;
		}
	} else {
		last_ft = table_get_last_ft(tbl);
		ret = table_connect_src_ft_to_miss_table(tbl,
							 last_ft,
							 tbl->default_miss.miss_tbl);
		if (ret) {
			HWS_ERR("Fatal: failed to disconnect last matcher");
			return ret;
		}
	}

	ret = matcher_shared_update_local_ft(tbl);
	if (ret) {
		HWS_ERR("Fatal: failed to update local_ft in shared table");
		return ret;
	}

	/* Removing first matcher, update connected miss tables if exists */
	if (prev_ft == tbl->ft) {
		ret = table_update_connected_miss_tables(tbl);
		if (ret) {
			HWS_ERR("Fatal: failed update connected miss table");
			return ret;
		}
	}

	ret = table_ft_set_default_next_ft(tbl, prev_ft);
	if (ret) {
		HWS_ERR("Fatal: failed to restore matcher ft default miss");
		return ret;
	}

	/* Failure to restore/modify FW results in a critical, unrecoverable error.
	 * Error handling is not applicable in this fatal scenario.
	 */
	return 0;
}

static bool matcher_supp_fw_wqe(struct mlx5dv_hws_matcher *matcher)
{
	struct cmd_query_caps *caps = matcher->tbl->ctx->caps;

	if (matcher->flags & MATCHER_FLAGS_HASH_DEFINER) {
		if (matcher->hash_definer->type == DEFINER_TYPE_MATCH &&
		    !IS_BIT_SET(caps->supp_ste_format_gen_wqe,
				MLX5_IFC_RTC_STE_FORMAT_8DW)) {
			HWS_ERR("Gen WQE MATCH format not supported");
			return false;
		}

		if (matcher->hash_definer->type == DEFINER_TYPE_JUMBO) {
			HWS_ERR("Gen WQE JUMBO format not supported");
			return false;
		}
	}

	if (matcher->attr.insert_mode != MLX5DV_HWS_MATCHER_INSERT_BY_HASH ||
	    matcher->attr.distribute_mode != MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH) {
		HWS_ERR("Gen WQE must be inserted and distribute by hash");
		return false;
	}

	if ((matcher->flags & MATCHER_FLAGS_RANGE_DEFINER) &&
	    !IS_BIT_SET(caps->supp_ste_format_gen_wqe,
			MLX5_IFC_RTC_STE_FORMAT_RANGE)) {
		HWS_ERR("Extended match gen wqe RANGE format not supported");
		return false;
	}

	if ((matcher->flags & MATCHER_FLAGS_COMPARE_DEFINER) &&
	    !IS_BIT_SET(caps->supp_ste_format_gen_wqe,
			MLX5_IFC_RTC_STE_FORMAT_4DW_RANGE)) {
		HWS_ERR("Gen WQE Compare match format not supported");
		return false;
	}

	if (!(caps->supp_type_gen_wqe & MLX5_GENERATE_WQE_TYPE_FLOW_UPDATE)) {
		HWS_ERR("Gen WQE command not supporting GTA");
		return false;
	}

	if (!caps->rtc_max_hash_def_gen_wqe) {
		HWS_ERR("Hash definer not supported");
		return false;
	}

	return true;
}

static int matcher_resize_init(struct mlx5dv_hws_matcher *src_matcher,
			       struct mlx5dv_hws_matcher *dst_matcher)
{
	struct matcher_resize_data *resize_data;

	resize_data = calloc(1, sizeof(*resize_data));
	if (!resize_data) {
		errno = ENOMEM;
		return -errno;
	}

	resize_data->max_stes = src_matcher->action_ste.max_stes;
	resize_data->action_stc = src_matcher->action_ste.stc;
	resize_data->action_rtc = src_matcher->action_ste.rtc;
	resize_data->action_ste = src_matcher->action_ste.ste;
	resize_data->action_ste_pool = src_matcher->action_ste.max_stes ?
				       src_matcher->action_ste.pool :
				       NULL;
	resize_data->num_of_at = src_matcher->num_of_at;
	resize_data->at = src_matcher->at;

	/* Place the new resized matcher on the dst matcher's list */
	list_add(&dst_matcher->resize_data, &resize_data->next);

	/* Move all the previous resized matchers to the dst matcher's list */
	while (!list_empty(&src_matcher->resize_data)) {
		resize_data = list_top(&src_matcher->resize_data,
				       struct matcher_resize_data, next);
		list_del(&resize_data->next);
		list_add(&dst_matcher->resize_data, &resize_data->next);
	}

	return 0;
}

static void matcher_resize_uninit(struct mlx5dv_hws_matcher *matcher)
{
	struct matcher_resize_data *resize_data;
	struct mlx5dv_hws_at *at;
	int i;

	if (!matcher_is_resizable(matcher))
		return;

	/* Remove previously resized action RTCs */
	while (!list_empty(&matcher->resize_data)) {
		resize_data = list_top(&matcher->resize_data,
				       struct matcher_resize_data, next);
		list_del(&resize_data->next);

		if (resize_data->max_stes) {
			action_free_single_stc(matcher->tbl->ctx,
					       matcher->tbl->type,
					       matcher->tbl->ib_port,
					       &resize_data->action_stc);

			cmd_destroy_obj(resize_data->action_rtc);

			if (resize_data->action_ste_pool)
				pool_destroy(resize_data->action_ste_pool);
		}

		for (i = 0; i < resize_data->num_of_at; i++) {
			at = &resize_data->at[i];
			action_template_unbind(matcher, at);
			free(at->actions);
		}
		if (resize_data->at)
			free(resize_data->at);

		free(resize_data);
	}
}

static int matcher_resize_precheck(struct mlx5dv_hws_matcher *src_matcher,
				   struct mlx5dv_hws_matcher *dst_matcher)
{
	int i;

	if (table_is_root(src_matcher->tbl) ||
	    table_is_root(dst_matcher->tbl)) {
		HWS_ERR("Src/dst matcher is root table - resize unsupported");
		goto out_einval;
	}

	if (src_matcher->tbl->type != dst_matcher->tbl->type) {
		HWS_ERR("Table type mismatch for src/dst matchers");
		goto out_einval;
	}

	if (matcher_req_fw_wqe(src_matcher) ||
	    matcher_req_fw_wqe(dst_matcher)) {
		HWS_ERR("Matchers require FW WQE - resize unsupported");
		goto out_einval;
	}

	if (!matcher_is_resizable(src_matcher) ||
	    !matcher_is_resizable(dst_matcher)) {
		HWS_ERR("Src/dst matcher is not resizable");
		goto out_einval;
	}

	if (matcher_is_insert_by_idx(src_matcher) !=
	    matcher_is_insert_by_idx(dst_matcher)) {
		HWS_ERR("Src/dst matchers insert mode mismatch");
		goto out_einval;
	}

	if (matcher_is_in_resize(src_matcher) ||
	    matcher_is_in_resize(dst_matcher)) {
		HWS_ERR("Src/dst matcher is already in resize");
		goto out_einval;
	}

	/* Compare match templates - make sure the definers are equivalent */
	if (src_matcher->num_of_mt != dst_matcher->num_of_mt) {
		HWS_ERR("Src/dst matcher match templates mismatch");
		goto out_einval;
	}

	if (src_matcher->action_ste.max_stes > dst_matcher->action_ste.max_stes) {
		HWS_ERR("Src/dst matcher max STEs mismatch");
		goto out_einval;
	}

	for (i = 0; i < src_matcher->num_of_mt; i++) {
		if (src_matcher->mt[i].definer->obj->object_id !=
		    dst_matcher->mt[i].definer->obj->object_id) {
			HWS_ERR("Src/dst matcher definers mismatch");
			goto out_einval;
		}
	}

	return 0;

out_einval:
	errno = EINVAL;
	return -errno;
}

int mlx5dv_hws_matcher_resize_set_target(struct mlx5dv_hws_matcher *src_matcher,
					 struct mlx5dv_hws_matcher *dst_matcher)
{
	int ret;

	pthread_spin_lock(&src_matcher->tbl->ctx->ctrl_lock);

	ret = matcher_resize_precheck(src_matcher, dst_matcher);
	if (ret) {
		HWS_ERR("Failed on resize precheck");
		goto out;
	}

	ret = matcher_resize_init(src_matcher, dst_matcher);
	if (ret) {
		HWS_ERR("Failed to init resize");
		goto out;
	}

	src_matcher->resize_dst = dst_matcher;

out:
	pthread_spin_unlock(&src_matcher->tbl->ctx->ctrl_lock);
	return ret;
}

int mlx5dv_hws_matcher_resize_rule_move(struct mlx5dv_hws_matcher *src_matcher,
					struct mlx5dv_hws_rule *rule,
					struct mlx5dv_hws_rule_attr *attr)
{
	if (unlikely(!matcher_is_in_resize(src_matcher))) {
		HWS_ERR("Matcher is not resizable or not in resize");
		goto out_einval;
	}

	if (unlikely(src_matcher != rule->matcher)) {
		HWS_ERR("Rule doesn't belong to src matcher");
		goto out_einval;
	}

	return rule_move_hws_add(rule, attr);

out_einval:
	errno = EINVAL;
	return -errno;
}

int matcher_get_match_rtc_id(struct mlx5dv_hws_matcher *matcher)
{
	if (!matcher)
		return 0;

	if (matcher->tbl->type == MLX5DV_HWS_TABLE_TYPE_FDB_TX)
		return matcher->match_ste.rtc_1->object_id;

	return matcher->match_ste.rtc_0->object_id;
}

static int matcher_action_rtc_create(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_devx_obj *ste_obj, *stc_obj, *rtc_obj;
	struct mlx5dv_hws_table *tbl = matcher->tbl;
	struct cmd_rtc_create_attr rtc_attr = {0};
	struct mlx5dv_hws_context *ctx = tbl->ctx;
	struct pool_chunk chunk = {0};
	struct pool *stc_pool;

	/* Get STE object from pool */
	ste_obj = pool_chunk_get_base_devx_obj(matcher->action_ste.pool, &chunk);

	/* Get STC, use any STC for the shared single base ID */
	chunk = tbl->common_res->default_stc->default_hit;
	stc_pool = context_get_stc_pool(ctx, tbl->type, tbl->ib_port);
	stc_obj = pool_chunk_get_base_devx_obj(stc_pool, &chunk);

	/* Create the action RTC */
	rtc_attr.pd = ctx->pd_num;
	rtc_attr.update_index_mode = MLX5_IFC_RTC_STE_UPDATE_MODE_BY_OFFSET;
	/* The action STEs use the default always hit definer */
	rtc_attr.match_definer_0 = ctx->caps->trivial_match_definer;
	rtc_attr.is_frst_jumbo = false;
	rtc_attr.miss_ft_id = 0;
	rtc_attr.reparse_mode = context_get_reparse_mode(ctx);
	rtc_attr.table_type = table_get_rtc_fw_ft_type(tbl->type);
	rtc_attr.stc_base = stc_obj->object_id;
	rtc_attr.ste_base = ste_obj->object_id;
	rtc_attr.log_size = ilog32(matcher->action_ste.max_stes - 1) +
			    matcher->attr.table.sz_row_log;
	rtc_attr.log_depth = 0;

	rtc_obj = cmd_rtc_create(ctx->ibv_ctx, &rtc_attr);
	if (!rtc_obj) {
		HWS_ERR("Failed to create matcher RTC of actions");
		return -errno;
	}

	matcher->action_ste.ste = ste_obj;
	matcher->action_ste.rtc = rtc_obj;

	return 0;
}

static void matcher_action_rtc_destroy(struct mlx5dv_hws_matcher *matcher)
{
	cmd_destroy_obj(matcher->action_ste.rtc);
}

static int matcher_match_dummy_rtc_create(struct mlx5dv_hws_matcher *matcher,
					  struct mlx5dv_devx_obj **rtc,
					  struct mlx5dv_devx_obj **ste,
					  enum mlx5_ifc_ft_type peer_ft_type)
{
	struct matcher_match_ste *match_ste = &matcher->match_ste;
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	struct cmd_rtc_create_attr rtc_attr = {0};
	struct cmd_stc_create_attr stc_attr = {0};
	struct cmd_ste_create_attr ste_attr = {0};

	ste_attr.log_obj_range = 0;
	ste_attr.table_type = peer_ft_type;
	*ste = cmd_ste_create(ctx->ibv_ctx, &ste_attr);
	if (!(*ste)) {
		HWS_ERR("Failed to create STE of dummy RTC");
		return -errno;
	}

	stc_attr.log_obj_range = 0;
	stc_attr.table_type = peer_ft_type;
	match_ste->dummy_stc = cmd_stc_create(ctx->ibv_ctx, &stc_attr);
	if (!match_ste->dummy_stc) {
		HWS_ERR("Failed to create STC of dummy RTC");
		goto free_ste;
	}

	rtc_attr.log_size = 0;
	rtc_attr.log_depth = 0;
	rtc_attr.pd = ctx->pd_num;
	rtc_attr.table_type = peer_ft_type;
	rtc_attr.miss_ft_id = matcher->end_ft->object_id;
	rtc_attr.reparse_mode = context_get_reparse_mode(ctx);
	rtc_attr.update_index_mode = MLX5_IFC_RTC_STE_UPDATE_MODE_BY_HASH;
	rtc_attr.match_definer_0 = ctx->caps->trivial_match_definer;
	rtc_attr.ste_base = (*ste)->object_id;
	rtc_attr.stc_base = match_ste->dummy_stc->object_id;
	*rtc = cmd_rtc_create(ctx->ibv_ctx, &rtc_attr);
	if (!(*rtc)) {
		HWS_ERR("Failed to create dummy RTC fw_ft [%d]",
			peer_ft_type);
		goto free_stc;
	}

	return 0;

free_stc:
	cmd_destroy_obj(match_ste->dummy_stc);
free_ste:
	cmd_destroy_obj(*ste);
	return -errno;
}

static int matcher_match_dummy_rtc_destroy(struct mlx5dv_hws_matcher *matcher,
					   struct mlx5dv_devx_obj *rtc,
					   struct mlx5dv_devx_obj *ste)
{
	cmd_destroy_obj(rtc);
	cmd_destroy_obj(matcher->match_ste.dummy_stc);
	cmd_destroy_obj(ste);
	return 0;
}

static int matcher_match_rtc_create(struct mlx5dv_hws_matcher *matcher)
{
	struct matcher_match_ste *match_ste = &matcher->match_ste;
	struct mlx5dv_hws_matcher_attr *attr = &matcher->attr;
	struct mlx5dv_hws_table *tbl = matcher->tbl;
	struct mlx5dv_hws_context *ctx = tbl->ctx;
	struct cmd_rtc_create_attr rtc_attr = {0};
	struct cmd_ste_create_attr ste_attr = {0};
	struct mlx5dv_devx_obj *rtc, *ste, *stc;
	struct mlx5dv_hws_mt *mt = matcher->mt;
	struct pool_chunk *tmp_stc;
	struct pool *stc_pool;
	int ret;

	ste_attr.vport = tbl->vport;
	ste_attr.other_vport = tbl->other_vport;
	ste_attr.table_type = table_get_rtc_fw_ft_type(tbl->type);
	ste_attr.log_obj_range = attr->table.sz_col_log +
				 attr->table.sz_row_log;

	/* Add additional rows due to additional range/compare STE */
	if (matcher_mt_is_range(mt) || matcher_mt_is_compare(mt))
		ste_attr.log_obj_range++;

	/* Allocate main hash table STEs */
	ste = cmd_ste_create(ctx->ibv_ctx ,&ste_attr);
	if (!(ste)) {
		HWS_ERR("Failed to allocate STE for match RTC");
		return -errno;
	}

	/* STC is a single resource (devx_obj), use any STC for the ID */
	stc_pool = context_get_stc_pool(ctx, tbl->type, tbl->ib_port);
	tmp_stc = &tbl->common_res->default_stc->default_hit;

	stc = pool_chunk_get_base_devx_obj(stc_pool, tmp_stc);

	rtc_attr.pd = ctx->pd_num;
	rtc_attr.vport = tbl->vport;
	rtc_attr.other_vport = tbl->other_vport;
	rtc_attr.ste_base = ste->object_id;
	rtc_attr.stc_base = stc->object_id;
	rtc_attr.table_type = table_get_rtc_fw_ft_type(tbl->type);
	rtc_attr.log_size = attr->table.sz_row_log;
	rtc_attr.log_depth = attr->table.sz_col_log;
	rtc_attr.reparse_mode = context_get_reparse_mode(ctx);
	rtc_attr.is_frst_jumbo = matcher_mt_is_jumbo(mt);
	rtc_attr.is_scnd_range = matcher_mt_is_range(mt);
	rtc_attr.is_scnd_compare = matcher_mt_is_compare(mt);
	rtc_attr.miss_ft_id = matcher->end_ft->object_id;

	if (attr->insert_mode == MLX5DV_HWS_MATCHER_INSERT_BY_HASH) {
		/* The usual Hash Table */
		rtc_attr.update_index_mode = MLX5_IFC_RTC_STE_UPDATE_MODE_BY_HASH;

		if (matcher->hash_definer) {
			/* Specify definer_id_0 is used for hashing */
			rtc_attr.fw_gen_wqe = true;
			rtc_attr.num_hash_definer = 1;
			rtc_attr.match_definer_0 =
				definer_get_id(matcher->hash_definer);
		} else {
			/* The first mt is used since all share the same definer */
			rtc_attr.match_definer_0 = definer_get_id(mt->definer);

			/* This is tricky, instead of passing two definers for
			 * match and range, we specify that this RTC uses a hash
			 * definer, this will allow us to use any range definer
			 * since only first STE is used for hashing anyways.
			 */
			if (matcher_is_range(matcher) ||
			    matcher_is_compare(matcher)) {
				rtc_attr.fw_gen_wqe = true;
				rtc_attr.num_hash_definer = 1;
			}
		}
	} else if (attr->insert_mode == MLX5DV_HWS_MATCHER_INSERT_BY_INDEX) {
		rtc_attr.update_index_mode =
			MLX5_IFC_RTC_STE_UPDATE_MODE_BY_OFFSET;
		rtc_attr.num_hash_definer = 1;

		if (attr->distribute_mode == MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH) {
			/* Hash Split Table (with or without match) */
			if (!matcher_is_always_hit(matcher))
				rtc_attr.num_hash_definer = 0;

			rtc_attr.access_index_mode =
				MLX5_IFC_RTC_STE_ACCESS_MODE_BY_HASH;
			rtc_attr.match_definer_0 = definer_get_id(mt->definer);
		} else if (attr->distribute_mode ==
			   MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_LINEAR) {
			/* Linear Lookup Table */
			rtc_attr.access_index_mode =
				MLX5_IFC_RTC_STE_ACCESS_MODE_LINEAR;
			rtc_attr.match_definer_0 =
				ctx->caps->linear_match_definer;
		}
	}

	rtc = cmd_rtc_create(ctx->ibv_ctx, &rtc_attr);
	if (!rtc) {
		HWS_ERR("Failed to create matcher RTC of type match");
		goto free_ste;
	}

	switch (tbl->type) {
	case MLX5DV_HWS_TABLE_TYPE_NIC_RX:
	case MLX5DV_HWS_TABLE_TYPE_NIC_TX:
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX:
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX:
		match_ste->rtc_0 = rtc;
		match_ste->ste_0 = ste;
		break;
	case MLX5DV_HWS_TABLE_TYPE_FDB:
		match_ste->rtc_0 = rtc;
		match_ste->ste_0 = ste;
		match_ste->rtc_1 = rtc;
		match_ste->ste_1 = ste;
		break;
	case MLX5DV_HWS_TABLE_TYPE_FDB_RX:
		match_ste->rtc_0 = rtc;
		match_ste->ste_0 = ste;
		ret = matcher_match_dummy_rtc_create(matcher,
						     &match_ste->rtc_1,
						     &match_ste->ste_1,
						     MLX5_IFC_FT_FDB_TX);
		if (ret) {
			HWS_ERR("Failed to create dummy TX RTC");
			goto free_rtc;
		}
		break;
	case MLX5DV_HWS_TABLE_TYPE_FDB_TX:
		match_ste->rtc_1 = rtc;
		match_ste->ste_1 = ste;
		ret = matcher_match_dummy_rtc_create(matcher,
						     &match_ste->rtc_0,
						     &match_ste->ste_0,
						     MLX5_IFC_FT_FDB_RX);
		if (ret) {
			HWS_ERR("Failed to create dummy RX RTC");
			goto free_rtc;
		}
		break;
	default:
		break;
	}

	return 0;

free_rtc:
	cmd_destroy_obj(rtc);
free_ste:
	cmd_destroy_obj(ste);
	return -errno;
}

static void matcher_match_rtc_destroy(struct mlx5dv_hws_matcher *matcher)
{
	switch (matcher->tbl->type) {
	case MLX5DV_HWS_TABLE_TYPE_NIC_RX:
	case MLX5DV_HWS_TABLE_TYPE_NIC_TX:
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX:
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX:
	case MLX5DV_HWS_TABLE_TYPE_FDB:
		cmd_destroy_obj(matcher->match_ste.rtc_0);
		cmd_destroy_obj(matcher->match_ste.ste_0);
		break;
	case MLX5DV_HWS_TABLE_TYPE_FDB_RX:
		matcher_match_dummy_rtc_destroy(matcher,
						matcher->match_ste.rtc_1,
						matcher->match_ste.ste_1);
		cmd_destroy_obj(matcher->match_ste.rtc_0);
		cmd_destroy_obj(matcher->match_ste.ste_0);
		break;
	case MLX5DV_HWS_TABLE_TYPE_FDB_TX:
		matcher_match_dummy_rtc_destroy(matcher,
						matcher->match_ste.rtc_0,
						matcher->match_ste.ste_0);
		cmd_destroy_obj(matcher->match_ste.rtc_1);
		cmd_destroy_obj(matcher->match_ste.ste_1);
		break;
	default:
		break;
	}
}

static int
matcher_check_attr_sz(struct cmd_query_caps *caps,
		      struct mlx5dv_hws_matcher_attr *attr)
{
	if (attr->table.sz_col_log > caps->rtc_log_depth_max) {
		HWS_ERR("Matcher depth exceeds limit %d",
			caps->rtc_log_depth_max);
		goto not_supported;
	}

	if (attr->table.sz_col_log + attr->table.sz_row_log > caps->ste_alloc_log_max) {
		HWS_ERR("Total matcher size exceeds limit %d",
			caps->ste_alloc_log_max);
		goto not_supported;
	}

	if (attr->table.sz_col_log + attr->table.sz_row_log <
	    caps->ste_alloc_log_gran) {
		HWS_ERR("Total matcher size below limit %d",
			caps->ste_alloc_log_gran);
		goto not_supported;
	}

	return 0;

not_supported:
	errno = EOPNOTSUPP;
	return -errno;
}

static int matcher_check_action_template(struct mlx5dv_hws_matcher *matcher,
					struct mlx5dv_hws_at *at_arr[],
					int num_of_at)
{
	int i, ret;

	for (i = 0; i < num_of_at; i++) {
		struct mlx5dv_hws_at *at = at_arr[i];

		ret = action_template_verify_actions(matcher, at);
		if (ret) {
			HWS_ERR("Invalid action template idx = %d", i);
			return -errno;
		}
	}

	return 0;
}

int matcher_action_ste_create(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_table *tbl = matcher->tbl;
	struct mlx5dv_hws_context *ctx = tbl->ctx;
	struct cmd_stc_modify_attr stc_attr = {0};
	struct pool_attr pool_attr = {0};
	int ret;

	/* Allocate action STE mempool */
	pool_attr.table_type = tbl->type;
	pool_attr.vport = tbl->vport;
	pool_attr.other_vport = tbl->other_vport;
	pool_attr.pool_type = POOL_TYPE_STE;
	pool_attr.flags = POOL_FLAGS_FOR_STE_ACTION_POOL;
	pool_attr.alloc_log_sz = ilog32(matcher->action_ste.max_stes - 1) +
					   matcher->attr.table.sz_row_log;

	matcher->action_ste.pool = pool_create(ctx, &pool_attr);
	if (!matcher->action_ste.pool) {
		HWS_ERR("Failed to create action ste pool");
		return -errno;
	}

	/* Allocate action RTC */
	ret = matcher_action_rtc_create(matcher);
	if (ret) {
		HWS_ERR("Failed to create action RTC");
		goto free_ste_pool;
	}

	/* Allocate STC for jumps to STE */
	stc_attr.action_offset = ACTION_OFFSET_HIT;
	stc_attr.action_type = MLX5_IFC_STC_ACTION_TYPE_JUMP_TO_STE_TABLE;
	stc_attr.reparse_mode = MLX5_IFC_STC_REPARSE_IGNORE;
	stc_attr.ste_table.ste_obj_id = matcher->action_ste.ste->object_id;
	stc_attr.ste_table.match_definer_id = ctx->caps->trivial_match_definer;

	ret = action_alloc_single_stc(ctx,
				      &stc_attr,
				      tbl->type,
				      tbl->ib_port,
				      &matcher->action_ste.stc);
	if (ret) {
		HWS_ERR("Failed to create action jump to table STC");
		goto free_rtc;
	}

	return 0;

free_rtc:
	matcher_action_rtc_destroy(matcher);
free_ste_pool:
	pool_destroy(matcher->action_ste.pool);

	return -errno;
}

static int matcher_at_process_init(struct mlx5dv_hws_matcher *matcher)
{
	uint32_t required_stes;
	bool is_jumbo;
	int i, ret;

	for (i = 0; i < matcher->num_of_at; i++) {
		struct mlx5dv_hws_at *at = &matcher->at[i];

		ret = action_template_bind(matcher, at);
		if (ret) {
			HWS_ERR("Invalid action template [%d]", i);
			goto clean_all;
		}

		if (!table_is_root(matcher->tbl)) {
			is_jumbo = matcher_mt_is_jumbo(matcher->mt);
			required_stes = at->num_of_action_stes - (!is_jumbo || at->only_term);
			matcher->action_ste.max_stes =
				max_t(uint8_t, matcher->action_ste.max_stes, required_stes);
		}
	}

	return 0;

clean_all:
	for (i = i - 1; i >= 0; i--)
		action_template_unbind(matcher, &matcher->at[i]);

	return ret;
}

static int matcher_at_process_uninit(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_at *at;
	int i;

	for (i = 0; i < matcher->num_of_at; i++) {
		at = &matcher->at[i];
		action_template_unbind(matcher, at);
	}

	return 0;
}

static int matcher_bind_at(struct mlx5dv_hws_matcher *matcher)
{
	int ret;

	if (matcher->flags & MATCHER_FLAGS_COLLISION)
		return 0;

	if (matcher->attr.max_num_of_at_attach && matcher_req_fw_wqe(matcher)) {
		HWS_ERR("FW extended matcher doesn't support additional at");
		errno = ENOTSUP;
		return -errno;
	}

	ret = matcher_at_process_init(matcher);
	if (ret) {
		HWS_ERR("Failed process at");
		errno = ENOTSUP;
		return -errno;
	}

	/* There are no additional STEs required for matcher */
	if (!matcher->action_ste.max_stes)
		return 0;

	if (matcher_req_fw_wqe(matcher)) {
		HWS_ERR("FW extended matcher cannot be binded to complex AT");
		errno = ENOTSUP;
		goto uninit_at;
	}

	ret = matcher_action_ste_create(matcher);
	if (ret) {
		HWS_ERR("Failed create action STE");
		goto uninit_at;
	}

	return 0;

uninit_at:
	matcher_at_process_uninit(matcher);

	return -errno;
}

static void matcher_unbind_at(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_table *tbl = matcher->tbl;

	if (matcher->flags & MATCHER_FLAGS_COLLISION ||
	    matcher_is_in_resize(matcher))
		return;

	matcher_at_process_uninit(matcher);

	if (!matcher->action_ste.max_stes)
		return;

	action_free_single_stc(tbl->ctx,
			       tbl->type,
			       tbl->ib_port,
			       &matcher->action_ste.stc);

	matcher_action_rtc_destroy(matcher);
	pool_destroy(matcher->action_ste.pool);
}

static int matcher_bind_mt(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	int ret;

	/* Calculate match, range, compare and hash definers */
	ret = definer_matcher_init(ctx, matcher);
	if (ret) {
		if (errno == E2BIG)
			HWS_DBG("Failed set match templates with match definers");
		else
			HWS_ERR("Failed set match templates with match definers");
		return ret;
	}

	if (matcher_req_fw_wqe(matcher) && !matcher_supp_fw_wqe(matcher)) {
		HWS_ERR("Matcher requires FW WQE which is not supported");
		errno = ENOTSUP;
		goto uninit_match_definer;
	}

	return 0;

uninit_match_definer:
	definer_matcher_uninit(matcher);
	return -errno;
}

static void matcher_unbind_mt(struct mlx5dv_hws_matcher *matcher)
{
	definer_matcher_uninit(matcher);
}

static int
matcher_validate_insert_mode(struct cmd_query_caps *caps,
			     struct mlx5dv_hws_matcher *matcher,
			     bool is_root)
{
	struct mlx5dv_hws_matcher_attr *attr = &matcher->attr;

	if (is_root) {
		if (attr->resource_mode != MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE) {
			HWS_ERR("Root matcher supports only rule resource mode");
			goto not_supported;
		}

		if (attr->insert_mode != MLX5DV_HWS_MATCHER_INSERT_BY_HASH) {
			HWS_ERR("Root matcher supports only insert by hash mode");
			goto not_supported;
		}

		if (attr->distribute_mode != MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH) {
			HWS_ERR("Root matcher supports only distribute by hash");
			goto not_supported;
		}

		return 0;
	}

	switch (attr->insert_mode) {
	case MLX5DV_HWS_MATCHER_INSERT_BY_HASH:
		if (attr->distribute_mode != MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH) {
			HWS_ERR("Invalid matcher distribute mode");
			goto not_supported;
		}

		if (attr->always_hit) {
			HWS_ERR("Always hit is not supported for insert by hash");
			goto not_supported;
		}
		break;

	case MLX5DV_HWS_MATCHER_INSERT_BY_INDEX:
		if (attr->table.sz_col_log) {
			HWS_ERR("INSERT_BY_INDEX supports only Nx1 table size");
			goto not_supported;
		}

		if (attr->distribute_mode == MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH) {
			/* Hash Split Table */
			if (matcher_is_always_hit(matcher)) {
				if (!caps->rtc_hash_split_table) {
					HWS_ERR("FW doesn't support index and hash");
					goto not_supported;
				}
			} else {
				if (!attr->isolated) {
					HWS_ERR("Insert by index and match supported only isolated matcher");
					goto not_supported;
				}
			}
		} else if (attr->distribute_mode ==
			   MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_LINEAR) {
			/* Linear Lookup Table */
			if (!caps->rtc_linear_lookup_table ||
			    !IS_BIT_SET(caps->access_index_mode,
					MLX5_IFC_RTC_STE_ACCESS_MODE_LINEAR)) {
				HWS_ERR("FW doesn't support index and linear");
				goto not_supported;
			}

			if (attr->table.sz_row_log >
			    MLX5_IFC_RTC_LINEAR_LOOKUP_TBL_LOG_MAX) {
				HWS_ERR("Matcher with linear: rows exceed limit %d",
				       MLX5_IFC_RTC_LINEAR_LOOKUP_TBL_LOG_MAX);
				goto not_supported;
			}

			if (!matcher_is_always_hit(matcher)) {
				HWS_ERR("Linear lookup require an empty MT (always-hit)");
				goto not_supported;
			}
		} else {
			HWS_ERR("Matcher has unsupported distribute mode");
			goto not_supported;
		}
		break;

	default:
		HWS_ERR("Matcher has unsupported insert mode");
		goto not_supported;
	}

	return 0;

not_supported:
	errno = EOPNOTSUPP;
	return -errno;
}

static int
matcher_process_attr(struct cmd_query_caps *caps,
		     struct mlx5dv_hws_matcher *matcher,
		     bool is_root)
{
	struct mlx5dv_hws_matcher_attr *attr = &matcher->attr;

	if (attr->comp_mask) {
		HWS_ERR("Invalid comp mask provided");
		goto not_supported;
	}

	if (matcher_validate_insert_mode(caps, matcher, is_root))
		goto not_supported;

	if (is_root) {
		if (attr->max_num_of_at_attach) {
			HWS_ERR("Root matcher does not support at attaching");
			goto not_supported;
		}

		if (attr->resizable) {
			HWS_ERR("Root matcher does not support resizing");
			goto not_supported;
		}

		if (attr->isolated) {
			HWS_ERR("Root matcher cannot be isolated");
			goto not_supported;
		}

		if (attr->always_hit) {
			HWS_ERR("Root matcher doesn't support always hit");
			goto not_supported;
		}

		return 0;
	}

	/* Convert number of rules to the required depth */
	if (attr->resource_mode == MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE &&
	    attr->insert_mode == MLX5DV_HWS_MATCHER_INSERT_BY_HASH)
		attr->table.sz_col_log =
			matcher_rules_to_tbl_depth(attr->rule.num_log);

	if (attr->isolated) {
		if (attr->insert_mode != MLX5DV_HWS_MATCHER_INSERT_BY_INDEX ||
		    attr->distribute_mode != MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH) {
			HWS_ERR("Isolated matcher requires mode: insert[%d] distribute[%d]",
				MLX5DV_HWS_MATCHER_INSERT_BY_INDEX,
				MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH);
			goto not_supported;
		}
	}

	matcher->flags |= attr->resizable ? MATCHER_FLAGS_RESIZABLE : 0;

	return matcher_check_attr_sz(caps, attr);

not_supported:
	errno = EOPNOTSUPP;
	return -errno;
}

static int matcher_create_and_connect(struct mlx5dv_hws_matcher *matcher)
{
	int ret;

	/* Select and create the definers for current matcher */
	ret = matcher_bind_mt(matcher);
	if (ret)
		return ret;

	/* Calculate and verify action combination */
	ret = matcher_bind_at(matcher);
	if (ret)
		goto unbind_mt;

	/* Create matcher end flow table anchor */
	ret = matcher_create_end_ft(matcher);
	if (ret)
		goto unbind_at;

	/* Allocate the RTC for the new matcher */
	ret = matcher_match_rtc_create(matcher);
	if (ret)
		goto destroy_end_ft;

	/* Allocate and set shared resources */
	ret = matcher_create_init_shared(matcher);
	if (ret)
		goto destroy_rtc;

	/* Connect the matcher to the matcher list */
	ret = matcher_connect(matcher);
	if (ret)
		goto destroy_shared;

	return 0;

destroy_shared:
	matcher_create_uninit_shared(matcher);
destroy_rtc:
	matcher_match_rtc_destroy(matcher);
destroy_end_ft:
	matcher_destroy_end_ft(matcher);
unbind_at:
	matcher_unbind_at(matcher);
unbind_mt:
	matcher_unbind_mt(matcher);
	return ret;
}

static void matcher_destroy_and_disconnect(struct mlx5dv_hws_matcher *matcher)
{
	matcher_resize_uninit(matcher);
	matcher_disconnect(matcher);
	matcher_create_uninit_shared(matcher);
	matcher_match_rtc_destroy(matcher);
	matcher_destroy_end_ft(matcher);
	matcher_unbind_at(matcher);
	matcher_unbind_mt(matcher);
}

static int
matcher_create_col_matcher(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	struct mlx5dv_hws_matcher *col_matcher;
	int ret;

	if (matcher->attr.resource_mode != MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE ||
	    matcher->attr.insert_mode == MLX5DV_HWS_MATCHER_INSERT_BY_INDEX)
		return 0;

	if (!matcher_requires_col_tbl(matcher->attr.rule.num_log))
		return 0;

	col_matcher = calloc(1, sizeof(*matcher));
	if (!col_matcher) {
		errno = ENOMEM;
		return -errno;
	}

	col_matcher->tbl = matcher->tbl;
	col_matcher->mt = matcher->mt;
	col_matcher->at = matcher->at;
	col_matcher->num_of_at = matcher->num_of_at;
	col_matcher->num_of_mt = matcher->num_of_mt;
	col_matcher->hash_definer = matcher->hash_definer;
	col_matcher->attr.priority = matcher->attr.priority;
	col_matcher->flags = matcher->flags | MATCHER_FLAGS_COLLISION;
	col_matcher->attr.resource_mode = MLX5DV_HWS_MATCHER_RESOURCE_MODE_HTABLE;
	col_matcher->attr.table.sz_row_log = matcher->attr.rule.num_log;
	col_matcher->attr.table.sz_col_log = MATCHER_ASSURED_COL_TBL_DEPTH;
	if (col_matcher->attr.table.sz_row_log > MATCHER_ASSURED_ROW_RATIO)
		col_matcher->attr.table.sz_row_log -= MATCHER_ASSURED_ROW_RATIO;

	col_matcher->attr.max_num_of_at_attach =
		matcher->attr.max_num_of_at_attach;

	list_head_init(&col_matcher->resize_data);

	ret = matcher_process_attr(ctx->caps, col_matcher, false);
	if (ret)
		goto free_col_matcher;

	ret = matcher_create_and_connect(col_matcher);
	if (ret)
		goto free_col_matcher;

	matcher->col_matcher = col_matcher;

	return 0;

free_col_matcher:
	free(col_matcher);
	HWS_ERR("Failed to create assured collision matcher");
	return ret;
}

static void
matcher_destroy_col_matcher(struct mlx5dv_hws_matcher *matcher)
{
	if (matcher->attr.resource_mode != MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE ||
	    matcher->attr.insert_mode == MLX5DV_HWS_MATCHER_INSERT_BY_INDEX)
		return;

	if (matcher->col_matcher) {
		matcher_destroy_and_disconnect(matcher->col_matcher);
		free(matcher->col_matcher);
	}
}

static int matcher_init(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	int ret;

	pthread_spin_lock(&ctx->ctrl_lock);

	/* Allocate matcher resource and connect to the packet pipe */
	ret = matcher_create_and_connect(matcher);
	if (ret)
		goto unlock_err;

	/* Create additional matcher for collision handling */
	ret = matcher_create_col_matcher(matcher);
	if (ret)
		goto destory_and_disconnect;

	pthread_spin_unlock(&ctx->ctrl_lock);

	return 0;

destory_and_disconnect:
	matcher_destroy_and_disconnect(matcher);
unlock_err:
	pthread_spin_unlock(&ctx->ctrl_lock);
	return ret;
}

static int matcher_uninit(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;

	pthread_spin_lock(&ctx->ctrl_lock);
	matcher_destroy_col_matcher(matcher);
	matcher_destroy_and_disconnect(matcher);
	pthread_spin_unlock(&ctx->ctrl_lock);

	return 0;
}

static int matcher_mt_get_match_criteria(uint8_t *mask)
{
	uint8_t match_bits[] = { MLX5_MATCH_CRITERIA_ENABLE_OUTER_BIT,
				 MLX5_MATCH_CRITERIA_ENABLE_MISC_BIT,
				 MLX5_MATCH_CRITERIA_ENABLE_INNER_BIT,
				 MLX5_MATCH_CRITERIA_ENABLE_MISC2_BIT,
				 MLX5_MATCH_CRITERIA_ENABLE_MISC3_BIT,
				 MLX5_MATCH_CRITERIA_ENABLE_MISC4_BIT,
				 MLX5_MATCH_CRITERIA_ENABLE_MISC5_BIT,
				 MLX5_MATCH_CRITERIA_ENABLE_MISC6_BIT };
	int spec_sz = DEVX_ST_SZ_BYTES(dr_match_spec);
	int match_criteria = 0;
	int i;

	for (i = 0; i < sizeof(match_bits) / sizeof(match_bits[0]); i++) {
		/* Check spec and set bit if it has mask */
		if (mask[0] || memcmp(&mask[0], &mask[1], spec_sz - 1))
			match_criteria |= 1 << match_bits[i];

		/* Move to next spec */
		mask += spec_sz;
	}

	return match_criteria;
}

static int matcher_init_root(struct mlx5dv_hws_matcher *matcher)
{
	enum mlx5dv_hws_table_type type = matcher->tbl->type;
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	struct mlx5dv_flow_matcher_attr attr = {0};
	struct mlx5dv_flow_match_parameters *mask;
	struct mlx5dv_hws_mt *mt = matcher->mt;
	struct ibv_context *ibv_ctx;
	uint8_t *match_buf;
	int ret;

	if (matcher->attr.priority > UINT16_MAX) {
		HWS_ERR("Root matcher priority exceeds allowed limit");
		errno = EINVAL;
		return -errno;
	}

	/* Check that match template is valid for root */
	ret = definer_mt_check_root(matcher->mt);
	if (ret) {
		HWS_ERR("Failed root match template checks");
		return ret;
	}

	mask = calloc(1, DEVX_ST_SZ_BYTES(dr_match_param) +
		      offsetof(struct mlx5dv_flow_match_parameters, match_buf));
	if (!mask) {
		errno = ENOMEM;
		return -errno;
	}

	/* Set user mask values to match param layout */
	match_buf = (uint8_t *)mask->match_buf;
	definer_create_tag_mask(mt->fcm, mt->fcm_sz, match_buf);
	definer_set_dependent_fields_root(mt->fcm, mt->fcm_sz, match_buf, true);

	mask->match_sz = DEVX_ST_SZ_BYTES(dr_match_param);
	attr.match_mask = mask;
	attr.type = IBV_FLOW_ATTR_NORMAL;
	attr.priority = matcher->attr.priority;
	attr.comp_mask = MLX5DV_FLOW_MATCHER_MASK_FT_TYPE;
	attr.match_criteria_enable = matcher_mt_get_match_criteria(match_buf);

	if (table_is_rdma_transport(type)) {
		attr.ib_port = matcher->tbl->ib_port;
		attr.comp_mask |= MLX5DV_FLOW_MATCHER_MASK_IB_PORT;
	}

	ret = table_conv_hws_tbl_type_to_dv_ft(type, &attr.ft_type);
	if (ret) {
		HWS_ERR("Failed to get DV ft_type for %d", type);
		goto free_mask;
	}

	ibv_ctx = context_get_local_ibv(ctx);
	matcher->dv_matcher = mlx5dv_create_flow_matcher(ibv_ctx, &attr);
	if (!matcher->dv_matcher) {
		HWS_ERR("Failed to create DV flow matcher");
		goto free_mask;
	}

	free(mask);

	pthread_spin_lock(&ctx->ctrl_lock);
	list_add(&matcher->tbl->head, &matcher->next);
	pthread_spin_unlock(&ctx->ctrl_lock);

	return 0;

free_mask:
	free(mask);
	return -errno;
}

static int matcher_uninit_root(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	int ret;

	pthread_spin_lock(&ctx->ctrl_lock);
	list_del(&matcher->next);
	pthread_spin_unlock(&ctx->ctrl_lock);

	ret = mlx5dv_destroy_flow_matcher(matcher->dv_matcher);
	if (ret)
		HWS_ERR("Failed to Destroy DV flow matcher");

	return ret;
}

static int matcher_copy_at(struct mlx5dv_hws_at *at,
			   struct mlx5dv_hws_at *n_at)
{
	int i;

	n_at->num_actions = at->num_actions;
	n_at->actions = calloc(at->num_actions,
			       sizeof(struct mlx5dv_hws_action));
	if (!n_at->actions) {
		HWS_ERR("Failed to allocate action template array");
		errno = ENOMEM;
		return -errno;
	}

	/* Deep copy of all the actions */
	for (i = 0; i < at->num_actions; i++)
		n_at->actions[i] = at->actions[i];

	return 0;
}

int mlx5dv_hws_matcher_attach_at(struct mlx5dv_hws_matcher *matcher,
				 struct mlx5dv_hws_at *at)
{
	bool is_jumbo = matcher_mt_is_jumbo(matcher->mt);
	struct mlx5dv_hws_at *new_at;
	uint32_t required_stes;
	int ret;

	if (!matcher->attr.max_num_of_at_attach) {
		HWS_ERR("Num of current at (%d) exceed allowed value",
			matcher->num_of_at);
		errno = ENOTSUP;
		return -errno;
	}

	new_at = calloc(1, sizeof(struct mlx5dv_hws_at));
	if (!new_at)
		return -ENOMEM;

	ret = matcher_copy_at(at, new_at);
	if (ret)
		goto free_new_at;

	pthread_spin_lock(&matcher->tbl->ctx->ctrl_lock);

	ret = action_template_bind(matcher, new_at);
	if (ret) {
		HWS_ERR("Failed to process new at");
		goto clean_new_at_actions;
	}

	required_stes = new_at->num_of_action_stes - (!is_jumbo || new_at->only_term);
	if (matcher->action_ste.max_stes < required_stes) {
		HWS_DBG("Required STEs [%d] exceeds action template STE [%d]",
			required_stes, matcher->action_ste.max_stes);
		errno = ENOMEM;
		goto unbind_at;
	}

	if (!matcher->attr.max_num_of_at_attach) {
		HWS_ERR("Num of current at (%d) exceed allowed value",
			matcher->num_of_at);
			errno = ENOTSUP;
			goto unbind_at;
	}

	matcher->at[matcher->num_of_at] = *new_at;
	matcher->num_of_at += 1;
	matcher->attr.max_num_of_at_attach -= 1;

	if (matcher->col_matcher)
		matcher->col_matcher->num_of_at = matcher->num_of_at;

	pthread_spin_unlock(&matcher->tbl->ctx->ctrl_lock);

	return 0;

unbind_at:
	action_template_unbind(matcher, new_at);
clean_new_at_actions:
	pthread_spin_unlock(&matcher->tbl->ctx->ctrl_lock);
	free(new_at->actions);
free_new_at:
	free(new_at);
	return -errno;
}

static bool matcher_mt_has_match(struct mlx5dv_hws_mt *mt[], uint8_t num_of_mt)
{
	int i;

	for (i = 0; i < num_of_mt; i++) {
		if (mt[i]->fcm_sz || mt[i]->fcr_sz || mt[i]->fcc_sz)
			return true;
	}

	return false;
}

static int
matcher_set_templates_mt(struct mlx5dv_hws_matcher *matcher,
			 struct mlx5dv_hws_mt *mt_arr[],
			 int num_of_mt)
{
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	struct mlx5dv_hws_mt *new_mt;
	struct definer_fc *fc;
	int i, j, fc_sz, ret;

	new_mt = calloc(num_of_mt, sizeof(struct mlx5dv_hws_mt));
	if (!new_mt) {
		HWS_ERR("Failed to allocate action template array");
		errno = ENOMEM;
		return -errno;
	}

	for (i = 0; i < num_of_mt; i++) {
		fc_sz = mt_arr[i]->fcm_sz + mt_arr[i]->fcr_sz + mt_arr[i]->fcc_sz;

		fc = calloc(fc_sz, sizeof(struct definer_fc));
		if (!fc) {
			HWS_ERR("Failed to alloc new fcs");
			errno = ENOMEM;
			goto free_mt;
		}

		/* Shallow copy */
		new_mt[i] = *mt_arr[i];

		/* Deep copy */
		new_mt[i].fcm = fc;
		new_mt[i].fcr = &fc[mt_arr[i]->fcm_sz];
		new_mt[i].fcc = &fc[mt_arr[i]->fcm_sz + mt_arr[i]->fcr_sz];

		for (j = 0; j < fc_sz; j++)
			fc[j] = mt_arr[i]->fcm[j];

		/* Align MT FCs */
		ret = definer_mt_align(ctx, &new_mt[i], table_is_root(matcher->tbl));
		if (ret) {
			errno = ENOTSUP;
			goto free_mt;
		}
	}

	matcher->mt = new_mt;
	matcher->num_of_mt = num_of_mt;

	return 0;

free_mt:
	for (i = 0; i < num_of_mt; i++)
		if (new_mt[i].fcm)
			free(new_mt[i].fcm);

	free(new_mt);
	return -errno;
}

static void
matcher_unset_templates_mt(struct mlx5dv_hws_matcher *matcher)
{
	int i;

	for (i = 0; i < matcher->num_of_mt; i++)
		free(matcher->mt[i].fcm);

	free(matcher->mt);
}

static int
matcher_set_templates_at(struct mlx5dv_hws_matcher *matcher,
			 struct mlx5dv_hws_at *at_arr[],
			 int num_of_at)
{
	struct mlx5dv_hws_at *at;
	int i, ret;

	at = calloc(num_of_at + matcher->attr.max_num_of_at_attach,
		    sizeof(struct mlx5dv_hws_at));
	if (!at) {
		HWS_ERR("Failed to allocate action template array");
		errno = ENOMEM;
		return -errno;
	}

	for (i = 0; i < num_of_at; i++) {
		at[i] = *at_arr[i];

		ret = matcher_copy_at(at_arr[i], &at[i]);
		if (ret)
			goto clean_actions;
	}

	matcher->at = at;
	matcher->num_of_at = num_of_at;

	return 0;

clean_actions:
	for (i = i - 1; i >= 0; i--)
		free(at[i].actions);
	free(at);
	return -errno;
}

static void
matcher_unset_templates_at(struct mlx5dv_hws_matcher *matcher)
{
	int i;

	if (matcher_is_in_resize(matcher))
		return;

	for (i = 0; i < matcher->num_of_at; i++)
		free(matcher->at[i].actions);

	free(matcher->at);
}

static int
matcher_set_templates(struct mlx5dv_hws_matcher *matcher,
		      struct mlx5dv_hws_mt *mt[],
		      uint8_t num_of_mt,
		      struct mlx5dv_hws_at *at[],
		      uint8_t num_of_at)
{
	bool is_root = table_is_root(matcher->tbl);
	int ret;

	if (matcher->attr.always_hit || !matcher_mt_has_match(mt, num_of_mt))
		matcher->flags |= MATCHER_FLAGS_ALWAYS_HIT;

	if (is_root) {
		if (mt[0]->fcr_sz) {
			HWS_ERR("Range not supported over root");
			errno = ENOTSUP;
			return -errno;
		}

		if (mt[0]->fcc_sz) {
			HWS_ERR("Compare not supported over root");
			errno = ENOTSUP;
			return -errno;
		}

		if (num_of_mt > MATCHER_MAX_MT_ROOT) {
			HWS_ERR("Number of match templates exceeds root limit");
			errno = ENOTSUP;
			return -errno;
		}
	}

	if (!num_of_mt || !num_of_at) {
		HWS_ERR("Number of action/match template cannot be zero");
		errno = ENOTSUP;
		return -errno;
	}

	ret = matcher_set_templates_mt(matcher, mt, num_of_mt);
	if (ret) {
		HWS_ERR("Failed to set match template");
		return ret;
	}

	ret = matcher_check_action_template(matcher, at, num_of_at);
	if (ret) {
		HWS_ERR("Failed check actions in template");
		goto unset_mt;
	}

	ret = matcher_set_templates_at(matcher, at, num_of_at);
	if (ret) {
		HWS_ERR("Failed to set action template");
		goto unset_mt;
	}

	return 0;

unset_mt:
	matcher_unset_templates_mt(matcher);
	return -errno;
}

static void
matcher_unset_templates(struct mlx5dv_hws_matcher *matcher)
{
	matcher_unset_templates_at(matcher);
	matcher_unset_templates_mt(matcher);
}

struct mlx5dv_hws_matcher *
mlx5dv_hws_matcher_create(struct mlx5dv_hws_table *tbl,
			  struct mlx5dv_hws_mt *mt[],
			  uint8_t num_of_mt,
			  struct mlx5dv_hws_at *at[],
			  uint8_t num_of_at,
			  struct mlx5dv_hws_matcher_attr *attr)
{
	struct mlx5dv_hws_matcher *matcher;
	bool is_root = table_is_root(tbl);
	int ret;

	matcher = calloc(1, sizeof(*matcher));
	if (!matcher) {
		errno = ENOMEM;
		return NULL;
	}

	matcher->tbl = tbl;
	matcher->attr = *attr;

	list_head_init(&matcher->resize_data);

	ret = matcher_set_templates(matcher, mt, num_of_mt, at, num_of_at);
	if (ret)
		goto free_matcher;

	ret = matcher_process_attr(tbl->ctx->caps, matcher, is_root);
	if (ret)
		goto unset_templates;

	if (is_root)
		ret = matcher_init_root(matcher);
	else
		ret = matcher_init(matcher);

	if (ret) {
		if (errno == E2BIG)
			HWS_DBG("Failed to initialise matcher: %d", ret);
		else
			HWS_ERR("Failed to initialise matcher: %d", ret);

		goto unset_templates;
	}

	return matcher;

unset_templates:
	matcher_unset_templates(matcher);
free_matcher:
	free(matcher);
	return NULL;
}

int mlx5dv_hws_matcher_destroy(struct mlx5dv_hws_matcher *matcher)
{
	if (table_is_root(matcher->tbl))
		matcher_uninit_root(matcher);
	else
		matcher_uninit(matcher);

	matcher_unset_templates(matcher);
	free(matcher);

	return 0;
}

struct mlx5dv_hws_mt *
mlx5dv_hws_match_template_create(struct mlx5dv_hws_context *ctx,
				 const struct mlx5dv_hws_item items[],
				 uint8_t num_of_items,
				 struct mlx5dv_hws_mt_attr *attr)
{
	struct mlx5dv_hws_mt *mt;
	int ret;

	if (attr->comp_mask) {
		HWS_ERR("Invalid comp mask provided");
		errno = EOPNOTSUPP;
		return NULL;
	}

	mt = calloc(1, sizeof(struct mlx5dv_hws_mt));
	if (!mt) {
		HWS_ERR("Failed allocating match template");
		errno = ENOMEM;
		return NULL;
	}

	/* Convert items to fc and fcr */
	ret = definer_mt_bind_items(ctx, mt, items, num_of_items);
	if (ret) {
		HWS_ERR("Failed to do initial items to fc convert");
		goto free_mt;
	}

	return mt;

free_mt:
	free(mt);
	return NULL;
}

int mlx5dv_hws_match_template_destroy(struct mlx5dv_hws_mt *mt)
{
	definer_mt_unbind_items(mt);
	free(mt);

	return 0;
}

uint32_t mlx5dv_hws_match_field_get_length(enum mlx5dv_hws_field_name fname)
{
	if (fname >= MLX5DV_HWS_FNAME_MAX) {
		errno = EINVAL;
		return 0;
	}

	return definer_get_field_width(fname);
}
