/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#ifndef _HWS_MATCHER_H_
#define _HWS_MATCHER_H_

/* Max supported match template */
#define MATCHER_MAX_MT_ROOT 1

/* We calculated that concatenating a collision table to the main table with
 * 3% of the main table rows will be enough resources for high insertion
 * success probability.
 *
 * The calculation: log2(2^x * 3 / 100) = log2(2^x) + log2(3/100) = x - 5.05 ~ 5
 */
#define MATCHER_ASSURED_ROW_RATIO 5
/* Thrashold to determine if amount of rules require a collision table */
#define MATCHER_ASSURED_RULES_TH 10
/* Required depth of an assured collision table */
#define MATCHER_ASSURED_COL_TBL_DEPTH 4
/* Required depth of the main large table */
#define MATCHER_ASSURED_MAIN_TBL_DEPTH 2

enum matcher_flags {
	MATCHER_FLAGS_RANGE_DEFINER    = 1 << 0,
	MATCHER_FLAGS_HASH_DEFINER     = 1 << 1,
	MATCHER_FLAGS_COLLISION        = 1 << 2,
	MATCHER_FLAGS_RESIZABLE        = 1 << 3,
	MATCHER_FLAGS_COMPARE_DEFINER  = 1 << 4,
	MATCHER_FLAGS_ALWAYS_HIT       = 1 << 5,
	MATCHER_FLAGS_CT_V4            = 1 << 6,
	MATCHER_FLAGS_CT_V6            = 1 << 7,
	MATCHER_FLAGS_CT_ONE_DIRECTION = 1 << 8,
};

struct matcher_match_ste {
	/* Ctrl */
	struct mlx5dv_devx_obj *rtc_0;
	struct mlx5dv_devx_obj *rtc_1;
	struct mlx5dv_devx_obj *ste_0;
	struct mlx5dv_devx_obj *ste_1;
	/* Dummy STC used for FDB RX and FDB TX */
	struct mlx5dv_devx_obj *dummy_stc;
	/* Currently not support FDB aliased */
	struct mlx5dv_devx_obj *aliased_rtc_0;
};

struct matcher_action_ste {
	/* DataPath*/
	struct mlx5dv_devx_obj *rtc;
	/* Ctrl */
	struct mlx5dv_devx_obj *ste;
	struct pool_chunk stc;
	struct pool *pool;
	uint8_t max_stes;
};

struct mlx5dv_hws_mt {
	uint8_t fcm_sz;
	uint8_t fcr_sz;
	uint8_t fcc_sz;
	struct definer_fc *fcm; /* Field copy match */
	struct definer_fc *fcr; /* Field copy range */
	struct definer_fc *fcc; /* Field copy compare */
	struct definer *definer;
	struct definer *range_definer;
	struct definer *compare_definer;
};

struct mlx5dv_hws_matcher {
	struct mlx5dv_hws_table *tbl;
	struct mlx5dv_hws_matcher_attr attr;
	struct mlx5dv_flow_matcher *dv_matcher;
	struct mlx5dv_hws_mt *mt;
	uint8_t num_of_mt;
	struct mlx5dv_hws_at *at;
	uint8_t num_of_at;
	/* enum matcher_flags */
	uint16_t flags;
	struct mlx5dv_devx_obj *end_ft;
	struct mlx5dv_hws_matcher *col_matcher;
	struct mlx5dv_hws_matcher *resize_dst;
	struct matcher_match_ste match_ste;
	struct matcher_action_ste action_ste;
	struct definer *hash_definer;
	struct list_node next;
	struct list_head resize_data;
};

struct matcher_resize_data {
	struct pool_chunk action_stc;
	struct mlx5dv_devx_obj *action_rtc;
	struct mlx5dv_devx_obj *action_ste;
	struct pool *action_ste_pool;
	struct mlx5dv_hws_at *at;
	uint8_t num_of_at;
	uint8_t max_stes;
	struct list_node next;
};

int matcher_get_match_rtc_id(struct mlx5dv_hws_matcher *matcher);

static inline bool matcher_mt_is_jumbo(struct mlx5dv_hws_mt *mt)
{
	return definer_is_jumbo(mt->definer);
}

static inline bool matcher_mt_is_range(struct mlx5dv_hws_mt *mt)
{
	return (!!mt->range_definer);
}

static inline bool matcher_mt_is_compare(struct mlx5dv_hws_mt *mt)
{
	return (!!mt->compare_definer);
}

static inline bool matcher_is_resizable(struct mlx5dv_hws_matcher *matcher)
{
	return !!(matcher->flags & MATCHER_FLAGS_RESIZABLE);
}

static inline bool matcher_is_in_resize(struct mlx5dv_hws_matcher *matcher)
{
	return !!matcher->resize_dst;
}

static inline bool matcher_is_insert_by_idx(struct mlx5dv_hws_matcher *matcher)
{
	return matcher->attr.insert_mode == MLX5DV_HWS_MATCHER_INSERT_BY_INDEX;
}

static inline bool matcher_is_range(struct mlx5dv_hws_matcher *matcher)
{
	return !!(matcher->flags & MATCHER_FLAGS_RANGE_DEFINER);
}

static inline bool matcher_is_compare(struct mlx5dv_hws_matcher *matcher)
{
	return !!(matcher->flags & MATCHER_FLAGS_COMPARE_DEFINER);
}

static inline bool matcher_is_always_hit(struct mlx5dv_hws_matcher *matcher)
{
	return !!(matcher->flags & MATCHER_FLAGS_ALWAYS_HIT);
}

static inline bool matcher_req_fw_wqe(struct mlx5dv_hws_matcher *matcher)
{
	/* HWS doesn't support hash different from match / range / compare */
	return unlikely(matcher->flags & (MATCHER_FLAGS_HASH_DEFINER |
					  MATCHER_FLAGS_RANGE_DEFINER |
					  MATCHER_FLAGS_COMPARE_DEFINER));
}

int matcher_free_rtc_pointing(struct mlx5dv_hws_table *tbl,
			      struct mlx5dv_devx_obj *ft);

int matcher_attach_at(struct mlx5dv_hws_matcher *matcher,
		      struct mlx5dv_hws_at *at);

int matcher_action_ste_create(struct mlx5dv_hws_matcher *matcher);
#endif /* _HWS_MATCHER_H_ */
