/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"

#define LOCAL_ACTION_TYP_MAX 33
static_assert(LOCAL_ACTION_TYP_MAX == ACTION_TYP_MAX,
	      "Missing new actions support in metric API");

/* 1 Hop costs ~30 clocks */
#define METRIC_COST_AVG_MISS_DIST_BY_HASH_SPLIT 30

/* Avg hops in a fully large populated table of having
 * a miss as first hop, second hop ...
 * (1 * 36%) + (2 * 64% * 72%) + (3 * 64% * 28% * 33%) +
 * (4 * 64% * 28% * 66% * 63%) + ~0.01 = ~1.77
 * Hop costs ~30 clocks => 1.77 * 30 = ~53
 */
#define METRIC_COST_AVG_MISS_DIST_BY_HASH 53

/* 1 Hop costs ~30 clocks */
#define METRIC_COST_AVG_MISS_DIST_BY_LINEAR 30

/* 1 Hop costs ~30 clocks */
#define METRIC_COST_AVG_HIT_DIST_BY_HASH_SPLIT 30

/* Avg hops in a fully large populated table of having
 * a hit as first hop, second hop ...
 * (1 * 64%) + (2 * 36% * 50%) + (3 * 36% * 50% * 66%) +
 * (4 * 36% * 50% * 33% * 75%) + ~0.2 = ~1.56
 * Hop costs ~30 clocks => 1.56 * 30 = ~47
 */
#define METRIC_COST_AVG_HIT_DIST_BY_HASH 47

/* 1 Hop costs ~30 clocks */
#define METRIC_COST_AVG_HIT_DIST_BY_LINEAR 30

/* 1 Hop costs ~30 clocks */
#define METRIC_COST_AVG_HIT_DIRECT_JUMP 30

/* 1 cache line fetch costs ~30 clocks */
#define METRIC_COST_AVG_CACHE_LINE_FETCH 30

/* Crypto actions cost on avg 300 clocks */
#define METRIC_COST_CRYPTO 300

/* On average writing through HW is 250 times faster than through FW */
#define METRIC_COST_AVG_INSERT_BY_FW_WQE 250

/* See METRIC_COST_AVG_HIT_DIST_BY_HASH calculation */
#define METRIC_COST_AVG_INSERT_BY_HASH 47

/* 1 Hop costs ~30 clocks */
#define METRIC_COST_AVG_INSERT_BY_INDEX 30

/* 1 Hop costs ~30 clocks */
#define METRIC_COST_AVG_INSERT_CACHE_LINE 30

#define METRIC_COST_MEM_NO_MEM_RESOURCE 0

/* ASO meter is 32B */
#define METRIC_COST_MEM_ASO_METER 32

/* ASO metric is 64B */
#define METRIC_COST_MEM_ASO_CT 64

/* DEK is 512B */
#define METRIC_COST_MEM_DEK 512

/* IPsec offload is 128B */
#define METRIC_COST_MEM_IPSEC_OFFLOAD 128

/* ASO IPsec is part of IPsec offload object */
#define METRIC_COST_MEM_ASO_IPSEC METRIC_COST_MEM_IPSEC_OFFLOAD

/* ASO first takes 1 bit, so will look at it as 0B */
#define METRIC_COST_MEM_ASO_FIRST_HIT 0

/* ASO counter is 16B */
#define METRIC_COST_MEM_ASO_CTR 16

#define METRIC_LARGE_TBL_LOG_SZ 20

/* Cost of doing extra SW operations (lock, unlock ...) */
#define METRIC_COST_NO_OPT_BY_IDX 100

/* On average on CX7 TLB miss will cost ~280 cycles */
#define METRIC_COST_LARGE_TBL 280

/* On average reparse costs 12 cycles */
#define METRIC_COST_REPARSE 12

/* TODO Action execution cost */
#define METRIC_COST_ENCAP 0
#define METRIC_COST_DECAP 0
#define METRIC_COST_TAG 0

static uint32_t
resource_arg_get_max_gran(struct mlx5dv_hws_resource **resource,
			  uint8_t num_of_resources)
{
	struct mlx5dv_hws_resource *res;
	uint32_t tmp_value, value = 0;
	int i;

	for (i = 0; i < num_of_resources; i++) {
		res = resource[i];
		tmp_value = resource_arg_gran_get(res->type);
		value = tmp_value > value ? tmp_value : value;
	}

	return value;
}

int mlx5dv_hws_metric_query_table(struct mlx5dv_hws_table *tbl,
				  uint64_t flags,
				  struct mlx5dv_hws_metric_table *out)
{
	if (flags != MLX5DV_HWS_METRIC_FLAG_INFO) {
		HWS_ERR("Unsupported flags were passed 0x%" PRIx64, flags);
		errno = EOPNOTSUPP;
		return -errno;
	}

	if (!tbl->level) {
		HWS_ERR("Unsupported for root level table");
		errno = EOPNOTSUPP;
		return -errno;
	}

	out->info.ft_id = tbl->ft->object_id;
	out->info.ft_fw_type = tbl->fw_ft_type;
	out->info.level = tbl->level;

	return 0;
}

static uint64_t
metric_query_matcher_mem_sz(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_matcher_attr *attr = &matcher->attr;
	struct mlx5dv_hws_matcher_attr *col_attr;
	uint64_t total_mem_sz;
	uint32_t mem_log_sz;

	/* Calculate matcher hash table size */
	mem_log_sz = attr->table.sz_row_log;
	mem_log_sz += attr->table.sz_col_log;

	/* Add additional rows due to additional range/compare STE */
	if (matcher_mt_is_range(matcher->mt) ||
	    matcher_mt_is_compare(matcher->mt))
		mem_log_sz++;

	total_mem_sz = (1ULL << mem_log_sz) * STE_SZ;

	/* Calculate matcher actions STE array */
	if (matcher->action_ste.max_stes) {
		mem_log_sz = ilog32(matcher->action_ste.max_stes - 1) +
			     attr->table.sz_row_log;
		total_mem_sz += (1ULL << mem_log_sz) * STE_SZ;
	}

	/* Calculate collision matcher size */
	if (matcher->col_matcher) {
		col_attr = &matcher->col_matcher->attr;
		mem_log_sz = col_attr->table.sz_row_log;
		mem_log_sz += col_attr->table.sz_col_log;
		total_mem_sz += (1ULL << mem_log_sz) * STE_SZ;
	}

	/* In case of FDB_RX/FDB_TX an extra dummy STE is allocated */
	if (matcher->tbl->type == MLX5DV_HWS_TABLE_TYPE_FDB_RX ||
	    matcher->tbl->type == MLX5DV_HWS_TABLE_TYPE_FDB_TX)
		total_mem_sz += STE_SZ;

	return total_mem_sz;
}

static void
metric_query_matcher_info(struct mlx5dv_hws_matcher *matcher,
			  struct mlx5dv_hws_metric_matcher *out)
{
	struct mlx5dv_hws_matcher_attr *attr = &matcher->attr;
	struct mlx5dv_hws_matcher_attr *col_attr;

	out->info.matcher_id = (uint64_t)(uintptr_t)matcher;
	out->info.priority = attr->priority;
	out->info.log_num_of_rows = attr->table.sz_row_log;
	out->info.log_num_of_columns = attr->table.sz_col_log;
	out->info.num_of_action_stes = matcher->action_ste.max_stes;
	out->info.matcher_tables_mem_sz = metric_query_matcher_mem_sz(matcher);
	out->info.is_jumbo = matcher_mt_is_jumbo(matcher->mt);

	if (matcher_mt_is_range(matcher->mt) ||
	    matcher_mt_is_compare(matcher->mt))
		out->info.num_of_match_stes = 2;
	else
		out->info.num_of_match_stes = 1;

	if (matcher->col_matcher) {
		col_attr = &matcher->col_matcher->attr;
		out->info.col_matcher_exist = true;
		out->info.col_matcher_log_num_of_rows = col_attr->table.sz_row_log;
		out->info.col_matcher_log_num_of_columns = col_attr->table.sz_col_log;
	} else {
		out->info.col_matcher_exist = false;
		out->info.col_matcher_log_num_of_rows = 0;
		out->info.col_matcher_log_num_of_columns = 0;
	}
}

static void
metric_query_action_resource_arg(struct mlx5dv_hws_action *action,
				 struct mlx5dv_hws_action_data *action_data,
				 uint32_t *res_gran_p,
				 uint32_t *data_sz_p)
{
	struct mlx5dv_hws_resource *res;
	uint32_t res_gran = 0;
	uint32_t data_sz = 0;
	uint8_t res_idx;

	switch(action->type) {
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
		if (action_data) {
			res_idx = action_data->reformat_l2_to_tnl_l2.resource_idx;
			res = action->reformat.resource[res_idx];
			res_gran = resource_arg_gran_get(res->type);
		} else {
			res_gran = resource_arg_get_max_gran(action->reformat.resource,
							     action->reformat.num_of_resources);
		}
		data_sz = action->reformat.header_size;
		break;
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
		if (action_data) {
			res_idx = action_data->reformat_l2_to_tnl_l3.resource_idx;
			res = action->reformat.resource[res_idx];
			res_gran = resource_arg_gran_get(res->type);
		} else {
			res_gran = resource_arg_get_max_gran(action->reformat.resource,
							     action->reformat.num_of_resources);
		}
		data_sz = action->reformat.header_size;
		break;
	case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
		if (action_data) {
			res_idx = action_data->reformat_tnl_l3_to_l2.resource_idx;
			res = action->modify_header.resource[res_idx];
			res_gran = resource_arg_gran_get(res->type);
		} else {
			res_gran = resource_arg_get_max_gran(action->modify_header.resource,
							     action->modify_header.num_of_resources);
		}
		data_sz = action->modify_header.num_of_actions * ACTION_MH_DOUBLE_SIZE;
		break;
	case ACTION_TYP_INSERT_HEADER:
		if (action_data) {
			res_idx = action_data->insert_header.resource_idx;
			res = action->insert_header.resource[res_idx];
			res_gran = resource_arg_gran_get(res->type);
		} else {
			res_gran = resource_arg_get_max_gran(action->insert_header.resource,
							     action->insert_header.num_of_resources);
		}
		data_sz = action->insert_header.header_size;
		break;
	case ACTION_TYP_MODIFY_HDR:
		if (action_data) {
			res_idx = action_data->modify_header.resource_idx;
			res = action->modify_header.resource[res_idx];
			res_gran = resource_arg_gran_get(res->type);
		} else {
			res_gran = resource_arg_get_max_gran(action->modify_header.resource,
							action->modify_header.num_of_resources);
		}
		data_sz = action->modify_header.num_of_actions * ACTION_MH_DOUBLE_SIZE;
		break;
	case ACTION_TYP_NAT64:
		res_gran = resource_arg_gran_get(MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_128B);
		/* The date is inserted 4 times to this arg size */
		data_sz = ACTION_NAT64_STAGES * ACTION_NAT64_REPLACE_DATA_SIZE;
		break;
	default:
		return;
	}

	if (res_gran_p)
		*res_gran_p = res_gran;

	if (data_sz_p)
		*data_sz_p = data_sz;
}

static uint64_t
metric_query_action_resource_arg_mem_cost(struct mlx5dv_hws_action *action,
					  struct mlx5dv_hws_action_data *action_data)
{
	uint32_t res_gran = 0;
	uint32_t data_sz = 0;
	uint32_t res_sz = 0;
	uint64_t cost = 0;

	metric_query_action_resource_arg(action, action_data, &res_gran, &data_sz);

	res_sz = res_gran * MLX5_IFC_ARG_SZ_DATA;
	cost = (data_sz / res_sz) * res_sz;
	if (data_sz % res_sz)
		cost += res_sz;

	return cost;
}

static uint64_t
metric_query_action_memory_cost(struct mlx5dv_hws_action *action,
				struct mlx5dv_hws_action_data *action_data)
{
	uint64_t cost = 0;
	size_t hdr_sz;
	int i;

	if (!action)
		return 0;

	switch (action->type) {
	case ACTION_TYP_DEST_DROP:
	case ACTION_TYP_DEST_MISS:
	case ACTION_TYP_DEST_TIR:
	case ACTION_TYP_DEST_TBL:
	case ACTION_TYP_DEST_ROOT:
	case ACTION_TYP_DEST_VPORT:
	case ACTION_TYP_DEST_MATCHER:
		break;
	case ACTION_TYP_DEST_ARRAY:
		for (i = 0; i < action->dest_array.num_dest; i++) {
			cost += STE_SZ;
			if (action->dest_array.dests[i].ext_reformat) {
				hdr_sz = action->dest_array.dests[i].ext_reformat_hdr_sz;
				cost += MLX5_IFC_ARG_SZ_DATA * (hdr_sz / MLX5_IFC_ARG_SZ_DATA);
				cost += MLX5_IFC_ARG_SZ_DATA * !!(hdr_sz % MLX5_IFC_ARG_SZ_DATA);
			}
		}
		break;
	case ACTION_TYP_TAG:
	case ACTION_TYP_POP_VLAN:
	case ACTION_TYP_PUSH_VLAN:
	case ACTION_TYP_REFORMAT_TNL_L2_TO_L2:
	case ACTION_TYP_REMOVE_HEADER:
	case ACTION_TYP_INLINE_SET:
	case ACTION_TYP_INLINE_ADD:
	case ACTION_TYP_INLINE_ADD_FIELD:
	case ACTION_TYP_INLINE_COPY:
	case ACTION_TYP_INLINE_INSERT:
	case ACTION_TYP_REFORMAT_TRAILER:
		cost += METRIC_COST_MEM_NO_MEM_RESOURCE;
		break;
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
	case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
	case ACTION_TYP_INSERT_HEADER:
	case ACTION_TYP_MODIFY_HDR:
		cost += metric_query_action_resource_arg_mem_cost(action, action_data);
		break;
	case ACTION_TYP_NAT64:
		cost += ACTION_NAT64_STAGES *
			metric_query_action_resource_arg_mem_cost(action, action_data);
		break;
	case ACTION_TYP_ASO_METER:
		cost += METRIC_COST_MEM_ASO_METER;
		break;
	case ACTION_TYP_ASO_CT:
		cost += METRIC_COST_MEM_ASO_CT;
		break;
	case ACTION_TYP_ASO_IPSEC:
		cost += METRIC_COST_MEM_ASO_IPSEC;
		break;
	case ACTION_TYP_ASO_FIRST_HIT:
		cost += METRIC_COST_MEM_ASO_FIRST_HIT;
		break;
	case ACTION_TYP_CTR:
		cost += METRIC_COST_MEM_ASO_CTR;
		break;
	case ACTION_TYP_CRYPTO_ENCRYPT:
	case ACTION_TYP_CRYPTO_DECRYPT:
		if (action->crypto.crypto_type == MLX5DV_HWS_ACTION_CRYPTO_TYPE_IPSEC)
			cost += METRIC_COST_MEM_IPSEC_OFFLOAD;

		cost += METRIC_COST_MEM_DEK;
		break;
	default:
		cost = 0;
		break;
	}

	return cost;
}

static uint64_t
metric_query_template_memory_cost(struct mlx5dv_hws_matcher *matcher,
				  struct mlx5dv_hws_metric_matcher_template_attr *attr)
{
	struct mlx5dv_hws_at *at = &matcher->at[attr->at_idx];
	struct mlx5dv_hws_mt *mt = &matcher->mt[attr->mt_idx];
	struct mlx5dv_hws_action_data *action_data;
	uint32_t mem_sz;
	int i;

	/* Calculate number of STEs */
	mem_sz = matcher_mt_is_range(mt) ? 2 : 1;
	mem_sz += at->num_of_action_stes - (!matcher_mt_is_jumbo(mt) || at->only_term);

	/* Multiply by STE size */
	mem_sz *= STE_SZ;

	/* Calculate memory cost for each action in the actions template */
	for (i = 0; i < at->num_actions; i++) {
		action_data = attr->actions_data ? &attr->actions_data[i] : NULL;
		mem_sz += metric_query_action_memory_cost(&at->actions[i],
							  action_data);
	}

	/* Calculate memory cost for dest action */
	mem_sz += metric_query_action_memory_cost(attr->dest_action, NULL);

	return mem_sz;
}

static uint64_t
metric_query_matcher_memory_cost(struct mlx5dv_hws_matcher *matcher)
{
	return metric_query_matcher_mem_sz(matcher);
}

static uint64_t metric_arg_data_sz_to_pp_cost(uint32_t data_sz)
{
	uint64_t args;

	args = data_sz / MLX5_IFC_ARG_SZ_DATA;
	args += !!(data_sz % MLX5_IFC_ARG_SZ_DATA);

	return args * METRIC_COST_AVG_CACHE_LINE_FETCH;
}

static uint64_t
metric_query_action_resource_arg_pkt_processing_cost(struct mlx5dv_hws_action *action,
						     struct mlx5dv_hws_action_data *action_data)
{
	uint32_t data_sz = 0;

	metric_query_action_resource_arg(action, action_data, NULL, &data_sz);

	return metric_arg_data_sz_to_pp_cost(data_sz);
}

static uint64_t
metric_query_action_pkt_processing_cost(struct mlx5dv_hws_action *action,
					struct mlx5dv_hws_action_data *action_data)
{
	uint64_t cost = 0;
	size_t hdr_sz;
	int i;

	if (!action)
		return 0;

	switch (action->type) {
	case ACTION_TYP_DEST_DROP:
	case ACTION_TYP_DEST_MISS:
	case ACTION_TYP_DEST_TIR:
	case ACTION_TYP_DEST_TBL:
	case ACTION_TYP_DEST_ROOT:
	case ACTION_TYP_DEST_VPORT:
	case ACTION_TYP_DEST_MATCHER:
		cost += METRIC_COST_AVG_HIT_DIRECT_JUMP;
		break;
	case ACTION_TYP_DEST_ARRAY:
		for (i = 0; i < action->dest_array.num_dest; i++) {
			cost += METRIC_COST_AVG_HIT_DIRECT_JUMP;
			if (action->dest_array.dests[i].ext_reformat) {
				hdr_sz = action->dest_array.dests[i].ext_reformat_hdr_sz;
				cost += metric_arg_data_sz_to_pp_cost(hdr_sz);
				cost += METRIC_COST_REPARSE;
				cost += METRIC_COST_ENCAP;
			}
		}
		break;
	case ACTION_TYP_TAG:
		cost += METRIC_COST_TAG;
		break;
	case ACTION_TYP_POP_VLAN:
	case ACTION_TYP_PUSH_VLAN:
		cost += METRIC_COST_REPARSE;
		break;
	case ACTION_TYP_REFORMAT_TNL_L2_TO_L2:
		cost += METRIC_COST_REPARSE;
		cost += METRIC_COST_DECAP;
		break;
	case ACTION_TYP_REMOVE_HEADER:
		cost += action->remove_header.reparse ? METRIC_COST_REPARSE : 0;
		cost += action->remove_header.decap ? METRIC_COST_DECAP : 0;
		break;
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
		cost += metric_query_action_resource_arg_pkt_processing_cost(action, action_data);
		cost += METRIC_COST_REPARSE;
		cost += METRIC_COST_ENCAP;
		break;
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
		cost += metric_query_action_resource_arg_pkt_processing_cost(action, action_data);
		cost += METRIC_COST_REPARSE;
		cost += METRIC_COST_ENCAP;
		break;
	case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
		cost += metric_query_action_resource_arg_pkt_processing_cost(action, action_data);
		cost += action->modify_header.reparse ? METRIC_COST_REPARSE : 0;
		break;
	case ACTION_TYP_INSERT_HEADER:
		cost += metric_query_action_resource_arg_pkt_processing_cost(action, action_data);
		cost += action->insert_header.reparse ? METRIC_COST_REPARSE : 0;
		cost += action->insert_header.encap ? METRIC_COST_ENCAP : 0;
		break;
	case ACTION_TYP_MODIFY_HDR:
		cost += metric_query_action_resource_arg_pkt_processing_cost(action, action_data);
		cost += action->modify_header.reparse ? METRIC_COST_REPARSE : 0;
		break;
	case ACTION_TYP_NAT64:
		cost += metric_arg_data_sz_to_pp_cost(ACTION_NAT64_COPY_DATA_SIZE);
		cost += action->nat64.need_reparse & BIT(ACTION_NAT64_STAGE_COPY) ?
			METRIC_COST_REPARSE : 0;
		cost += metric_arg_data_sz_to_pp_cost(ACTION_NAT64_REPLACE_DATA_SIZE);
		cost += action->nat64.need_reparse & BIT(ACTION_NAT64_STAGE_REPLACE) ?
			METRIC_COST_REPARSE : 0;
		cost += metric_arg_data_sz_to_pp_cost(ACTION_NAT64_COPY_PROTOCOL_DATA_SIZE);
		cost += action->nat64.need_reparse & BIT(ACTION_NAT64_STAGE_COPY_PROTOCOL) ?
			METRIC_COST_REPARSE : 0;
		cost += metric_arg_data_sz_to_pp_cost(ACTION_NAT64_COPYBACK_DATA_SIZE);
		cost += action->nat64.need_reparse & BIT(ACTION_NAT64_STAGE_COPYBACK) ?
			METRIC_COST_REPARSE : 0;
		break;
	case ACTION_TYP_INLINE_SET:
	case ACTION_TYP_INLINE_ADD:
	case ACTION_TYP_INLINE_ADD_FIELD:
	case ACTION_TYP_INLINE_COPY:
	case ACTION_TYP_INLINE_INSERT:
		cost += action->inline_modify_header.reparse ? METRIC_COST_REPARSE : 0;
		break;
	case ACTION_TYP_ASO_METER:
	case ACTION_TYP_ASO_CT:
	case ACTION_TYP_ASO_IPSEC:
	case ACTION_TYP_ASO_FIRST_HIT:
	case ACTION_TYP_CTR:
		cost += METRIC_COST_AVG_CACHE_LINE_FETCH;
		/* TODO - To check execution ASO cost */
		break;
	case ACTION_TYP_CRYPTO_ENCRYPT:
	case ACTION_TYP_CRYPTO_DECRYPT:
		cost += METRIC_COST_CRYPTO;
		break;
	case ACTION_TYP_REFORMAT_TRAILER:
		/* TODO - To check reformat trailer cost */
		break;
	default:
		break;
	}

	return cost;
}

static uint32_t
metric_query_template_pkt_processing_cost(struct mlx5dv_hws_matcher *matcher,
					  struct mlx5dv_hws_action *dest_action,
					  uint8_t mt_idx,
					  uint8_t at_idx,
					  struct mlx5dv_hws_action_data actions_data[])
{
	struct mlx5dv_hws_matcher_attr *attr = &matcher->attr;
	struct mlx5dv_hws_at *at = &matcher->at[at_idx];
	struct mlx5dv_hws_mt *mt = &matcher->mt[mt_idx];
	struct mlx5dv_hws_action_data *action_data;
	uint32_t needed_action_stes;
	uint32_t cost = 0;
	int i;

	/* Cost of hitting first STE */
	switch (attr->distribute_mode) {
	case MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH:
		if (attr->table.sz_col_log) {
			cost += METRIC_COST_AVG_HIT_DIST_BY_HASH;
		} else {
			/* Hash split table cost */
			cost += METRIC_COST_AVG_HIT_DIST_BY_HASH_SPLIT;
		}
		break;
	case MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_LINEAR:
		cost += METRIC_COST_AVG_HIT_DIST_BY_LINEAR;
		break;
	default:
		cost += 1;
		break;
	}

	/* In case of range/compare add cost to jump directly the second STE */
	if (matcher_mt_is_range(mt) || matcher_mt_is_compare(mt))
		cost += METRIC_COST_AVG_HIT_DIRECT_JUMP;

	/* Add cost of jumping to action STEs */
	needed_action_stes = at->num_of_action_stes -
			     (!matcher_mt_is_jumbo(mt) || at->only_term);
	cost += needed_action_stes * METRIC_COST_AVG_HIT_DIRECT_JUMP;

	/* Add actions cost */
	for (i = 0; i < at->num_actions; i++) {
		action_data = actions_data ? &actions_data[i] : NULL;
		cost += metric_query_action_pkt_processing_cost(&at->actions[i],
								action_data);
	}

	cost += metric_query_action_pkt_processing_cost(dest_action, NULL);

	/* Extra cost for large table TLB misses */
	if (attr->table.sz_row_log + attr->table.sz_row_log > METRIC_LARGE_TBL_LOG_SZ)
		cost += METRIC_COST_LARGE_TBL;

	return (uint32_t)cost;
}

static uint32_t
metric_query_matcher_pkt_processing_miss_cost(struct mlx5dv_hws_matcher *matcher)
{
	struct mlx5dv_hws_matcher_attr *attr = &matcher->attr;
	uint32_t cost = 0;

	switch (attr->distribute_mode) {
	case MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH:
		if (attr->table.sz_col_log) {
			cost += METRIC_COST_AVG_MISS_DIST_BY_HASH;
		} else {
			/* Hash split table miss cost */
			cost += METRIC_COST_AVG_MISS_DIST_BY_HASH_SPLIT;
		}
		break;
	case MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_LINEAR:
		cost += METRIC_COST_AVG_MISS_DIST_BY_LINEAR;
		break;
	default:
		cost += 1;
		break;
	}

	/* Matcher end anchor hop */
	cost += METRIC_COST_AVG_HIT_DIRECT_JUMP;

	/* In case of collision matcher we have an extra hop for it's end anchor */
	if (matcher->col_matcher)
		cost += METRIC_COST_AVG_HIT_DIRECT_JUMP;

	/* Extra cost for large table TLB misses */
	if (attr->table.sz_row_log + attr->table.sz_row_log > METRIC_LARGE_TBL_LOG_SZ)
		cost += METRIC_COST_LARGE_TBL;

	return (uint32_t)cost;
}

static uint64_t
metric_query_action_resource_arg_insertion_cost(struct mlx5dv_hws_action *action,
						struct mlx5dv_hws_action_data *action_data)
{
	uint32_t data_sz = 0;
	uint64_t cost = 0;

	metric_query_action_resource_arg(action, action_data, NULL, &data_sz);

	cost = data_sz / MLX5_IFC_ARG_SZ_DATA;
	if (data_sz % MLX5_IFC_ARG_SZ_DATA)
		cost++;

	return cost * METRIC_COST_AVG_INSERT_CACHE_LINE;
}

static uint64_t
metric_query_action_insertion_cost(struct mlx5dv_hws_action *action,
				   struct mlx5dv_hws_action_data *action_data)
{
	uint64_t cost = 0;

	if (!action)
		return 0;

	switch (action->type) {
	case ACTION_TYP_DEST_DROP:
	case ACTION_TYP_DEST_MISS:
	case ACTION_TYP_DEST_TIR:
	case ACTION_TYP_DEST_TBL:
	case ACTION_TYP_DEST_ROOT:
	case ACTION_TYP_DEST_VPORT:
	case ACTION_TYP_DEST_ARRAY:
	case ACTION_TYP_DEST_MATCHER:
	case ACTION_TYP_TAG:
	case ACTION_TYP_POP_VLAN:
	case ACTION_TYP_PUSH_VLAN:
	case ACTION_TYP_REFORMAT_TNL_L2_TO_L2:
	case ACTION_TYP_REMOVE_HEADER:
	case ACTION_TYP_INLINE_SET:
	case ACTION_TYP_INLINE_ADD:
	case ACTION_TYP_INLINE_ADD_FIELD:
	case ACTION_TYP_INLINE_COPY:
	case ACTION_TYP_INLINE_INSERT:
	case ACTION_TYP_CRYPTO_ENCRYPT:
	case ACTION_TYP_CRYPTO_DECRYPT:
	case ACTION_TYP_REFORMAT_TRAILER:
		break;
	case ACTION_TYP_ASO_METER:
		if (action_data && action_data->aso_meter.aso_wqe)
			cost += METRIC_COST_AVG_INSERT_CACHE_LINE;
		break;
	case ACTION_TYP_ASO_IPSEC:
		if (action_data && action_data->aso_ipsec.aso_wqe)
			cost += METRIC_COST_AVG_INSERT_CACHE_LINE;
		break;
	case ACTION_TYP_ASO_CT:
		if (action_data && action_data->aso_ct.aso_wqe)
			cost += METRIC_COST_AVG_INSERT_CACHE_LINE;
		break;
	case ACTION_TYP_ASO_FIRST_HIT:
		if (action_data && action_data->aso_first_hit.aso_wqe)
			cost += METRIC_COST_AVG_INSERT_CACHE_LINE;
		break;
	case ACTION_TYP_CTR:
		if (action_data && action_data->counter.aso_wqe)
			cost += METRIC_COST_AVG_INSERT_CACHE_LINE;
		break;
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
	case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
	case ACTION_TYP_INSERT_HEADER:
	case ACTION_TYP_MODIFY_HDR:
		cost += metric_query_action_resource_arg_insertion_cost(action, action_data);
		break;
	case ACTION_TYP_NAT64:
		/* Writing 4 arguments on per STE */
		cost += ACTION_NAT64_STAGES *
			metric_query_action_resource_arg_insertion_cost(action, action_data);
		break;
	default:
		break;
	}

	return cost;
}

static uint32_t
metric_query_template_insertion_cost(struct mlx5dv_hws_matcher *matcher,
				     struct mlx5dv_hws_action *dest_action,
				     uint8_t mt_idx,
				     uint8_t at_idx,
				     struct mlx5dv_hws_action_data actions_data[])
{
	struct mlx5dv_hws_matcher_attr *attr = &matcher->attr;
	struct mlx5dv_hws_at *at = &matcher->at[at_idx];
	struct mlx5dv_hws_mt *mt = &matcher->mt[mt_idx];
	struct mlx5dv_hws_action_data *action_data;
	uint32_t needed_action_stes;
	uint32_t cost = 0;
	int i;

	/* Cost of inserting first STE */
	if (attr->insert_mode == MLX5DV_HWS_MATCHER_INSERT_BY_HASH)
		cost += METRIC_COST_AVG_INSERT_BY_HASH;
	else
		cost += METRIC_COST_AVG_INSERT_BY_INDEX;

	/* Cost of inserting second STE in case of range/compare */
	if (matcher_mt_is_range(mt) || matcher_mt_is_compare(mt))
		cost += METRIC_COST_AVG_INSERT_BY_INDEX;

	/* In case of FW WQE multiply this by GTA WQE to FW WQE performance ratio */
	if (matcher_req_fw_wqe(matcher))
		cost *= METRIC_COST_AVG_INSERT_BY_FW_WQE;

	/* Cost of inserting actions STEs */
	needed_action_stes = at->num_of_action_stes -
			     (!matcher_mt_is_jumbo(mt) || at->only_term);
	cost += needed_action_stes * METRIC_COST_AVG_INSERT_BY_INDEX;

	for (i = 0; i < at->num_actions; i++) {
		action_data = actions_data ? &actions_data[i] : NULL;
		cost += metric_query_action_insertion_cost(&at->actions[i], action_data);
	}

	cost += metric_query_action_insertion_cost(dest_action, NULL);

	/* Extra cost for large table TLB misses */
	if (attr->table.sz_row_log + attr->table.sz_row_log > METRIC_LARGE_TBL_LOG_SZ)
		cost += METRIC_COST_LARGE_TBL;

	/* Extra SW work for locking */
	if (!attr->optimize_using_rule_idx && needed_action_stes)
		cost += METRIC_COST_NO_OPT_BY_IDX;

	return (uint32_t)cost;
}

int mlx5dv_hws_metric_query_matcher(struct mlx5dv_hws_matcher *matcher,
				    uint64_t flags,
				    struct mlx5dv_hws_metric_matcher *out)
{
	if (flags &
	    ~(MLX5DV_HWS_METRIC_FLAG_INFO | MLX5DV_HWS_METRIC_FLAG_COST)) {
		HWS_ERR("Unsupported flags were passed 0x%" PRIx64, flags);
		errno = EOPNOTSUPP;
		return -errno;
	}

	if (matcher->tbl->level == TABLE_ROOT_LEVEL) {
		HWS_ERR("Unsupported for root level matcher");
		errno = EOPNOTSUPP;
		return -errno;
	}

	if (flags & MLX5DV_HWS_METRIC_FLAG_INFO)
		metric_query_matcher_info(matcher, out);

	if (flags & MLX5DV_HWS_METRIC_FLAG_COST) {
		out->cost.memory = metric_query_matcher_memory_cost(matcher);
		out->cost.miss_pkt_processing =
			metric_query_matcher_pkt_processing_miss_cost(matcher);
	}

	return 0;
}

static void
metric_query_action_resource(struct mlx5dv_hws_action *action,
			     struct mlx5dv_hws_metric_matcher_template *out,
			     struct mlx5dv_hws_action_data *action_data)
{
	struct mlx5dv_hws_resource *res;
	uint32_t value = 0;
	uint8_t res_idx;

	if (!action)
		return;

	switch (action->type) {
	case ACTION_TYP_DEST_DROP:
	case ACTION_TYP_DEST_MISS:
	case ACTION_TYP_DEST_TIR:
	case ACTION_TYP_DEST_TBL:
	case ACTION_TYP_DEST_ROOT:
	case ACTION_TYP_DEST_VPORT:
	case ACTION_TYP_DEST_MATCHER:
	case ACTION_TYP_TAG:
	case ACTION_TYP_POP_VLAN:
	case ACTION_TYP_PUSH_VLAN:
	case ACTION_TYP_REFORMAT_TNL_L2_TO_L2:
	case ACTION_TYP_REMOVE_HEADER:
	case ACTION_TYP_DEST_ARRAY:
	case ACTION_TYP_INLINE_SET:
	case ACTION_TYP_INLINE_ADD:
	case ACTION_TYP_INLINE_ADD_FIELD:
	case ACTION_TYP_INLINE_COPY:
	case ACTION_TYP_INLINE_INSERT:
	case ACTION_TYP_REFORMAT_TRAILER:
		break;
	case ACTION_TYP_ASO_METER:
	case ACTION_TYP_ASO_CT:
	case ACTION_TYP_ASO_IPSEC:
	case ACTION_TYP_ASO_FIRST_HIT:
		out->resource.aso++;
		break;
	case ACTION_TYP_CTR:
		out->resource.counters++;
		break;
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L2:
		if (action_data) {
			res_idx = action_data->reformat_l2_to_tnl_l2.resource_idx;
			res = action->reformat.resource[res_idx];
			value = resource_arg_gran_get(res->type);
		} else {
			value = resource_arg_get_max_gran(action->reformat.resource,
							  action->reformat.num_of_resources);
		}
		out->resource.args += value;
		break;
	case ACTION_TYP_REFORMAT_L2_TO_TNL_L3:
		if (action_data) {
			res_idx = action_data->reformat_l2_to_tnl_l3.resource_idx;
			res = action->reformat.resource[res_idx];
			value = resource_arg_gran_get(res->type);
		} else {
			value = resource_arg_get_max_gran(action->reformat.resource,
							  action->reformat.num_of_resources);
		}
		out->resource.args += value;
		break;
	case ACTION_TYP_REFORMAT_TNL_L3_TO_L2:
		if (action_data) {
			res_idx = action_data->reformat_tnl_l3_to_l2.resource_idx;
			res = action->modify_header.resource[res_idx];
			value = resource_arg_gran_get(res->type);
		} else {
			value = resource_arg_get_max_gran(action->modify_header.resource,
							  action->modify_header.num_of_resources);
		}
		out->resource.args += value;
		break;
	case ACTION_TYP_INSERT_HEADER:
		if (action_data) {
			res_idx = action_data->insert_header.resource_idx;
			res = action->insert_header.resource[res_idx];
			value = resource_arg_gran_get(res->type);
		} else {
			value = resource_arg_get_max_gran(action->insert_header.resource,
							  action->insert_header.num_of_resources);
		}
		out->resource.args += value;
		break;
	case ACTION_TYP_MODIFY_HDR:
		if (action_data) {
			res_idx = action_data->modify_header.resource_idx;
			res = action->modify_header.resource[res_idx];
			value = resource_arg_gran_get(res->type);
		} else {
			value = resource_arg_get_max_gran(action->modify_header.resource,
							  action->modify_header.num_of_resources);
		}
		out->resource.args += value;
		break;
	case ACTION_TYP_NAT64:
		value = ACTION_NAT64_STAGES *
			resource_arg_gran_get(MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_128B);
		out->resource.args += value;
		break;
	case ACTION_TYP_CRYPTO_ENCRYPT:
	case ACTION_TYP_CRYPTO_DECRYPT:
		/* IPsec uses DEK & IPsec offload objects */
		if (action->crypto.crypto_type == MLX5DV_HWS_ACTION_CRYPTO_TYPE_IPSEC)
			out->resource.crypto += 2;
		else
			out->resource.crypto++;
		break;
	default:
		break;
	}
}

static void
metric_query_mt_resource(struct mlx5dv_hws_matcher *matcher,
			 struct mlx5dv_hws_metric_matcher_template_attr *attr,
			 struct mlx5dv_hws_metric_matcher_template *out)
{
	struct mlx5dv_hws_at *at = &matcher->at[attr->at_idx];
	struct mlx5dv_hws_mt *mt = &matcher->mt[attr->mt_idx];
	struct mlx5dv_hws_action_data *action_data;
	int i;

	out->resource.match_stes = (matcher_mt_is_range(mt) ||
				    matcher_mt_is_compare(mt)) ? 2 : 1;
	out->resource.action_stes = at->num_of_action_stes -
				    (!matcher_mt_is_jumbo(mt) || at->only_term);

	/* Initialize resources */
	out->resource.counters = 0;
	out->resource.aso = 0;
	out->resource.args = 0;
	out->resource.crypto = 0;

	for (i = 0; i < at->num_actions; i++) {
		action_data = attr->actions_data ? &attr->actions_data[i] : NULL;
		metric_query_action_resource(&at->actions[i], out, action_data);
	}

	metric_query_action_resource(attr->dest_action, out, NULL);
}

int mlx5dv_hws_metric_query_matcher_template(struct mlx5dv_hws_matcher *matcher,
					     struct mlx5dv_hws_metric_matcher_template_attr *attr,
					     struct mlx5dv_hws_metric_matcher_template *out)
{
	if (attr->flags &
	    ~(MLX5DV_HWS_METRIC_FLAG_RESOURCE | MLX5DV_HWS_METRIC_FLAG_COST)) {
		HWS_ERR("Unsupported flags were passed 0x%" PRIx64, attr->flags);
		errno = EOPNOTSUPP;
		return -errno;
	}

	if (matcher->tbl->level == TABLE_ROOT_LEVEL) {
		HWS_ERR("Unsupported for root level matcher");
		errno = EOPNOTSUPP;
		return -errno;
	}

	if (attr->mt_idx >= matcher->num_of_mt ||
	    attr->at_idx >= matcher->num_of_at) {
		HWS_ERR("Invalid indexes");
		errno = EINVAL;
		return -errno;
	}

	if (attr->flags & MLX5DV_HWS_METRIC_FLAG_RESOURCE)
		metric_query_mt_resource(matcher, attr, out);

	if (attr->flags & MLX5DV_HWS_METRIC_FLAG_COST) {
		out->cost.memory =
			metric_query_template_memory_cost(matcher, attr);
		out->cost.pkt_processing =
			metric_query_template_pkt_processing_cost(matcher,
								  attr->dest_action,
								  attr->mt_idx,
								  attr->at_idx,
								  attr->actions_data);
		out->cost.insertion =
			metric_query_template_insertion_cost(matcher,
							     attr->dest_action,
							     attr->mt_idx,
							     attr->at_idx,
							     attr->actions_data);
	}

	return 0;
}
