/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"

static void parser_graph_node_unbind(struct mlx5dv_hws_parser_node *node)
{
	struct mlx5dv_hws_parser_sampler *sampler;

	cmd_destroy_obj(node->obj);
	list_for_each(&node->sampler_head, sampler, next) {
		sampler->binded--;
	}

	node->obj = NULL;
	node->binded = false;
}

static int parser_graph_flex_sample_info_query(struct mlx5dv_hws_parser_node *node)
{
	struct mlx5dv_hws_parser_sampler *sampler;
	int ret;

	list_for_each(&node->sampler_head, sampler, next) {
		ret = cmd_flex_sample_info_query(node->graph->ctx->ibv_ctx,
						 sampler->id,
						 &sampler->info_query);
		if (ret)
			return ret;
	}

	return 0;
}

static int parser_graph_node_bind(struct mlx5dv_hws_parser_node *node)
{
	int ret;
	/* TODO: change input from node into a cmd struct */
	node->obj = cmd_flex_node_create(node->graph->ctx, node);

	if (!node->obj)
		return -EINVAL;

	ret = parser_graph_flex_sample_info_query(node);
	if (ret) {
		parser_graph_node_unbind(node);
		return ret;
	}

	node->binded = true;

	return 0;
}

static int parser_graph_node_bind_upper(struct mlx5dv_hws_parser_node *node, int iter)
{
	struct mlx5dv_hws_parser_arc *arc;
	int err;

	if (node->graph->refcount < iter) {
		HWS_ERR("Cycles are not allowd in the graph");
		return -EINVAL;
	}

	list_for_each(&node->arc_src_head, arc, src_next) {
		if (arc->cfg.src_node.type == MLX5DV_HWS_PARSER_NODE_TYPE_NATIVE)
			continue;

		if (arc->cfg.src_node.flex.node->binded)
			continue;

		err = parser_graph_node_bind_upper(arc->cfg.src_node.flex.node, iter + 1);
		if (err)
			return err;
	}

	return parser_graph_node_bind(node);
}

static void parser_graph_node_unbind_lower(struct mlx5dv_hws_parser_node *node)
{
	struct mlx5dv_hws_parser_arc *arc;

	list_for_each(&node->arc_dst_head, arc, dst_next) {
		if (arc->cfg.dst_node.type == MLX5DV_HWS_PARSER_NODE_TYPE_NATIVE)
			continue;

		if (!arc->cfg.dst_node.flex.node->binded)
			continue;

		parser_graph_node_unbind_lower(arc->cfg.dst_node.flex.node);
	}

	parser_graph_node_unbind(node);
}

static void parser_graph_unbind_from_hw(struct mlx5dv_hws_parser_graph *graph)
{
	struct mlx5dv_hws_parser_node *node;

	list_for_each(&graph->head, node, next) {
		if (!node->binded)
			continue;

		parser_graph_node_unbind_lower(node);
	}
}

/* When a node is created, it is appended to the graph node list. On bind, we
 * iterate on that list recursively, ensuring all its sources are already
 * binded. If the iteration passes the number of nodes in the graph, we are in a
 * cycle.
 */
static int parser_graph_bind_to_hw(struct mlx5dv_hws_parser_graph *graph)
{
	struct mlx5dv_hws_parser_node *node;
	int err;

	list_for_each(&graph->head, node, next) {
		if (node->binded)
			continue;

		err = parser_graph_node_bind_upper(node, 0);
		if (err)
			goto err_node;
	}

	return 0;

err_node:
	parser_graph_unbind_from_hw(graph);
	return err;
}

struct mlx5dv_hws_parser_graph *
mlx5dv_hws_parser_graph_create(struct mlx5dv_hws_context *ctx)
{
	struct mlx5dv_hws_parser_graph *graph;

	graph = calloc(1, sizeof(*graph));
	if (!graph) {
		HWS_ERR("Failed to allocate memory for graph node");
		errno = ENOMEM;
		return NULL;
	}

	list_head_init(&graph->head);
	graph->ctx = ctx;

	return graph;
}

int mlx5dv_hws_parser_graph_destroy(struct mlx5dv_hws_parser_graph *graph)
{
	if (graph->binded) {
		HWS_ERR("Graph is still binded");
		errno = EINVAL;
		return -errno;
	}

	if (graph->refcount) {
		HWS_ERR("Graph refcount isn't zero");
		errno = EINVAL;
		return -errno;
	}

	free(graph);

	return 0;
}


struct mlx5dv_hws_parser_node *
mlx5dv_hws_parser_node_create(struct mlx5dv_hws_parser_graph *graph,
			      struct mlx5dv_hws_parser_node_cfg *node_cfg)
{
	struct mlx5dv_hws_parser_node *node;

	if (graph->binded) {
		HWS_ERR("Graph is binded");
		errno = EINVAL;
		return NULL;
	}

	if (node_cfg->hdr_len.mode == MLX5DV_HWS_PARSER_HDR_LEN_MODE_FIELD &&
	    (!node_cfg->hdr_len.field.multiplier ||
	     (node_cfg->hdr_len.field.multiplier - 1) &
	     node_cfg->hdr_len.field.multiplier)) {
		HWS_ERR("Unsupported multiplier for header length");
		errno = EINVAL;
		return NULL;
	}

	if (node_cfg->has_options &&
	    node_cfg->option_len.mode == MLX5DV_HWS_PARSER_HDR_LEN_MODE_FIELD &&
	    (!node_cfg->option_len.field.multiplier ||
	     (node_cfg->option_len.field.multiplier - 1) &
	     node_cfg->option_len.field.multiplier)) {
		HWS_ERR("Unsupported multiplier for option length");
		errno = EINVAL;
		return NULL;
	}

	node = calloc(1, sizeof(*node));
	if (!node) {
		HWS_ERR("Failed to allocate memopry for node");
		errno = ENOMEM;
		return NULL;
	}

	list_head_init(&node->arc_src_head);
	list_head_init(&node->arc_dst_head);
	list_head_init(&node->sampler_head);

	node->graph = graph;
	node->cfg = *node_cfg;
	list_add(&graph->head, &node->next);
	graph->refcount++;

	return node;
}

int mlx5dv_hws_parser_node_destroy(struct mlx5dv_hws_parser_node *node)
{
	if (node->graph->binded) {
		HWS_ERR("Graph is binded");
		errno = EINVAL;
		return -errno;
	}

	if (node->refcount) {
		HWS_ERR("Node is in use");
		errno = EINVAL;
		return -errno;
	}

	node->graph->refcount--;
	list_del(&node->next);
	free(node);

	return 0;
}

struct mlx5dv_hws_parser_arc *
mlx5dv_hws_parser_arc_create(struct mlx5dv_hws_parser_graph *graph,
			     struct mlx5dv_hws_parser_arc_cfg *arc_cfg)
{
	struct mlx5dv_hws_parser_arc *arc;

	if (graph->binded) {
		HWS_ERR("Graph is already binded");
		errno = EINVAL;
		return NULL;
	}

	arc = calloc(1, sizeof(*arc));
	if (!arc) {
		HWS_ERR("Failed to allocate memopry for arc");
		errno = ENOMEM;
		return NULL;
	}

	if (arc_cfg->src_node.type == MLX5DV_HWS_PARSER_NODE_TYPE_FLEX &&
		arc_cfg->src_node.flex.node->graph != graph) {
		HWS_ERR("Src node graph donesn't match passed graph");
		errno = ENOMEM;
		goto free_arc;
	}

	if (arc_cfg->dst_node.type == MLX5DV_HWS_PARSER_NODE_TYPE_FLEX &&
		arc_cfg->dst_node.flex.node->graph != graph) {
		HWS_ERR("Dst node graph donesn't match passed graph");
		errno = ENOMEM;
		goto free_arc;
	}

	arc->graph = graph;
	arc->cfg = *arc_cfg;

	if (arc->cfg.src_node.type == MLX5DV_HWS_PARSER_NODE_TYPE_FLEX) {
		arc->cfg.src_node.flex.node->refcount++;
		list_add(&arc->cfg.src_node.flex.node->arc_dst_head,
			 &arc->dst_next);
	}

	if (arc->cfg.dst_node.type == MLX5DV_HWS_PARSER_NODE_TYPE_FLEX) {
		arc->cfg.dst_node.flex.node->refcount++;
		list_add(&arc->cfg.dst_node.flex.node->arc_src_head,
			 &arc->src_next);
	}

	return arc;
free_arc:
	free(arc);
	return NULL;
}

int mlx5dv_hws_parser_arc_destroy(struct mlx5dv_hws_parser_arc *arc)
{
	if (arc->graph->binded) {
		HWS_ERR("Graph is binded");
		errno = EINVAL;
		return -errno;
	}

	if (arc->cfg.src_node.type == MLX5DV_HWS_PARSER_NODE_TYPE_FLEX) {
		arc->cfg.src_node.flex.node->refcount--;
		list_del(&arc->dst_next);
	}

	if (arc->cfg.dst_node.type == MLX5DV_HWS_PARSER_NODE_TYPE_FLEX) {
		arc->cfg.dst_node.flex.node->refcount--;
		list_del(&arc->src_next);
	}

	free(arc);

	return 0;
}

struct mlx5dv_hws_parser_sampler *
mlx5dv_hws_parser_sampler_create(struct mlx5dv_hws_parser_node *node,
				 struct mlx5dv_hws_parser_sampler_cfg *sampler_cfg)
{
	struct mlx5dv_hws_parser_sampler *sampler;
	struct mlx5dv_hws_parser_graph *graph;

	graph = node->graph;

	if (graph->binded) {
		HWS_ERR("Graph is binded");
		errno = EINVAL;
		return NULL;
	}

	if (sampler_cfg->node_encap_level > 1) {
		HWS_ERR("Unsupported node encap level");
		errno = EINVAL;
		return NULL;
	}

	if (sampler_cfg->field_cfg.bit_length != 32) {
		HWS_ERR("Unsupported sampler length");
		errno = EINVAL;
		return NULL;
	}

	if (sampler_cfg->field_cfg.bit_offset % BITS_IN_BYTE) {
		HWS_ERR("Bit offset should indicte byte offset");
		errno = EINVAL;
		return NULL;
	}

	if (sampler_cfg->share_hw_resources) {
		HWS_ERR("Sharing hw resources isn't supported");
		errno = ENOMEM;
		return NULL;
	}

	if (!node->cfg.has_options && sampler_cfg->sample_from_option) {
		HWS_ERR("Can't sample from options if node doesn't have options enabled");
		errno = EINVAL;
		return NULL;
	}

	sampler = calloc(1, sizeof(*sampler));
	if (!sampler) {
		HWS_ERR("Failed to allocate memory for sampler");
		errno = ENOMEM;
		return NULL;
	}

	sampler->node = node;
	sampler->cfg = *sampler_cfg;
	sampler->node->refcount++;
	list_add(&sampler->node->sampler_head, &sampler->next);

	return sampler;
}

int mlx5dv_hws_parser_sampler_query(struct mlx5dv_hws_parser_sampler *sampler,
				    enum mlx5dv_hws_field_name *parser_fname,
				    enum mlx5dv_hws_field_name *parser_ok_bit_fname,
				    enum mlx5dv_hws_field_header *field_heder,
				    uint32_t *modify_header_field)
{
	uint32_t bit_sz, bit_off;

	/* Validate sampler creation via geneve options or graph bind */
	if (!sampler->geneve_obj && !sampler->node->graph->binded) {
		errno = EINVAL;
		return -errno;
	}

	*modify_header_field = sampler->info_query.modify_field_id;

	bit_sz = BITS_IN_DW;
	bit_off = sampler->info_query.field_format_select_dw * BITS_IN_DW;
	*parser_fname = definer_hl_to_fname(MLX5DV_HWS_HEADER_NONE, bit_off, bit_sz);

	bit_sz = 1;
	/* ok_bit_offset is from LSB and struct mlx5_ifc_definer_hl_bits offset is from MSB */
	bit_off = sampler->info_query.ok_bit_format_select_dw * (BITS_IN_DW) +
		  BITS_IN_DW - sampler->info_query.ok_bit_offset - bit_sz;
	*parser_ok_bit_fname = definer_hl_to_fname(MLX5DV_HWS_HEADER_NONE, bit_off, bit_sz);
	/* Currently limited to MLX5DV_HWS_HEADER_NONE for parser_fname;
	 * other header types not yet supported.
	 */
	*field_heder = MLX5DV_HWS_HEADER_NONE;

	return 0;
}

int mlx5dv_hws_parser_sampler_destroy(struct mlx5dv_hws_parser_sampler *sampler)
{
	if (sampler->geneve_obj) {
		int ret;

		ret = cmd_destroy_obj(sampler->geneve_obj);
		free(sampler);

		return ret;
	}

	if (sampler->node->graph->binded) {
		HWS_ERR("Graph is still binded");
		errno = EINVAL;
		return -errno;
	}

	sampler->node->refcount--;
	free(sampler);

	return 0;
}

int mlx5dv_hws_parser_graph_bind(struct mlx5dv_hws_parser_graph *graph)
{
	int err;

	if (graph->binded) {
		HWS_ERR("Graph is already binded");
		errno = EINVAL;
		return -errno;
	}

	err = parser_graph_bind_to_hw(graph);
	if (!err)
		graph->binded = true;

	return err;
}

int mlx5dv_hws_parser_graph_unbind(struct mlx5dv_hws_parser_graph *graph)
{
	if (!graph->binded) {
		HWS_ERR("Graph isn't binded");
		errno = EINVAL;
		return -errno;
	}

	parser_graph_unbind_from_hw(graph);
	graph->binded = false;

	return 0;
}

struct mlx5dv_hws_parser_sampler *
mlx5dv_hws_parser_geneve_tlv_options_create(struct mlx5dv_hws_context *ctx,
					    struct mlx5dv_hws_parser_geneve_option_cfg *cfg)
{
	struct mlx5dv_hws_parser_sampler *sampler;
	int ret = 0;

	if (!ctx->caps->geneve_tlv_opt) {
		errno = EOPNOTSUPP;
		return NULL;
	}

	sampler = calloc(1, sizeof(*sampler));
	if (!sampler) {
		HWS_ERR("Failed to allocate memory for sampler");
		errno = ENOMEM;
		return NULL;
	}

	sampler->geneve_obj = cmd_geneve_tlv_option_create(ctx->ibv_ctx,
							   cfg->option_class,
							   cfg->option_type,
							   cfg->option_data_len,
							   cfg->option_class_ignore,
							   cfg->offset_valid,
							   cfg->sample_offset);
	if (!sampler->geneve_obj)
		goto free_sampler;

	if (ctx->caps->flex_parser_id_geneve_opt_0) {
		/* In FLEX_PARSER_PROFILE_ENABLE 0 the sampler is predefined */
		sampler->id = ctx->caps->flex_parser_id_geneve_opt_0;
	} else {
		/* In FLEX_PARSER_PROFILE_ENABLE 8 the sampler should be queried */
		ret = cmd_geneve_tlv_option_query(ctx->ibv_ctx,
						  sampler->geneve_obj,
						  &sampler->id);
	}

	if (ret)
		goto destroy_geneve_obj;

	ret = cmd_flex_sample_info_query(ctx->ibv_ctx, sampler->id, &sampler->info_query);
	if (ret)
		goto destroy_geneve_obj;

	return sampler;

destroy_geneve_obj:
	cmd_destroy_obj(sampler->geneve_obj);
free_sampler:
	free(sampler);
	return NULL;
}
