/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"

bool arg_check_idx_aligned(uint32_t arg_idx, size_t data_sz)
{
	size_t min_arg_sz = MLX5_IFC_ARG_SZ_DATA;
	uint32_t valid_gran;

	/*  64B, valid granularity is 1
	 * 128B, valid granularity is 2
	 * 256B, valid granularity is 4
	 * 512B, valid granularity is 8
	 * ...
	 * Note: -1 is done since ilog32 does log2above.
	 */

	valid_gran = 1 << (ilog32(max(data_sz, min_arg_sz) - 1) - 6);

	return !(arg_idx % valid_gran);
}

bool pat_require_reparse(__be64 *actions, uint16_t num_of_actions)
{
	uint16_t i, field, src_field;
	uint8_t action_id;

	for (i = 0; i < num_of_actions; i++) {
		action_id = DEVX_GET(set_action_in, &actions[i], action_type);

		switch (action_id) {
		case MLX5_MODIFICATION_TYPE_NOP:
			field = MLX5_MODI_OUT_NONE;
			break;

		case MLX5_MODIFICATION_TYPE_SET:
		case MLX5_MODIFICATION_TYPE_ADD:
			field = DEVX_GET(set_action_in, &actions[i], field);
			break;

		case MLX5_MODIFICATION_TYPE_COPY:
			field = DEVX_GET(copy_action_in, &actions[i], dst_field);
			if (field == MLX5_MODI_OUT_IP_PROTOCOL) {
				/* W/A. should be resolved by reparse control */
				src_field = DEVX_GET(copy_action_in, &actions[i], src_field);
				if (src_field == MLX5_MODI_OUT_IPSEC_NEXT_HDR ||
				    src_field == MLX5_MODI_PSP_HEADER_0)
					return false;
			}
			break;
		case MLX5_MODIFICATION_TYPE_ADD_FIELD:
			field = DEVX_GET(copy_action_in, &actions[i], dst_field);
			break;

		default:
			/* Insert/Remove/Unknown actions require reparse */
			return true;
		}

		/* Below fields can change packet structure require a reparse */
		if (field == MLX5_MODI_OUT_ETHERTYPE ||
		    field == MLX5_MODI_OUT_IP_PROTOCOL)
			return true;
	}

	return false;
}

/* Cache and cache element handling */
int pat_cache_init(struct pattern_cache **cache)
{
	struct pattern_cache *new_cache;

	new_cache = calloc(1, sizeof(*new_cache));
	if (!new_cache) {
		errno = ENOMEM;
		return errno;
	}
	list_head_init(&new_cache->head);
	pthread_spin_init(&new_cache->lock, PTHREAD_PROCESS_PRIVATE);

	*cache = new_cache;

	return 0;
}

void pat_cache_uninit(struct pattern_cache *cache)
{
	free(cache);
}

static bool pat_compare_pattern(int cur_num_of_actions,
				__be64 cur_actions[],
				int num_of_actions,
				__be64 actions[])
{
	int i;

	if (cur_num_of_actions != num_of_actions)
		return false;

	for (i = 0; i < num_of_actions; i++) {
		u8 action_id =
			DEVX_GET(set_action_in, &actions[i], action_type);

		if (action_id == MLX5_MODIFICATION_TYPE_COPY ||
		    action_id == MLX5_MODIFICATION_TYPE_ADD_FIELD) {
			if (actions[i] != cur_actions[i])
				return false;
		} else {
			/* Compare just the control, not the values */
			if ((__force __be32)actions[i] !=
			    (__force __be32)cur_actions[i])
				return false;
		}
	}

	return true;
}

static struct pattern_cache_item *
pat_find_cached_pattern(struct pattern_cache *cache,
			uint16_t num_of_actions,
			__be64 *actions)
{
	struct pattern_cache_item *cached_pat;

	list_for_each(&cache->head, cached_pat, next) {
		if (pat_compare_pattern(cached_pat->mh_data.num_of_actions,
					(__be64 *)cached_pat->mh_data.data,
					num_of_actions,
					actions))
			return cached_pat;
	}

	return NULL;
}

static struct pattern_cache_item *
pat_get_existing_cached_pattern(struct pattern_cache *cache,
				uint16_t num_of_actions,
				__be64 *actions)
{
	struct pattern_cache_item *cached_pattern;

	cached_pattern = pat_find_cached_pattern(cache, num_of_actions, actions);
	if (cached_pattern) {
		/* LRU: move it to be first in the list */
		list_del(&cached_pattern->next);
		list_add(&cache->head, &cached_pattern->next);
		cached_pattern->refcount++;
	}

	return cached_pattern;
}

static struct pattern_cache_item *
pat_add_pattern_to_cache(struct pattern_cache *cache,
			 struct mlx5dv_devx_obj *pattern_obj,
			 uint16_t num_of_actions,
			 __be64 *actions)
{
	struct pattern_cache_item *cached_pattern;

	cached_pattern = calloc(1, sizeof(*cached_pattern));
	if (!cached_pattern) {
		HWS_ERR("Failed to allocate cached_pattern");
		errno = ENOMEM;
		return NULL;
	}

	cached_pattern->mh_data.num_of_actions = num_of_actions;
	cached_pattern->mh_data.pattern_obj = pattern_obj;
	cached_pattern->mh_data.data =
		malloc(num_of_actions * MLX5_IFC_ARG_SZ_ACTION);
	if (!cached_pattern->mh_data.data) {
		HWS_ERR("Failed to allocate mh_data.data");
		errno = ENOMEM;
		goto free_cached_obj;
	}

	memcpy(cached_pattern->mh_data.data, actions,
	       num_of_actions * MLX5_IFC_ARG_SZ_ACTION);

	list_add(&cache->head, &cached_pattern->next);
	cached_pattern->refcount = 1;

	return cached_pattern;

free_cached_obj:
	free(cached_pattern);
	return NULL;
}

static struct pattern_cache_item *
pat_find_cached_pattern_by_obj(struct pattern_cache *cache,
			       struct mlx5dv_devx_obj *pat_obj)
{
	struct pattern_cache_item *cached_pattern;

	list_for_each(&cache->head, cached_pattern, next) {
		if (cached_pattern->mh_data.pattern_obj->object_id == pat_obj->object_id)
			return cached_pattern;
	}

	return NULL;
}

static void
pat_remove_pattern(struct pattern_cache_item *cached_pattern)
{
	list_del(&cached_pattern->next);
	free(cached_pattern->mh_data.data);
	free(cached_pattern);
}

void pat_put_pattern(struct mlx5dv_hws_context *ctx,
		     struct mlx5dv_devx_obj *pat_obj)
{
	struct pattern_cache *cache = ctx->pattern_cache;
	struct pattern_cache_item *cached_pattern;

	pthread_spin_lock(&cache->lock);
	cached_pattern = pat_find_cached_pattern_by_obj(cache, pat_obj);
	if (!cached_pattern) {
		HWS_ERR("Failed to find pattern according to action with pt");
		assert(false);
		goto out;
	}

	if (--cached_pattern->refcount)
		goto out;

	pat_remove_pattern(cached_pattern);
	cmd_destroy_obj(pat_obj);

out:
	pthread_spin_unlock(&cache->lock);
}

void pat_get_pattern_by_pat_obj(struct mlx5dv_hws_context *ctx,
				struct mlx5dv_devx_obj *pat_obj)
{
	struct pattern_cache *cache = ctx->pattern_cache;
	struct pattern_cache_item *cached_pattern;

	pthread_spin_lock(&cache->lock);
	cached_pattern = pat_find_cached_pattern_by_obj(cache, pat_obj);
	if (!cached_pattern) {
		HWS_ERR("Failed to find pattern according to pat object");
		assert(false);
		goto out;
	}

	cached_pattern->refcount++;

out:
	pthread_spin_unlock(&cache->lock);

}

struct mlx5dv_devx_obj *
pat_get_pattern(struct mlx5dv_hws_context *ctx,
		__be64 *pattern,
		size_t pattern_sz)
{
	uint16_t num_of_actions = pattern_sz / MLX5_IFC_ARG_SZ_ACTION;
	struct pattern_cache_item *cached_pattern;
	struct mlx5dv_devx_obj *pat_obj = NULL;

	pthread_spin_lock(&ctx->pattern_cache->lock);

	cached_pattern = pat_get_existing_cached_pattern(ctx->pattern_cache,
							num_of_actions,
							pattern);
	if (cached_pattern) {
		pat_obj = cached_pattern->mh_data.pattern_obj;
		goto out_unlock;
	}

	pat_obj = cmd_header_modify_pattern_create(ctx->ibv_ctx,
						   pattern_sz,
						   (uint8_t *)pattern);
	if (!pat_obj) {
		HWS_ERR("Failed to create pattern FW object");
		goto out_unlock;
	}

	cached_pattern = pat_add_pattern_to_cache(ctx->pattern_cache,
						  pat_obj,
						  num_of_actions,
						  pattern);
	if (!cached_pattern) {
		HWS_ERR("Failed to add pattern to cache");
		goto clean_pattern;
	}

	pthread_spin_unlock(&ctx->pattern_cache->lock);
	return pat_obj;

clean_pattern:
	cmd_destroy_obj(pat_obj);
	pat_obj = NULL;
out_unlock:
	pthread_spin_unlock(&ctx->pattern_cache->lock);
	return pat_obj;
}

static void
arg_init_send_attr(struct send_engine_post_attr *send_attr,
		   void *comp_data,
		   uint32_t arg_idx)
{
	send_attr->opcode = SEND_WQE_OPCODE_TBL_ACCESS;
	send_attr->opmod = SEND_WQE_GTA_OPMOD_MOD_ARG;
	send_attr->len = SEND_WQE_SZ_GTA_CTRL + SEND_WQE_SZ_GTA_DATA;
	send_attr->id = arg_idx;
	send_attr->user_data = comp_data;
}

void
arg_prepare_decap_l3_data(uint8_t *src, uint8_t *dst, uint16_t num_of_actions)
{
	uint8_t *e_src;
	int i;

	/* Action order:
	 * 1) Remove L2L3 from L2L3[L3...]
	 * 2) 4/5 inserts for new L2
	 * 3) Remove extra 2 bytes
	 * -Copy from end of src to the start of dst.
	 * -Move to the end, 2 is the leftover from 14B or 18B
	 */
	if (num_of_actions == ACTION_MH_DECAP_L3_NUM_ACTIONS_W_NO_VLAN)
		e_src = src + ACTION_HDR_LEN_L2;
	else
		e_src = src + ACTION_HDR_LEN_L2_W_VLAN;

	/* Move dst over the (1) remove action + zero data */
	dst += ACTION_MH_DOUBLE_SIZE;

	/* Move dst over the first insert ctrl action */
	dst += ACTION_MH_DOUBLE_SIZE / 2;

	/* No vlan:   RemoveHeader, Insert4B x4, Remove2B.
	 * With vlan: RemoveHeader, Insert4B x5, Remove2B.
	 * the loop is without the last insertion.
	 */
	for (i = 0; i < num_of_actions - 3; i++) {
		e_src -= ACTION_MH_INLINE_DATA_SIZE;
		memcpy(dst, e_src, ACTION_MH_INLINE_DATA_SIZE); /* data */
		dst += ACTION_MH_DOUBLE_SIZE;
	}

	/* Copy the last 2 bytes after a gap of 2 bytes which will be removed */
	e_src -= ACTION_MH_INLINE_DATA_SIZE / 2;
	dst += ACTION_MH_INLINE_DATA_SIZE / 2;
	memcpy(dst, e_src, 2);
}

void arg_decapl3_write(struct send_engine *queue,
		       uint32_t arg_idx,
		       uint8_t *arg_data,
		       uint16_t num_of_actions)
{
	struct send_engine_post_attr send_attr = {0};
	struct wqe_gta_data_seg_arg *wqe_arg;
	struct send_engine_post_ctrl ctrl;
	struct wqe_gta_ctrl_seg *wqe_ctrl;
	size_t wqe_len;

	arg_init_send_attr(&send_attr, NULL, arg_idx);

	ctrl = send_engine_post_start(queue);
	send_engine_post_req_wqe(&ctrl, (void *)&wqe_ctrl, &wqe_len);
	memset(wqe_ctrl, 0, wqe_len);
	send_engine_post_req_wqe(&ctrl, (void *)&wqe_arg, &wqe_len);
	memset(wqe_arg, 0, wqe_len);
	arg_prepare_decap_l3_data(arg_data, (uint8_t *)wqe_arg, num_of_actions);
	send_engine_post_end(&ctrl, &send_attr);
}

void arg_write(struct send_engine *queue,
	       void *comp_data,
	       uint32_t arg_idx,
	       uint8_t *arg_data,
	       size_t data_size)
{
	struct send_engine_post_attr send_attr = {0};
	struct send_wqe_gta_data_seg_arg *wqe_arg;
	struct send_engine_post_ctrl ctrl;
	struct send_wqe_gta_ctrl_seg *wqe_ctrl;
	int i, full_iter, leftover;
	size_t wqe_len;

	arg_init_send_attr(&send_attr, comp_data, arg_idx);

	/* Each WQE can hold 64B of data, it might require multiple iteration */
	full_iter = data_size / MLX5_IFC_ARG_SZ_DATA;
	leftover = data_size & (MLX5_IFC_ARG_SZ_DATA - 1);

	for (i = 0; i < full_iter; i++) {
		ctrl = send_engine_post_start(queue);
		send_engine_post_req_wqe(&ctrl, (void *)&wqe_ctrl, &wqe_len);
		memset(wqe_ctrl, 0, wqe_len);
		send_engine_post_req_wqe(&ctrl, (void *)&wqe_arg, &wqe_len);
		memcpy(wqe_arg, arg_data, wqe_len);
		send_attr.id = arg_idx++;
		send_engine_post_end(&ctrl, &send_attr);

		/* Move to next argument data */
		arg_data += MLX5_IFC_ARG_SZ_DATA;
	}

	if (leftover) {
		ctrl = send_engine_post_start(queue);
		send_engine_post_req_wqe(&ctrl, (void *)&wqe_ctrl, &wqe_len);
		memset(wqe_ctrl, 0, wqe_len);
		send_engine_post_req_wqe(&ctrl, (void *)&wqe_arg, &wqe_len);
		memcpy(wqe_arg, arg_data, leftover);
		send_attr.id = arg_idx;
		send_engine_post_end(&ctrl, &send_attr);
	}
}

bool pat_verify_actions(__be64 pattern[], size_t sz)
{
	size_t i;

	for (i = 0; i < sz / MLX5_IFC_ARG_SZ_ACTION; i++) {
		u8 action_id =
			DEVX_GET(set_action_in, &pattern[i], action_type);
		if (action_id >= MLX5_MODIFICATION_TYPE_MAX) {
			HWS_ERR("Invalid action %u", action_id);
			return false;
		}
	}

	return true;
}
