/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"
#include "hws_buddy.h"

static void pool_resource_free_one(struct pool_resource *resource)
{
	cmd_destroy_obj(resource->devx_obj);
	free(resource);
}

static void
pool_resource_free(struct pool *pool, int resource_idx)
{
	pool_resource_free_one(pool->resource[resource_idx]);
	pool->resource[resource_idx] = NULL;
}

static struct pool_resource *
pool_create_one_resource(struct pool *pool,
			 uint32_t log_range,
			 uint32_t fw_ft_type)
{
	struct mlx5dv_devx_obj *devx_obj = NULL;
	struct cmd_ste_create_attr ste_attr;
	struct cmd_stc_create_attr stc_attr;
	struct pool_resource *resource;

	resource = malloc(sizeof(*resource));
	if (!resource) {
		errno = ENOMEM;
		return NULL;
	}

	switch (pool->type) {
	case POOL_TYPE_STE:
		ste_attr.vport = pool->vport;
		ste_attr.other_vport = pool->other_vport;
		ste_attr.log_obj_range = log_range;
		ste_attr.table_type = fw_ft_type;
		devx_obj = cmd_ste_create(pool->ctx->ibv_ctx, &ste_attr);
		break;
	case POOL_TYPE_STC:
		stc_attr.vport = pool->vport;
		stc_attr.other_vport = pool->other_vport;
		stc_attr.log_obj_range = log_range;
		stc_attr.table_type = fw_ft_type;
		devx_obj = cmd_stc_create(pool->ctx->ibv_ctx, &stc_attr);
		break;
	default:
		assert(0);
		break;
	}

	if (!devx_obj) {
		HWS_ERR("Failed to allocate resource objects");
		goto free_resource;
	}

	resource->pool = pool;
	resource->devx_obj = devx_obj;
	resource->range = 1 << log_range;

	return resource;

free_resource:
	free(resource);
	return NULL;
}

static int
pool_resource_alloc(struct pool *pool, uint32_t log_range, int idx)
{
	struct pool_resource *resource;
	uint32_t fw_ft_type;

	fw_ft_type = table_get_rtc_fw_ft_type(pool->tbl_type);
	resource = pool_create_one_resource(pool, log_range, fw_ft_type);
	if (!resource) {
		HWS_ERR("Failed allocating resource");
		return -errno;
	}
	pool->resource[idx] = resource;

	return 0;
}

static int pool_bitmap_get_free_slot(unsigned long *bitmap, uint32_t sz, uint32_t *iidx)
{
	unsigned long idx;

	idx = bitmap_find_first_bit(bitmap, 0, sz);
	if (idx == sz)
		return -ENOMEM;

	bitmap_clear_bit(bitmap, idx);

	*iidx = idx;

	return 0;
}

static void pool_buddy_db_put_chunk(struct pool *pool,
				    struct pool_chunk *chunk)
{
	struct buddy_mem *buddy;

	buddy = pool->db.buddy_manager->buddies[chunk->resource_idx];
	if (!buddy) {
		assert(false);
		HWS_ERR("No such buddy (%d)", chunk->resource_idx);
		return;
	}

	buddy_free_mem(buddy, chunk->offset, chunk->order);
}

static struct buddy_mem *
pool_buddy_get_next_buddy(struct pool *pool, int idx,
			  uint32_t order, bool *is_new_buddy)
{
	static struct buddy_mem *buddy;
	uint32_t new_buddy_size;

	buddy = pool->db.buddy_manager->buddies[idx];
	if (buddy)
		return buddy;

	new_buddy_size = max_t(uint32_t, pool->alloc_log_sz, order);
	*is_new_buddy = true;
	buddy = buddy_create(new_buddy_size);
	if (!buddy) {
		HWS_ERR("Failed to create buddy order: %d index: %d",
			new_buddy_size, idx);
		return NULL;
	}

	if (pool_resource_alloc(pool, new_buddy_size, idx) != 0) {
		HWS_ERR("Failed to create resource type: %d: size %d index: %d",
			pool->type, new_buddy_size, idx);
		buddy_cleanup(buddy);
		return NULL;
	}

	pool->db.buddy_manager->buddies[idx] = buddy;

	return buddy;
}

static int pool_buddy_get_mem_chunk(struct pool *pool,
				    int order,
				    uint32_t *buddy_idx,
				    int *seg)
{
	struct buddy_mem *buddy;
	bool new_mem = false;
	int err = 0;
	int i;

	*seg = -1;

	/* Find the next free place from the buddy array */
	while (*seg == -1) {
		for (i = 0; i < POOL_RESOURCE_ARR_SZ; i++) {
			buddy = pool_buddy_get_next_buddy(pool, i, order, &new_mem);
			if (!buddy) {
				err = -errno;
				goto out;
			}

			*seg = buddy_alloc_mem(buddy, order);
			if (*seg != -1)
				goto found;

			if (pool->flags & POOL_FLAGS_ONE_RESOURCE) {
				HWS_ERR("Fail to allocate seg for one resource pool");
				errno = ENOMEM;
				err = -errno;
				goto out;
			}

			if (new_mem) {
				/* We have new memory pool, should be place for us */
				assert(false);
				HWS_ERR("No memory for order: %d with buddy no: %d", order, i);
				errno = ENOMEM;
				err = -errno;
				goto out;
			}
		}
	}

found:
	*buddy_idx = i;
out:
	return err;
}

static int pool_buddy_db_get_chunk(struct pool *pool,
				   struct pool_chunk *chunk)
{
	int ret = 0;

	/* Go over the buddies and find next free slot */
	ret = pool_buddy_get_mem_chunk(pool, chunk->order,
				       &chunk->resource_idx,
				       &chunk->offset);
	if (ret)
		HWS_ERR("Failed to get free slot for chunk with order: %d",
			chunk->order);

	return ret;
}

static void pool_buddy_db_uninit(struct pool *pool)
{
	struct buddy_mem *buddy;
	int i;

	for (i = 0; i < POOL_RESOURCE_ARR_SZ; i++) {
		buddy = pool->db.buddy_manager->buddies[i];
		if (buddy) {
			buddy_cleanup(buddy);
			free(buddy);
			pool->db.buddy_manager->buddies[i] = NULL;
		}
	}

	free(pool->db.buddy_manager);
}

static int pool_buddy_db_init(struct pool *pool, uint32_t log_range)
{
	pool->db.buddy_manager = calloc(1, sizeof(*pool->db.buddy_manager));
	if (!pool->db.buddy_manager) {
		HWS_ERR("No mem for buddy_manager with log_range: %d",
			log_range);
		errno = ENOMEM;
		return -errno;
	}

	if (pool->flags & POOL_FLAGS_ALLOC_MEM_ON_CREATE) {
		bool new_buddy;

		if (!pool_buddy_get_next_buddy(pool, 0, log_range, &new_buddy)) {
			HWS_ERR("Failed allocating memory on create log_sz: %d",
				log_range);
			free(pool->db.buddy_manager);
			return -errno;
		}
	}

	pool->p_db_uninit = &pool_buddy_db_uninit;
	pool->p_get_chunk = &pool_buddy_db_get_chunk;
	pool->p_put_chunk = &pool_buddy_db_put_chunk;

	return 0;
}

static int pool_create_resource_on_index(struct pool *pool,
					 uint32_t alloc_size, int idx)
{
	if (pool_resource_alloc(pool, alloc_size, idx) != 0) {
		HWS_ERR("Failed to create resource type: %d: size %d index: %d",
			pool->type, alloc_size, idx);
		return -errno;
	}

	return 0;
}

static struct pool_elements *
pool_element_create_new_elem(struct pool *pool, uint32_t order, int idx)
{
	struct pool_elements *elem;
	uint32_t alloc_size;

	alloc_size = pool->alloc_log_sz;

	elem = calloc(1, sizeof(*elem));
	if (!elem) {
		HWS_ERR("Failed to create elem order: %d index: %d",
			order, idx);
		errno = ENOMEM;
		return NULL;
	}

	/* Sharing the same resource, also means that all the elements are with size 1*/
	if ((pool->flags & POOL_FLAGS_FIXED_SIZE_OBJECTS) &&
	    !(pool->flags & POOL_FLAGS_RESOURCE_PER_CHUNK)) {
		 /* Currently all chunks in size 1 */
		elem->bitmap_sz = 1 << (alloc_size - order);
		elem->bitmap =  bitmap_alloc1(elem->bitmap_sz);
		if (!elem->bitmap) {
			HWS_ERR("Failed to create bitmap type: %d: size %d index: %d",
				pool->type, alloc_size, idx);
			goto free_elem;
		}
	}

	if (pool_create_resource_on_index(pool, alloc_size, idx)) {
		HWS_ERR("Failed to create resource type: %d: size %d index: %d",
			pool->type, alloc_size, idx);
		goto free_db;
	}

	pool->db.element_manager->elements[idx] = elem;

	return elem;

free_db:
	free(elem->bitmap);
free_elem:
	free(elem);
	return NULL;
}

static int pool_element_find_seg(struct pool_elements *elem, int *seg)
{
	if (pool_bitmap_get_free_slot(elem->bitmap, elem->bitmap_sz, (uint32_t *)seg)) {
		elem->is_full = true;
		return -ENOMEM;
	}

	return 0;
}

static int
pool_onesize_element_get_mem_chunk(struct pool *pool,
				   uint32_t order,
				   uint32_t *idx,
				   int *seg)
{
	struct pool_elements *elem;

	elem = pool->db.element_manager->elements[0];
	if (!elem)
		elem = pool_element_create_new_elem(pool, order, 0);
	if (!elem)
		goto err_no_elem;

	*idx = 0;

	if (pool_element_find_seg(elem, seg) != 0) {
		HWS_ERR("No more resources (last request order: %d)",
			order);
		errno = ENOMEM;
		return -errno;
	}

	elem->num_of_elements++;
	return 0;

err_no_elem:
	HWS_ERR("Failed to allocate element for order: %d", order);
	return -ENOMEM;
}

static int
pool_general_element_get_mem_chunk(struct pool *pool,
				   uint32_t order,
				   uint32_t *idx,
				   int *seg)
{
	int ret;
	int i;

	for (i = 0; i < POOL_RESOURCE_ARR_SZ; i++) {
		if (!pool->resource[i]) {
			ret = pool_create_resource_on_index(pool, order, i);
			if (ret)
				goto err_no_res;
			*idx = i;
			*seg = 0; /* One memory slot in that element */
			return 0;
		}
	}

	HWS_ERR("No more resources (last request order: %d)", order);
	errno = ENOMEM;
	return -errno;

err_no_res:
	HWS_ERR("Failed to allocate element for order: %d", order);
	return -ENOMEM;
}

static int pool_general_element_db_get_chunk(struct pool *pool,
					     struct pool_chunk *chunk)
{
	int ret;

	/* Go over all memory elements and find/allocate free slot */
	ret = pool_general_element_get_mem_chunk(pool, chunk->order,
						 &chunk->resource_idx,
						 &chunk->offset);
	if (ret)
		HWS_ERR("Failed to get free slot for chunk with order: %d",
			chunk->order);

	return ret;
}

static void pool_general_element_db_put_chunk(struct pool *pool,
					      struct pool_chunk *chunk)
{
	assert(pool->resource[chunk->resource_idx]);

	if (pool->flags & POOL_FLAGS_RELEASE_FREE_RESOURCE)
		pool_resource_free(pool, chunk->resource_idx);
}

static void pool_general_element_db_uninit(struct pool *pool)
{
	(void)pool;
}

/* This memory management works as the following:
 * - At start doesn't allocate no mem at all.
 * - When new request for chunk arrived:
 *	allocate resource and give it.
 * - When free that chunk:
 *	the resource is freed.
 */
static int pool_general_element_db_init(struct pool *pool)
{
	pool->p_db_uninit = &pool_general_element_db_uninit;
	pool->p_get_chunk = &pool_general_element_db_get_chunk;
	pool->p_put_chunk = &pool_general_element_db_put_chunk;

	return 0;
}

static int pool_onesize_element_db_get_chunk(struct pool *pool,
					     struct pool_chunk *chunk)
{
	int ret = 0;

	/* Go over all memory elements and find/allocate free slot */
	ret = pool_onesize_element_get_mem_chunk(pool,
						 chunk->order,
						 &chunk->resource_idx,
						 &chunk->offset);
	if (ret)
		HWS_ERR("Failed to get free slot for chunk with order: %d",
			chunk->order);

	return ret;
}

static void pool_onesize_element_db_destroy_element(struct pool *pool,
						    struct pool_elements *elem,
						    struct pool_chunk *chunk)
{
	assert(pool->resource[chunk->resource_idx]);

	pool_resource_free(pool, chunk->resource_idx);

	free(elem);
	pool->db.element_manager->elements[chunk->resource_idx] = NULL;
}

static void pool_onesize_element_db_put_chunk(struct pool *pool,
					      struct pool_chunk *chunk)
{
	struct pool_elements *elem;

	assert(chunk->resource_idx == 0);

	elem = pool->db.element_manager->elements[chunk->resource_idx];
	if (!elem) {
		assert(0);
		HWS_ERR("No such element (%d)", chunk->resource_idx);
		return;
	}

	bitmap_set_bit(elem->bitmap, chunk->offset);
	elem->is_full = false;
	elem->num_of_elements--;

	if (pool->flags & POOL_FLAGS_RELEASE_FREE_RESOURCE && !elem->num_of_elements)
		pool_onesize_element_db_destroy_element(pool, elem, chunk);
}

static void pool_onesize_element_db_uninit(struct pool *pool)
{
	struct pool_elements *elem;
	int i;

	for (i = 0; i < POOL_RESOURCE_ARR_SZ; i++) {
		elem = pool->db.element_manager->elements[i];
		if (elem) {
			free(elem->bitmap);
			free(elem);
			pool->db.element_manager->elements[i] = NULL;
		}
	}
	free(pool->db.element_manager);
}

/* This memory management works as the following:
 * - At start doesn't allocate memory at all.
 * - When new request for chunk arrived:
 *   aloocate the first and only slot of memory/resource.
 * - When memory is depleted return error.
 */
static int pool_onesize_element_db_init(struct pool *pool)
{
	pool->db.element_manager = calloc(1, sizeof(*pool->db.element_manager));
	if (!pool->db.element_manager) {
		HWS_ERR("No memory for general elemnt_manager");
		errno = ENOMEM;
		return -errno;
	}

	pool->p_db_uninit = &pool_onesize_element_db_uninit;
	pool->p_get_chunk = &pool_onesize_element_db_get_chunk;
	pool->p_put_chunk = &pool_onesize_element_db_put_chunk;

	return 0;
}

static int pool_db_init(struct pool *pool, enum pool_db_type db_type)
{
	int ret;

	switch (db_type) {
	case POOL_DB_TYPE_GENERAL_SIZE:
		ret = pool_general_element_db_init(pool);
		break;
	case POOL_DB_TYPE_ONE_SIZE_RESOURCE:
		ret = pool_onesize_element_db_init(pool);
		break;
	case POOL_DB_TYPE_BUDDY:
		ret = pool_buddy_db_init(pool, pool->alloc_log_sz);
		break;
	default:
		assert(0);
		break;
	}

	if (ret) {
		HWS_ERR("Failed to init general db : %d (ret: %d)",
			db_type, ret);
		return ret;
	}

	return 0;
}

static void pool_db_unint(struct pool *pool)
{
	pool->p_db_uninit(pool);
}

int pool_chunk_alloc(struct pool *pool, struct pool_chunk *chunk)
{
	int ret;

	pthread_spin_lock(&pool->lock);
	ret = pool->p_get_chunk(pool, chunk);
	pthread_spin_unlock(&pool->lock);

	return ret;
}

void pool_chunk_free(struct pool *pool, struct pool_chunk *chunk)
{
	pthread_spin_lock(&pool->lock);
	pool->p_put_chunk(pool, chunk);
	pthread_spin_unlock(&pool->lock);
}

struct pool *pool_create(struct mlx5dv_hws_context *ctx,
			 struct pool_attr *pool_attr)
{
	enum pool_db_type res_db_type;
	struct pool *pool;

	pool = calloc(1, sizeof(*pool));
	if (!pool) {
		errno = ENOMEM;
		return NULL;
	}

	pool->ctx = ctx;
	pool->type = pool_attr->pool_type;
	pool->alloc_log_sz = pool_attr->alloc_log_sz;
	pool->flags = pool_attr->flags;
	pool->tbl_type = pool_attr->table_type;
	pool->vport = pool_attr->vport;
	pool->other_vport = pool_attr->other_vport;

	pthread_spin_init(&pool->lock, PTHREAD_PROCESS_PRIVATE);

	/* Support general db */
	if (pool->flags == (POOL_FLAGS_RELEASE_FREE_RESOURCE |
			    POOL_FLAGS_RESOURCE_PER_CHUNK))
		res_db_type = POOL_DB_TYPE_GENERAL_SIZE;
	else if (pool->flags == (POOL_FLAGS_ONE_RESOURCE |
				 POOL_FLAGS_FIXED_SIZE_OBJECTS))
		res_db_type = POOL_DB_TYPE_ONE_SIZE_RESOURCE;
	else
		res_db_type = POOL_DB_TYPE_BUDDY;

	pool->alloc_log_sz = pool_attr->alloc_log_sz;

	if (pool_db_init(pool, res_db_type))
		goto free_pool;

	return pool;

free_pool:
	pthread_spin_destroy(&pool->lock);
	free(pool);
	return NULL;
}

int pool_destroy(struct pool *pool)
{
	int i;

	for (i = 0; i < POOL_RESOURCE_ARR_SZ; i++)
		if (pool->resource[i])
			pool_resource_free(pool, i);

	pool_db_unint(pool);

	pthread_spin_destroy(&pool->lock);
	free(pool);
	return 0;
}
