/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#ifndef _HWS_POOL_H_
#define _HWS_POOL_H_

#define POOL_STC_LOG_SZ 15
#define POOL_RESOURCE_ARR_SZ 100

enum pool_type {
	POOL_TYPE_STE,
	POOL_TYPE_STC,
};

struct pool_chunk {
	uint32_t resource_idx;
	/* Internal offset, relative to base index */
	int offset;
	int order;
};

struct pool_resource {
	struct pool *pool;
	struct mlx5dv_devx_obj *devx_obj;
	uint32_t range;
};

enum pool_flags {
	/* Only one resource in that pool */
	POOL_FLAGS_ONE_RESOURCE = 1 << 0,
	POOL_FLAGS_RELEASE_FREE_RESOURCE = 1 << 1,
	/* No sharing resources between chunks */
	POOL_FLAGS_RESOURCE_PER_CHUNK = 1 << 2,
	/* All objects are in the same size */
	POOL_FLAGS_FIXED_SIZE_OBJECTS = 1 << 3,
	/* Manged by buddy allocator */
	POOL_FLAGS_BUDDY_MANAGED = 1 << 4,
	/* Allocate pool_type memory on pool creation */
	POOL_FLAGS_ALLOC_MEM_ON_CREATE = 1 << 5,

	/* These values should be used by the caller */
	POOL_FLAGS_FOR_STC_POOL =
		POOL_FLAGS_ONE_RESOURCE |
		POOL_FLAGS_FIXED_SIZE_OBJECTS,
	POOL_FLAGS_FOR_MATCHER_STE_POOL =
		POOL_FLAGS_RELEASE_FREE_RESOURCE |
		POOL_FLAGS_RESOURCE_PER_CHUNK,
	POOL_FLAGS_FOR_STE_ACTION_POOL =
		POOL_FLAGS_ONE_RESOURCE |
		POOL_FLAGS_BUDDY_MANAGED |
		POOL_FLAGS_ALLOC_MEM_ON_CREATE,
};

struct pool_attr {
	enum pool_flags flags;
	enum pool_type pool_type;
	enum mlx5dv_hws_table_type table_type;
	uint16_t vport;
	bool other_vport;
	/* Allocation size once memory is depleted */
	size_t alloc_log_sz;
};

enum pool_db_type {
	/* Uses for allocating chunk of big memory, each element has its own resource in the FW */
	POOL_DB_TYPE_GENERAL_SIZE,
	/* One resource only, all the elements are with same one size */
	POOL_DB_TYPE_ONE_SIZE_RESOURCE,
	/* Many resources, the memory allocated with buddy mechanism */
	POOL_DB_TYPE_BUDDY,
};

struct buddy_manager {
	struct buddy_mem *buddies[POOL_RESOURCE_ARR_SZ];
};

struct pool_elements {
	uint32_t num_of_elements;
	unsigned long *bitmap;
	uint32_t bitmap_sz;
	bool is_full;
};

struct element_manager {
	struct pool_elements *elements[POOL_RESOURCE_ARR_SZ];
};

struct pool_db {
	enum pool_db_type type;
	union {
		struct element_manager *element_manager;
		struct buddy_manager *buddy_manager;
	};
};

typedef int (*pool_db_get_chunk)(struct pool *pool,
					struct pool_chunk *chunk);
typedef void (*pool_db_put_chunk)(struct pool *pool,
					 struct pool_chunk *chunk);
typedef void (*pool_unint_db)(struct pool *pool);

struct pool {
	struct mlx5dv_hws_context *ctx;
	enum pool_type type;
	enum pool_flags flags;
	pthread_spinlock_t lock;
	size_t alloc_log_sz;
	enum mlx5dv_hws_table_type tbl_type;
	uint16_t vport;
	bool other_vport;
	struct pool_resource *resource[POOL_RESOURCE_ARR_SZ];
	/* DB */
	struct pool_db db;
	/* Functions */
	pool_unint_db p_db_uninit;
	pool_db_get_chunk p_get_chunk;
	pool_db_put_chunk p_put_chunk;
};

struct pool *pool_create(struct mlx5dv_hws_context *ctx, struct pool_attr *pool_attr);

int pool_destroy(struct pool *pool);

int pool_chunk_alloc(struct pool *pool, struct pool_chunk *chunk);

void pool_chunk_free(struct pool *pool, struct pool_chunk *chunk);

static inline struct mlx5dv_devx_obj *
pool_chunk_get_base_devx_obj(struct pool *pool,
			     struct pool_chunk *chunk)
{
	return pool->resource[chunk->resource_idx]->devx_obj;
}
#endif /* _HWS_POOL_H_ */
