/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include <util/bitmap.h>
#include "hws_internal.h"
#include "hws_buddy.h"

#define RESOURCE_MAX_CHUNK_SIZE (1 << 10)
#define RESOURCE_MIN_CHUNK_SIZE (1 << 2)

enum resource_type_aso_alignment {
	RESOURCE_TYPE_ASO_ALIGNMENT_COUNTER = 4,
	RESOURCE_TYPE_ASO_ALIGNMENT_CONNTRACK = 1,
	RESOURCE_TYPE_ASO_ALIGNMENT_FLOW_HIT = 512,
	RESOURCE_TYPE_ASO_ALIGNMENT_METER = 2,
	RESOURCE_TYPE_ASO_ALIGNMENT_IPSEC_OFFLOAD = 1,
	RESOURCE_TYPE_ASO_ALIGNMENT_DEK = 1,
};

struct resource_index_pool_attr {
	/* Index pool size. Must be a power of 2 */
	uint32_t size;
	/* Number of queues for cache allocation */
	uint32_t n_queues;
	/* Chunk entry size. Must be a power of 2 */
	uint32_t chunk_size;
	/* Whether to allow variable block size allocations */
	bool block_alloc_support;
};

struct resource_ipool_chunk_fixed {
	unsigned long *lbmap;
};

struct resource_ipool_chunk_buddy {
	struct buddy_mem buddy;
};

struct resource_ipool_chunk {
	uint32_t base_idx;
	size_t free_count;
	size_t chunk_size;
	struct list_node next;
	union {
		struct resource_ipool_chunk_fixed fixed;
		struct resource_ipool_chunk_buddy buddy;
	};
};

struct resource_ipool_chunk_ops {
	int (*init)(struct resource_ipool_chunk *chunk, size_t chunk_size);
	void (*destroy)(struct resource_ipool_chunk *chunk);
	int (*get_idx)(struct resource_ipool_chunk *chunk, uint8_t log_sz);
	int (*put_idx)(struct resource_ipool_chunk *chunk, uint8_t log_sz,
		       uint32_t offset);
};

struct resource_ipool_per_qid {
	struct list_head available;
	struct list_head taken;
};

struct resource_index_pool {
	struct resource_ipool_per_qid *queue_cache;
	struct resource_ipool_chunk *chunk_arr;
	struct resource_index_pool_attr attr;
	pthread_spinlock_t lock;
	struct list_head chunk_list;
	const struct resource_ipool_chunk_ops *chunk_ops;
};

static int
resource_ipool_chunk_full(const struct resource_ipool_chunk *ipool_chunk)
{
	return ipool_chunk->free_count == ipool_chunk->chunk_size;
}

static int
resource_ipool_chunk_empty(const struct resource_ipool_chunk *ipool_chunk)
{
	return ipool_chunk->free_count == 0;
}

static int resource_ipool_fixed_init(struct resource_ipool_chunk *ipool_chunk,
				     size_t chunk_size)
{
	ipool_chunk->fixed.lbmap = bitmap_alloc1(chunk_size);
	if (!ipool_chunk->fixed.lbmap) {
		errno = ENOMEM;
		return -errno;
	}

	return 0;
}

static void
resource_ipool_fixed_destroy(struct resource_ipool_chunk *ipool_chunk)
{
	if (ipool_chunk->fixed.lbmap)
		free(ipool_chunk->fixed.lbmap);
}

static int
resource_ipool_fixed_get_idx(struct resource_ipool_chunk *ipool_chunk,
			     uint8_t log_sz)
{
	int idx;

	assert(log_sz == 0);

	idx = bitmap_find_first_bit(ipool_chunk->fixed.lbmap, 0,
				    ipool_chunk->chunk_size);
	bitmap_clear_bit(ipool_chunk->fixed.lbmap, idx);
	ipool_chunk->free_count--;

	return idx + ipool_chunk->base_idx;
}

static int
resource_ipool_fixed_put_idx(struct resource_ipool_chunk *ipool_chunk,
			     uint8_t log_sz, uint32_t offset)
{
	uint32_t lbmap_idx = offset - ipool_chunk->base_idx;

	assert(log_sz == 0);
	bitmap_set_bit(ipool_chunk->fixed.lbmap, lbmap_idx);
	ipool_chunk->free_count++;

	return 0;
}

static const struct resource_ipool_chunk_ops fixed_chunk_ops = {
	.init = resource_ipool_fixed_init,
	.destroy = resource_ipool_fixed_destroy,
	.get_idx = resource_ipool_fixed_get_idx,
	.put_idx = resource_ipool_fixed_put_idx,
};

static int resource_ipool_buddy_init(struct resource_ipool_chunk *ipool_chunk,
				     size_t chunk_size)
{
	const uint8_t log_sz = 31 - __builtin_clz(chunk_size);

	return buddy_init(&ipool_chunk->buddy.buddy, log_sz);
}

static void
resource_ipool_buddy_destroy(struct resource_ipool_chunk *ipool_chunk)
{
	buddy_cleanup(&ipool_chunk->buddy.buddy);
}

static int
resource_ipool_buddy_get_idx(struct resource_ipool_chunk *ipool_chunk,
			     uint8_t log_sz)
{
	int idx = buddy_alloc_mem(&ipool_chunk->buddy.buddy, log_sz);

	if (idx == -1)
		return -1;

	ipool_chunk->free_count -= (1 << log_sz);

	return idx + ipool_chunk->base_idx;
}

static int
resource_ipool_buddy_put_idx(struct resource_ipool_chunk *ipool_chunk,
			     uint8_t log_sz, uint32_t offset)
{
	uint32_t local_idx = offset - ipool_chunk->base_idx;

	buddy_free_mem(&ipool_chunk->buddy.buddy, local_idx, log_sz);
	ipool_chunk->free_count += (1 << log_sz);

	return 0;
}

static const struct resource_ipool_chunk_ops buddy_chunk_ops = {
	.init = resource_ipool_buddy_init,
	.destroy = resource_ipool_buddy_destroy,
	.get_idx = resource_ipool_buddy_get_idx,
	.put_idx = resource_ipool_buddy_put_idx,
};

static bool resource_ipool_precheck_valid(struct resource_index_pool_attr *attr)
{
	if (!attr || !attr->size)
		return false;

	if (!attr->chunk_size || attr->n_queues < 1)
		return false;

	if (!is_power_of_2(attr->size) || !is_power_of_2(attr->chunk_size))
		return false;

	return true;
}

static struct resource_index_pool *
resource_ipool_create(struct resource_index_pool_attr *attr)
{
	struct resource_ipool_chunk *chunk_arr;
	struct resource_index_pool *ipool;
	int i, n_entries, ret;

	if (!resource_ipool_precheck_valid(attr)) {
		errno = EINVAL;
		goto out_err;
	}

	ipool = calloc(1, sizeof(*ipool));
	if (!ipool) {
		errno = ENOMEM;
		goto out_err;
	}

	ipool->chunk_ops =
		attr->block_alloc_support ? &buddy_chunk_ops : &fixed_chunk_ops;
	ipool->attr = *attr;
	ipool->queue_cache =
		calloc(attr->n_queues, sizeof(*ipool->queue_cache));
	if (!ipool->queue_cache) {
		errno = ENOMEM;
		goto free_ipool;
	}

	for (i = 0; i < attr->n_queues; i++) {
		struct resource_ipool_per_qid *per_qid;

		per_qid = &ipool->queue_cache[i];
		list_head_init(&per_qid->available);
		list_head_init(&per_qid->taken);
	}

	list_head_init(&ipool->chunk_list);
	n_entries = attr->size / attr->chunk_size;
	chunk_arr = calloc(n_entries, sizeof(*chunk_arr));
	if (!chunk_arr) {
		errno = ENOMEM;
		goto free_queue_cache;
	}

	ipool->chunk_arr = chunk_arr;
	for (i = 0; i < n_entries; i++) {
		struct resource_ipool_chunk *chunk = &chunk_arr[i];

		chunk->base_idx = i * attr->chunk_size;
		chunk->free_count = attr->chunk_size;
		chunk->chunk_size = attr->chunk_size;
		ret = ipool->chunk_ops->init(chunk, attr->chunk_size);
		if (ret)
			goto free_chunk_arr;

		list_add_tail(&ipool->chunk_list, &chunk_arr[i].next);
	}

	pthread_spin_init(&ipool->lock, PTHREAD_PROCESS_PRIVATE);

	return ipool;

free_chunk_arr:
	for (i = 0; i < n_entries; i++)
		ipool->chunk_ops->destroy(&chunk_arr[i]);

	free(chunk_arr);
free_queue_cache:
	free(ipool->queue_cache);
free_ipool:
	free(ipool);
out_err:
	return NULL;
}

static void resource_ipool_destroy(struct resource_index_pool *ipool)
{
	int i, n_entries;

	n_entries = ipool->attr.size / ipool->attr.chunk_size;
	for (i = 0; i < n_entries; i++)
		ipool->chunk_ops->destroy(&ipool->chunk_arr[i]);

	free(ipool->chunk_arr);
	free(ipool->queue_cache);
	free(ipool);
}

static uint32_t resource_ipool_get_offset(struct resource_index_pool *ipool,
					  uint16_t qid, uint8_t log_sz,
					  uint32_t *resource_offset)
{
	struct resource_ipool_per_qid *per_qid = &ipool->queue_cache[qid];
	struct resource_ipool_chunk *ipool_chunk, *c;
	int idx;

	/* For fixed-size, there will be at most one available chunk. If there
	 * is one, allocation from it will never fail by definition, since
	 * empty chunks are promptly moved to the `taken` list.
	 *
	 * For buddy chunking, a chunk may have available offsets, but
	 * of an order smaller than desired. In this case we move on to
	 * other available chunks.
	 */
	ipool_chunk = NULL;
	list_for_each(&per_qid->available, c, next) {
		idx = ipool->chunk_ops->get_idx(c, log_sz);
		if (idx != -1) {
			ipool_chunk = c;
			break;
		}
	}

	if (unlikely(!ipool_chunk)) {
		/* Allocate from ipool */
		pthread_spin_lock(&ipool->lock);
		ipool_chunk = list_top(&ipool->chunk_list,
				       struct resource_ipool_chunk, next);
		if (unlikely(!ipool_chunk)) {
			pthread_spin_unlock(&ipool->lock);
			errno = ENOMEM;
			return -errno;
		}

		list_del(&ipool_chunk->next);
		pthread_spin_unlock(&ipool->lock);
		list_add(&per_qid->available, &ipool_chunk->next);
		/* Guaranteed to succeed because we got an empty chunk
		 * and we check upstack that log_sz fits in a chunk.
		 */
		idx = ipool->chunk_ops->get_idx(ipool_chunk, log_sz);
	}

	if (unlikely(resource_ipool_chunk_empty(ipool_chunk))) {
		/* No more available indexes in chunk (idx is last) */
		list_del(&ipool_chunk->next);
		list_add(&per_qid->taken, &ipool_chunk->next);
	}

	*resource_offset = idx;

	return 0;
}

int mlx5dv_hws_resource_get_offset(struct mlx5dv_hws_resource *resource,
				   uint16_t qid, uint32_t *resource_offset)
{
	return resource_ipool_get_offset(resource->ipool, qid, 0,
					 resource_offset);
}

int mlx5dv_hws_resource_get_offset_block(struct mlx5dv_hws_resource *resource,
					 uint16_t qid, uint8_t log_sz,
					 uint32_t *resource_offset)
{
	if (log_sz != 0 && !resource->ipool->attr.block_alloc_support) {
		errno = EINVAL;
		return -errno;
	}

	if ((1 << log_sz) > resource->ipool->attr.chunk_size) {
		errno = ENOMEM;
		return -errno;
	}

	return resource_ipool_get_offset(resource->ipool, qid, log_sz,
					 resource_offset);
}

static inline uint32_t resource_ipool_get_base_index(uint32_t offset,
						     uint32_t chunk_size)
{
	/* Substracting lbmap offset and dividing in chunk size */
	return (offset - (offset & (chunk_size - 1))) / chunk_size;
}

static int resource_ipool_put_offset(struct resource_index_pool *ipool,
				     uint16_t qid, uint8_t log_sz,
				     uint32_t offset)
{
	struct resource_ipool_per_qid *per_qid = &ipool->queue_cache[qid];
	struct resource_ipool_chunk *ipool_chunk;
	uint32_t base_index;
	int err;

	base_index = resource_ipool_get_base_index(offset,
						   ipool->attr.chunk_size);
	ipool_chunk = &ipool->chunk_arr[base_index];

	/* All indexes were used, move to available list */
	if (unlikely(resource_ipool_chunk_empty(ipool_chunk))) {
		list_del(&ipool_chunk->next);
		list_add(&per_qid->available, &ipool_chunk->next);
	}

	err = ipool->chunk_ops->put_idx(ipool_chunk, log_sz, offset);
	if (err)
		return err;

	/* All indexes are unused, return to ipool */
	if (unlikely(resource_ipool_chunk_full(ipool_chunk))) {
		list_del(&ipool_chunk->next);
		pthread_spin_lock(&ipool->lock);
		list_add(&ipool->chunk_list, &ipool_chunk->next);
		pthread_spin_unlock(&ipool->lock);
	}

	return 0;
}

int mlx5dv_hws_resource_put_offset(struct mlx5dv_hws_resource *resource,
				   uint16_t qid, uint32_t resource_offset)
{
	return resource_ipool_put_offset(resource->ipool, qid, 0,
					 resource_offset);
}

int mlx5dv_hws_resource_put_offset_block(struct mlx5dv_hws_resource *resource,
					 uint16_t qid, uint8_t log_sz,
					 uint32_t resource_offset)
{
	return resource_ipool_put_offset(resource->ipool, qid, log_sz,
					 resource_offset);
}

int mlx5dv_hws_resource_max_block_size(
	const struct mlx5dv_hws_resource *resource)
{
	return resource->ipool->attr.block_alloc_support ?
		       31 - __builtin_clz(resource->ipool->attr.chunk_size) :
			     0;
}

static uint8_t resource_aso_get_opmod(struct mlx5dv_hws_resource *resource)
{
	switch (resource->type) {
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER:
		return resource->ctx->caps->flow_counter_access_aso_opc_mod;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK:
		return SEND_ASO_OPC_MOD_CONNECTION_TRACKING;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_FLOW_HIT:
		return SEND_ASO_OPC_MOD_FLOW_HIT;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_METER:
		return SEND_ASO_OPC_MOD_FLOW_METER;
	default:
		assert(0);
	}
	/* Silence compilation warning */
	assert(0);
	return 0;
}

static size_t resource_aso_get_aligment(struct mlx5dv_hws_resource *resource)
{
	switch (resource->type) {
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER:
		return RESOURCE_TYPE_ASO_ALIGNMENT_COUNTER;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK:
		return RESOURCE_TYPE_ASO_ALIGNMENT_CONNTRACK;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_FLOW_HIT:
		return RESOURCE_TYPE_ASO_ALIGNMENT_FLOW_HIT;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_METER:
		return RESOURCE_TYPE_ASO_ALIGNMENT_METER;
	case MLX5DV_HWS_RESOURCE_TYPE_IPSEC_OFFLOAD:
		return RESOURCE_TYPE_ASO_ALIGNMENT_IPSEC_OFFLOAD;
	case MLX5DV_HWS_RESOURCE_TYPE_DEK:
		return RESOURCE_TYPE_ASO_ALIGNMENT_DEK;
	default:
		return 0;
	}
}

static bool hws_resource_devx_type_valid(enum mlx5dv_hws_resource_type rsc_type,
					 enum mlx5_devx_obj_type devx_type)
{
	switch (rsc_type) {
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER:
		return devx_type == MLX5_DEVX_FLOW_COUNTER;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK:
		return devx_type == MLX5_DEVX_ASO_CT;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_FLOW_HIT:
		return devx_type == MLX5_DEVX_ASO_FIRST_HIT;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_METER:
		return devx_type == MLX5_DEVX_FLOW_METER;
	case MLX5DV_HWS_RESOURCE_TYPE_DEK:
		return devx_type == MLX5_DEVX_DEK;
	case MLX5DV_HWS_RESOURCE_TYPE_IPSEC_OFFLOAD:
		return devx_type == MLX5_DEVX_IPSEC_OFFLOAD;
	case MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B:
	case MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_128B:
	case MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_256B:
		return devx_type == MLX5_DEVX_MODIFY_ARGUMENT;
	default:
		return false;
	}
	return false;
}

static uint32_t resource_calc_chunk_size(uint32_t size, uint32_t n_queues)
{
	uint32_t chunk_size;

	/* Size is the maximal number of offsets the pool will manage.
	 * We want to let any qid cache chunks that will minimize fragmentation
	 * and allow fast performance per qid. The closest (bottom) power of two
	 * of 25% from size / n_queues is a heuristic number that should
	 * reflect a balanced choice that follows the requirements.
	 *
	 * This is further clamped by a min and max for the following reasons:
	 * 1. Too large chunks have poor performance when allocating / freeing,
	 *    as the internal control structures grow.
	 * 2. Too small chunks lead to surprising behavior, e.g. allocating 32
	 *    offsets on 24 queues will yield a chunk_size of 1 and effectively
	 *    mean we can't use block allocation at all.
	 */
	chunk_size = 1 << (31 - __builtin_clz(0.25 * size / n_queues));
	if (chunk_size > RESOURCE_MAX_CHUNK_SIZE)
		chunk_size = RESOURCE_MAX_CHUNK_SIZE;
	if (chunk_size < RESOURCE_MIN_CHUNK_SIZE)
		chunk_size = RESOURCE_MIN_CHUNK_SIZE;
	/* But make sure that chunk size does not exceed the original size. */
	if (chunk_size > size)
		chunk_size = size;

	return chunk_size;
}

struct mlx5dv_hws_resource *
mlx5dv_hws_resource_alloc(struct mlx5dv_hws_context *ctx,
			  struct mlx5dv_hws_resource_attr *attr)
{
	struct resource_index_pool_attr pool_attr;
	struct mlx5dv_hws_resource *resource;
	uint16_t aso_ctx_bulk_log_size = 0;
	struct mlx5dv_devx_obj *devx_obj;

	if (attr->comp_mask) {
		HWS_ERR("Invalid comp mask provided");
		errno = ENOTSUP;
		return NULL;
	}

	resource = calloc(1, sizeof(*resource));
	if (!resource) {
		HWS_ERR("Failed to allocate resource");
		errno = ENOMEM;
		return NULL;
	}

	resource->ctx = ctx;
	resource->type = attr->type;
	pool_attr.size = 1 << attr->bulk_log_size;
	pool_attr.n_queues = ctx->queues;
	pool_attr.block_alloc_support = attr->block_alloc_support;
	pool_attr.chunk_size =
		resource_calc_chunk_size(pool_attr.size, pool_attr.n_queues);
	resource->ipool = resource_ipool_create(&pool_attr);
	if (!resource->ipool)
		goto free_resource;

	if (attr->obj) {
		if (!hws_resource_devx_type_valid(resource->type, attr->obj->type)) {
			HWS_ERR("User input devx object is different than resource type");
			goto free_index_pool;
		}

		resource->user_devx_obj = true;
		resource->devx_obj = attr->obj;
		return resource;
	}

	switch (attr->type) {
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER:
		if (!ctx->caps->flow_counter_access_aso) {
			HWS_ERR("Flow counter access ASO is not supported");
			errno = ENOTSUP;
			goto free_index_pool;
		}

		devx_obj = cmd_flow_counter_create(ctx->ibv_ctx,
						   attr->bulk_log_size,
						   ctx->pd_num);
		break;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK:
		devx_obj = cmd_conn_track_create(ctx->ibv_ctx,
						 attr->bulk_log_size,
						 ctx->pd_num);
		break;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_FLOW_HIT:
		if (attr->bulk_log_size >= MLX5_ASO_FIRST_HIT_LOG_NUM_PER_OBJ)
			aso_ctx_bulk_log_size = attr->bulk_log_size -
						MLX5_ASO_FIRST_HIT_LOG_NUM_PER_OBJ;
		devx_obj = cmd_flow_hit_create(ctx->ibv_ctx,
					       aso_ctx_bulk_log_size,
					       ctx->pd_num);
		break;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_METER:
		if (attr->bulk_log_size >= MLX5_ASO_METER_LOG_NUM_PER_OBJ)
			aso_ctx_bulk_log_size = attr->bulk_log_size -
						MLX5_ASO_METER_LOG_NUM_PER_OBJ;
		devx_obj = cmd_flow_meter_create(ctx->ibv_ctx,
						 aso_ctx_bulk_log_size,
						 ctx->pd_num);
		break;
	case MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B:
		devx_obj = cmd_arg_create(ctx->ibv_ctx,
					  attr->bulk_log_size,
					  ctx->pd_num);
		break;
	case MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_128B:
		devx_obj = cmd_arg_create(ctx->ibv_ctx,
					  (attr->bulk_log_size + 1),
					  ctx->pd_num);
		break;
	case MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_256B:
		devx_obj = cmd_arg_create(ctx->ibv_ctx,
					  (attr->bulk_log_size + 2),
					  ctx->pd_num);
		break;
	case MLX5DV_HWS_RESOURCE_TYPE_IPSEC_OFFLOAD:
		HWS_ERR("devx_obj missing for IPSEC");
		errno = ENOTSUP;
		break;
	case MLX5DV_HWS_RESOURCE_TYPE_DEK:
		HWS_ERR("devx_obj missing for PSP");
		errno = ENOTSUP;
		goto free_index_pool;
	default:
		goto free_index_pool;
	}

	if (!devx_obj) {
		HWS_ERR("Failed to create resource");
		goto free_index_pool;
	}

	resource->user_devx_obj = false;
	resource->devx_obj = devx_obj;

	return resource;

free_index_pool:
	resource_ipool_destroy(resource->ipool);
free_resource:
	free(resource);
	return NULL;
}

int mlx5dv_hws_resource_free(struct mlx5dv_hws_resource *resource)
{
	if (!resource->user_devx_obj)
		cmd_destroy_obj(resource->devx_obj);
	resource_ipool_destroy(resource->ipool);
	free(resource);

	return 0;
}

int mlx5dv_hws_resource_get_devx_obj(struct mlx5dv_hws_resource *resource,
				     struct mlx5dv_devx_obj **ret_obj,
				     uint32_t *ret_obj_id)
{
	if (!resource || !ret_obj || !ret_obj_id) {
		errno = EINVAL;
		return -errno;
	}

	*ret_obj = resource->devx_obj;
	*ret_obj_id = resource->devx_obj->object_id;
	return 0;
}

static int
resource_enqueue_aso_wqe_check(struct mlx5dv_hws_resource *resource,
			       size_t wqe_len,
			       struct mlx5dv_hws_resource_enqueue_aso_attr *attr,
			       struct send_engine *queue)
{
	if (unlikely(attr->comp_mask)) {
		HWS_ERR("Invalid comp mask provided");
		errno = ENOTSUP;
		return -errno;
	}

	if (unlikely(attr->offset &
		     (resource_aso_get_aligment(resource) - 1))) {
		HWS_ERR("Invalid argument, Offset is not aligned");
		errno = EINVAL;
		return -errno;
	}

	if (unlikely(wqe_len != sizeof(struct send_wqe_aso_ctrl_seg) &&
		     wqe_len != sizeof(struct send_wqe_aso_ctrl_seg) +
				sizeof(((struct send_wqe_aso *)0)->data))) {
		HWS_ERR("Invalid wqe_len value (48B or 112B)");
		errno = EINVAL;
		return -errno;
	}

	if (unlikely(send_engine_err(queue))) {
		HWS_ERR("Queue error detected");
		errno = EIO;
		return -errno;
	}

	if (unlikely(send_engine_full(queue))) {
		HWS_ERR("Queue is full");
		errno = EBUSY;
		return -errno;
	}

	return 0;
}

int
resource_enqueue_aso_wqe(struct mlx5dv_hws_resource *resource,
			 struct send_engine *queue,
			 __be32 *aso_wqe,
			 size_t wqe_len,
			 struct mlx5dv_hws_resource_enqueue_aso_attr *attr)
{
	struct send_engine_post_attr send_attr = {0};
	struct send_engine_post_ctrl ctrl;
	void *wqe_data, *wqe_ctrl;
	size_t wqe_len_wr;

	if (resource_enqueue_aso_wqe_check(resource, wqe_len, attr, queue))
		return -errno;

	send_attr.opmod = resource_aso_get_opmod(resource);
	send_attr.opcode = SEND_WQE_OPCODE_ACCESS_ASO;
	send_attr.len = MLX5_SEND_WQE_BB * 2 - sizeof(struct send_wqe_ctrl_seg);

	switch (resource->type) {
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER:
		/* Offset by counter id (divided by 4), and not by ASO Context index */
		send_attr.id = (resource->devx_obj->object_id + attr->offset) /
				MLX5_ASO_FLOW_COUNTER_NUM_PER_OBJ;
		break;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_METER:
		send_attr.id = resource->devx_obj->object_id +
			       (attr->offset / MLX5_ASO_METER_NUM_PER_OBJ);
		break;
	case MLX5DV_HWS_RESOURCE_TYPE_ASO_FLOW_HIT:
		send_attr.id = resource->devx_obj->object_id +
			       (attr->offset / MLX5_ASO_FIRST_HIT_NUM_PER_OBJ);
		break;
	default:
		send_attr.id = resource->devx_obj->object_id + attr->offset;
		break;
	}

	send_attr.notify_hw = !attr->burst;
	send_attr.user_data = attr->user_data;

	ctrl = send_engine_post_start(queue);
	send_engine_post_req_wqe(&ctrl, (void *)&wqe_ctrl, &wqe_len_wr);
	send_engine_post_req_wqe(&ctrl, (void *)&wqe_data, &wqe_len_wr);
	memcpy(wqe_ctrl, (void *)aso_wqe,
	       sizeof(struct send_wqe_aso_ctrl_seg));
	if (wqe_len == sizeof(struct send_wqe_aso_ctrl_seg) +
		       sizeof(((struct send_wqe_aso *)0)->data)) {
		memcpy(wqe_data,
		       (void *)aso_wqe +
		       sizeof(struct send_wqe_aso_ctrl_seg),
		       sizeof(((struct send_wqe_aso *)0)->data));
	}

	send_engine_post_end(&ctrl, &send_attr);

	if (send_attr.user_data)
		send_engine_inc_entry(queue);

	/* Send dependent WQEs */
	if (!attr->burst)
		send_all_dep_wqe(queue);

	return 0;
}

uint32_t resource_arg_gran_get(enum mlx5dv_hws_resource_type resource_type)
{
	return (1 << (resource_type - MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B));
}

int
mlx5dv_hws_resource_enqueue_aso_wqe(struct mlx5dv_hws_resource *resource,
				    uint16_t qid,
				    __be32 *aso_wqe,
				    size_t wqe_len,
				    struct mlx5dv_hws_resource_enqueue_aso_attr *attr)
{
	struct send_engine *queue;

	queue = &resource->ctx->send_queue[qid];

	return resource_enqueue_aso_wqe(resource, queue, aso_wqe, wqe_len, attr);
}
