/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"
#include <string.h>

size_t mlx5dv_hws_rule_get_handle_size(void)
{
	return sizeof(struct mlx5dv_hws_rule);
}

static void
rule_update_copy_tag(struct mlx5dv_hws_rule *rule,
		     struct send_wqe_gta_data_seg_ste *wqe_data,
		     bool is_jumbo)
{
	if (is_jumbo)
		memcpy(wqe_data->jumbo, rule->tag.jumbo, RULE_MATCH_TAG_SZ_JUMBO);
	else
		memcpy(wqe_data->tag, rule->tag.match, RULE_MATCH_TAG_SZ_MATCH);
}

void rule_init_dep_wqe(struct send_ring_dep_wqe *dep_wqe,
		       struct mlx5dv_hws_rule *rule,
		       struct mlx5dv_hws_rule_attr *attr,
		       bool is_update)
{
	struct mlx5dv_hws_matcher *matcher = rule->matcher;

	dep_wqe->rule = rule;
	dep_wqe->used_id = &rule->rtc;
	dep_wqe->user_data = attr->user_data;
	dep_wqe->direct_index = matcher_is_insert_by_idx(matcher) ?
		attr->rule_idx : 0;

	if (is_update) {
		dep_wqe->rtc = rule->rtc;
		dep_wqe->retry_rtc = 0;
		return;
	}

	dep_wqe->rtc = matcher_get_match_rtc_id(matcher);
	dep_wqe->retry_rtc = matcher_get_match_rtc_id(matcher->col_matcher);
}

static void rule_move_get_rtc(struct mlx5dv_hws_rule *rule,
			      struct send_ste_attr *ste_attr)
{
	struct mlx5dv_hws_matcher *dst_matcher = rule->matcher->resize_dst;

	ste_attr->send_attr.id = matcher_get_match_rtc_id(dst_matcher);
	ste_attr->send_attr.retry_id = matcher_get_match_rtc_id(dst_matcher->col_matcher);
}

static void rule_gen_comp(struct send_engine *queue,
			  struct mlx5dv_hws_rule *rule,
			  void *user_data,
			  enum mlx5dv_hws_queue_op_status op_status,
			  enum rule_status rule_status)
{
	rule->status = rule_status;

	send_engine_inc_entry(queue);
	send_engine_gen_comp(queue, user_data, op_status);
}

static void
rule_save_resize_info(struct mlx5dv_hws_rule *rule,
		      struct send_ste_attr *ste_attr)
{
	struct rule_resize_info *resize_info;
	struct mlx5dv_hws_matcher *matcher;

	if (likely(!matcher_is_resizable(rule->matcher)))
		return;

	resize_info = calloc(1, sizeof(*rule->resize_info));
	if (unlikely(!resize_info)) {
		assert(resize_info);
		errno = ENOMEM;
	}

	memcpy(resize_info->ctrl_seg, ste_attr->wqe_ctrl,
	       sizeof(resize_info->ctrl_seg));
	memcpy(resize_info->data_seg, ste_attr->wqe_data,
	       sizeof(resize_info->data_seg));

	matcher = rule->matcher;

	resize_info->max_stes = matcher->action_ste.max_stes;
	resize_info->action_ste_pool = matcher->action_ste.pool;

	rule->resize_info = resize_info;
}

void rule_clear_resize_info(struct mlx5dv_hws_rule *rule)
{
	if (unlikely(matcher_is_resizable(rule->matcher) && rule->resize_info)) {
		free(rule->resize_info);
		rule->resize_info = NULL;
	}

}

static void
rule_save_delete_info(struct mlx5dv_hws_rule *rule,
		      struct send_ste_attr *ste_attr)
{
	struct mlx5dv_hws_mt *mt = rule->matcher->mt;
	bool is_jumbo = matcher_mt_is_jumbo(mt);

	if (unlikely(matcher_req_fw_wqe(rule->matcher))) {
		/* Save match definer id and tag for delete */
		rule->tag_ptr = calloc(2, sizeof(*rule->tag_ptr));
		assert(rule->tag_ptr);

		if (is_jumbo)
			memcpy(rule->tag_ptr[0].jumbo, ste_attr->wqe_data->action,
			       RULE_MATCH_TAG_SZ_JUMBO);
		else
			memcpy(rule->tag_ptr[0].match, ste_attr->wqe_data->tag,
			       RULE_MATCH_TAG_SZ_MATCH);

		rule->tag_ptr[1].reserved[0] = ste_attr->send_attr.match_definer_id;

		/* Save range/compare definer id and tag for delete */
		if (ste_attr->scnd_wqe_data) {
			memcpy(rule->tag_ptr[1].match, ste_attr->scnd_wqe_data->tag,
			       RULE_MATCH_TAG_SZ_MATCH);
			rule->tag_ptr[1].reserved[1] = ste_attr->send_attr.scnd_definer_id;
		}
		return;
	}

	if (likely(!matcher_is_resizable(rule->matcher))) {
		if (is_jumbo)
			memcpy(&rule->tag.jumbo, ste_attr->wqe_data->action,
			       RULE_MATCH_TAG_SZ_JUMBO);
		else
			memcpy(&rule->tag.match, ste_attr->wqe_data->tag,
			       RULE_MATCH_TAG_SZ_MATCH);
		return;
	}
}

static void rule_clear_delete_info(struct mlx5dv_hws_rule *rule)
{
	if (unlikely(matcher_req_fw_wqe(rule->matcher))) {
		free(rule->tag_ptr);
		return;
	}
}

static void rule_load_delete_info(struct mlx5dv_hws_rule *rule,
				  struct send_ste_attr *ste_attr)
{
	if (unlikely(matcher_req_fw_wqe(rule->matcher))) {
		/* Load match definer id and tag for delete */
		ste_attr->wqe_tag = &rule->tag_ptr[0];
		ste_attr->send_attr.match_definer_id = rule->tag_ptr[1].reserved[0];

		/* Load range/compare definer id and tag for delete */
		if (matcher_is_range(rule->matcher) ||
		    matcher_is_compare(rule->matcher)) {
			ste_attr->scnd_wqe_tag = &rule->tag_ptr[1];
			ste_attr->send_attr.scnd_definer_id = rule->tag_ptr[1].reserved[1];
		}
	} else if (likely(!matcher_is_resizable(rule->matcher))) {
		ste_attr->wqe_tag = &rule->tag;
	} else {
		ste_attr->wqe_tag = (struct rule_match_tag *)
			&(rule->resize_info->data_seg[SEND_WQE_GTA_STE_SZ_CTRL]);
	}
}

static int rule_alloc_action_ste(struct mlx5dv_hws_rule *rule,
				 struct mlx5dv_hws_rule_attr *attr)
{
	struct mlx5dv_hws_matcher *matcher = rule->matcher;
	int ret;

	/* Use rule_idx for locking optimzation, otherwise allocate from pool */
	if (matcher->attr.optimize_using_rule_idx ||
	    matcher_is_insert_by_idx(matcher)) {
		rule->action_ste_idx = attr->rule_idx * matcher->action_ste.max_stes;
	} else {
		struct pool_chunk ste = {0};

		ste.order = ilog32(matcher->action_ste.max_stes - 1);
		ret = pool_chunk_alloc(matcher->action_ste.pool, &ste);
		if (ret) {
			HWS_ERR("Failed to allocate STE for rule actions");
			return ret;
		}
		rule->action_ste_idx = ste.offset;
	}
	return 0;
}

void rule_free_action_ste_idx(struct mlx5dv_hws_rule *rule)
{
	struct mlx5dv_hws_matcher *matcher = rule->matcher;
	struct pool *pool;
	uint8_t max_stes;

	if (rule->action_ste_idx > -1 &&
	    !matcher->attr.optimize_using_rule_idx &&
	    !matcher_is_insert_by_idx(matcher)) {
		struct pool_chunk ste = {0};

		if (unlikely(matcher_is_resizable(matcher))) {
			/* Free the original action pool if rule was resized */
			max_stes = rule->resize_info->max_stes;
			pool = rule->resize_info->action_ste_pool;
		} else {
			max_stes = matcher->action_ste.max_stes;
			pool = matcher->action_ste.pool;
		}

		/* This release is safe only when the rule match part was deleted */
		ste.order = ilog32(max_stes - 1);
		ste.offset = rule->action_ste_idx;

		pool_chunk_free(pool, &ste);
	}
}

void rule_create_init_ste_attr(struct mlx5dv_hws_rule *rule,
			       struct send_ste_attr *ste_attr,
			       bool is_update)
{
	/* Init rule before reuse, in update we need current rtc */
	if (!is_update) {
		rule->rtc = 0;
		rule->rtc_ext = 0;
	}

	rule->pending_wqes = 0;
	rule->action_ste_idx = -1;
	rule->status = RULE_STATUS_CREATING;

	/* Init default send STE attributes */
	ste_attr->gta_opcode = SEND_WQE_GTA_OP_ACTIVATE;
	ste_attr->send_attr.opmod = SEND_WQE_GTA_OPMOD_STE;
	ste_attr->send_attr.opcode = SEND_WQE_OPCODE_TBL_ACCESS;
	ste_attr->send_attr.len = SEND_WQE_SZ_GTA_CTRL + SEND_WQE_SZ_GTA_DATA;
}

static void rule_create_init(struct mlx5dv_hws_rule *rule,
			     struct send_ste_attr *ste_attr,
			     struct actions_apply_data *apply,
			     bool is_update)
{
	struct mlx5dv_hws_matcher *matcher = rule->matcher;
	struct mlx5dv_hws_table *tbl = matcher->tbl;

	rule_create_init_ste_attr(rule, ste_attr, is_update);

	/* Init default action apply */
	apply->tbl_type = tbl->type;
	apply->default_stcs = tbl->common_res->default_stc;
	apply->jump_to_action_stc = matcher->action_ste.stc.offset;
	apply->require_dep = 0;
	apply->rule = rule;
}

static void rule_move_init(struct mlx5dv_hws_rule *rule,
			   struct mlx5dv_hws_rule_attr *attr)
{
	/* Save the old RTC IDs to be later used in match STE delete */
	rule->resize_info->rtc = rule->rtc;
	rule->resize_info->rule_idx = attr->rule_idx;

	rule->rtc = 0;
	rule->pending_wqes = 0;
	rule->action_ste_idx = -1;
	rule->status = RULE_STATUS_CREATING;
	rule->resize_info->state = RULE_RESIZE_STATE_WRITING;
}

bool rule_move_in_progress(struct mlx5dv_hws_rule *rule)
{
	return matcher_is_in_resize(rule->matcher) &&
	       rule->resize_info &&
	       rule->resize_info->state != RULE_RESIZE_STATE_IDLE;
}

static int rule_create_hws_fw_wqe(struct mlx5dv_hws_rule *rule,
				  struct mlx5dv_hws_rule_attr *attr,
				  uint8_t mt_idx,
				  struct mlx5dv_hws_item_data item_data[],
				  uint8_t at_idx,
				  struct mlx5dv_hws_action_data actions_data[],
				  struct mlx5dv_hws_action *dest_action)
{
	struct mlx5dv_hws_mt *mt = &rule->matcher->mt[mt_idx];
	struct mlx5dv_hws_at *at = &rule->matcher->at[at_idx];
	struct mlx5dv_hws_matcher *matcher = rule->matcher;
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	bool is_compare = matcher_mt_is_compare(mt);
	struct send_ring_dep_wqe second_wqe = {0};
	struct send_ring_dep_wqe first_wqe = {0};
	bool is_range = matcher_mt_is_range(mt);
	bool is_jumbo = matcher_mt_is_jumbo(mt);
	struct send_ste_attr ste_attr = {0};
	struct actions_apply_data apply;
	struct send_engine *queue;

	queue = &ctx->send_queue[attr->queue_id];
	if (unlikely(send_engine_err(queue))) {
		errno = EIO;
		return -errno;
	}

	rule_create_init(rule, &ste_attr, &apply, false);
	rule_init_dep_wqe(&first_wqe, rule, attr, false);
	rule_init_dep_wqe(&second_wqe, rule, attr, false);

	ste_attr.direct_index = 0;

	ste_attr.send_attr.rule = first_wqe.rule;
	ste_attr.send_attr.user_data = first_wqe.user_data;
	ste_attr.send_attr.id = first_wqe.rtc;
	ste_attr.send_attr.used_id = &rule->rtc;
	ste_attr.send_attr.retry_id = first_wqe.retry_rtc;

	ste_attr.send_attr.fence = 1;
	ste_attr.send_attr.notify_hw = 1;
	ste_attr.wqe_tag_is_jumbo = is_jumbo;

	/* Prepare match STE TAG */
	ste_attr.wqe_ctrl = &first_wqe.wqe_ctrl;
	ste_attr.wqe_data = &first_wqe.wqe_data;
	ste_attr.send_attr.match_definer_id = definer_get_id(mt->definer);

	definer_create_tag(item_data,
			   mt->fcm,
			   mt->fcm_sz,
			  (uint8_t *)first_wqe.wqe_data.action);

	/* Prepare range/compare STE TAG */
	if (is_range) {
		ste_attr.scnd_wqe_data = &second_wqe.wqe_data;
		ste_attr.send_attr.len += SEND_WQE_SZ_GTA_DATA;
		ste_attr.send_attr.scnd_definer_id = definer_get_id(mt->range_definer);

		definer_create_tag_range(item_data,
					 mt->fcr,
					 mt->fcr_sz,
					 (uint8_t *)second_wqe.wqe_data.action);
	} else if (is_compare) {
		ste_attr.scnd_wqe_data = &second_wqe.wqe_data;
		ste_attr.send_attr.len += SEND_WQE_SZ_GTA_DATA;
		ste_attr.send_attr.scnd_definer_id = definer_get_id(mt->compare_definer);

		definer_create_tag_compare(item_data,
					   mt->fcc,
					   mt->fcc_sz,
					   (uint8_t *)second_wqe.wqe_data.action);
	}

	/* Apply the actions on the last STE */
	apply.queue = queue;
	apply.next_direct_idx = 0;
	apply.action_data = actions_data;
	apply.dest_action = dest_action;
	apply.wqe_ctrl = &first_wqe.wqe_ctrl;
	apply.wqe_data = (__be32 *)(is_range || is_compare ?
				    &second_wqe.wqe_data :
				    &first_wqe.wqe_data);

	apply.at = at;

	/* Skip setters[0] used for jumbo STE since not support with FW WQE */
	action_apply_setter(&apply, &at->setters[1], false);

	/* Send WQEs to FW */
	send_stes_fw(queue, &ste_attr);

	/* Backup TAG on the rule for deletion */
	rule_save_delete_info(rule, &ste_attr);
	send_engine_inc_entry(queue);

	/* Send dependent WQEs */
	if (!attr->burst)
		send_all_dep_wqe(queue);

	return 0;
}

static int rule_create_hws(struct mlx5dv_hws_rule *rule,
			   struct mlx5dv_hws_rule_attr *attr,
			   uint8_t mt_idx,
			   struct mlx5dv_hws_item_data item_data[],
			   uint8_t at_idx,
			   struct mlx5dv_hws_action_data actions_data[],
			   struct mlx5dv_hws_action *dest_action,
			   bool is_update)
{
	struct mlx5dv_hws_at *at = &rule->matcher->at[at_idx];
	struct mlx5dv_hws_mt *mt = &rule->matcher->mt[mt_idx];
	struct mlx5dv_hws_matcher *matcher = rule->matcher;
	struct mlx5dv_hws_context *ctx = matcher->tbl->ctx;
	bool is_jumbo = matcher_mt_is_jumbo(mt);
	struct send_ste_attr ste_attr = {0};
	struct actions_wqe_setter *setter;
	struct send_ring_dep_wqe *dep_wqe;
	struct actions_apply_data apply;
	uint8_t total_stes, action_stes;
	struct send_engine *queue;
	int i, ret;

	/* Insert rule using FW WQE if cannot use GTA WQE */
	if (unlikely(matcher_req_fw_wqe(matcher) && !is_update))
		return rule_create_hws_fw_wqe(rule, attr, mt_idx, item_data,
					      at_idx, actions_data, dest_action);

	queue = &ctx->send_queue[attr->queue_id];
	if (unlikely(send_engine_err(queue))) {
		errno = EIO;
		return -errno;
	}

	rule_create_init(rule, &ste_attr, &apply, is_update);

	/* Allocate dependent match WQE since rule might have dependent writes.
	 * The queued dependent WQE can be later aborted or kept as a dependency.
	 * dep_wqe buffers (ctrl, data) are also reused for all STE writes.
	 */
	dep_wqe = send_add_new_dep_wqe(queue);
	rule_init_dep_wqe(dep_wqe, rule, attr, is_update);

	ste_attr.wqe_ctrl = &dep_wqe->wqe_ctrl;
	ste_attr.wqe_data = &dep_wqe->wqe_data;
	apply.wqe_ctrl = &dep_wqe->wqe_ctrl;
	apply.wqe_data = (__be32 *)&dep_wqe->wqe_data;
	apply.action_data = actions_data;
	apply.dest_action = dest_action;
	apply.queue = queue;
	apply.at = at;

	setter = &at->setters[at->num_of_action_stes];
	total_stes = at->num_of_action_stes + (is_jumbo && !at->only_term);
	action_stes = total_stes - 1;

	if (action_stes) {
		/* Allocate action STEs for complex rules */
		ret = rule_alloc_action_ste(rule, attr);
		if (ret) {
			HWS_ERR("Failed to allocate action memory %d", ret);
			send_abort_new_dep_wqe(queue);
			return ret;
		}
		/* Action STEs are written to a specific index last to first */
		ste_attr.direct_index = rule->action_ste_idx + action_stes;
		ste_attr.send_attr.id = matcher->action_ste.rtc->object_id;

		apply.next_direct_idx = ste_attr.direct_index;
	} else {
		apply.next_direct_idx = 0;
	}

	for (i = total_stes; i-- > 0;) {
		action_apply_setter(&apply, setter--, !i && is_jumbo);

		if (i == 0) {
			/* Handle last match STE.
			 * For hash split / linear lookup RTCs, packets reaching any STE
			 * will always match and perform the specified actions, which
			 * makes the tag irrelevant.
			 */
			if (likely(!matcher_is_always_hit(matcher) && !is_update))
				definer_create_tag(item_data, mt->fcm, mt->fcm_sz,
						   (uint8_t *)dep_wqe->wqe_data.action);
			else if (unlikely(is_update))
				rule_update_copy_tag(rule, &dep_wqe->wqe_data, is_jumbo);

			/* Rule has dependent WQEs, match dep_wqe is queued */
			if (action_stes || apply.require_dep)
				break;

			/* Rule has no dependencies, abort dep_wqe and send WQE now */
			send_abort_new_dep_wqe(queue);
			ste_attr.wqe_tag_is_jumbo = is_jumbo;
			ste_attr.send_attr.notify_hw = !attr->burst;
			ste_attr.send_attr.user_data = dep_wqe->user_data;
			ste_attr.send_attr.rule = dep_wqe->rule;
			ste_attr.send_attr.id = dep_wqe->rtc;
			ste_attr.send_attr.used_id = &rule->rtc;
			ste_attr.send_attr.retry_id = dep_wqe->retry_rtc;
			ste_attr.direct_index = dep_wqe->direct_index;
		} else {
			apply.next_direct_idx = --ste_attr.direct_index;
		}

		send_ste(queue, &ste_attr);
	}

	/* Backup TAG on the rule for deletion and resize info for
	 * moving rules to a new matcher, only after insertion.
	 */
	if (!is_update) {
		rule_save_delete_info(rule, &ste_attr);
		rule_save_resize_info(rule, &ste_attr);
	}

	send_engine_inc_entry(queue);

	/* Send dependent WQEs */
	if (!attr->burst)
		send_all_dep_wqe(queue);

	return 0;
}

void rule_destroy_failed_hws(struct mlx5dv_hws_rule *rule,
			     struct mlx5dv_hws_rule_attr *attr)
{
	struct mlx5dv_hws_context *ctx = rule->matcher->tbl->ctx;
	struct send_engine *queue;

	queue = &ctx->send_queue[attr->queue_id];

	rule_gen_comp(queue, rule, attr->user_data,
		      MLX5DV_HWS_QUEUE_OP_SUCCESS,
		      RULE_STATUS_DELETED);

	/* Rule failed now we can safely release action STEs */
	rule_free_action_ste_idx(rule);

	/* Clear complex tag */
	rule_clear_delete_info(rule);

	/* Clear info that was saved for resizing */
	rule_clear_resize_info(rule);

	/* If a rule that was indicated as burst (need to trigger HW) has failed
	 * insertion we won't ring the HW as nothing is being written to the WQ.
	 * In such case update the last WQE and ring the HW with that work
	 */
	if (attr->burst)
		return;

	send_all_dep_wqe(queue);
	send_engine_flush_queue(queue);
}

static int rule_destroy_hws(struct mlx5dv_hws_rule *rule,
			    struct mlx5dv_hws_rule_attr *attr)
{
	struct mlx5dv_hws_context *ctx = rule->matcher->tbl->ctx;
	struct mlx5dv_hws_matcher *matcher = rule->matcher;
	bool is_compare = matcher_mt_is_compare(matcher->mt);
	bool is_range = matcher_mt_is_range(matcher->mt);
	bool is_jumbo = matcher_mt_is_jumbo(matcher->mt);
	struct send_wqe_gta_ctrl_seg wqe_ctrl = {0};
	bool fw_wqe = matcher_req_fw_wqe(matcher);
	struct send_ste_attr ste_attr = {0};
	struct send_engine *queue;

	queue = &ctx->send_queue[attr->queue_id];

	if (unlikely(send_engine_err(queue))) {
		rule_destroy_failed_hws(rule, attr);
		return 0;
	}

	/* Rule failed and doesn't require cleanup */
	if (rule->status == RULE_STATUS_FAILED) {
		rule_destroy_failed_hws(rule, attr);
		return 0;
	}

	/* Rule is not completed yet */
	if (rule->status != RULE_STATUS_CREATED &&
	    rule->status != RULE_STATUS_FAILING) {
		errno = EBUSY;
		return -errno;
	}

	send_engine_inc_entry(queue);

	/* Send dependent WQE */
	if (!attr->burst)
		send_all_dep_wqe(queue);

	rule->status = RULE_STATUS_DELETING;

	ste_attr.send_attr.opmod = SEND_WQE_GTA_OPMOD_STE;
	ste_attr.send_attr.opcode = SEND_WQE_OPCODE_TBL_ACCESS;
	ste_attr.send_attr.len = SEND_WQE_SZ_GTA_CTRL + SEND_WQE_SZ_GTA_DATA;
	if (unlikely(is_range || is_compare))
		ste_attr.send_attr.len += SEND_WQE_SZ_GTA_DATA;

	ste_attr.send_attr.rule = rule;
	ste_attr.send_attr.notify_hw = !attr->burst;
	ste_attr.send_attr.user_data = attr->user_data;
	ste_attr.send_attr.id = rule->rtc;
	ste_attr.send_attr.used_id = &rule->rtc;

	ste_attr.wqe_ctrl = &wqe_ctrl;
	ste_attr.wqe_tag_is_jumbo = is_jumbo;
	ste_attr.gta_opcode = SEND_WQE_GTA_OP_DEACTIVATE;
	if (unlikely(matcher_is_insert_by_idx(matcher)))
		ste_attr.direct_index = attr->rule_idx;

	rule_load_delete_info(rule, &ste_attr);

	if (unlikely(fw_wqe))
		send_stes_fw(queue, &ste_attr);
	else
		send_ste(queue, &ste_attr);

	rule_clear_delete_info(rule);

	return 0;
}

int rule_create_root_no_comp(struct mlx5dv_hws_rule *rule,
			     struct mlx5dv_hws_item_data item_data[],
			     struct mlx5dv_hws_action *dest_action,
			     uint8_t at_idx,
			     struct mlx5dv_hws_action_data actions_data[])
{
	struct mlx5dv_flow_matcher *dv_matcher = rule->matcher->dv_matcher;
	uint8_t num_actions = rule->matcher->at[at_idx].num_actions;
	struct mlx5dv_hws_mt *mt = rule->matcher->mt;
	struct mlx5dv_flow_action_attr *action_attr;
	struct mlx5dv_flow_match_parameters *value;
	int ret;

	/* Num of actions + dest_action */
	action_attr = calloc(num_actions + 1, sizeof(*action_attr));
	if (!action_attr) {
		errno = ENOMEM;
		return -errno;
	}

	value = calloc(1, DEVX_ST_SZ_BYTES(dr_match_param) +
			  offsetof(struct mlx5dv_flow_match_parameters, match_buf));
	if (!value) {
		errno = ENOMEM;
		goto free_attr;
	}

	/* Set item data to match param format */
	definer_create_tag(item_data,
			   mt->fcm, mt->fcm_sz,
			   (uint8_t *)value->match_buf);

	/* Set dependent fields as requested by FW */
	definer_set_dependent_fields_root(mt->fcm, mt->fcm_sz,
				     (uint8_t *)value->match_buf, false);

	/* Convert actions to verb action attr */
	ret = action_root_build_attr(actions_data,
				     &rule->matcher->at[at_idx],
				     dest_action,
				     action_attr);
	if (ret)
		goto free_value;

	/* Create verb flow */
	value->match_sz = DEVX_ST_SZ_BYTES(dr_match_param);
	/* Num_actions + dest_action */
	rule->flow = mlx5dv_create_flow(dv_matcher, value, num_actions + 1, action_attr);

	free(value);
	free(action_attr);

	return 0;

free_value:
	free(value);
free_attr:
	free(action_attr);
	return -errno;
}

static int rule_create_root(struct mlx5dv_hws_rule *rule,
			    struct mlx5dv_hws_rule_attr *rule_attr,
			    struct mlx5dv_hws_item_data item_data[],
			    uint8_t at_idx,
			    struct mlx5dv_hws_action_data actions_data[],
			    struct mlx5dv_hws_action *dest_action)
{
	struct mlx5dv_hws_context *ctx = rule->matcher->tbl->ctx;
	int ret;

	ret = rule_create_root_no_comp(rule,
				       item_data,
				       dest_action,
				       at_idx,
				       actions_data);
	if (ret)
		return -errno;

	rule_gen_comp(&ctx->send_queue[rule_attr->queue_id],
		      rule,
		      rule_attr->user_data,
		      rule->flow ? MLX5DV_HWS_QUEUE_OP_SUCCESS :
				   MLX5DV_HWS_QUEUE_OP_ERROR,
		      rule->flow ? RULE_STATUS_CREATED : RULE_STATUS_FAILED);

	return 0;
}

int rule_destroy_root_no_comp(struct mlx5dv_hws_rule *rule)
{
	if (rule->flow)
		return ibv_destroy_flow(rule->flow);

	return 0;
}

static int rule_destroy_root(struct mlx5dv_hws_rule *rule,
			     struct mlx5dv_hws_rule_attr *attr)
{
	struct mlx5dv_hws_context *ctx = rule->matcher->tbl->ctx;
	int err;

	err = rule_destroy_root_no_comp(rule);

	rule_gen_comp(&ctx->send_queue[attr->queue_id],
		      rule,
		      attr->user_data,
		      !err ? MLX5DV_HWS_QUEUE_OP_SUCCESS :
			     MLX5DV_HWS_QUEUE_OP_ERROR,
		      !err ? RULE_STATUS_DELETED : RULE_STATUS_FAILED);

	return 0;
}

static int rule_enqueue_precheck(struct mlx5dv_hws_rule *rule,
				 struct mlx5dv_hws_rule_attr *attr)
{
	struct mlx5dv_hws_context *ctx = rule->matcher->tbl->ctx;

	if (unlikely(!attr->user_data)) {
		errno = EINVAL;
		return -errno;
	}

	/* Check if there is room in queue */
	if (unlikely(send_engine_full(&ctx->send_queue[attr->queue_id]))) {
		errno = EBUSY;
		return -errno;
	}

	return 0;
}

static int rule_enqueue_precheck_move(struct mlx5dv_hws_rule *rule,
				      struct mlx5dv_hws_rule_attr *attr)
{
	if (unlikely(rule->status != RULE_STATUS_CREATED)) {
		errno = EINVAL;
		return -errno;
	}

	return rule_enqueue_precheck(rule, attr);
}

static int rule_enqueue_precheck_create(struct mlx5dv_hws_rule *rule,
					struct mlx5dv_hws_rule_attr *attr)
{
	if (unlikely(matcher_is_in_resize(rule->matcher))) {
		/* Matcher in resize - new rules are not allowed */
		errno = EAGAIN;
		return -errno;
	}

	return rule_enqueue_precheck(rule, attr);
}

static int rule_enqueue_precheck_update(struct mlx5dv_hws_rule *rule,
					struct mlx5dv_hws_rule_attr *attr)
{
	struct mlx5dv_hws_matcher *matcher = rule->matcher;

	if (unlikely(table_is_root(matcher->tbl) ||
		     matcher_req_fw_wqe(matcher))) {
		HWS_ERR("Rule update is not supported on current matcher");
		errno = ENOTSUP;
		return -errno;
	}

	if (unlikely(!matcher->attr.optimize_using_rule_idx &&
		     !matcher_is_insert_by_idx(matcher))) {
		HWS_ERR("Rule update requires optimize by idx matcher");
		errno = ENOTSUP;
		return -errno;
	}

	if (unlikely(matcher_is_resizable(rule->matcher))) {
		HWS_ERR("Rule update is not supported on resizable matcher");
		errno = ENOTSUP;
		return -errno;
	}

	if (unlikely(rule->status != RULE_STATUS_CREATED)) {
		HWS_ERR("Current rule status does not allow update");
		errno = EBUSY;
		return -errno;
	}

	return rule_enqueue_precheck_create(rule, attr);
}

int rule_move_hws_remove(struct mlx5dv_hws_rule *rule,
			 void *queue_ptr,
			 void *user_data)
{
	bool is_jumbo = matcher_mt_is_jumbo(rule->matcher->mt);
	struct mlx5dv_hws_matcher *matcher = rule->matcher;
	struct send_wqe_gta_ctrl_seg empty_wqe_ctrl = {0};
	struct send_engine *queue = queue_ptr;
	struct send_ste_attr ste_attr = {0};

	/* Send dependent WQEs */
	send_all_dep_wqe(queue);

	rule->resize_info->state = RULE_RESIZE_STATE_DELETING;

	ste_attr.send_attr.fence = 0;
	ste_attr.send_attr.opmod = SEND_WQE_GTA_OPMOD_STE;
	ste_attr.send_attr.opcode = SEND_WQE_OPCODE_TBL_ACCESS;
	ste_attr.send_attr.len = SEND_WQE_SZ_GTA_CTRL + SEND_WQE_SZ_GTA_DATA;
	ste_attr.send_attr.rule = rule;
	ste_attr.send_attr.notify_hw = 1;
	ste_attr.send_attr.user_data = user_data;
	ste_attr.send_attr.id = rule->resize_info->rtc;
	ste_attr.send_attr.used_id = &rule->resize_info->rtc;
	ste_attr.wqe_ctrl = &empty_wqe_ctrl;
	ste_attr.wqe_tag_is_jumbo = is_jumbo;
	ste_attr.gta_opcode = SEND_WQE_GTA_OP_DEACTIVATE;

	if (unlikely(matcher_is_insert_by_idx(matcher)))
		ste_attr.direct_index = rule->resize_info->rule_idx;

	rule_load_delete_info(rule, &ste_attr);
	send_ste(queue, &ste_attr);

	return 0;
}

int rule_move_hws_add(struct mlx5dv_hws_rule *rule,
		      struct mlx5dv_hws_rule_attr *attr)
{
	struct mlx5dv_hws_context *ctx = rule->matcher->tbl->ctx;
	bool is_jumbo = matcher_mt_is_jumbo(rule->matcher->mt);
	struct mlx5dv_hws_matcher *matcher = rule->matcher;
	struct send_ste_attr ste_attr = {0};
	struct send_engine *queue;

	if (unlikely(rule_enqueue_precheck_move(rule, attr)))
		return -errno;

	queue = &ctx->send_queue[attr->queue_id];

	if (unlikely(send_engine_err(queue))) {
		errno = EIO;
		return -errno;
	}

	rule_move_init(rule, attr);
	rule_move_get_rtc(rule, &ste_attr);

	ste_attr.send_attr.opmod = SEND_WQE_GTA_OPMOD_STE;
	ste_attr.send_attr.opcode = SEND_WQE_OPCODE_TBL_ACCESS;
	ste_attr.send_attr.len = SEND_WQE_SZ_GTA_CTRL + SEND_WQE_SZ_GTA_DATA;
	ste_attr.gta_opcode = SEND_WQE_GTA_OP_ACTIVATE;
	ste_attr.wqe_tag_is_jumbo = is_jumbo;

	ste_attr.send_attr.rule = rule;
	ste_attr.send_attr.fence = 0;
	ste_attr.send_attr.notify_hw = !attr->burst;
	ste_attr.send_attr.user_data = attr->user_data;
	ste_attr.send_attr.used_id = &rule->rtc;

	ste_attr.wqe_ctrl = (struct send_wqe_gta_ctrl_seg *)rule->resize_info->ctrl_seg;
	ste_attr.wqe_data = (struct send_wqe_gta_data_seg_ste *)rule->resize_info->data_seg;
	ste_attr.direct_index = matcher_is_insert_by_idx(matcher) ?
				attr->rule_idx : 0;

	send_ste(queue, &ste_attr);
	send_engine_inc_entry(queue);

	/* Send dependent WQEs */
	if (!attr->burst)
		send_all_dep_wqe(queue);

	return 0;
}

int mlx5dv_hws_rule_create(struct mlx5dv_hws_matcher *matcher,
			   uint8_t mt_idx,
			   struct mlx5dv_hws_item_data item_data[],
			   uint8_t at_idx,
			   struct mlx5dv_hws_action_data actions_data[],
			   struct mlx5dv_hws_action *dest_action,
			   struct mlx5dv_hws_rule_attr *attr,
			   struct mlx5dv_hws_rule *rule_handle)
{
	int ret;

	if (attr->comp_mask) {
		HWS_ERR("Rule create invalid comp mask provided");
		errno = ENOTSUP;
		return -errno;
	}

	rule_handle->matcher = matcher;

	if (unlikely(rule_enqueue_precheck_create(rule_handle, attr)))
		return -errno;

	assert(matcher->num_of_mt >= mt_idx);
	assert(matcher->num_of_at >= at_idx);

	if (unlikely(table_is_root(matcher->tbl)))
		ret = rule_create_root(rule_handle,
				       attr,
				       item_data,
				       at_idx,
				       actions_data,
				       dest_action);
	else
		ret = rule_create_hws(rule_handle,
				      attr,
				      mt_idx,
				      item_data,
				      at_idx,
				      actions_data,
				      dest_action,
				      false);
	return ret;
}

int mlx5dv_hws_rule_destroy(struct mlx5dv_hws_rule *rule,
			    struct mlx5dv_hws_rule_attr *attr)
{
	int ret;

	if (attr->comp_mask) {
		HWS_ERR("Rule destroy invalid comp mask provided");
		errno = ENOTSUP;
		return -errno;
	}

	if (unlikely(rule_enqueue_precheck(rule, attr)))
		return -errno;

	if (unlikely(table_is_root(rule->matcher->tbl)))
		ret = rule_destroy_root(rule, attr);
	else
		ret = rule_destroy_hws(rule, attr);

	return ret;
}

int mlx5dv_hws_rule_action_data_update(struct mlx5dv_hws_rule *rule_handle,
				       uint8_t at_idx,
				       struct mlx5dv_hws_action_data actions_data[],
				       struct mlx5dv_hws_action *dest_action,
				       struct mlx5dv_hws_rule_attr *attr)
{
	if (attr->comp_mask) {
		HWS_ERR("Rule action update invalid comp mask provided");
		errno = ENOTSUP;
		return -errno;
	}

	if (unlikely(rule_enqueue_precheck_update(rule_handle, attr)))
		return -errno;

	return rule_create_hws(rule_handle,
			       attr,
			       0,
			       NULL,
			       at_idx,
			       actions_data,
			       dest_action,
			       true);
}

int mlx5dv_hws_rule_query(struct mlx5dv_hws_rule *rule,
			  struct mlx5dv_hws_rule_query_data *ret_query)
{
	struct mlx5dv_hws_matcher *matcher;

	if (!rule) {
		errno = EINVAL;
		return -errno;
	}

	if (ret_query->comp_mask) {
		HWS_ERR("Rule query invalid comp mask provided");
		errno = ENOTSUP;
		return -errno;
	}

	matcher = rule->matcher;

	if (matcher_req_fw_wqe(matcher)) {
		HWS_ERR("Not supported for rules requiring FW WQE");
		goto err_enotsup;
	}

	if (table_is_root(matcher->tbl)) {
		HWS_ERR("Not supported for root level");
		goto err_enotsup;
	}

	if (matcher_mt_is_jumbo(matcher->mt)) {
		ret_query->match_tag_sz = RULE_MATCH_TAG_SZ_JUMBO;
		memcpy(ret_query->match_tag, rule->tag.jumbo, RULE_MATCH_TAG_SZ_JUMBO);
	} else {
		ret_query->match_tag_sz = RULE_MATCH_TAG_SZ_MATCH;
		memcpy(ret_query->match_tag, rule->tag.match, RULE_MATCH_TAG_SZ_MATCH);
	}

	return 0;

err_enotsup:
	errno = ENOTSUP;
	return -errno;
}

int mlx5dv_hws_rule_hash_calculate(struct mlx5dv_hws_matcher *matcher,
				   struct mlx5dv_hws_item_data item_data[],
				   uint8_t mt_idx,
				   uint32_t *ret_hash_idx,
				   uint32_t *ret_hash_raw)
{
	struct rule_match_tag tag = {0};
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_hws_mt *mt;
	uint32_t raw_hash;

	if (!matcher) {
		errno = EINVAL;
		return -errno;
	}

	mt = &matcher->mt[mt_idx];
	ctx = matcher->tbl->ctx;

	if (matcher_req_fw_wqe(matcher)) {
		HWS_ERR("Not supported for rules requiring FW WQE");
		goto err_enotsup;
	}

	if (table_is_root(matcher->tbl)) {
		HWS_ERR("Not supported for root level");
		goto err_enotsup;
	}

	if (ctx->caps->flow_table_hash_type != MLX5_FLOW_TABLE_HASH_TYPE_CRC32) {
		HWS_ERR("Current HW hash calculation algorithm not supported");
		goto err_enotsup;
	}

	if (matcher->attr.distribute_mode ==
	    MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_LINEAR) {
		HWS_ERR("Matcher doesn't distribute packets by hash");
		goto err_enotsup;
	}

	definer_create_tag(item_data, mt->fcm, mt->fcm_sz, tag.jumbo);

	if (matcher_mt_is_jumbo(mt))
		raw_hash = crc32_calc(tag.jumbo, RULE_MATCH_TAG_SZ_JUMBO);
	else
		raw_hash = crc32_calc(tag.match, RULE_MATCH_TAG_SZ_MATCH);

	*ret_hash_raw = raw_hash;
	*ret_hash_idx = raw_hash & (BIT(matcher->attr.table.sz_row_log) - 1);

	return 0;

err_enotsup:
	errno = ENOTSUP;
	return -errno;
}
