/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include <util/mmio.h>
#include "hws_internal.h"

struct send_ring_dep_wqe *
send_add_new_dep_wqe(struct send_engine *queue)
{
	struct send_ring_sq *send_sq = &queue->send_ring->send_sq;
	unsigned int idx;

	idx = send_sq->head_dep_idx++ & (queue->num_entries - 1);

	memset(&send_sq->dep_wqe[idx].wqe_data.tag, 0, SEND_WQE_GTA_STE_SZ_MATCH_TAG);

	return &send_sq->dep_wqe[idx];
}

void send_abort_new_dep_wqe(struct send_engine *queue)
{
	queue->send_ring->send_sq.head_dep_idx--;
}

void send_all_dep_wqe(struct send_engine *queue)
{
	struct send_ring_sq *send_sq = &queue->send_ring->send_sq;
	struct send_ste_attr ste_attr = {0};
	struct send_ring_dep_wqe *dep_wqe;

	ste_attr.send_attr.opmod = SEND_WQE_GTA_OPMOD_STE;
	ste_attr.send_attr.opcode = SEND_WQE_OPCODE_TBL_ACCESS;
	ste_attr.send_attr.len = SEND_WQE_SZ_GTA_CTRL + SEND_WQE_SZ_GTA_DATA;
	ste_attr.gta_opcode = SEND_WQE_GTA_OP_ACTIVATE;

	/* Fence first from previous depend WQEs  */
	ste_attr.send_attr.fence = 1;

	while (send_sq->head_dep_idx != send_sq->tail_dep_idx) {
		dep_wqe = &send_sq->dep_wqe[send_sq->tail_dep_idx++ & (queue->num_entries - 1)];

		/* Notify HW on the last WQE */
		ste_attr.send_attr.notify_hw = (send_sq->tail_dep_idx == send_sq->head_dep_idx);
		ste_attr.send_attr.user_data = dep_wqe->user_data;
		ste_attr.send_attr.rule = dep_wqe->rule;
		ste_attr.send_attr.id = dep_wqe->rtc;
		ste_attr.send_attr.used_id = dep_wqe->used_id;
		ste_attr.send_attr.retry_id = dep_wqe->retry_rtc;

		ste_attr.wqe_ctrl = &dep_wqe->wqe_ctrl;
		ste_attr.wqe_data = &dep_wqe->wqe_data;
		ste_attr.direct_index = dep_wqe->direct_index;

		send_ste(queue, &ste_attr);

		/* Fencing is done only on the first WQE */
		ste_attr.send_attr.fence = 0;
	}
}

struct send_engine_post_ctrl
send_engine_post_start(struct send_engine *queue)
{
	struct send_engine_post_ctrl ctrl;

	ctrl.queue = queue;
	/* Currently only one send ring is supported */
	ctrl.send_ring = &queue->send_ring[0];
	ctrl.num_wqebbs = 0;

	return ctrl;
}

void send_engine_post_req_wqe(struct send_engine_post_ctrl *ctrl,
			      char **buf, size_t *len)
{
	struct send_ring_sq *send_sq = &ctrl->send_ring->send_sq;
	unsigned int idx;

	idx = (send_sq->cur_post + ctrl->num_wqebbs) & send_sq->buf_mask;

	*buf = send_sq->buf + (idx << MLX5_SEND_WQE_SHIFT);
	*len = MLX5_SEND_WQE_BB;

	if (!ctrl->num_wqebbs) {
		*buf += sizeof(struct send_wqe_ctrl_seg);
		*len -= sizeof(struct send_wqe_ctrl_seg);
	}

	ctrl->num_wqebbs++;
}

static void send_engine_post_ring(struct send_ring_sq *sq,
				  struct mlx5dv_devx_uar *uar,
				  struct send_wqe_ctrl_seg *wqe_ctrl)
{
	udma_to_device_barrier();
	sq->db[MLX5_SND_DBR] = htobe32(sq->cur_post);

	mmio_wc_start();
	mmio_write64_be((uint8_t *)uar->reg_addr, *(__be64 *)wqe_ctrl);
	mmio_flush_writes();
}

static void send_wqe_set_tag(struct send_wqe_gta_data_seg_ste *wqe_data,
			     struct rule_match_tag *tag,
			     bool is_jumbo)
{
	if (is_jumbo) {
		/* Clear previous possibly dirty control */
		memset(wqe_data, 0, SEND_WQE_GTA_STE_SZ_CTRL);
		memcpy(wqe_data->jumbo, tag->jumbo, SEND_WQE_GTA_STE_SZ_JUMBO_TAG);
	} else {
		/* Clear previous possibly dirty control and actions */
		memset(wqe_data, 0, SEND_WQE_GTA_STE_SZ_CTRL + SEND_WQE_GTA_STE_SZ_MATCH_ACTIONS);
		memcpy(wqe_data->tag, tag->match, SEND_WQE_GTA_STE_SZ_MATCH_TAG);
	}
}

void send_engine_post_end(struct send_engine_post_ctrl *ctrl,
			  struct send_engine_post_attr *attr)
{
	struct send_wqe_ctrl_seg *wqe_ctrl;
	struct send_ring_sq *sq;
	uint32_t flags = 0;
	unsigned int idx;

	sq = &ctrl->send_ring->send_sq;
	idx = sq->cur_post & sq->buf_mask;
	sq->last_idx = idx;

	wqe_ctrl = (void *)(sq->buf + (idx << MLX5_SEND_WQE_SHIFT));

	wqe_ctrl->opmod_idx_opcode =
			htobe32((attr->opmod << 24) |
				((sq->cur_post & 0xffff) << 8) |
				attr->opcode);
	wqe_ctrl->qpn_ds = htobe32((attr->len + sizeof(struct send_wqe_ctrl_seg)) / 16 | sq->sqn << 8);

	wqe_ctrl->imm = htobe32(attr->id);

	flags |= attr->notify_hw ? MLX5_WQE_CTRL_CQ_UPDATE : 0;
	flags |= attr->fence ? MLX5_WQE_CTRL_INITIATOR_SMALL_FENCE : 0;
	wqe_ctrl->flags = htobe32(flags);

	sq->wr_priv[idx].id = attr->id;
	sq->wr_priv[idx].retry_id = attr->retry_id;

	sq->wr_priv[idx].rule = attr->rule;
	sq->wr_priv[idx].user_data = attr->user_data;
	sq->wr_priv[idx].num_wqebbs = ctrl->num_wqebbs;

	if (attr->rule) {
		sq->wr_priv[idx].rule->pending_wqes++;
		sq->wr_priv[idx].used_id = attr->used_id;
	}

	sq->cur_post += ctrl->num_wqebbs;

	if (attr->notify_hw)
		send_engine_post_ring(sq, ctrl->queue->uar, wqe_ctrl);
}

static void send_wqe(struct send_engine *queue,
		     struct send_engine_post_attr *send_attr,
		     struct send_wqe_gta_ctrl_seg *send_wqe_ctrl,
		     void *send_wqe_data,
		     void *send_wqe_tag,
		     bool is_jumbo,
		     uint8_t gta_opcode,
		     uint32_t direct_index)
{
	struct send_wqe_gta_data_seg_ste *wqe_data;
	struct send_wqe_gta_ctrl_seg *wqe_ctrl;
	struct send_engine_post_ctrl ctrl;
	size_t wqe_len;

	ctrl = send_engine_post_start(queue);
	send_engine_post_req_wqe(&ctrl, (void *)&wqe_ctrl, &wqe_len);
	send_engine_post_req_wqe(&ctrl, (void *)&wqe_data, &wqe_len);

	wqe_ctrl->op_dirix = htobe32(gta_opcode << 28 | direct_index);
	memcpy(wqe_ctrl->stc_ix, send_wqe_ctrl->stc_ix, sizeof(send_wqe_ctrl->stc_ix));

	/* Make sure no bits left from any other WQE type here (like ASO ctrl)*/
	memset(wqe_ctrl->rsvd0, 0, sizeof(send_wqe_ctrl->rsvd0));

	if (send_wqe_data)
		memcpy(wqe_data, send_wqe_data, sizeof(*wqe_data));
	else
		send_wqe_set_tag(wqe_data, send_wqe_tag, is_jumbo);

	send_engine_post_end(&ctrl, send_attr);
}

void send_ste(struct send_engine *queue,
	      struct send_ste_attr *ste_attr)
{
	struct send_engine_post_attr *send_attr = &ste_attr->send_attr;

	send_wqe(queue, send_attr,
		 ste_attr->wqe_ctrl,
		 ste_attr->wqe_data,
		 ste_attr->wqe_tag,
		 ste_attr->wqe_tag_is_jumbo,
		 ste_attr->gta_opcode,
		 ste_attr->direct_index);
}

static int send_wqe_fw(struct ibv_context *ibv_ctx,
		       uint32_t pd_num,
		       struct send_engine_post_attr *send_attr,
		       struct send_wqe_gta_ctrl_seg *send_wqe_ctrl,
		       void *send_wqe_match_data,
		       void *send_wqe_match_tag,
		       void *send_wqe_scnd_data,
		       void *send_wqe_scnd_tag,
		       bool is_jumbo,
		       uint8_t gta_opcode)
{
	bool has_match = send_wqe_match_data || send_wqe_match_tag;
	bool has_scnd = send_wqe_scnd_data || send_wqe_scnd_tag;
	struct send_wqe_gta_data_seg_ste gta_wqe_data0 = {0};
	struct send_wqe_gta_data_seg_ste gta_wqe_data1 = {0};
	struct send_wqe_gta_ctrl_seg gta_wqe_ctrl = {0};
	struct cmd_generate_wqe_attr attr = {0};
	struct send_wqe_ctrl_seg wqe_ctrl = {0};
	struct mlx5_cqe64 cqe;
	uint32_t flags = 0;
	int ret;

	/* Set WQE control */
	wqe_ctrl.opmod_idx_opcode = htobe32((send_attr->opmod << 24) | send_attr->opcode);
	wqe_ctrl.qpn_ds = htobe32((send_attr->len + sizeof(struct send_wqe_ctrl_seg)) / 16);
	flags |= send_attr->notify_hw ? MLX5_WQE_CTRL_CQ_UPDATE : 0;
	wqe_ctrl.flags = htobe32(flags);
	wqe_ctrl.imm = htobe32(send_attr->id);

	/* Set GTA WQE CTRL */
	memcpy(gta_wqe_ctrl.stc_ix, send_wqe_ctrl->stc_ix, sizeof(send_wqe_ctrl->stc_ix));
	gta_wqe_ctrl.op_dirix = htobe32(gta_opcode << 28);

	/* Set GTA match WQE DATA */
	if (has_match) {
		if (send_wqe_match_data)
			memcpy(&gta_wqe_data0, send_wqe_match_data, sizeof(gta_wqe_data0));
		else
			send_wqe_set_tag(&gta_wqe_data0, send_wqe_match_tag, is_jumbo);

		gta_wqe_data0.rsvd1_definer = htobe32(send_attr->match_definer_id << 8);
		attr.gta_data_0 = (uint8_t *)&gta_wqe_data0;
	}

	/* Set GTA range/compare WQE DATA */
	if (has_scnd) {
		if (send_wqe_scnd_data)
			memcpy(&gta_wqe_data1, send_wqe_scnd_data, sizeof(gta_wqe_data1));
		else
			send_wqe_set_tag(&gta_wqe_data1, send_wqe_scnd_tag, false);

		gta_wqe_data1.rsvd1_definer = htobe32(send_attr->scnd_definer_id << 8);
		attr.gta_data_1 = (uint8_t *)&gta_wqe_data1;
	}

	attr.pdn = pd_num;
	attr.wqe_ctrl = (uint8_t *)&wqe_ctrl;
	attr.gta_ctrl = (uint8_t *)&gta_wqe_ctrl;

send_wqe:
	ret = cmd_generate_wqe(ibv_ctx, &attr, &cqe);
	if (ret) {
		HWS_ERR("Failed to write WQE using command");
		return ret;
	}

	if ((mlx5dv_get_cqe_opcode(&cqe) == MLX5_CQE_REQ) &&
	    (be32toh(cqe.byte_cnt) >> 31 == 0)) {
		*send_attr->used_id = send_attr->id;
		return 0;
	}

	/* Retry if rule failed */
	if (send_attr->retry_id) {
		wqe_ctrl.imm = htobe32(send_attr->retry_id);
		send_attr->id = send_attr->retry_id;
		send_attr->retry_id = 0;
		goto send_wqe;
	}

	return -1;
}

void send_stes_fw(struct send_engine *queue,
		  struct send_ste_attr *ste_attr)
{
	struct send_engine_post_attr *send_attr = &ste_attr->send_attr;
	struct mlx5dv_hws_rule *rule = send_attr->rule;
	struct mlx5dv_hws_context *ctx;
	struct ibv_context *ibv_ctx;
	uint16_t queue_id;
	uint32_t pdn;
	int ret;

	ctx = rule->matcher->tbl->ctx;
	queue_id = queue - ctx->send_queue;
	ibv_ctx = ctx->ibv_ctx;
	pdn = ctx->pd_num;

	/* Writing through FW can't HW fence, therefore we drain the queue */
	if (send_attr->fence)
		mlx5dv_hws_queue_execute_op(ctx,
					    queue_id,
					    MLX5DV_HWS_QUEUE_OP_DRAIN_SYNC);

	ret = send_wqe_fw(ibv_ctx, pdn, send_attr,
			  ste_attr->wqe_ctrl,
			  ste_attr->wqe_data,
			  ste_attr->wqe_tag,
			  ste_attr->scnd_wqe_data,
			  ste_attr->scnd_wqe_tag,
			  ste_attr->wqe_tag_is_jumbo,
			  ste_attr->gta_opcode);
	if (ret)
		goto fail_rule;

	/* Increase the status, this only works on good flow as the enum
	 * is arrange it away creating -> created -> deleting -> deleted
	 */
	rule->status++;
	send_engine_gen_comp(queue, send_attr->user_data, MLX5DV_HWS_QUEUE_OP_SUCCESS);
	return;

fail_rule:
	rule->status = RULE_STATUS_FAILED;
	send_engine_gen_comp(queue, send_attr->user_data, MLX5DV_HWS_QUEUE_OP_ERROR);
}

static void send_engine_retry_post_send(struct send_engine *queue,
					struct send_ring_priv *priv,
					uint16_t wqe_cnt)
{
	struct send_engine_post_attr send_attr = {0};
	struct send_wqe_gta_data_seg_ste *wqe_data;
	struct send_wqe_gta_ctrl_seg *wqe_ctrl;
	struct send_engine_post_ctrl ctrl;
	struct send_ring_sq *send_sq;
	unsigned int idx;
	size_t wqe_len;
	char *p;

	send_attr.rule = priv->rule;
	send_attr.opcode = SEND_WQE_OPCODE_TBL_ACCESS;
	send_attr.opmod = SEND_WQE_GTA_OPMOD_STE;
	send_attr.len = MLX5_SEND_WQE_BB * 2 - sizeof(struct send_wqe_ctrl_seg);
	send_attr.notify_hw = 1;
	send_attr.fence = 0;
	send_attr.user_data = priv->user_data;
	send_attr.id = priv->retry_id;
	send_attr.used_id = priv->used_id;

	ctrl = send_engine_post_start(queue);
	send_engine_post_req_wqe(&ctrl, (void *)&wqe_ctrl, &wqe_len);
	send_engine_post_req_wqe(&ctrl, (void *)&wqe_data, &wqe_len);

	send_sq = &ctrl.send_ring->send_sq;
	idx = wqe_cnt & send_sq->buf_mask;
	p = send_sq->buf + (idx << MLX5_SEND_WQE_SHIFT);

	/* Copy old gta ctrl */
	memcpy(wqe_ctrl, p + sizeof(struct send_wqe_ctrl_seg),
	       MLX5_SEND_WQE_BB - sizeof(struct send_wqe_ctrl_seg));

	idx = (wqe_cnt + 1) & send_sq->buf_mask;
	p = send_sq->buf + (idx << MLX5_SEND_WQE_SHIFT);

	/* Copy old gta data */
	memcpy(wqe_data, p, MLX5_SEND_WQE_BB);

	send_engine_post_end(&ctrl, &send_attr);
}

void send_engine_flush_queue(struct send_engine *queue)
{
	struct send_ring_sq *sq = &queue->send_ring[0].send_sq;
	struct send_wqe_ctrl_seg *wqe_ctrl;

	wqe_ctrl = (void *)(sq->buf + (sq->last_idx << MLX5_SEND_WQE_SHIFT));

	wqe_ctrl->flags |= htobe32(MLX5_WQE_CTRL_CQ_UPDATE);

	send_engine_post_ring(sq, queue->uar, wqe_ctrl);
}

static void send_engine_update_rule_resize(struct send_engine *queue,
					   struct send_ring_priv *priv,
					   enum mlx5dv_hws_queue_op_status *status)
{
	switch (priv->rule->resize_info->state) {
	case RULE_RESIZE_STATE_WRITING:
		if (priv->rule->status == RULE_STATUS_FAILED) {
			priv->rule->rtc = priv->rule->resize_info->rtc;
			priv->rule->resize_info->rtc = 0;
			priv->rule->resize_info->state = RULE_RESIZE_STATE_IDLE;
			priv->rule->status = RULE_STATUS_CREATED;
			return;
		}
		/* Remove old rule from src_matcher */
		rule_move_hws_remove(priv->rule, queue, priv->user_data);
		break;

	case RULE_RESIZE_STATE_DELETING:
		/* In case delete from src_matcher fail we cannot recover,
		 * prefer to leak rule in src_matcher than fail.
		 */
		if (unlikely(priv->rule->status == RULE_STATUS_FAILED))
			HWS_ERR("Rule move fail to remove rule from src_matcher");

		/* If rule delete from src_matcher failed we ignore */
		*status = MLX5DV_HWS_QUEUE_OP_SUCCESS;
		priv->rule->matcher = priv->rule->matcher->resize_dst;
		priv->rule->resize_info->state = RULE_RESIZE_STATE_IDLE;
		priv->rule->status = RULE_STATUS_CREATED;
		break;

	default:
		break;
	}
}

static void
send_engine_update_rule(struct send_engine *queue,
			struct send_ring_priv *priv,
			uint16_t wqe_cnt,
			enum mlx5dv_hws_queue_op_status *status)
{
	priv->rule->pending_wqes--;

	if (*status == MLX5DV_HWS_QUEUE_OP_ERROR) {
		if (priv->retry_id) {
			send_engine_retry_post_send(queue, priv, wqe_cnt);
			return;
		}

		/* Some part of the rule failed */
		priv->rule->status = RULE_STATUS_FAILING;

		*priv->used_id = 0;
	} else {
		*priv->used_id = priv->id;
	}

	/* Update rule status for the last completion */
	if (!priv->rule->pending_wqes) {
		if (unlikely(rule_move_in_progress(priv->rule))) {
			send_engine_update_rule_resize(queue, priv, status);
			return;
		}

		if (unlikely(priv->rule->status == RULE_STATUS_FAILING)) {
			/* Rule completely failed and doesn't require cleanup */
			if (!priv->rule->rtc && !priv->rule->rtc_ext)
				priv->rule->status = RULE_STATUS_FAILED;

			*status = MLX5DV_HWS_QUEUE_OP_ERROR;
		} else {
			/* Increase the status, this only works on good flow as the enum
			* is arrange it away creating -> created -> deleting -> deleted
			*/
			priv->rule->status++;
			*status = MLX5DV_HWS_QUEUE_OP_SUCCESS;
			/* Rule was deleted now we can safely release action STEs
				and clear resize info */
			if (priv->rule->status == RULE_STATUS_DELETED) {
				rule_free_action_ste_idx(priv->rule);
				rule_clear_resize_info(priv->rule);
			}
		}
	}
}

static void send_engine_update(struct send_engine *queue,
			       struct mlx5_cqe64 *cqe,
			       struct send_ring_priv *priv,
			       struct mlx5dv_hws_queue_op_result res[],
			       int64_t *i,
			       uint32_t res_nb,
			       uint16_t wqe_cnt)
{
	enum mlx5dv_hws_queue_op_status status;

	if (!cqe || (likely(be32toh(cqe->byte_cnt) >> 31 == 0) &&
	    likely(mlx5dv_get_cqe_opcode(cqe) == MLX5_CQE_REQ))) {
		status = MLX5DV_HWS_QUEUE_OP_SUCCESS;
	} else {
		status = MLX5DV_HWS_QUEUE_OP_ERROR;
	}

	if (priv->user_data) {
		if (priv->rule) {
			send_engine_update_rule(queue, priv, wqe_cnt, &status);
			if (priv->rule->pending_wqes)
				return;
		}

		if (*i < res_nb) {
			res[*i].user_data = priv->user_data;
			res[*i].status = status;
			(*i)++;
			send_engine_dec_entry(queue);
		} else {
			send_engine_gen_comp(queue, priv->user_data, status);
		}
	}
}

static void send_engine_poll_cq(struct send_engine *queue,
				struct send_ring *send_ring,
				struct mlx5dv_hws_queue_op_result res[],
				int64_t *i,
				uint32_t res_nb)
{
	struct send_ring_cq *cq = &send_ring->send_cq;
	struct send_ring_sq *sq = &send_ring->send_sq;
	struct send_ring_priv *priv;
	struct mlx5_cqe64 *cqe;
	uint32_t offset_cqe64;
	uint8_t cqe_opcode;
	uint8_t cqe_owner;
	uint16_t wqe_cnt;
	uint32_t cq_idx;
	uint8_t sw_own;

	cq_idx = cq->cons_index & cq->ncqe_mask;
	/* CQE size can be 64B or 128B */
	offset_cqe64 = (cq->cqe_sz == SEND_CQE_SZ_64) ? 0 : SEND_CQE_PAD_64;
	cqe = (void *)(cq->buf + (cq_idx << cq->cqe_log_sz) + offset_cqe64);

	sw_own = (cq->cons_index & cq->ncqe) ? 1 : 0;
	cqe_opcode = mlx5dv_get_cqe_opcode(cqe);
	cqe_owner = mlx5dv_get_cqe_owner(cqe);

	if (cqe_opcode == MLX5_CQE_INVALID ||
	    cqe_owner != sw_own)
		return;

	if (unlikely(cqe_opcode != MLX5_CQE_REQ)) {
		struct mlx5_err_cqe *err_cqe = (struct mlx5_err_cqe *)cqe;
		HWS_ERR("CQE ERR:0x%x, Ven_ERR:0x%x, HW-synd:0x%x, OP:0x%x, QPN:0x%x, WQE_CNT:0x%x",
			err_cqe->syndrome,
			err_cqe->vendor_err_synd,
			err_cqe->rsvd1[16], cqe_opcode,
			(be32toh((__force __be32)err_cqe->s_wqe_opcode_qpn) & 0xffffff),
			be16toh((__force __be16)err_cqe->wqe_counter));
		queue->err = true;
	}

	udma_from_device_barrier();

	wqe_cnt = be16toh(cqe->wqe_counter) & sq->buf_mask;

	while (cq->poll_wqe != wqe_cnt) {
		priv = &sq->wr_priv[cq->poll_wqe];
		send_engine_update(queue, NULL, priv, res, i, res_nb, 0);
		cq->poll_wqe = (cq->poll_wqe + priv->num_wqebbs) & sq->buf_mask;
	}

	priv = &sq->wr_priv[wqe_cnt];
	cq->poll_wqe = (wqe_cnt + priv->num_wqebbs) & sq->buf_mask;
	send_engine_update(queue, cqe, priv, res, i, res_nb, wqe_cnt);
	cq->cons_index++;
	*cq->db = htobe32(cq->cons_index & 0xffffff);
}

static void send_engine_poll_cqs(struct send_engine *queue,
				 struct mlx5dv_hws_queue_op_result res[],
				 int64_t *polled,
				 uint32_t res_nb)
{
	volatile int j; /* o3 optimization might fall here, so volatile*/

	for (j = 0; j < SEND_NUM_SEND_RINGS; j++)
		send_engine_poll_cq(queue, &queue->send_ring[j],
				    res, polled, res_nb);
}

static void send_engine_poll_list(struct send_engine *queue,
				  struct mlx5dv_hws_queue_op_result res[],
				  int64_t *polled,
				  uint32_t res_nb)
{
	struct send_completed_poll *comp = &queue->completed;

	while (comp->ci != comp->pi) {
		if (*polled < res_nb) {
			res[*polled].status =
				comp->entries[comp->ci].status;
			res[*polled].user_data =
				comp->entries[comp->ci].user_data;
			(*polled)++;
			comp->ci = (comp->ci + 1) & comp->mask;
			send_engine_dec_entry(queue);
		} else {
			return;
		}
	}
}

static int send_engine_poll(struct send_engine *queue,
			    struct mlx5dv_hws_queue_op_result res[],
			    uint32_t res_nb)
{
	int64_t polled = 0;

	send_engine_poll_list(queue, res, &polled, res_nb);

	if (polled >= res_nb)
		return polled;

	send_engine_poll_cqs(queue, res, &polled, res_nb);

	return polled;
}

int mlx5dv_hws_queue_poll(struct mlx5dv_hws_context *ctx,
			  uint16_t queue_id,
			  struct mlx5dv_hws_queue_op_result res[],
			  uint32_t res_nb)
{
	return send_engine_poll(&ctx->send_queue[queue_id], res, res_nb);
}

static int send_ring_create_sq_obj(struct mlx5dv_hws_context *ctx,
				   struct send_engine *queue,
				   struct send_ring_sq *sq,
				   struct send_ring_cq *cq,
				   size_t log_wq_sz)
{
	struct cmd_sq_create_attr attr = {0};
	int err;

	attr.cqn = cq->cqn;
	attr.pdn = ctx->pd_num;
	attr.page_id = queue->uar->page_id;
	attr.dbr_id = sq->db_umem->umem_id;
	attr.wq_id = sq->buf_umem->umem_id;
	attr.log_wq_sz = log_wq_sz;
	if (ctx->caps->sq_ts_format == MLX5_HCA_CAP_TIMESTAMP_FORMAT_FR)
		attr.ts_format = MLX5_QPC_TIMESTAMP_FORMAT_FREE_RUNNING;
	else
		attr.ts_format = MLX5_QPC_TIMESTAMP_FORMAT_DEFAULT;

	sq->obj = cmd_sq_create(ctx->ibv_ctx, &attr);
	if (!sq->obj)
		return -errno;

	sq->sqn = sq->obj->object_id;

	err = cmd_sq_modify_rdy(sq->obj);
	if (err)
		goto free_sq;

	return 0;

free_sq:
	cmd_destroy_obj(sq->obj);

	return err;
}

static int send_ring_open_sq(struct mlx5dv_hws_context *ctx,
			     struct send_engine *queue,
			     struct send_ring_sq *sq,
			     struct send_ring_cq *cq)
{
	size_t sq_log_buf_sz;
	size_t buf_aligned;
	size_t sq_buf_sz;
	size_t page_size;
	size_t buf_sz;
	int err;

	buf_sz = queue->num_entries * SEND_MAX_WQES_PER_RULE;
	sq_log_buf_sz = ilog32(buf_sz - 1);
	sq_buf_sz = 1 << (sq_log_buf_sz + ilog32(MLX5_SEND_WQE_BB - 1));

	page_size = sysconf(_SC_PAGESIZE);
	buf_aligned = align(sq_buf_sz, page_size);
	err = posix_memalign((void **)&sq->buf, page_size, buf_aligned);
	if (err) {
		errno = ENOMEM;
		return err;
	}
	memset(sq->buf, 0, buf_aligned);

	err = posix_memalign((void **)&sq->db, 8, 8);
	if (err)
		goto free_buf;

	sq->buf_umem = mlx5dv_devx_umem_reg(ctx->ibv_ctx, sq->buf, sq_buf_sz, 0);

	if (!sq->buf_umem) {
		err = -errno;
		goto free_db;
	}

	sq->db_umem = mlx5dv_devx_umem_reg(ctx->ibv_ctx, sq->db, 8, 0);
	if (!sq->db_umem) {
		err = -errno;
		goto free_buf_umem;
	}

	err = send_ring_create_sq_obj(ctx, queue, sq, cq, sq_log_buf_sz);

	if (err)
		goto free_db_umem;

	sq->wr_priv = malloc(sizeof(*sq->wr_priv) * buf_sz);
	if (!sq->wr_priv) {
		err = ENOMEM;
		goto destroy_sq_obj;
	}

	sq->dep_wqe = calloc(queue->num_entries, sizeof(*sq->dep_wqe));
	if (!sq->dep_wqe) {
		err = ENOMEM;
		goto destroy_wr_priv;
	}

	sq->buf_mask = buf_sz - 1;

	return 0;

destroy_wr_priv:
	free(sq->wr_priv);
destroy_sq_obj:
	cmd_destroy_obj(sq->obj);
free_db_umem:
	mlx5dv_devx_umem_dereg(sq->db_umem);
free_buf_umem:
	mlx5dv_devx_umem_dereg(sq->buf_umem);
free_db:
	free(sq->db);
free_buf:
	free(sq->buf);
	errno = -err;
	return err;
}

static void send_ring_close_sq(struct send_ring_sq *sq)
{
	free(sq->dep_wqe);
	cmd_destroy_obj(sq->obj);
	mlx5dv_devx_umem_dereg(sq->db_umem);
	mlx5dv_devx_umem_dereg(sq->buf_umem);
	free(sq->wr_priv);
	free(sq->db);
	free(sq->buf);
}

static int send_ring_open_cq(struct mlx5dv_hws_context *ctx,
			     struct send_engine *queue,
			     struct send_ring_cq *cq)
{
	struct mlx5dv_cq mlx5_cq = {0};
	struct mlx5dv_obj obj;
	struct ibv_cq *ibv_cq;
	size_t cq_size;
	int err;

	cq_size = queue->num_entries;
	ibv_cq = ibv_create_cq(ctx->ibv_ctx, cq_size, NULL, NULL, 0);
	if (!ibv_cq) {
		HWS_ERR("Failed to create CQ");
		return -errno;
	}

	obj.cq.in = ibv_cq;
	obj.cq.out = &mlx5_cq;
	err = mlx5dv_init_obj(&obj, MLX5DV_OBJ_CQ);
	if (err) {
		err = errno;
		goto close_cq;
	}

	cq->buf = mlx5_cq.buf;
	cq->db = mlx5_cq.dbrec;
	cq->ncqe = mlx5_cq.cqe_cnt;
	cq->cqe_sz = mlx5_cq.cqe_size;
	cq->cqe_log_sz = ilog32(cq->cqe_sz - 1);
	cq->ncqe_mask = cq->ncqe - 1;
	cq->buf_sz = cq->cqe_sz * cq->ncqe;
	cq->cqn = mlx5_cq.cqn;
	cq->ibv_cq = ibv_cq;

	return 0;

close_cq:
	ibv_destroy_cq(ibv_cq);
	errno = err;
	return -err;
}

static void send_ring_close_cq(struct send_ring_cq *cq)
{
	ibv_destroy_cq(cq->ibv_cq);
}

static void send_ring_close(struct send_ring *ring)
{
	send_ring_close_sq(&ring->send_sq);
	send_ring_close_cq(&ring->send_cq);
}

static int send_ring_open(struct mlx5dv_hws_context *ctx,
			  struct send_engine *queue,
			  struct send_ring *ring)
{
	int err;

	err = send_ring_open_cq(ctx, queue, &ring->send_cq);
	if (err)
		return err;

	err = send_ring_open_sq(ctx, queue, &ring->send_sq, &ring->send_cq);
	if (err)
		goto close_cq;

	return err;

close_cq:
	send_ring_close_cq(&ring->send_cq);

	return err;
}

static void __send_rings_close(struct send_engine *queue,
			       uint16_t i)
{
	while (i--)
		send_ring_close(&queue->send_ring[i]);
}

static void send_rings_close(struct send_engine *queue)
{
	__send_rings_close(queue, queue->rings);
}

static int send_rings_open(struct mlx5dv_hws_context *ctx,
			   struct send_engine *queue)
{
	uint16_t i;
	int err;

	for (i = 0; i < queue->rings; i++) {
		err = send_ring_open(ctx, queue, &queue->send_ring[i]);
		if (err)
			goto free_rings;
	}

	return 0;

free_rings:
	__send_rings_close(queue, i);

	return err;
}

void send_queue_close(struct send_engine *queue)
{
	send_rings_close(queue);
	free(queue->completed.entries);
	mlx5dv_devx_free_uar(queue->uar);
}

int send_queue_open(struct mlx5dv_hws_context *ctx,
		    struct send_engine *queue,
		    uint16_t queue_size)
{
	struct mlx5dv_devx_uar *uar;
	int err;

	uar = mlx5dv_devx_alloc_uar(ctx->ibv_ctx, MLX5_IB_UAPI_UAR_ALLOC_TYPE_NC);
	if (!uar)
		return -errno;

	queue->uar = uar;
	queue->rings = SEND_NUM_SEND_RINGS;
	queue->num_entries = 1 << ilog32(queue_size - 1);
	queue->used_entries = 0;

	queue->completed.entries = calloc(queue->num_entries, sizeof(queue->completed.entries[0]));
	if (!queue->completed.entries) {
		errno = ENOMEM;
		goto free_uar;
	}
	queue->completed.pi = 0;
	queue->completed.ci = 0;
	queue->completed.mask = queue->num_entries - 1;

	err = send_rings_open(ctx, queue);
	if (err)
		goto free_completed_entries;

	return 0;

free_completed_entries:
	free(queue->completed.entries);
free_uar:
	mlx5dv_devx_free_uar(uar);
	return -errno;
}

static void __send_queues_close(struct mlx5dv_hws_context *ctx,
				uint16_t queues)
{
	while (queues--)
		send_queue_close(&ctx->send_queue[queues]);
}

void send_queues_close(struct mlx5dv_hws_context *ctx)
{
	__send_queues_close(ctx, ctx->queues);
	free(ctx->send_queue);
}

int send_queues_open(struct mlx5dv_hws_context *ctx,
		     uint16_t queues,
		     uint16_t queue_size)
{
	int err = 0;
	uint32_t i;

	/* Open one extra queue for control path */
	ctx->queues = queues + 1;

	ctx->send_queue = calloc(ctx->queues, sizeof(*ctx->send_queue));
	if (!ctx->send_queue) {
		errno = ENOMEM;
		return -errno;
	}

	for (i = 0; i < ctx->queues; i++) {
		err = send_queue_open(ctx, &ctx->send_queue[i], queue_size);
		if (err)
			goto close_send_queues;
	}

	return 0;

close_send_queues:
	 __send_queues_close(ctx, i);

	free(ctx->send_queue);

	return err;
}

int mlx5dv_hws_queue_execute_op(struct mlx5dv_hws_context *ctx,
				uint16_t queue_id,
				uint32_t queue_op)
{
	struct send_ring_sq *send_sq;
	struct send_engine *queue;
	bool wait_comp = false;
	int64_t polled = 0;

	queue = &ctx->send_queue[queue_id];
	send_sq = &queue->send_ring->send_sq;

	switch (queue_op) {
	case MLX5DV_HWS_QUEUE_OP_DRAIN_SYNC:
		wait_comp = true;
		/* FALLTHROUGH */
	case MLX5DV_HWS_QUEUE_OP_DRAIN_ASYNC:
		if (send_sq->head_dep_idx != send_sq->tail_dep_idx)
			/* Send dependent WQEs to drain the queue */
			send_all_dep_wqe(queue);
		else
			/* Signal on the last posted WQE */
			send_engine_flush_queue(queue);

		/* Poll queue until empty */
		while (wait_comp && !send_engine_empty(queue))
			send_engine_poll_cqs(queue, NULL, &polled, 0);

		break;
	default:
		errno = EINVAL;
		return -errno;
	}

	return 0;
}

int mlx5dv_hws_queue_pending(struct mlx5dv_hws_context *ctx, uint16_t queue_id)
{
	return ctx->send_queue[queue_id].used_entries;
}
