/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#ifndef _HWS_SEND_H_
#define _HWS_SEND_H_

#define SEND_NUM_SEND_RINGS 1

/* As a single operation requires at least two WQEBBS.
 * This means a maximum of 16 such operations per rule.
 */
#define SEND_MAX_WQES_PER_RULE 32

#define SEND_CQE_SZ_64 64
#define SEND_CQE_PAD_64 64

enum send_wqe_opcode {
	SEND_WQE_OPCODE_TBL_ACCESS = 0x2c,
	SEND_WQE_OPCODE_ACCESS_ASO = 0x2d,
};

enum send_wqe_opmod {
	SEND_WQE_OPMOD_GTA_STE = 0,
	SEND_WQE_OPMOD_GTA_MOD_ARG = 1,
};

enum send_wqe_gta_opcode {
	SEND_WQE_GTA_OP_ACTIVATE = 0,
	SEND_WQE_GTA_OP_DEACTIVATE = 1,
};

enum send_wqe_gta_opmod {
	SEND_WQE_GTA_OPMOD_STE = 0,
	SEND_WQE_GTA_OPMOD_MOD_ARG = 1,
};

enum send_wqe_gta_sz {
	SEND_WQE_SZ_GTA_CTRL = 48,
	SEND_WQE_SZ_GTA_DATA = 64,
};

enum send_wqe_gta_ste_sz {
	SEND_WQE_GTA_STE_SZ_CTRL = 20,
	SEND_WQE_GTA_STE_SZ_MATCH_ACTIONS = 12,
	SEND_WQE_GTA_STE_SZ_MATCH_TAG = 32,
	SEND_WQE_GTA_STE_SZ_JUMBO_TAG = 44,
};

enum send_aso_opc_mod {
	SEND_ASO_OPC_MOD_IPSEC = 0x0,
	SEND_ASO_OPC_MOD_CONNECTION_TRACKING = 0x1,
	SEND_ASO_OPC_MOD_FLOW_METER = 0x2,
	SEND_ASO_OPC_MOD_RACE_AVOIDANCE = 0x3,
	SEND_ASO_OPC_MOD_FLOW_HIT = 0x4,
};

/* WQE Control segment. */
struct send_wqe_ctrl_seg {
	__be32 opmod_idx_opcode;
	__be32 qpn_ds;
	__be32 flags;
	__be32 imm;
};

struct send_wqe_aso_ctrl_seg {
	__be64 va_r;
	__be32 lkey;
	__be32 operand_masks;
	__be32 condition_0_data;
	__be32 condition_0_mask;
	__be32 condition_1_data;
	__be32 condition_1_mask;
	__be64 bitwise_data;
	__be64 data_mask;
};

struct send_wqe_aso {
	struct send_wqe_ctrl_seg general_cseg;
	struct send_wqe_aso_ctrl_seg aso_cseg;
	__be64 data[8];
};

struct send_wqe_gta_ctrl_seg {
	__be32 op_dirix;
	__be32 stc_ix[5];
	__be32 rsvd0[6];
};

struct send_wqe_gta_data_seg_ste {
	__be32 rsvd0_ctr_id;
	__be32 rsvd1_definer;
	__be32 rsvd2[3];
	union {
		struct {
			__be32 action[3];
			__be32 tag[8];
		};
		__be32 jumbo[11];
	};
};

struct send_wqe_gta_data_seg_arg {
	__be32 action_args[8];
};

struct send_wqe_gta {
	struct send_wqe_gta_ctrl_seg gta_ctrl;
	union {
		struct send_wqe_gta_data_seg_ste seg_ste;
		struct send_wqe_gta_data_seg_arg seg_arg;
	};
};

struct send_ring_cq {
	uint8_t *buf;
	uint32_t cons_index;
	uint32_t ncqe_mask;
	uint32_t buf_sz;
	uint32_t ncqe;
	uint32_t cqe_log_sz;
	__be32 *db;
	uint16_t poll_wqe;
	struct ibv_cq *ibv_cq;
	uint32_t cqn;
	uint32_t cqe_sz;
};

struct send_ring_priv {
	struct mlx5dv_hws_rule *rule;
	void *user_data;
	uint32_t num_wqebbs;
	uint32_t id;
	uint32_t retry_id;
	uint32_t *used_id;
};

struct send_ring_dep_wqe {
	struct send_wqe_gta_ctrl_seg wqe_ctrl;
	struct send_wqe_gta_data_seg_ste wqe_data;
	struct mlx5dv_hws_rule *rule;
	uint32_t rtc;
	uint32_t retry_rtc;
	uint32_t direct_index;
	void *user_data;
	uint32_t *used_id;
};

struct send_ring_sq {
	char *buf;
	uint32_t sqn;
	__be32 *db;
	uint16_t cur_post;
	uint16_t buf_mask;
	struct send_ring_priv *wr_priv;
	unsigned int last_idx;
	struct send_ring_dep_wqe *dep_wqe;
	unsigned int head_dep_idx;
	unsigned int tail_dep_idx;
	struct mlx5dv_devx_obj *obj;
	struct mlx5dv_devx_umem *buf_umem;
	struct mlx5dv_devx_umem *db_umem;
};

struct send_ring {
	struct send_ring_cq send_cq;
	struct send_ring_sq send_sq;
};

struct send_completed_poll_entry {
	void *user_data;
	enum mlx5dv_hws_queue_op_status status;
};

struct send_completed_poll {
	struct send_completed_poll_entry *entries;
	uint16_t ci;
	uint16_t pi;
	uint16_t mask;
};

struct send_engine {
	struct send_ring send_ring[SEND_NUM_SEND_RINGS]; /* For now 1:1 mapping */
	struct mlx5dv_devx_uar *uar; /* Uar is shared between rings of a queue */
	struct send_completed_poll completed;
	uint16_t used_entries;
	uint16_t rings;
	uint16_t num_entries;
	bool err;
} __attribute__((__aligned__(64)));//TODO

struct send_engine_post_ctrl {
	struct send_engine *queue;
	struct send_ring *send_ring;
	size_t num_wqebbs;
};

struct send_engine_post_attr {
	uint8_t opcode;
	uint8_t opmod;
	uint8_t notify_hw;
	uint8_t fence;
	uint8_t match_definer_id;
	uint8_t scnd_definer_id;
	size_t len;
	struct mlx5dv_hws_rule *rule;
	uint32_t id;
	uint32_t retry_id;
	uint32_t *used_id;
	void *user_data;
};

struct send_ste_attr {
	bool wqe_tag_is_jumbo;
	uint8_t gta_opcode;
	uint32_t direct_index;
	struct send_engine_post_attr send_attr;
	struct rule_match_tag *wqe_tag;
	struct rule_match_tag *scnd_wqe_tag;
	struct send_wqe_gta_ctrl_seg *wqe_ctrl;
	struct send_wqe_gta_data_seg_ste *wqe_data;
	struct send_wqe_gta_data_seg_ste *scnd_wqe_data;
};

struct send_ring_dep_wqe *
send_add_new_dep_wqe(struct send_engine *queue);

void send_abort_new_dep_wqe(struct send_engine *queue);

void send_all_dep_wqe(struct send_engine *queue);

void send_queue_close(struct send_engine *queue);

int send_queue_open(struct mlx5dv_hws_context *ctx,
		    struct send_engine *queue,
		    uint16_t queue_size);

void send_queues_close(struct mlx5dv_hws_context *ctx);

int send_queues_open(struct mlx5dv_hws_context *ctx,
		     uint16_t queues,
		     uint16_t queue_size);

struct send_engine_post_ctrl
send_engine_post_start(struct send_engine *queue);

void send_engine_post_req_wqe(struct send_engine_post_ctrl *ctrl,
			      char **buf, size_t *len);

void send_engine_post_end(struct send_engine_post_ctrl *ctrl,
			  struct send_engine_post_attr *attr);

void send_ste(struct send_engine *queue,
	      struct send_ste_attr *ste_attr);

void send_stes_fw(struct send_engine *queue,
		  struct send_ste_attr *ste_attr);

void send_engine_flush_queue(struct send_engine *queue);

static inline bool
send_engine_empty(struct send_engine *queue)
{
	struct send_ring_sq *send_sq = &queue->send_ring->send_sq;
	struct send_ring_cq *send_cq = &queue->send_ring->send_cq;

	return ((send_sq->cur_post & send_sq->buf_mask) == send_cq->poll_wqe);
}

static inline bool
send_engine_full(struct send_engine *queue)
{
	return queue->used_entries >= queue->num_entries;
}

static inline void
send_engine_inc_entry(struct send_engine *queue)
{
	queue->used_entries++;
}

static inline void
send_engine_dec_entry(struct send_engine *queue)
{
	queue->used_entries--;
}

static inline void
send_engine_gen_comp(struct send_engine *queue,
		     void *user_data,
		     int comp_status)
{
	struct send_completed_poll *comp = &queue->completed;

	comp->entries[comp->pi].status = comp_status;
	comp->entries[comp->pi].user_data = user_data;

	comp->pi = (comp->pi + 1) & comp->mask;
}

static inline bool
send_engine_err(struct send_engine *queue)
{
	return queue->err;
}

#endif /* _HWS_SEND_H_ */
