/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#include "hws_internal.h"

static void table_init_next_ft_attr(struct mlx5dv_hws_table *tbl,
				    struct cmd_ft_create_attr *ft_attr)
{
	ft_attr->rtc_valid = true;
	ft_attr->vport = tbl->vport;
	ft_attr->other_vport = tbl->other_vport;
	ft_attr->type = tbl->fw_ft_type;
	ft_attr->level = table_is_fdb_any(tbl->type) ?
		tbl->ctx->caps->fdb_ft.max_level - 1 :
		tbl->ctx->caps->nic_ft.max_level - 1;
}

static int table_default_fdb_miss_tbl_get(struct mlx5dv_hws_table *tbl)
{
	struct mlx5dv_hws_context *ctx = tbl->ctx;
	struct cmd_ft_create_attr ft_attr = {0};
	struct cmd_set_fte_attr fte_attr = {0};
	struct cmd_forward_tbl *default_miss;
	struct cmd_set_fte_dest dest = {0};
	uint8_t tbl_type = tbl->type;

	if (!table_is_fdb_any(tbl->type))
		return 0;

	if (tbl->common_res->default_miss) {
		tbl->common_res->default_miss->refcount++;
		return 0;
	}

	ft_attr.type = tbl->fw_ft_type;
	ft_attr.level = tbl->ctx->caps->fdb_ft.max_level; /* The last level */
	ft_attr.rtc_valid = false;

	dest.destination_type = MLX5_FLOW_DESTINATION_TYPE_VPORT;
	dest.destination_id = ctx->caps->eswitch_manager_vport_number;

	fte_attr.action_flags = MLX5_IFC_FLOW_CONTEXT_ACTION_FWD_DEST;
	fte_attr.dests_num = 1;
	fte_attr.dests = &dest;

	default_miss = cmd_forward_tbl_create(context_get_local_ibv(ctx),
					      &ft_attr,
					      &fte_attr);
	if (!default_miss) {
		HWS_ERR("Failed to create default miss table type: 0x%x", tbl_type);
		return -errno;
	}

	tbl->common_res->default_miss = default_miss;
	tbl->common_res->default_miss->refcount++;

	return 0;
}

static void table_default_fdb_miss_tbl_put(struct mlx5dv_hws_table *tbl)
{
	struct cmd_forward_tbl *default_miss;

	if (!table_is_fdb_any(tbl->type))
		return;

	default_miss = tbl->common_res->default_miss;
	if (--default_miss->refcount)
		return;

	cmd_forward_tbl_destroy(default_miss);
	tbl->common_res->default_miss = NULL;
}

static int
table_connect_to_default_miss_tbl(struct mlx5dv_hws_table *tbl,
				  struct mlx5dv_devx_obj *ft)
{
	struct cmd_ft_modify_attr ft_attr = {0};
	int ret;

	assert(table_is_fdb_any(tbl->type));

	cmd_set_attr_connect_miss_tbl(tbl->ctx,
				      tbl->fw_ft_type,
				      tbl->type,
				      tbl->other_vport,
				      tbl->vport,
				      &ft_attr);

	/* Connect to next */
	ret = cmd_flow_table_modify(ft, &ft_attr);
	if (ret) {
		HWS_ERR("Failed to connect FT to default FDB FT");
		return ret;
	}

	return 0;
}

struct mlx5dv_devx_obj *table_create_default_ft(struct ibv_context *ibv,
						struct mlx5dv_hws_table *tbl)
{
	struct cmd_ft_create_attr ft_attr = {0};
	struct mlx5dv_devx_obj *ft_obj;
	int ret;

	table_init_next_ft_attr(tbl, &ft_attr);

	ft_obj = cmd_flow_table_create(ibv, &ft_attr);
	if (ft_obj && table_is_fdb_any(tbl->type)) {
		/* Take/create ref over the default miss */
		ret = table_default_fdb_miss_tbl_get(tbl);
		if (ret) {
			HWS_ERR("Failed to get default fdb miss for type: %d\n",
			       tbl->type);
			goto free_ft_obj;
		}

		ret = table_connect_to_default_miss_tbl(tbl, ft_obj);
		if (ret) {
			HWS_ERR("Failed connecting to default miss tbl (type: %d)",
			       tbl->type);
			goto down_miss_tbl;
		}
	}

	return ft_obj;

down_miss_tbl:
	table_default_fdb_miss_tbl_put(tbl);
free_ft_obj:
	cmd_destroy_obj(ft_obj);
	return NULL;
}

void table_destroy_default_ft(struct mlx5dv_hws_table *tbl,
			      struct mlx5dv_devx_obj *ft_obj)
{
	cmd_destroy_obj(ft_obj);
	table_default_fdb_miss_tbl_put(tbl);
}

static int
table_shared_gvmi_resource_create(struct mlx5dv_hws_context *ctx,
				  enum mlx5dv_hws_table_type type,
				  struct context_shared_gvmi_res *gvmi_res)
{
	struct cmd_ft_create_attr ft_attr = {0};
	uint32_t calculated_ft_id = 0; /*Fix compilation warning*/
	struct mlx5dv_devx_obj *obj;

	if (!context_shared_gvmi_used(ctx))
		return 0;

	ft_attr.type = table_get_rtc_fw_ft_type(type);
	ft_attr.level = ctx->caps->nic_ft.max_level - 1;
	ft_attr.rtc_valid = true;

	gvmi_res->end_ft = cmd_flow_table_create(context_get_local_ibv(ctx),
						 &ft_attr);
	if (!gvmi_res->end_ft) {
		HWS_ERR("Failed to create end-ft");
		return -errno;
	}

	/* Due to FW BWC issue, provide FT type in the MSB of the ID */
	table_get_fw_ft_type(type, &calculated_ft_id);
	calculated_ft_id = calculated_ft_id << 24;
	calculated_ft_id |= gvmi_res->end_ft->object_id;

	/* Create alias to that FT */
	obj = cmd_allow_and_create_aliased_obj(ctx->local_ibv_ctx,
					       ctx->ibv_ctx,
					       ctx->caps->vhca_id,
					       calculated_ft_id,
					       MLX5_GENERAL_OBJ_TYPE_FT_ALIAS);
	if (!obj) {
		HWS_ERR("Failed to create alias end-ft");
		goto free_end_ft;
	}

	gvmi_res->aliased_end_ft = obj;

	return 0;

free_end_ft:
	cmd_destroy_obj(gvmi_res->end_ft);
	return -errno;
}

static void
table_shared_gvmi_resourse_destroy(struct mlx5dv_hws_context *ctx,
				   struct context_shared_gvmi_res *gvmi_res)
{
	if (!context_shared_gvmi_used(ctx))
		return;

	if (gvmi_res->aliased_end_ft) {
		cmd_destroy_obj(gvmi_res->aliased_end_ft);
		gvmi_res->aliased_end_ft = NULL;
	}
	if (gvmi_res->end_ft) {
		cmd_destroy_obj(gvmi_res->end_ft);
		gvmi_res->end_ft = NULL;
	}
}

static struct context_shared_gvmi_res *
table_get_shared_gvmi_res(struct mlx5dv_hws_table *tbl)
{
	struct mlx5dv_hws_context *ctx = tbl->ctx;
	struct context_shared_gvmi_res *gvmi_res;
	int ret;

	if (!context_shared_gvmi_used(ctx))
		return NULL;

	gvmi_res = tbl->common_res->gvmi_res;
	if (gvmi_res) {
		gvmi_res->refcount++;
		return gvmi_res;
	}

	gvmi_res = calloc(1, sizeof(struct context_shared_gvmi_res));
	if (!gvmi_res) {
		errno = ENOMEM;
		return NULL;
	}

	ret = table_shared_gvmi_resource_create(ctx, tbl->type, gvmi_res);
	if (ret) {
		HWS_ERR("Failed to create shared gvmi res for type: %d", tbl->type);
		goto free_gvmi_res;
	}

	tbl->common_res->gvmi_res = gvmi_res;
	tbl->common_res->gvmi_res->refcount = 1;

	return gvmi_res;

free_gvmi_res:
	free(gvmi_res);
	return NULL;
}

static void table_put_shared_gvmi_res(struct mlx5dv_hws_table *tbl)
{
	struct mlx5dv_hws_context *ctx = tbl->ctx;

	if (!context_shared_gvmi_used(ctx))
		return;

	if (--tbl->common_res->gvmi_res->refcount)
		return;

	table_shared_gvmi_resourse_destroy(ctx, tbl->common_res->gvmi_res);

	free(tbl->common_res->gvmi_res);
	tbl->common_res->gvmi_res = NULL;
}

static void table_shared_ctx_res_uninit(struct mlx5dv_hws_table *tbl)
{
	struct mlx5dv_hws_context *ctx = tbl->ctx;

	if (!context_shared_gvmi_used(ctx))
		return;

	cmd_destroy_obj(tbl->local_ft);

	table_put_shared_gvmi_res(tbl);
}

static int table_shared_ctx_res_init(struct mlx5dv_hws_context *ctx,
				     struct mlx5dv_hws_table *tbl)
{
	struct cmd_ft_modify_attr ft_attr = {0};
	int ret;

	if (!context_shared_gvmi_used(ctx))
		return 0;

	/* Create local-ft for root access */
	tbl->local_ft = table_create_default_ft(context_get_local_ibv(ctx), tbl);
	if (!tbl->local_ft) {
		HWS_ERR("Failed to create local-ft");
		return -errno;
	}

	if (!table_get_shared_gvmi_res(tbl)) {
		HWS_ERR("Failed to shared gvmi resources");
		goto clean_local_ft;
	}

	/* On shared gvmi the default behavior is jump to alias end ft */
	cmd_set_attr_connect_miss_tbl(tbl->ctx,
				      tbl->fw_ft_type,
				      tbl->type,
				      tbl->other_vport,
				      tbl->vport,
				      &ft_attr);

	ret = cmd_flow_table_modify(tbl->ft, &ft_attr);
	if (ret) {
		HWS_ERR("Failed to point table to its default miss");
		goto clean_shared_res;
	}

	return 0;

clean_shared_res:
	table_put_shared_gvmi_res(tbl);
clean_local_ft:
	table_destroy_default_ft(tbl, tbl->local_ft);
	return -errno;
}

static int table_init(struct mlx5dv_hws_table *tbl)
{
	struct mlx5dv_hws_context *ctx = tbl->ctx;
	int ret;

	table_get_fw_ft_type(tbl->type, &tbl->fw_ft_type);

	if (table_is_rdma_transport(tbl->type)) {
		ret = context_query_ib_port(ctx,
					    tbl->ib_port,
					    &tbl->vport,
					    &tbl->other_vport);
		if (ret) {
			HWS_ERR("Failed to query IB port");
			return ret;
		}
	}

	if (table_is_root(tbl))
		return 0;

	pthread_spin_lock(&ctx->ctrl_lock);

	tbl->common_res = context_get_common_res(ctx, tbl->type, tbl->ib_port);
	if (!tbl->common_res) {
		HWS_ERR("Failed to get common resources");
		goto out_unlock;
	}

	tbl->ft = table_create_default_ft(tbl->ctx->ibv_ctx, tbl);
	if (!tbl->ft) {
		HWS_ERR("Failed to create flow table devx object");
		goto out_unlock;
	}

	ret = table_shared_ctx_res_init(ctx, tbl);
	if (ret)
		goto tbl_destroy;

	ret = action_get_default_stc(ctx, tbl->type, tbl->ib_port);
	if (ret)
		goto free_shared_ctx;

	pthread_spin_unlock(&ctx->ctrl_lock);

	return 0;

free_shared_ctx:
	table_shared_ctx_res_uninit(tbl);
tbl_destroy:
	table_destroy_default_ft(tbl, tbl->ft);
out_unlock:
	pthread_spin_unlock(&ctx->ctrl_lock);
	return -errno;
}

static void table_uninit(struct mlx5dv_hws_table *tbl)
{
	if (table_is_root(tbl))
		return;

	pthread_spin_lock(&tbl->ctx->ctrl_lock);
	action_put_default_stc(tbl->ctx, tbl->type, tbl->ib_port);
	table_shared_ctx_res_uninit(tbl);
	table_destroy_default_ft(tbl, tbl->ft);
	pthread_spin_unlock(&tbl->ctx->ctrl_lock);
}

static int table_process_attr(struct mlx5dv_hws_context *ctx,
			      struct mlx5dv_hws_table_attr *attr)
{
	if (attr->comp_mask) {
		HWS_ERR("Invalid comp mask provided");
		goto out_no_supp;
	}

	if (attr->type >= MLX5DV_HWS_TABLE_TYPE_MAX) {
		HWS_ERR("Invalid table type %d", attr->type);
		goto out_no_supp;
	}

	if (attr->ib_port && !table_is_rdma_transport(attr->type)) {
		HWS_ERR("IB port is valid only for RDMA transport tables");
		goto out_no_supp;
	}

	if (attr->level == 0) {
		if (attr->type != MLX5DV_HWS_TABLE_TYPE_NIC_RX &&
		    attr->type != MLX5DV_HWS_TABLE_TYPE_NIC_TX &&
		    attr->type != MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX &&
		    attr->type != MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX &&
		    attr->type != MLX5DV_HWS_TABLE_TYPE_FDB) {
			HWS_ERR("Table type not supported by root table");
			goto out_no_supp;
		}
	} else {
		if (!(ctx->flags & CONTEXT_FLAG_HWS_SUPPORT)) {
			HWS_ERR("HWS not supported, cannot create HWS table");
			goto out_no_supp;
		}

		if (context_shared_gvmi_used(ctx) && !table_is_nic(attr->type)) {
			HWS_ERR("Only NIC tables support shared port resources");
			goto out_no_supp;
		}

		if (attr->type == MLX5DV_HWS_TABLE_TYPE_FDB && !ctx->caps->fdb_unified_en) {
			HWS_ERR("Unified FDB not supported by current FW");
			goto out_no_supp;
		}
	}

	return 0;

out_no_supp:
	errno = EOPNOTSUPP;
	return -errno;
}

struct mlx5dv_hws_table *
mlx5dv_hws_table_create(struct mlx5dv_hws_context *ctx,
			struct mlx5dv_hws_table_attr *attr)
{
	struct mlx5dv_hws_table *tbl;
	int ret;

	ret = table_process_attr(ctx, attr);
	if (ret) {
		HWS_ERR("Failed to process table attr");
		return NULL;
	}

	tbl = calloc(1, sizeof(*tbl));
	if (!tbl) {
		errno = ENOMEM;
		return NULL;
	}

	tbl->ctx = ctx;
	tbl->type = attr->type;
	tbl->level = attr->level;
	tbl->ib_port = attr->ib_port;
	list_head_init(&tbl->head);
	list_head_init(&tbl->default_miss.head);
	list_head_init(&tbl->isolated_matchers);

	ret = table_init(tbl);
	if (ret) {
		HWS_ERR("Failed to initialise table type[%d]", attr->type);
		goto free_tbl;
	}

	pthread_spin_lock(&ctx->ctrl_lock);
	list_add(&ctx->head, &tbl->next);
	pthread_spin_unlock(&ctx->ctrl_lock);

	return tbl;

free_tbl:
	free(tbl);
	return NULL;
}

int mlx5dv_hws_table_destroy(struct mlx5dv_hws_table *tbl)
{
	struct mlx5dv_hws_context *ctx = tbl->ctx;

	pthread_spin_lock(&ctx->ctrl_lock);
	if (!list_empty(&tbl->head) || !list_empty(&tbl->isolated_matchers)) {
		HWS_ERR("Cannot destroy table containing matchers");
		errno = EBUSY;
		goto unlock_err;
	}

	if (!list_empty(&tbl->default_miss.head)) {
		HWS_ERR("Cannot destroy table pointed by default miss");
		errno = EBUSY;
		goto unlock_err;
	}

	list_del(&tbl->next);
	pthread_spin_unlock(&ctx->ctrl_lock);
	table_uninit(tbl);
	free(tbl);

	return 0;

unlock_err:
	pthread_spin_unlock(&ctx->ctrl_lock);
	return -errno;
}

struct mlx5dv_devx_obj *table_get_last_ft(struct mlx5dv_hws_table *tbl)
{
	struct mlx5dv_devx_obj *last_ft = tbl->ft;
	struct mlx5dv_hws_matcher *matcher;

	list_for_each(&tbl->head, matcher, next)
		last_ft = matcher->end_ft;

	return last_ft;
}

int table_ft_set_default_next_ft(struct mlx5dv_hws_table *tbl,
				 struct mlx5dv_devx_obj *ft_obj)
{
	struct cmd_ft_modify_attr ft_attr = {0};
	int ret;

	/* Due to FW limitation, resetting the flow table to default action will
	 * disconnect RTC when ignore_flow_level_rtc_valid is not supported.
	 */
	if (!tbl->ctx->caps->nic_ft.ignore_flow_level_rtc_valid)
		return 0;

	if (table_is_fdb_any(tbl->type))
		return table_connect_to_default_miss_tbl(tbl, ft_obj);

	ft_attr.modify_fs = MLX5_IFC_MODIFY_FLOW_TABLE_MISS_ACTION;
	ft_attr.type = tbl->fw_ft_type;
	ft_attr.vport = tbl->vport;
	ft_attr.other_vport = tbl->other_vport;
	ft_attr.table_miss_action = MLX5_IFC_MODIFY_FLOW_TABLE_MISS_ACTION_DEFAULT;

	ret = cmd_flow_table_modify(ft_obj, &ft_attr);
	if (ret) {
		HWS_ERR("Failed to set FT default miss action");
		return ret;
	}

	return 0;
}

static int table_ft_set_next_ft(struct mlx5dv_devx_obj *ft,
				uint32_t fw_ft_type,
				bool other_vport,
				uint16_t vport,
				uint32_t next_ft_id)
{
	struct cmd_ft_modify_attr ft_attr = {0};

	ft_attr.modify_fs = MLX5_IFC_MODIFY_FLOW_TABLE_MISS_ACTION;
	ft_attr.type = fw_ft_type;
	ft_attr.vport = vport;
	ft_attr.other_vport = other_vport;
	ft_attr.table_miss_id = next_ft_id;
	ft_attr.table_miss_action = MLX5_IFC_MODIFY_FLOW_TABLE_MISS_ACTION_GOTO_TBL;

	return cmd_flow_table_modify(ft, &ft_attr);
}

int table_update_connected_miss_tables(struct mlx5dv_hws_table *dst_tbl)
{
	struct mlx5dv_hws_table *src_tbl;
	int ret;

	if (list_empty(&dst_tbl->default_miss.head))
		return 0;

	list_for_each(&dst_tbl->default_miss.head, src_tbl, default_miss.next) {
		ret = table_connect_to_miss_table(src_tbl, dst_tbl);
		if (ret) {
			HWS_ERR("Failed to update source miss table, unexpected behavior");
			return ret;
		}
	}

	return 0;
}

int table_connect_src_ft_to_miss_table(struct mlx5dv_hws_table *src_tbl,
				       struct mlx5dv_devx_obj *src_ft,
				       struct mlx5dv_hws_table *dst_tbl)
{
	struct mlx5dv_hws_matcher *matcher;
	int ret;

	if (dst_tbl) {
		if (list_empty(&dst_tbl->head)) {
			/* Connect src_tbl ft to dst_tbl start anchor */
			ret = table_ft_set_next_ft(src_ft,
						   src_tbl->fw_ft_type,
						   src_tbl->other_vport,
						   src_tbl->vport,
						   dst_tbl->ft->object_id);
			if (ret)
				return ret;

			/* Reset ft RTC to default RTC */
			ret = cmd_flow_table_modify_rtc(src_ft,
							src_tbl->fw_ft_type,
							src_tbl->other_vport,
							src_tbl->vport,
							NULL,
							NULL);
			if (ret)
				return ret;
		} else {
			/* Connect src_tbl ft to dst_tbl first matcher RTC */
			matcher = list_top(&dst_tbl->head,
					   struct mlx5dv_hws_matcher,
					   next);
			ret = cmd_flow_table_modify_rtc(src_ft,
							src_tbl->fw_ft_type,
							src_tbl->other_vport,
							src_tbl->vport,
							matcher->match_ste.rtc_0,
							matcher->match_ste.rtc_1);
			if (ret)
				return ret;

			/* Reset next miss FT to default */
			ret = table_ft_set_default_next_ft(src_tbl, src_ft);
			if (ret)
				return ret;
		}
	} else {
		/* Reset next miss FT to default */
		ret = table_ft_set_default_next_ft(src_tbl, src_ft);
		if (ret)
			return ret;

		/* Reset ft RTC to default RTC */
		ret = cmd_flow_table_modify_rtc(src_ft,
						src_tbl->fw_ft_type,
						src_tbl->other_vport,
						src_tbl->vport,
						NULL, NULL);
		if (ret)
			return ret;
	}

	return 0;
}

int table_connect_to_miss_table(struct mlx5dv_hws_table *src_tbl,
				struct mlx5dv_hws_table *dst_tbl)
{
	struct mlx5dv_hws_matcher *matcher;
	struct mlx5dv_devx_obj *src_ft;
	int ret;

	/* Connect last FT in the src_tbl matchers chain */
	src_ft = table_get_last_ft(src_tbl);
	ret = table_connect_src_ft_to_miss_table(src_tbl, src_ft, dst_tbl);
	if (ret)
		return ret;

	/* Connect isolated matchers FT */
	list_for_each(&src_tbl->isolated_matchers, matcher, next) {
		src_ft = matcher->end_ft;
		ret = table_connect_src_ft_to_miss_table(src_tbl, src_ft, dst_tbl);
		if (ret)
			return ret;
	}

	src_tbl->default_miss.miss_tbl = dst_tbl;

	return 0;
}

static int table_set_default_miss_not_valid(struct mlx5dv_hws_table *tbl,
					    struct mlx5dv_hws_table *miss_tbl)
{
	if (!tbl->ctx->caps->nic_ft.ignore_flow_level_rtc_valid ||
	    context_shared_gvmi_used(tbl->ctx)) {
		HWS_ERR("Default miss table is not supported");
		errno = EOPNOTSUPP;
		return -errno;
	}

	if (table_is_root(tbl) ||
	    (miss_tbl && table_is_root(miss_tbl)) ||
	    (miss_tbl && miss_tbl->type != tbl->type)) {
		HWS_ERR("Invalid arguments");
		errno = EINVAL;
		return -errno;
	}

	return 0;
}

int mlx5dv_hws_table_set_default_miss(struct mlx5dv_hws_table *tbl,
				      struct mlx5dv_hws_table *miss_tbl)
{
	struct mlx5dv_hws_context *ctx = tbl->ctx;
	struct mlx5dv_hws_table *old_miss_tbl;
	int ret;

	ret = table_set_default_miss_not_valid(tbl, miss_tbl);
	if (ret)
		return ret;

	pthread_spin_lock(&ctx->ctrl_lock);
	old_miss_tbl = tbl->default_miss.miss_tbl;
	ret = table_connect_to_miss_table(tbl, miss_tbl);
	if (ret)
		goto out;

	if (old_miss_tbl)
		list_del(&tbl->default_miss.next);

	if (miss_tbl)
		list_add(&miss_tbl->default_miss.head, &tbl->default_miss.next);
	else
		list_del(&tbl->default_miss.next);

	pthread_spin_unlock(&ctx->ctrl_lock);
	return 0;
out:
	pthread_spin_unlock(&ctx->ctrl_lock);
	return ret;
}

int
table_conv_hws_tbl_type_to_dv_ft(enum mlx5dv_hws_table_type tbl_type,
				 enum mlx5dv_flow_table_type *ft_type)
{
	switch (tbl_type) {
	case MLX5DV_HWS_TABLE_TYPE_NIC_RX:
		*ft_type = MLX5DV_FLOW_TABLE_TYPE_NIC_RX;
		break;
	case MLX5DV_HWS_TABLE_TYPE_NIC_TX:
		*ft_type = MLX5DV_FLOW_TABLE_TYPE_NIC_TX;
		break;
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX:
		*ft_type = MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_RX;
		break;
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX:
		*ft_type = MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_TX;
		break;
	case MLX5DV_HWS_TABLE_TYPE_FDB:
		*ft_type = MLX5DV_FLOW_TABLE_TYPE_FDB;
		break;
	default:
		errno = ENOTSUP;
		return -errno;
	}
	return 0;
}
