/* SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

#ifndef _HWS_TABLE_H_
#define _HWS_TABLE_H_

#define TABLE_ROOT_LEVEL 0

struct mlx5dv_hws_default_miss {
	/* My miss table */
	struct mlx5dv_hws_table *miss_tbl;
	struct list_node next;
	/* Tables missing to my table */
	struct list_head head;
};

struct mlx5dv_hws_table {
	struct mlx5dv_hws_context *ctx;
	struct mlx5dv_devx_obj *ft;
	struct mlx5dv_devx_obj *local_ft;
	struct context_common_res *common_res;
	enum mlx5dv_hws_table_type type;
	enum mlx5_ifc_ft_type fw_ft_type;
	uint32_t level;
	uint32_t ib_port;
	uint16_t vport;
	bool other_vport;
	struct list_head head;
	struct list_node next;
	struct list_head isolated_matchers;
	struct mlx5dv_hws_default_miss default_miss;
};

static inline bool table_is_fw_fdb_any(uint32_t fw_tbl_type)
{
	return fw_tbl_type == MLX5_IFC_FT_FDB_TX ||
	       fw_tbl_type == MLX5_IFC_FT_FDB_RX ||
	       fw_tbl_type == MLX5_IFC_FT_FDB_UNIFIED;
}

static inline bool table_is_nic(enum mlx5dv_hws_table_type tbl_type)
{
	return tbl_type == MLX5DV_HWS_TABLE_TYPE_NIC_RX ||
	       tbl_type == MLX5DV_HWS_TABLE_TYPE_NIC_TX;
}

static inline bool table_is_rdma_transport(enum mlx5dv_hws_table_type tbl_type)
{
	return tbl_type == MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX ||
	       tbl_type == MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX;
}

static inline bool table_is_fdb_any(enum mlx5dv_hws_table_type tbl_type)
{
	return tbl_type == MLX5DV_HWS_TABLE_TYPE_FDB ||
	       tbl_type == MLX5DV_HWS_TABLE_TYPE_FDB_TX ||
	       tbl_type == MLX5DV_HWS_TABLE_TYPE_FDB_RX;
}

/* This function is used for flow-table objects, when there is no FDB subdomain types */
static inline
void table_get_fw_ft_type(enum mlx5dv_hws_table_type type,
			  enum mlx5_ifc_ft_type *ret_type)
{
	switch (type) {
	case MLX5DV_HWS_TABLE_TYPE_NIC_RX:
		*ret_type = MLX5_IFC_FT_NIC_RX;
		break;
	case MLX5DV_HWS_TABLE_TYPE_NIC_TX:
		*ret_type = MLX5_IFC_FT_NIC_TX;
		break;
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX:
		*ret_type = MLX5_IFC_FT_RDMA_TRAN_RX;
		break;
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX:
		*ret_type = MLX5_IFC_FT_RDMA_TRAN_TX;
		break;
	case MLX5DV_HWS_TABLE_TYPE_FDB:
	case MLX5DV_HWS_TABLE_TYPE_FDB_RX:
	case MLX5DV_HWS_TABLE_TYPE_FDB_TX:
		*ret_type = MLX5_IFC_FT_FDB;
		break;
	default:
		assert(0);
	}
}

/* For flow-table there is only FDB fw type for all the FDB domains
 * (FDB_RX, FDB_TX, UNIFIED), for all other resources that connected to RTC
 * (RTC / STE / STC) there is the distingue between the types.
 * This function return the real fw_ft_type, not like the other
 * function table_get_fw_ft_type that uses for tables resources only.
 */
static inline uint32_t
table_get_rtc_fw_ft_type(enum mlx5dv_hws_table_type tbl_type)
{
	switch (tbl_type) {
	case MLX5DV_HWS_TABLE_TYPE_NIC_RX:
		return MLX5_IFC_FT_NIC_RX;
	case MLX5DV_HWS_TABLE_TYPE_NIC_TX:
		return MLX5_IFC_FT_NIC_TX;
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX:
		return MLX5_IFC_FT_RDMA_TRAN_RX;
	case MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX:
		return MLX5_IFC_FT_RDMA_TRAN_TX;
	case MLX5DV_HWS_TABLE_TYPE_FDB_RX:
		return MLX5_IFC_FT_FDB_RX;
	case MLX5DV_HWS_TABLE_TYPE_FDB_TX:
		return MLX5_IFC_FT_FDB_TX;
	case MLX5DV_HWS_TABLE_TYPE_FDB:
		return MLX5_IFC_FT_FDB_UNIFIED;
	default:
		assert(0);
		return 0;
	}
}

static inline bool table_is_root(struct mlx5dv_hws_table *tbl)
{
	return (tbl->level == TABLE_ROOT_LEVEL);
}

struct mlx5dv_devx_obj *table_create_default_ft(struct ibv_context *ibv,
						struct mlx5dv_hws_table *tbl);

void table_destroy_default_ft(struct mlx5dv_hws_table *tbl,
			      struct mlx5dv_devx_obj *ft_obj);

int table_connect_to_miss_table(struct mlx5dv_hws_table *src_tbl,
				struct mlx5dv_hws_table *dst_tbl);

int table_connect_src_ft_to_miss_table(struct mlx5dv_hws_table *src_tbl,
				       struct mlx5dv_devx_obj *src_ft,
				       struct mlx5dv_hws_table *dst_tbl);

int table_update_connected_miss_tables(struct mlx5dv_hws_table *dst_tbl);

struct mlx5dv_devx_obj *table_get_last_ft(struct mlx5dv_hws_table *tbl);

int table_ft_set_default_next_ft(struct mlx5dv_hws_table *tbl,
				 struct mlx5dv_devx_obj *ft_obj);

int table_ft_set_next_rtc(struct mlx5dv_devx_obj *ft,
			  uint32_t fw_ft_type,
			  struct mlx5dv_devx_obj *rtc_0,
			  struct mlx5dv_devx_obj *rtc_1);

int table_conv_hws_tbl_type_to_dv_ft(enum mlx5dv_hws_table_type tbl_type,
				     enum mlx5dv_flow_table_type *ft_type);

#endif /* _HWS_TABLE_H_ */
